/*
 * BindExperimentController.java
 *
 * Copyright (c) 2005 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

package net.emulab.netlab.client;

import java.awt.Color;
import java.awt.Image;
import java.awt.MediaTracker;
import java.awt.Rectangle;
import java.io.IOException;
import java.net.URL;
import java.util.Arrays;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.List;
import java.util.Vector;

import net.emulab.netlab.client.emulab.EmulabExperimentProxy;
import net.emulab.netlab.client.emulab.EmulabXmlRpcException;
import net.emulab.ns.DBState;
import net.emulab.ns.NSFile;
import net.emulab.ns.NSInterface;
import net.emulab.ns.NSLan;
import net.emulab.ns.NSLink;
import net.emulab.ns.NSNode;
import net.emulab.ns.NSObject;
import net.emulab.ns.NSTopology;
import net.emulab.util.FileDiff;
import net.emulab.util.ServerProperties;
import thinlet.ArrayController;
import thinlet.AutoFillController;
import thinlet.ThinletAPI;
import thinlet.ThinletContent;
import thinlet.ThinletKeyValue;

/**
 * Controller for the "Open" and "Modify" experiment dialogs.
 */
public class BindExperimentController extends NetlabArrayController {
	private static final int X_OFFSET = 50;
	private static final int Y_OFFSET = 25;
	private static final double SCALE = 1.75;

	public static final Color REMOVE_COLOR = new Color(225, 150, 170);
	public static final Color ADD_COLOR = new Color(150, 225, 170);
	public static final Color NUMBER_COLOR = new Color(75, 75, 75);
	
	public static final String COMPARE_EXPLANATION =
		"Check the differences below and click 'Ok' only if they are acceptable "
		+ "to you.  "
		+ "Lines that will be added to the file are colored green and "
		+ "lines that will be removed are colored red.  "
		+ "To unhighlight lines, double click on them or select them and use the "
		+ "'Hide Changes' button.  "
		+ "Alternatively, you can choose to create a new experiment later instead "
		+ "of overwriting this one.";
	
	/**
	 * The keys that should be searched when filtering experiment models.
	 */
	private static final String EXP_FILTER_KEYS[] = { "name", "gid",
			"description", "state", "expt_head", };

	private static final String DIFF_FILTER_KEYS[] = { "line" };

	/**
	 * Content converter that colors the rows in the dialog that shows the
	 * differences between the old and new NS files.
	 */
	public static final ThinletContent ENTRY_CONVERTER = new ThinletContent() {
		
		private void setRowColor(ThinletAPI thinlet, Object row, Color color) {
			Object cells[] = thinlet.getItems(row);
			Color oldColor;
			int lpc;

			oldColor = thinlet.getColor(cells[0], "background");
			if ((oldColor == null) || !oldColor.equals(color)) {
				for (lpc = 0; lpc < cells.length; lpc++) {
					thinlet.setColor(cells[lpc], "background", color);
					if (lpc == 1) {
						thinlet.setColor(cells[lpc], "foreground", NUMBER_COLOR);
					}
				}
			}
		}

		public void initItem(ThinletAPI thinlet, Object item, Object content) {
			FileDiff fd = (FileDiff) content;
			Color color;
			
			if ("-".equals(fd.getType())) {
				color = REMOVE_COLOR;
			}
			else if ("+".equals(fd.getType())) {
				color = ADD_COLOR;
			}
			else {
				color = null;
			}

			this.setRowColor(thinlet, item, color);
		}

		public void replaceItem(ThinletAPI thinlet, Object item, Object content) {
			this.initItem(thinlet, item, content);
		}
	};

	private Hashtable nsToComp = new Hashtable();

	protected boolean synced;

	public Object defaultAction;
	
	/**
	 * The error dialog, used to display any errors that occur during a modify.
	 */
	private Object errorDialog;
	public Object errorDetails;
	public String nsfilestr;
	
	public Object logButton;
	
	public Object difftable;
	public Object hideChangesButton;
	
	public String mode = "";
	public String modeText = "";
	
	private NSTopology nst;

	/**
	 * The compare dialog, used to show differences between the loaded NS file and
	 * the one we'll create.
	 */
	private Object compareDialog;

	public DBState dbs = new DBState();

	/**
	 * Controller for the project selector.
	 */
	public final ArrayController pc;

	/**
	 * Controller for the search box.
	 */
	public final AutoFillController sc;
	
	public final ArrayController dc;

	public BindExperimentController(NetlabClient netlab) {
		super(netlab);

		super.addFilterKeys(EXP_FILTER_KEYS);
		
		this.pc = new ArrayController(netlab);
		this.sc = new AutoFillController(netlab);
		this.dc = new ArrayController(netlab);
		this.dc.addFilterKeys(DIFF_FILTER_KEYS);
		
		/* Bind the project selector to the experiment list. */
		this.tkv.observeKeyValue(this.pc, "selection.eids",
				ThinletKeyValue.CHANGE_ANY, this, "content");
		
		this.tkv.observeKeyValue(this, "dbs.diffs",
				ThinletKeyValue.CHANGE_ANY, this.dc, "content");
		
		this.tkv.addKeyDependency(this, "modifyWarning", "selection");
	}

	public Controller modelChanged(Object component) {
		return this;
	}

	public Controller setView(Object view) {
		super.setView(view);
		
		try {
			this.compareDialog = netlab.parse("expcompare.xml", this);
			this.errorDialog = this.netlab.parse("parseerror.xml", this);
		}
		catch (IOException e) {
			throw new InternalError();
		}
		
		return this;
	}

	public Controller triggeredBy(Object component) throws Exception {
		this.tkv.setKeyValue(this, "mode", this.netlab.getProperty(component, "mode"));
		this.tkv.setKeyValue(this, "modeText", "  " + this.mode + "  ");
		this.netlab.setString(this.defaultAction, "state", this.mode);

		/* 
		 * Create the ns file early so any errors are signalled before we add the
		 * dialog.
		 */
		if ("Modify".equals(this.mode)) {
			NSFile nsf = new NSFile();
			
			this.tkv.setKeyValue(this, "nsfilestr",
					nsf.toNSString(netlab.getTopologyController().getNSTopology()));
		}
		
		super.triggeredBy(component);
		
		return this;
	}
	
	/**
	 * Method that processes the experiment listing returned by Emulab into
	 * something we can use.
	 * 
	 * @param el The experiment listing as returned by experiment.getlist()
	 */
	private void setExperimentList(Hashtable el) {
		Hashtable allProjects = new Hashtable();
		Vector allEids = new Vector();
		Vector root = new Vector();
		Object newsel = null;
		Enumeration penum;

		penum = el.keys();
		while (penum.hasMoreElements()) {
			Hashtable projectContents = new Hashtable();
			String project = (String) penum.nextElement();
			Vector projEids = new Vector();
			Enumeration genum;
			Hashtable sub;

			sub = (Hashtable) el.get(project);
			genum = sub.keys();
			while (genum.hasMoreElements()) {
				String group = (String) genum.nextElement();
				int lpc, size;
				Vector eids;

				eids = (Vector) sub.get(group);
				size = eids.size();
				for (lpc = 0; lpc < size; lpc++) {
					Hashtable ht = (Hashtable) eids.get(lpc);
					int actual = 0, max = 0;
					String name;
					
					name = (String)ht.get("name");
					if (project.equals(this.netlab.defaultPid) &&
							name.equals(this.netlab.defaultEid)) {
						newsel = ht;
					}
					ht.put("thumbnail", this.netlab.getImageManager()
							.getObjectImage("coming-soon"));
					ht.put("needsThumbnail", Boolean.TRUE);
					try {
						actual = this.tkv.getKeyValueAsInteger(ht, "actual_nodes");
						max = this.tkv.getKeyValueAsInteger(ht, "maximum_nodes");
					}
					catch (IllegalArgumentException e) {
					}
					this.tkv.setKeyValue(ht, "node_count", actual > 0 ? actual : max);
				}
				projEids.addAll(eids);
				allEids.addAll(eids);
			}

			projectContents.put("name", "  " + project + "  ");
			projectContents.put("eids", projEids);

			root.add(projectContents);
		}
		allProjects.put("name", "  (All)  ");
		allProjects.put("eids", allEids);
		root.add(0, allProjects);

		this.tkv.setKeyValue(this.pc, "content", root);
		this.tkv.setKeyValue(this.pc, "selectedIndex", 0);

		this.synced = true;
		
		if (newsel != null) {
			this.tkv.setKeyValue(this, "selection", newsel);
			this.newListSelection();
		}
	}

	public void showDialog() {
		if (!this.synced) {
			this.syncExperimentList();
		}
	}

	private void syncExperimentList() {
		final ProgressController pc;

		pc = this.netlab.showProgressDialog("Downloading Experiment List...", false);
		new Downloader(this.netlab, pc) {
			public void download() throws Throwable {
				EmulabExperimentProxy eep;
				Hashtable ht;
				
				eep = new EmulabExperimentProxy(tkv, super.netlab.getPool());
				ht = eep.getlist("full");
				if (netlab.defaultPid != null && netlab.defaultEid != null) {
					Hashtable group, md;
					Vector exps;

					md = eep.metadata(netlab.defaultPid, netlab.defaultEid);
					md.put("name", md.get("eid"));
					md.put("description", md.get("expt_name"));
					md.put("expt_head", md.get("expt_head_uid"));
					group = (Hashtable)ht.get(netlab.defaultPid);
					if (group == null ||
							(exps = (Vector)group.get(md.get("gid"))) == null) {
						netlab.defaultPid = null;
						netlab.defaultEid = null;
					}
					else {
						List list = tkv.getKeyValueAsList(exps, "name");
						
						if (!list.contains(md.get("eid"))) {
							exps.add(md);
						}
					}
				}
				BindExperimentController.this.setExperimentList(ht);
			}

			public void downloadCancelled() {
				super.netlab.remove(BindExperimentController.this.view);
			}
			
			public void finished() {
				super.finished();
				
				if (synced && this.netlab.defaultPid != null && this.netlab.defaultEid != null) {
					ok();
				}
			}
		}.start();
	}

	private void addNodes(ProgressController pc, Object field, Hashtable viz)
			throws Exception {
		NSNode nsn[];
		List list;
		int lpc;

		list = this.nst.getNodes();
		nsn = (NSNode[]) list.toArray(new NSNode[list.size()]);
		pc.setStage(null, "Adding nodes...");
		pc.setContentLength(null, nsn.length);
		for (lpc = 0; lpc < nsn.length; lpc++) {
			ElementController ec;
			Object component;
			Hashtable point;

			point = (Hashtable) viz.get(nsn[lpc].getName().toString());
			ec = (ElementController) this.netlab.getControllerManager()
					.getController("net.emulab.netlab.client.NodeController");
			component = ec.copyElement(nsn[lpc]);
			this.nsToComp.put(nsn[lpc], component);
			this.netlab.add(field, component);
			this.netlab.setRectangle(component, "bounds",
					(int) (((Double) point.get("x")).doubleValue() * SCALE + X_OFFSET),
					(int) (((Double) point.get("y")).doubleValue() * SCALE + Y_OFFSET),
					-1, 0);
			pc.updateContentRemaining(null, nsn.length - lpc);
		}
	}

	private void addLans(ProgressController pc, Object field, Hashtable viz)
			throws Exception {
		NSLan nsl[];
		List list;
		int lpc;

		list = this.nst.getLans();
		nsl = (NSLan[]) list.toArray(new NSLan[list.size()]);
		pc.setStage(null, "Adding lans...");
		pc.setContentLength(null, nsl.length);
		for (lpc = 0; lpc < nsl.length; lpc++) {
			double ax = 0.0, ay = 0.0;
			ElementController ec;
			Object component;
			Hashtable point;
			Vector members;
			int lpc2, size;

			point = (Hashtable) viz.get(nsl[lpc].getName().toString());
			ec = (ElementController) this.netlab.getControllerManager()
					.getController("net.emulab.netlab.client.LanController");
			component = ec.copyElement(nsl[lpc]);
			this.nsToComp.put(nsl[lpc], component);
			this.netlab.add(field, component);

			members = nsl[lpc].getMembers();
			size = members.size();
			for (lpc2 = 0; lpc2 < size; lpc2++) {
				Rectangle bounds;
				NSInterface nsi;
				Object connector, ocomp;
				NSObject nso;

				nsi = (NSInterface) members.elementAt(lpc2);
				nso = nsi.getNSObject();
				ocomp = this.nsToComp.get(nso);
				bounds = this.netlab.getRectangle(ocomp, "bounds");
				ax += bounds.x;
				ay += bounds.y;
				connector = this.netlab.gotConnect(component, ocomp);
				this.netlab.putProperty(connector, "model", nsi);
			}
			
			if (point == null) {
				ax /= size;
				ay /= size;
				this.netlab.setRectangle(component, "bounds",
						(int) ax,
						(int) ay,
						-1, 0);
			}
			else {
				this.netlab.setRectangle(component, "bounds",
						(int) (((Double) point.get("x")).doubleValue() * SCALE + X_OFFSET),
						(int) (((Double) point.get("y")).doubleValue() * SCALE + Y_OFFSET),
						-1, 0);
			}
			
			pc.updateContentRemaining(null, nsl.length - lpc);
		}
	}

	private void addLinks(ProgressController pc, Object field) throws Exception {
		NSLink nslink[];
		List list;
		int lpc;

		list = this.nst.getLinks();
		nslink = (NSLink[]) list.toArray(new NSLink[list.size()]);
		pc.setStage(null, "Adding links...");
		pc.setContentLength(null, nslink.length);
		for (lpc = 0; lpc < nslink.length; lpc++) {
			NSObject nsleft, nsright;
			Object connector;

			nsleft = nslink[lpc].getLeftMember().getNSObject();
			nsright = nslink[lpc].getRightMember().getNSObject();
			connector = this.netlab.gotConnect(this.nsToComp.get(nsleft),
					this.nsToComp.get(nsright));
			this.netlab.getTopologyController().replaceNSObject(
					(NSObject) this.netlab.getProperty(connector, "model"),
					nslink[lpc]);
			this.netlab.putProperty(connector, "model", nslink[lpc]);
			pc.updateContentRemaining(null, nslink.length - lpc);
		}
	}

	public void syncTopology(ProgressController pc, String pid, String eid)
			throws Exception {
		EmulabExperimentProxy eep;
		Hashtable topo;
		DBState ndbs;
		String file;
		
		this.nst = new NSTopology();
		ndbs = new DBState();
		
		eep = new EmulabExperimentProxy(tkv, this.netlab.getPool());
		topo = eep.virtual_topology(pid, eid, new Vector(Arrays.asList(new String[] {
				"routes"
		})));
		file = eep.nsfile(pid, eid);

		ndbs.fromVirtualTopology(this.tkv, this.nst, topo, file);
		
		pc.done();
		
		this.tkv.setKeyValue(this, "dbs", ndbs);
		
		if (dbs.getWarnings().size() == 0 &&
				dbs.getDiffs().size() == 0) {
			this.finishSync(null);
		}
		else {
			this.netlab.setString(this.compareDialog, "text",
					"NS Comparison for " + pid + "/" + eid);
			this.netlab.setBoolean(this.hideChangesButton, "enabled", false);
			this.tkv.setKeyValue(this, "diffSortDirection", null);
			this.tkv.setKeyValue(this.dc, "activeKey", null);
			this.tkv.setKeyValue(this.dc, "searchKey", null);
			this.netlab.add(this.compareDialog);
		}
	}
	
	public void cancelCompare(Object sender) {
		this.netlab.remove(this.compareDialog);
		this.netlab.add(this.view);
	}
	
	public void finishSync(Object sender) {
		if (sender != null)
			this.netlab.remove(this.compareDialog);
		
		this.netlab.newFile(null);
		
		ProgressController pc = this.netlab.showProgressDialog("Downloading Topology...", false);
		new Downloader(this.netlab, pc) {
			public void download() throws Throwable {
				EmulabExperimentProxy eep = new EmulabExperimentProxy(tkv, netlab.getPool());
				Hashtable viz;

				String pid = tkv.getKeyValueAsString(BindExperimentController.this, "selection.pid");
				String eid = tkv.getKeyValueAsString(BindExperimentController.this, "selection.name");

				netlab.defaultPid = pid;
				netlab.defaultEid = eid;
				
				pc.setTitle("Downloading Visualization...");
				pc.setContentLength(null, 1);
				
				viz = eep.getviz(pid, eid);
				
				pc.setTitle("Downloading NS file...");
				pc.setContentLength(null, 1);
				
				pc.setTitle("Updating display...");
				pc.setProgressHalter(null, null);
				
				thinlet.remove(this.netlab.field);
				
				addNodes(pc, this.netlab.field, viz);

				addLans(pc, this.netlab.field, viz);
				
				addLinks(pc, this.netlab.field);
		
				tkv.setKeyValue(this.netlab.getTopologyController(), "content", nst);

				thinlet.add(this.netlab.deck, this.netlab.field);
				netlab.deselect(null);
				
				pc.done();
			}
			
			public void downloadCancelled() {
			}
		}.start();
	}

	public void downloadThumbnail() {
		final ProgressController pc;

		pc = this.netlab.showProgressDialog("Downloading Experiment List...");
		new Downloader(this.netlab, pc) {
			public void download() throws Throwable {
				MediaTracker mediatracker;
				EmulabExperimentProxy eep;
				byte thumbnailBits[];
				Image thumbnail;
				String pid, eid;

				pid = tkv.getKeyValueAsString(BindExperimentController.this, "selection.pid");
				eid = tkv.getKeyValueAsString(BindExperimentController.this, "selection.name");
				eep = new EmulabExperimentProxy(tkv, super.netlab.getPool());
				thumbnailBits = eep.thumbnail(pid, eid);
				thumbnail = super.netlab.getToolkit()
						.createImage(thumbnailBits);
				mediatracker = new MediaTracker(super.netlab);
				mediatracker.addImage(thumbnail, 1);
				try {
					mediatracker.waitForID(1, 5000);
				} catch (InterruptedException ie) {
				}
				tkv.setKeyValue(BindExperimentController.this, "selection.thumbnail", thumbnail);
				tkv.setKeyValue(BindExperimentController.this, "selection.needsThumbnail", false);
			}

			public void downloadCancelled() {
			}
		}.start();
	}

	public void retry() {
		if (this.netlab.getParent(this.errorDialog) != null)
			this.netlab.remove(this.errorDialog);
		
		this.ok();
	}
	
	public void setDiffSortDirection(String dir) {
		Object rows[] = this.netlab.getSelectedItems(this.difftable);
		int lpc;
		
		for (lpc = 0; lpc < rows.length; lpc++) {
			this.netlab.setBoolean(rows[lpc], "selected", false);
		}
		this.dc.setSortDirection(dir);
	}
	
	public String getDiffSortDirection() {
		return this.dc.sortDirection();
	}
	
	public void diffTableAction(Object sender) {
		Object rows[] = this.netlab.getSelectedItems(this.difftable);
		boolean enabled = false;
		int lpc;
		
		for (lpc = 0; lpc < rows.length; lpc++) {
			Object cell = this.netlab.getItem(rows[lpc], 0);
			
			if (this.netlab.getColor(cell, "background") != null) {
				enabled = true;
				break;
			}
		}
		
		this.netlab.setBoolean(this.hideChangesButton, "enabled", enabled);
	}
	
	public void hideChanges(Object sender) {
		Object rows[] = this.netlab.getItems(this.difftable);
		int lpc, start = -1, end = -1;
		
		for (lpc = 0; lpc < rows.length; lpc++) {
			if (this.netlab.getBoolean(rows[lpc], "selected")) {
				this.netlab.setBoolean(rows[lpc], "selected", false);
				if (start == -1) {
					start = end = lpc;
				}
				else {
					end = lpc;
				}
			}
		}

		List ar = this.dc.arrangedObjects();
		
		for (lpc = start; lpc <= end; lpc++) {
			FileDiff fd = (FileDiff)ar.get(lpc);
			
			if ("+".equals(fd.getType())) {
				int actual = this.dbs.getDiffs().indexOf(fd);
				
				fd.setType("");
				this.tkv.notifyKeyValue(this, "dbs.diffs",
						ThinletKeyValue.CHANGE_REPLACE, null, new int[] { actual, actual + 1 });
			}
			else if ("-".equals(fd.getType())) {
				this.dc.removeObject(fd);
			}
		}
		
		this.netlab.setBoolean(this.hideChangesButton, "enabled", false);
	}
	
	public String getModifyWarning() {
		String pid, eid;

		pid = this.tkv.getKeyValueAsString(this, "selection.pid");
		eid = this.tkv.getKeyValueAsString(this, "selection.name");
		
		return "Are you sure you want to modify " + pid + "/" + eid + "?";
	}

	public void ok() {
		final ProgressController pc;
		final String pid, eid;

		pid = this.tkv.getKeyValueAsString(this, "selection.pid");
		eid = this.tkv.getKeyValueAsString(this, "selection.name");

		this.netlab.remove(this.view);
		
		if ("Open".equals(this.mode)) {
			pc = this.netlab.showProgressDialog("Downloading topology...", false);
			new Downloader(this.netlab, pc) {
				public void download() throws Throwable {
					syncTopology(super.pc, pid, eid);
				}
				
				public void downloadCancelled() {
				}
			}.start();
		}
		else if ("Modify".equals(this.mode)) {
			pc = this.netlab.showProgressDialog("Modifying experiment...", false);
			new Downloader(this.netlab, pc) {
				public void download() throws Throwable {
					EmulabExperimentProxy exp = new EmulabExperimentProxy(tkv, netlab.getPool());
					
					exp.modify(pid, eid, nsfilestr, false); // XXX

					netlab.invoke(logButton, null, "action");
				}

				public void downloadError(EmulabXmlRpcException e) {
					netlab.setString(errorDetails, "text", e.getMessage());
					netlab.add(errorDialog);
				}
				
				public void downloadCancelled() {
				}
			}.start();
		}
	}
	
	public void newListSelection() {
		String pid, eid;
		
		pid = this.tkv.getKeyValueAsString(this, "selection.pid");
		eid = this.tkv.getKeyValueAsString(this, "selection.name");
		
		this.netlab.putProperty(this.logButton, "pid", pid);
		this.netlab.putProperty(this.logButton, "eid", eid);
	}

	public void cancel() {
		if (this.netlab.getParent(this.view) != null)
			this.netlab.remove(this.view);
		if (this.netlab.getParent(this.errorDialog) != null)
			this.netlab.remove(this.errorDialog);
	}

	public String toString() {
		return "CreateExperimentController[" + super.toString() + "; synced="
				+ this.synced + "; pc=" + this.pc + "; sc="
				+ this.sc + "]";
	}
}
