/*
 * NodePropertiesController.java
 *
 * Copyright (c) 2005 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

package net.emulab.netlab.client;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;
import java.util.SortedSet;
import java.util.Vector;

import net.emulab.netlab.client.emulab.EmulabNodeProxy;
import net.emulab.ns.DBState;
import net.emulab.ns.NSNode;
import net.emulab.ns.NSObject;
import net.emulab.ns.NSProgram;
import net.emulab.ns.NSTopology;
import net.emulab.ns.NodeTraffic;
import net.emulab.ns.Pair;
import thinlet.ArrayController;
import thinlet.AutoFillController;
import thinlet.IntersectArrayController;
import thinlet.NumberFormatter;
import thinlet.SizeFormatter;
import thinlet.ThinletKeyValue;

public class CoSPropertiesController extends PropertiesController {
	public final class ProgramArrayController extends ArrayController {
		private final NetlabClient netlab;
		
		private ProgramArrayController(NetlabClient netlab) {
			super(netlab);
			
			this.netlab = netlab;
		}

		public Object newObject() {
			TopologyController tc = netlab.getTopologyController();
			NSProgram retval = new NSProgram();
			
			retval.setName(tc.getNamer().suggest(PROGRAM_NAME_BASE));
			retval.setNode((NSNode) getNSObject());
			
			tc.addNSObject(retval);
		
			return retval;
		}

		public void removeObject(Object value) {
			TopologyController tc = netlab.getTopologyController();
			NSProgram nsp = (NSProgram) value;
			
			super.removeObject(value);
		
			tc.getNamer().remove(nsp.getName());
			tc.removeNSObject(nsp);
		}

		public void setSelection(Object selection) {
			super.setSelection(selection);
			
			if (selection instanceof NSProgram) {
				progNameFormatter.setNSObject((NSObject)selection);
			}
			else {
				progNameFormatter.setNSObject(null);
			}
		}
	}

	/**
	 * Initial auto-fill values for the OS ID field.
	 */
	private static final List OS_AUTO_FILL = Arrays.asList(new String[] {
			"FBSD-STD", "FC4-STD", "RHL-STD", "WINXP-UPDATE" , "ncSCADA-CoS" });

	private static final List FAILURE_AUTO_FILL = Arrays.asList(new String[] {
			"nonfatal" });
	
	/**
	 * Initial auto-fill values for the hardware field.
	 */
	private static final List HARDWARE_AUTO_FILL = Arrays.asList(new String[] {
			"pc" });

	public static final NumberFormatter TIMEOUT_FORMATTER = new NumberFormatter()
			.setMinimum(0.0).setEmpty(new Integer(-1));

	public static final NumberFormatter EEC_FORMATTER = new NumberFormatter()
			.setEmpty(new Integer(0));

	public static final NumberFormatter RATE_FORMATTER = new SizeFormatter(
			null, "bps").setMinimum(0.0);

	public static final NumberFormatter PACKET_SIZE_FORMATTER = new NumberFormatter()
			.setMinimum(0.0);

	public static final NumberFormatter INTERVAL_FORMATTER = new NumberFormatter()
			.setMinimum(0.0);

	public static final NumberFormatter COST_FORMATTER = new NumberFormatter()
                        .setMinimum(-1);

	private static final String PROGRAM_NAME_BASE = "prog";

	private static final String TRAFFIC_NAME_BASE = "tg";

	public final UniqueNameFormatter progNameFormatter;

	public final UniqueNameFormatter trafficNameFormatter;

	public Object trafficDialog;

	public boolean trafficDialogOpened;

	/**
	 * Controller for the OS ID field.
	 */
	public final AutoFillController osc;

	/**
	 * Controller for the startup script field.
	 */
	public final AutoFillController ssc;

	/**
	 * Controller for the hardware field.
	 */
	public final AutoFillController hc;

	/**
	 * Controller for the node field.
	 */
	public final AutoFillController nc;
	
	public final AutoFillController fac;
	
	private Object progDialog;
	private boolean progDialogOpened;

	public Object tarButton;

	public Object rpmButton;
	
	public boolean tarEnabled;
	public boolean rpmEnabled;

	public final ArrayController pc;
	public final AutoFillController pcc;
	public final AutoFillController pdc;

	public final TrafficArrayController tc;
	public final ArrayController tmc;

	public final ArrayController dc;
	
	/**
	 * Construct a NodePropertiesController with the given value.
	 */
	public CoSPropertiesController(final NetlabClient netlab) {
		super(netlab);

		this.trafficNameFormatter = new UniqueNameFormatter(netlab.getTopologyController());
		
		this.setObjectClass(NSNode.class);

		this.vis.put("phys", Boolean.TRUE);
		this.vis.put("software", Boolean.TRUE);
		this.vis.put("programs", Boolean.TRUE);
		this.vis.put("misc", Boolean.FALSE);
		this.vis.put("traf", Boolean.FALSE);
		
		this.progNameFormatter = new UniqueNameFormatter(netlab.getTopologyController());
		
		this.osc = new AutoFillController(netlab, new Vector(OS_AUTO_FILL));
		this.ssc = new AutoFillController(netlab);
		this.hc = new AutoFillController(netlab);
		this.hc.setActiveKey("type");
		this.hc.setFillKey("type");
		this.nc = new AutoFillController(netlab);
		this.fac = new AutoFillController(netlab, new Vector(FAILURE_AUTO_FILL));
		this.pc = new ProgramArrayController(netlab);
		this.pcc = new AutoFillController(netlab);
		this.pdc = new AutoFillController(netlab);
		
		this.dc = new IntersectArrayController(netlab, "desires");
		this.dc.setObjectClass(Pair.class);

		this.tc = new TrafficArrayController(netlab, this.trafficNameFormatter) {
			public Object newObject() {
				TopologyController tc = this.netlab.getTopologyController();
				List list = (List)CoSPropertiesController.this.selection;
				NodeTraffic nt;
				Object retval;

				nt = new NodeTraffic();
				nt.setName(tc.getNamer().suggest(CoSPropertiesController.TRAFFIC_NAME_BASE));
				nt.setType(TrafficArrayController.DEFAULT_TYPE);
				nt.getAttributes().putAll(TrafficArrayController.DEFAULT_ATTRIBUTES);
				nt.setMembers(list);
				nt.setNode((NSNode)list.get(0));
				retval = nt;

				this.netlab.getTopologyController().addAggregate(nt);

				return retval;
			}
		};
		this.tmc = new ArrayController(this.netlab);

		this.tkv.observeKeyValue(this.tc, "selection.members", ThinletKeyValue.CHANGE_ANY, this.tmc, "content");
		this.tkv.observeKeyValue(this.tc, "selection.node", ThinletKeyValue.CHANGE_ANY, this.tmc, "selection");
		this.tkv.observeKeyValue(this.tmc, "selection", ThinletKeyValue.CHANGE_ANY, this.tc, "selection.node");
		this.tkv.observeKeyValue(this.tc, "selection.attributes", new TrafficAttrObserver(this.tc, this.tkv));
		
				
		this.tkv.addKeyDependency(this, "trafficEmptyText", "selection");
		
		if (this.netlab.getPoolValue("applet") != null) {
			ProgressController dpc = netlab
					.showProgressDialog("Downloading Node Types...");
			new Downloader(netlab, dpc) {
				private final String OPSPID = "emulab-ops-";

				public void download() throws Throwable {
					List list = new ArrayList();
					EmulabNodeProxy node;
					Enumeration en;
					Hashtable ht;

					node = new EmulabNodeProxy(tkv, netlab.getPool());
					DBState.NODE_TYPE_INFO = ht = node.typeinfo();
					en = ht.elements();
					while (en.hasMoreElements()) {
						Hashtable sub = (Hashtable) en.nextElement();
						String osid;

						osid = tkv.getKeyValueAsString(sub, "osid");
						if (osid.startsWith(OPSPID)) {
							tkv.setKeyValue(sub, "osid", osid.substring(OPSPID
									.length()));
						}
						list.add(sub);
					}
					tkv.setKeyValue(hc, "content", list);
				}

				public void downloadCancelled() {
				}

				public void finished() {
					super.finished();

					netlab.requestFocus();
					netlab.requestFocus(netlab.field);
				}
			}.start();
		}
	}

	public void setTopologySelection(Object selection) {
		super.setTopologySelection(selection);

		NSTopology nst = this.netlab.getTopologyController().getNSTopology();
		List tccontents = new ArrayList();
		List pccontents = new ArrayList();
		List dccontents = null;
		Object files;

		files = this.tkv.getKeyValueAsSingleton(this, "selection.tarFiles");
		if (!(files instanceof SortedSet))
			files = null;
		this.tkv.setKeyValue(this, "tarEnabled", files != null);
		this.thinlet.putProperty(this.tarButton, "objcontroller", this);
		
		files = this.tkv.getKeyValueAsSingleton(this, "selection.RPMFiles");
		if (!(files instanceof SortedSet))
			files = null;
		this.tkv.setKeyValue(this, "rpmEnabled", files != null);
		this.thinlet.putProperty(this.rpmButton, "objcontroller", this);

		if (this.selection instanceof List) {
			List list = (List)this.selection;
			Iterator iter;
			
			for (iter = list.iterator(); iter.hasNext(); ) {
				NSNode nsn = (NSNode)iter.next();
				Collection progs;
				List aggs;
				
				aggs = nst.findAggregates(nsn, NodeTraffic.class);
				aggs.removeAll(tccontents);
				tccontents.addAll(aggs);
				
				progs = nsn.getPrograms();
				progs.removeAll(pccontents);
				pccontents.addAll(progs);
			}
			dccontents = (List)this.selection;
		}
		else if (this.selection instanceof NSObject) {
			tccontents.addAll(nst.findAggregates((NSObject)this.selection, NodeTraffic.class));
			pccontents.addAll(((NSNode)this.selection).getPrograms());
			dccontents = Collections.singletonList(this.selection);
		}
		this.tkv.setKeyValue(this.tc, "content", tccontents);
		this.tkv.setKeyValue(this.pc, "content", pccontents);
		this.tkv.setKeyValue(this.dc, "superContent", dccontents);
	}
	
	public Controller triggeredBy(Object component) {
		super.triggeredBy(component);

		return this;
	}

	/**
	 * @see net.emulab.netlab.client.Controller#setView(java.lang.Object)
	 */
	public Controller setView(Object view) {
		super.setView(view);

		try {
			this.progDialog = this.thinlet.parse("progdialog.xml", this);
			this.trafficDialog = this.thinlet.parse("nodetrafficdialog.xml", this);
		} catch (IOException e) {
			throw new InternalError();
		}
		return this;
	}

	public void openProgDialog(Object sender) {
		if (!this.progDialogOpened) {
			this.thinlet.add(this.progDialog);
			this.progDialogOpened = true;
		}
	}

	public void closeProgDialog() {
		this.thinlet.remove(this.progDialog);
		this.progDialogOpened = false;
	}

	public void openTrafficDialog(Object sender) {
		if (!this.trafficDialogOpened) {
			this.thinlet.add(this.trafficDialog);
			this.trafficDialogOpened = true;
		}
	}

	public void closeTrafficDialog() {
		if (this.trafficDialogOpened) {
			this.thinlet.remove(this.trafficDialog);
			this.trafficDialogOpened = false;
		}
	}
	
	public String getTrafficEmptyText() {
		String retval;
		
		if (this.isMultipleNSObjectSelection())
			retval = "";
		else
			retval = "Select multiple nodes to add traffic.";
		
		return retval;
	}

	public boolean verifyElement(ProgressController pc, Object component)
			throws Exception {
		boolean retval = true;

		return retval;
	}

	public String toString() {
		return "CoSPropertiesController[" + super.toString() + "]";
	}
}
