/*
 * CreateExperimentController.java
 *
 * Copyright (c) 2005 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

package net.emulab.netlab.client;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.List;

import net.emulab.netlab.client.emulab.EmulabExperimentProxy;
import net.emulab.netlab.client.emulab.EmulabUserProxy;
import net.emulab.netlab.client.emulab.EmulabXmlRpcException;
import net.emulab.ns.NSFile;
import thinlet.ArrayController;

public class CreateExperimentController extends ControllerBase {
	public static class ProjectGroupList {
		public String pid;
		public List groups;
		
		public ProjectGroupList(String pid, List groups) {
			this.pid = pid;
			this.groups = groups;
		}
	}
	
	private final NSFile nsf = new NSFile();

	public final ArrayController pc;
	public final ArrayController gc;

	public final ExperimentParameters ep = new ExperimentParameters();

	protected boolean synced = false;

	public boolean advanced = false;
	
	public Object projectSelection;
	public Object errorDetails;
	public Object errorDialog;
	
	public String nsfilestr = "";
	
	public String project = "";
	
	public String group = "";
	
	public String name = "";
	
	public String description = "";
	
	public boolean swapInImmediately = false;
	
	public boolean batch = false;
	
	public boolean idleSwap = true;
	
	public double idleSwapHours = 4.0;
	
	public String noIdleSwapReason = "";
	
	public boolean doMaxDuration = true;
	
	public double maxDuration = 16.0;

	public CreateExperimentController(NetlabClient netbuild) {
		super(netbuild);
		
		this.tkv.addKeyDependency(this, "canAccept", "name");
		this.tkv.addKeyDependency(this, "canAccept", "description");
		
		this.pc = new ArrayController(netbuild) {
			public boolean selectsNewContent() {
				return true;
			}
		};
		this.pc.setActiveKey("pid");

		this.gc = new ArrayController(netbuild) {
			public boolean selectsNewContent() {
				return true;
			}
		};
		this.gc.bindContentTo(this.pc, "selection.groups");
	}

	public Controller setView(Object view) {
		super.setView(view);
		
		try {
			this.errorDialog = this.netlab.parse("parseerror.xml", this);
		}
		catch (IOException e) {
			throw new InternalError();
		}
		
		return this;
	}

	public void editOk(Object component) throws Exception {
		this.netlab.remove(this.netlab.getParent(component));
		this.triggeredBy(null);
	}

	public Controller triggeredBy(Object component) throws Exception {
		super.triggeredBy(component);
		
		this.tkv.setKeyValue(this, "nsfilestr",
				this.nsf.toNSString(this.netlab.getTopologyController().getNSTopology()));
		
		return this;
	}

	public void showDialog(final Object sender) {
		if (!this.synced) {
			final ProgressController pc;

			pc = this.netlab.showProgressDialog("Downloading User Info...");
			new Downloader(this.netlab, pc) {
				public void download() throws Throwable {
					List list = new ArrayList();
					Object newsel = null;
					EmulabUserProxy eup;
					Enumeration en;
					Hashtable ht;
					
					eup = new EmulabUserProxy(tkv, super.netlab.getPool());
					ht = eup.membership(EmulabUserProxy.PERMISSION_CREATE_EXPERIMENT);
					en = ht.keys();
					while (en.hasMoreElements()) {
						String pid = (String)en.nextElement();
						List groups = (List)ht.get(pid);
						ProjectGroupList pgl;
						
						if (groups.contains(pid)) {
							groups.remove(pid);
							groups.add("(default)");
						}
						pgl = new ProjectGroupList(pid, groups);
						if (pid.equals(this.netlab.defaultPid))
							newsel = pgl;
						list.add(pgl);
					}
					tkv.setKeyValue(CreateExperimentController.this.pc, "content", list);
					
					if (newsel != null)
						tkv.setKeyValue(CreateExperimentController.this.pc, "selection", newsel);
					
					synced = true;
				}

				public void downloadCancelled() {
					super.netlab.remove(sender);
				}
			}.start();
		}
	}
	
	public boolean canAccept() {
		return (this.name.length() > 0 &&
				this.description.length() > 0);
	}
	
	public void retry() {
		if (this.netlab.getParent(this.errorDialog) != null)
			this.netlab.remove(this.errorDialog);
		
		this.ok();
	}

	public void ok() {
		this.ep.setProject(project);
		if (!"(default)".equals(group) && !"".equals(group))
			this.ep.setGroup(group);
		else
			this.ep.remove("group");
		this.ep.setName(this.name);
		this.ep.setDescription(this.description);
		this.ep.setNoSwapIn(true);
		this.ep.setBatch(batch);
		if (idleSwap) {
			this.ep.setIdleSwap((int) (this.idleSwapHours * 60.0));
			this.ep.remove("noidleswap_reason");
		} else {
			this.ep.setIdleSwap(0);
			this.ep.setNoIdleSwapReason(noIdleSwapReason);
		}
		if (doMaxDuration) {
			this.ep.setMaxDuration((int) (maxDuration * 60.0));
		} else {
			this.ep.remove("max_duration");
		}
		this.ep.setWait(true);
		this.ep.setNSFileString(this.nsfilestr);
		this.netlab.remove(this.view);
		
		final ProgressController pc;
		
		pc = this.netlab.showProgressDialog("Creating Experiment...", false);
		new Downloader(this.netlab, pc) {
			public void download() throws Throwable {
				EmulabExperimentProxy eep;
				ExperimentParameters ep;
				
				ep = CreateExperimentController.this.ep;
				
				super.pc.setStage(null, "Setting up local state...");
				super.pc.setContentLength(null, 1);
				
				eep = new EmulabExperimentProxy(tkv, super.netlab.getPool());
				eep.startexp(ep);
				if (swapInImmediately) {
					super.pc.setTitle("Swapping experiment in...");
					super.pc.setContentLength(null, 1);
					eep.swapexp(project, name, "in", false);
					super.netlab.getNoticeController().showDialog(
							"Your new experiment is swapping in and "
							+ "you should receive an e-mail when the "
							+ "process completes in a few minutes.");
				}
				super.pc.done();
				
				netlab.openLink("showexp.php3?pid=" + project
						+ "&eid=" + name, "_self");
			}
			
			public void downloadError(EmulabXmlRpcException e) {
				netlab.setString(errorDetails, "text", e.getMessage());
				netlab.add(errorDialog);
			}
			
			public void downloadCancelled() {
			}
		}.start();
	}

	public void cancel() {
		if (this.netlab.getParent(this.view) != null)
			this.netlab.remove(this.view);
		if (this.netlab.getParent(this.errorDialog) != null)
			this.netlab.remove(this.errorDialog);
	}

	public String toString() {
		return "CreateExperimentController[" + super.toString() + "]";
	}
}
