/*
 * DirectoryEntry.java
 *
 * Copyright (c) 2005 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

package net.emulab.netlab.client;

import java.io.File;
import java.util.Date;
import java.util.Vector;

/**
 * Class used to encapsulate individual directory entries.
 */
public class DirectoryEntry implements Comparable {
	/**
	 * The array offset of the 'name' value.
	 */
	public static final int OFFSET_NAME = 0;

	/**
	 * The array offset of the 'type' value.
	 */
	public static final int OFFSET_TYPE = 1;

	/**
	 * The array offset of the 'mode' value.
	 */
	public static final int OFFSET_MODE = 2;

	/**
	 * The array offset of the 'uid' value.
	 */
	public static final int OFFSET_UID = 3;

	/**
	 * The array offset of the 'gid' value.
	 */
	public static final int OFFSET_GID = 4;

	/**
	 * The array offset of the 'size' value.
	 */
	public static final int OFFSET_SIZE = 5;

	/**
	 * The array offset of the 'atime' value.
	 */
	public static final int OFFSET_ATIME = 6;

	/**
	 * The array offset of the 'mtime' value.
	 */
	public static final int OFFSET_MTIME = 7;

	/**
	 * The array offset of the 'ctime' value.
	 */
	public static final int OFFSET_CTIME = 8;

	/**
	 * The number of milliseconds in a second.
	 */
	public static final long MILLIS_PER_SECOND = 1000L;

	/**
	 * The entry name.
	 */
	public String name;

	/**
	 * The entry type: 'd' = directory, 'f' = regular file, 'u' = unknown.
	 */
	public char type;

	/**
	 * The file access mode.
	 */
	public int mode;

	/**
	 * The user owner of the entry.
	 */
	public String uid;

	/**
	 * The group owner of the entry.
	 */
	public String gid;

	/**
	 * The size of the entry, or -1 if it is a directory.
	 */
	public long size;

	/**
	 * The last access time, if available.
	 */
	public Date atime;

	/**
	 * The last modified time.
	 */
	public Date mtime;

	/**
	 * The creation time, if available.
	 */
	public Date ctime;

	/**
	 * Construct a DirectoryEntry from the given File object.
	 * 
	 * @param file
	 *            The File object to encapsulate.
	 */
	public DirectoryEntry(File file) {
		this.name = file.getName();
		if (file.isDirectory()) {
			this.type = 'd';
		} else if (file.isFile()) {
			this.type = 'f';
		} else {
			this.type = 'u';
		}
		if (this.type == 'd') {
			this.name += File.separator;
			this.size = -1;
		} else {
			this.size = file.length();
		}
		this.mtime = new Date(file.lastModified());
	}

	/**
	 * Construct a DirectoryEntry from the given vector.
	 * 
	 * @param vector
	 *            A directory entry encoded as a vector where the values are
	 *            ordered: name, type, mode, uid, gid, size, atime, mtime, and
	 *            ctime.
	 */
	public DirectoryEntry(Vector vector) {
		this.name = ((String) vector.get(OFFSET_NAME));
		this.type = ((String) vector.get(OFFSET_TYPE)).charAt(0);
		this.mode = ((Integer) vector.get(OFFSET_MODE)).intValue();
		this.uid = ((String) vector.get(OFFSET_UID));
		this.gid = ((String) vector.get(OFFSET_GID));
		this.size = ((Integer) vector.get(OFFSET_SIZE)).intValue();
		this.atime = new Date(((Integer) vector.get(OFFSET_ATIME)).longValue()
				* MILLIS_PER_SECOND);
		this.mtime = new Date(((Integer) vector.get(OFFSET_MTIME)).longValue()
				* MILLIS_PER_SECOND);
		this.ctime = new Date(((Integer) vector.get(OFFSET_CTIME)).longValue()
				* MILLIS_PER_SECOND);

		if (this.type == 'd') {
			this.name += "/";
			this.size = -1;
		}
	}

	/**
	 * Construct a DirectoryEntry with reasonable defaults.
	 */
	public DirectoryEntry() {
		this.name = "New File";
		this.type = 'f';
	}

	/**
	 * Comparison function for the entry's file type. We only compare the
	 * entry's type here because entries can be sorted on different attributes
	 * (e.g. name, size, date). However, there is generally a secondary sorting
	 * to separate files from directories, that is the intention of this
	 * function.
	 * 
	 * @param obj
	 *            The object to compare against this one.
	 * @return Zero if both objects are directories or files, negative one if
	 *         this object is a directory and the other is a file, or positive
	 *         one if the reverse is true.
	 */
	public int compareTo(Object obj) {
		DirectoryEntry de;
		int retval;

		if (!(obj instanceof DirectoryEntry))
			throw new IllegalArgumentException("obj is not a DirectoryEntry");

		de = (DirectoryEntry) obj;
		if (this.type == de.type)
			retval = 0;
		else if (this.type == 'd')
			retval = -1;
		else
			retval = 1;

		return retval;
	}

	/**
	 * @see java.lang.Object#toString()
	 */
	public String toString() {
		return "DirectoryEntry[name=" + this.name + "; type=" + this.type
				+ "; mode=" + Integer.toOctalString(this.mode) + "; uid="
				+ this.uid + "; gid=" + this.gid + "; size=" + this.size
				+ "; atime=" + this.atime + "; mtime=" + this.mtime
				+ "; ctime=" + this.ctime + "]";
	}
}
