/*
 * Downloader.java
 *
 * Copyright (c) 2005 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */
package net.emulab.netlab.client;

import net.emulab.netlab.client.emulab.EmulabXmlRpcException;

/**
 * Helper class used to manage an interruptible sequence of events, for example,
 * doing an XML-RPC.
 */
public abstract class Downloader extends Thread {
	/**
	 * The NetlabClient instance this controller is attached to.
	 */
	protected final NetlabClient netlab;

	protected final ProgressController pc;
	
	/**
	 * Construct a Downloader with the given values.
	 * 
	 * @param net.emulab.netlab.client
	 *            The NetlabClient instance this controller is attached to.
	 * @param pc
	 *            The ProgressController that is monitoring the sequence of
	 *            events.
	 */
	public Downloader(NetlabClient netlab, ProgressController pc) {
		this.netlab = netlab;
		this.pc = pc;
		
		this.setDaemon(true);
	}

	/**
	 * Callback used to start downloading/rpc'ing/etc while updating the
	 * ProgressController.
	 * 
	 * @throws Throwable
	 *             if anything goes awry.
	 */
	public abstract void download() throws Throwable;

	/**
	 * Handle an Emulab specific XML-RPC exception. Default behavior is to show
	 * the error dialog to the user with the exception message.
	 * 
	 * @param e
	 *            The exception thrown by the XML-RPC proxy.
	 */
	public void downloadError(EmulabXmlRpcException e) {
		this.netlab.getNoticeController().showDialog(e);
	}

	/**
	 * Callback used to signal that the download failed in some fashion and the
	 * state needs to be cleaned up.
	 */
	public abstract void downloadCancelled();

	/**
	 * The download has successfully/unsuccessfully finished. Default behavior
	 * is to close the ProgressController.
	 */
	public void finished() {
		this.pc.done();
	}

	/**
	 * @see java.lang.Runnable#run()
	 */
	public void run() {
		try {
			/* Always do the thing the user requested first and then */
			this.download();
		} catch (EmulabXmlRpcException e) {
			/* The RPC failed... */
			this.downloadError(e);
		} catch (Throwable th) {
			System.err.println("Downloader failed with:");
			th.printStackTrace(System.err);
			/* Check if the user clicked "stop" */
			if (!pc.wasHalted()) {
				pc.stop();
				this.netlab.getNoticeController().showDialog(th);
			}
			this.downloadCancelled();
		}
		this.finished();
	}

	/**
	 * @see java.lang.Object#toString()
	 */
	public String toString() {
		return "Downloader[" + super.toString() + "; net.emulab.netlab.client="
				+ this.netlab + "; pc=" + this.pc + "]";
	}
}
