/*
 * EmulabFileDialogController.java
 *
 * Copyright (c) 2005 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

package net.emulab.netlab.client;

import java.awt.Color;
import java.net.URL;
import java.util.Arrays;
import java.util.Vector;

import net.emulab.netlab.client.emulab.EmulabFSProxy;
import net.emulab.netlab.client.emulab.EmulabXmlRpcException;
import net.emulab.util.ProgressHalter;
import net.emulab.util.ProgressObserver;
import thinlet.NumberFormatter;
import thinlet.SizeFormatter;
import thinlet.ThinletKeyValueObserver;

/**
 * Controller for the Emulab file browser.
 */
public class EmulabFileDialogController extends FileDialogController {
	/**
	 * The dialog box title.
	 */
	public static final String DIALOG_TITLE = "Select a file on the Emulab NFS server";

	/**
	 * The face text for the "spare" button, which corresponds to the NFS export
	 * listing.
	 */
	public static final String SPARE_TEXT = "Shared Dirs";

	/**
	 * The tooltip text for the "spare" button, which corresponds to the NFS
	 * export listing.
	 */
	public static final String SPARE_TOOLTIP = "Show the list of NFS exports potentially available to nodes";

	/**
	 * The column header for the "spare" listing.
	 */
	public static final String SPARE_HEADER = "NFS Export";

	/**
	 * Indicates whether or not the controller is synced with the Emulab NFS
	 * export listing.
	 */
	protected boolean synced;

	/**
	 * Reference to the text field that should be filled in when the user
	 * accepts a file.
	 */
	protected Object model;

	/**
	 * The current directory name or null if the spare listing is being
	 * displayed.
	 */
	public String dir = "";

	/**
	 * The directory loader thread.
	 */
	private DirectoryLoader loader;

	/**
	 * The Thinlet component that focus should be returned to when directory
	 * loading has completed.
	 */
	private Object returnFocus;

	/**
	 * The directory loading implementation.
	 */
	private class DirectoryLoader extends Thread implements ProgressObserver {
		public final NumberFormatter SIZE_FORMATTER = new SizeFormatter("", "");
		
		/**
		 * The path for the directory currently being displayed.
		 */
		private final String oldDir;

		/**
		 * The path for the directory to be loaded and displayed.
		 */
		private final String dir;

		/**
		 * The amount of data to download from Emulab.
		 */
		private long contentLength;

		/**
		 * The current index in the ACTIVITY_SEQUENCE color array.
		 */
		private int colorIndex = 0;

		/**
		 * The object that can be used to halt an in-progress download.
		 */
		public ProgressHalter ph;

		/**
		 * Construct a DirectoryLoader with the given parameters.
		 * 
		 * @param oldDir
		 *            The path for the directory currently being displayed.
		 * @param dir
		 *            The path for the directory to be loaded and displayed.
		 */
		public DirectoryLoader(String oldDir, String dir) {
			this.oldDir = oldDir;
			this.dir = dir;
		}

		public void run() {
			Color newColor = ACTIVITY_PLAIN;
			String newDirPath = this.dir;

			try {
				String newIOError = null;
				EmulabFSProxy efp;
				Vector vector;
				Object oldpo;

				thinlet.setInteger(statusBar, "value", 0);
				thinlet.setString(barText, "text", "Downloading ...");
				oldpo = netlab.getPoolValue("server.observer");
				netlab.setPoolValue("server.observer", this);
				efp = new EmulabFSProxy(tkv, netlab.getPool());
				netlab.setPoolValue("server.observer", oldpo);
				try {
					thinlet.setString(
							EmulabFileDialogController.this.statusText, "text",
							"Downloading directory listing");
					vector = new Vector(Arrays.asList(efp.listdir(this.dir)));
					tkv.setKeyValue(EmulabFileDialogController.this, "ioError",
							null);
				} catch (EmulabXmlRpcException e) {
					vector = new Vector();
					newIOError = e.getMessage();
					newColor = ACTIVITY_ERROR;
				}
				tkv.setKeyValue(EmulabFileDialogController.this, "content",
						vector);
				tkv.setKeyValue(EmulabFileDialogController.this, "ioError",
						newIOError);
			} catch (Throwable th) {
				EmulabFileDialogController.this.dir = this.oldDir;
				newDirPath = this.oldDir;
			} finally {
				thinlet.setObject(dirComponent, "text", newDirPath);
				thinlet.setInteger(dirComponent, "start", 0);
				thinlet.setInteger(dirComponent, "end", newDirPath.length());
				thinlet.setColor(activity, "background", newColor);
				tkv.setKeyValue(EmulabFileDialogController.this, "loading",
						false);
				if (returnFocus != null)
					thinlet.requestFocus(returnFocus);
				returnFocus = null;
			}
		}

		public void setProgressHalter(URL url, ProgressHalter ph) {
			this.ph = ph;
		}

		public void setStage(URL url, String stageName) {
			thinlet.setString(statusText, "text", stageName);
		}

		public void setContentLength(URL url, long contentLength) {
			this.contentLength = contentLength;
			thinlet.setString(statusText, "text", "Downloading 0/"
					+ SIZE_FORMATTER.stringForObjectValue(null, null, new Long(this.contentLength)) + "Bytes");
			thinlet.setInteger(statusBar, "maximum", (int)contentLength);
		}

		public void updateContentRemaining(URL url, long contentRemaining) {
			String str;
			
			thinlet.setColor(activity, "background",
					ACTIVITY_SEQUENCE[this.colorIndex
							% ACTIVITY_SEQUENCE.length]);
			str = SIZE_FORMATTER.stringForObjectValue(null, null, new Long(this.contentLength - contentRemaining));
			thinlet.setString(statusText, "text", "Downloading "
					+ str + "/"
					+ SIZE_FORMATTER.stringForObjectValue(null, null, new Long(this.contentLength)) + "Bytes");
			thinlet.setInteger(statusBar, "value", (int)(this.contentLength - contentRemaining));
			thinlet.setString(barText, "text", str + "Bytes");
			this.colorIndex += 1;
		}
	}

	/**
	 * Construct an EmulabFileDialogController controller with the given
	 * argument.
	 * 
	 * @param netlab
	 *            The NetlabClient this controller is working for.
	 */
	public EmulabFileDialogController(final NetlabClient netlab) {
		super(netlab);

		this.tkv.observeKeyValue(this, "dir", new ThinletKeyValueObserver() {
			public void observedKeyValue(Object obj, String keyPath, int flags,
					Object oldValue, Object newValue) {
				boolean empty = "".equals(newValue);

				if (empty) {
					tkv.setKeyValue(obj, "showSpare", true);
				} else if (newValue == null) {
				} else {
					tkv.setKeyValue(obj, "showSpare", false);
					tkv.setKeyValue(obj, "loading", true);
					loader = new DirectoryLoader((String) oldValue,
							(String) newValue);
					loader.start();
				}
			}
		});
	}

	/**
	 * Extended to set the icon for the dialog to the Emulab logo.
	 * 
	 * @see net.emulab.netlab.client.Controller#setView(java.lang.Object)
	 */
	public Controller setView(Object view) {
		super.setView(view);

		this.thinlet.setIcon(view, "icon", this.netlab.getImageManager()
				.getObjectImage("favicon"));

		return this;
	}

	public Controller triggeredBy(Object component) throws Exception {
		String modelName;

		modelName = (String) this.netlab.getProperty(component, "model");
		this.model = this.netlab.find(component, modelName);

		return super.triggeredBy(component);
	}

	/**
	 * Method called just before the window is displayed on the screen. Just
	 * used to sync the controller's view of the NFS export listing with what
	 * Emulab says.
	 */
	public void showDialog() {
		if (!this.synced) {
			final ProgressController pc;

			pc = this.netlab.showProgressDialog("Downloading NFS exports...");
			new Downloader(this.netlab, pc) {
				public void download() throws Throwable {
					EmulabFSProxy efp;

					efp = new EmulabFSProxy(tkv, netlab.getPool());
					tkv.setKeyValue(sc, "content", efp.exports());
					synced = true;
				}

				public void downloadCancelled() {
				}
			}.start();
		}
	}

	/**
	 * Method called when the user double clicks on the file listing. If the
	 * selected object is a directory, the browser will descend into that
	 * directory. Otherwise, it will behave as if the user clicked the "Ok"
	 * button.
	 */
	public void acceptFile() {
		DirectoryEntry de = (DirectoryEntry) this.selection;
		String separator = "";
		String newPath;

		if (!this.dir.endsWith("/"))
			separator = "/";
		newPath = this.dir + separator + de.name;

		switch (de.type) {
		case 'd':
			this.netlab.setString(this.fileSearchBox, "text", "");
			this.tkv.setKeyValue(this, "filePath", "");
			this.commitDirectory(newPath, 0);
			this.returnFocus = this.fileList;
			break;
		default:
			this.ok();
			break;
		}
	}

	public void acceptFileField() {
		DirectoryEntry de = (DirectoryEntry) this.selection;
		String separator = "";
		String newPath;

		if (!this.dir.endsWith("/"))
			separator = "/";
		newPath = this.dir + separator + de.name;

		switch (de.type) {
		case 'd':
			this.netlab.setString(this.fileSearchBox, "text", "");
			this.tkv.setKeyValue(this, "filePath", "");
			this.commitDirectory(newPath, 0);
			this.returnFocus = this.fileField;
			break;
		default:
			this.ok();
			break;
		}
	}

	public void cancelLoading(Object sender) {
		super.cancelLoading(sender);

		if (this.loading) {
			this.loader.ph.haltProgress();
			this.loader = null;
			this.tkv.setKeyValue(this, "loading", false);
		}
	}

	/**
	 * Method called when the user double clicks on the spare listing. The only
	 * behavior is that the browser will switch to displaying the contents of
	 * the selected directory.
	 */
	public void acceptSpare() {
		this.netlab.setString(this.fileSearchBox, "text", "");
		this.tkv.setKeyValue(this, "dir", this.sc.selection());
		this.thinlet.requestFocus(this.fileList);
	}

	/**
	 * Method called when the user hits return or tabs out of the directory text
	 * field. If the directory in the text field is different from the displayed
	 * listing, it will be updated accordingly.
	 * 
	 * @param dir
	 *            The text in the directory field.
	 */
	public void commitDirectory(String dir, int reload) {
		if (!this.loading) {
			if (reload != 0) {
				this.tkv.setKeyValue(this, "dir", null); // Clear the old
			}
			if (!dir.equals(this.dir)) {
				this.tkv.setKeyValue(this, "dir", dir);
				super.commitDirectory(dir, reload);
			}
		}
	}

	/**
	 * Method called when the user clicks the "Ok" button. Currently, this will
	 * close the window and update the text field with the selected file.
	 */
	public void ok() {
		DirectoryEntry de = (DirectoryEntry) this.selection;
		String separator = "";
		String newPath;

		/* Join the directory and file paths, */
		if (!this.dir.endsWith("/"))
			separator = "/";
		newPath = this.dir + separator + de.name;

		/* ... close the window, */
		this.thinlet.remove(this.view);

		/* ... update the text, */
		this.thinlet.setString(this.model, "text", newPath);
		/* ... selected the text, and */
		this.thinlet.setInteger(this.model, "start", 0);
		this.thinlet.setInteger(this.model, "end", newPath.length());
		/* ... return focus to the text field. */
		this.thinlet.requestFocus(this.model);
	}

	/**
	 * Method called when the user clicks the "NFS Exports" button, causing the
	 * directory and file text fields to be cleared, and the listing to be
	 * switched from displaying files to the NFS exports.
	 */
	public void spare() {
		this.netlab.setString(this.fileSearchBox, "text", "");
		this.tkv.setKeyValue(this, "dir", "");
		this.tkv.setKeyValue(this, "filePath", "");
	}

	/**
	 * Method called when the user clicks the "Parent" button, causing the file
	 * listing to display the parent directory.
	 */
	public void parent() {
		if (!"".equals(this.dir) && !"/".equals(this.dir)) {
			int fromIndex, index;
			String newPath;

			this.netlab.setString(this.fileSearchBox, "text", "");
			
			/* Find the last '/' character, */
			fromIndex = this.dir.length() - 1;
			if (this.dir.endsWith("/"))
				fromIndex -= 1;
			index = this.dir.lastIndexOf('/', fromIndex);
			/* ... strip off the directory name, and */
			if (index <= 0)
				newPath = "/";
			else
				newPath = this.dir.substring(0, index);
			/* ... update the 'dir' value. */
			this.tkv.setKeyValue(this, "dir", newPath);
		}
	}

	/**
	 * @see java.lang.Object#toString()
	 */
	public String toString() {
		return "EmulabFileDialogController[" + super.toString() + "; synced="
				+ this.synced + "; model=" + this.model + "; dir=" + this.dir
				+ "; loader=" + this.loader + "; returnFocus="
				+ this.returnFocus + "]";
	}
}
