/*
 * FileDialogController.java
 *
 * Copyright (c) 2005 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

package net.emulab.netlab.client;

import java.awt.Color;
import java.util.Collections;
import java.util.List;

import thinlet.ArrayController;
import thinlet.AutoFillController;
import thinlet.DateFormatter;
import thinlet.SizeFormatter;
import thinlet.ThinletAPI;
import thinlet.ThinletContent;
import thinlet.ThinletController;
import thinlet.ThinletKeyValueObserver;

/**
 * Abstract controller for the file browser.
 */
public abstract class FileDialogController extends NetlabArrayController {
	/**
	 * The filter keys for this controller.
	 */
	private static final String FILTER_KEYS[] = { "name", "uid", "gid" };

	/**
	 * Color sequence that simulates a "throbbing" blue light. The subclass is
	 * expected to update the activity light with a new color whenever it makes
	 * progress doing some task.
	 */
	protected static final Color ACTIVITY_SEQUENCE[] = new Color[17];

	public static final ThinletContent ENTRY_CONVERTER = new ThinletContent() {
		private void setRowColor(ThinletAPI thinlet, Object row, Color color) {
			Object cells[] = thinlet.getItems(row);
			Color oldColor;
			int lpc;

			oldColor = thinlet.getColor(cells[0], "foreground");
			if ((oldColor == null) || !oldColor.equals(color)) {
				for (lpc = 0; lpc < cells.length; lpc++) {
					thinlet.setColor(cells[lpc], "foreground", color);
				}
			}
		}

		public void initItem(ThinletAPI thinlet, Object item, Object content) {
			DirectoryEntry de = (DirectoryEntry) content;
			Color color;

			switch (de.type) {
			case 'd':
				color = Color.blue;
				break;
			case 'u':
				color = Color.red;
				break;
			default:
				color = Color.black;
				break;
			}
			this.setRowColor(thinlet, item, color);
		}

		public void replaceItem(ThinletAPI thinlet, Object item, Object content) {
			this.initItem(thinlet, item, content);
		}
	};

	static {
		int lpc = 0;

		/* Create a gradient that transitions from a blue to a cyan and then */
		ACTIVITY_SEQUENCE[lpc++] = new Color(0x3080ff);
		ACTIVITY_SEQUENCE[lpc++] = new Color(0x3088ff);
		ACTIVITY_SEQUENCE[lpc++] = new Color(0x3090ff);
		ACTIVITY_SEQUENCE[lpc++] = new Color(0x3098ff);
		ACTIVITY_SEQUENCE[lpc++] = new Color(0x30a0ff);
		ACTIVITY_SEQUENCE[lpc++] = new Color(0x30a8ff);
		ACTIVITY_SEQUENCE[lpc++] = new Color(0x30b0ff);
		ACTIVITY_SEQUENCE[lpc++] = new Color(0x30b8ff);
		ACTIVITY_SEQUENCE[lpc++] = new Color(0x30c0ff);
		/* ... back again. */
		for (; lpc < ACTIVITY_SEQUENCE.length; lpc++) {
			ACTIVITY_SEQUENCE[lpc] = ACTIVITY_SEQUENCE[ACTIVITY_SEQUENCE.length
					- lpc];
		}
	}

	/**
	 * Plain color used when the dialog is inactive.
	 */
	protected static final Color ACTIVITY_PLAIN = new Color(0xe6e6e6);

	/**
	 * Red color used to indicate an error with the user input.
	 */
	protected static final Color ACTIVITY_ERROR = new Color(0xff4444);

	/**
	 * Formatter for the last modified time.
	 */
	public static final DateFormatter MTIME_FORMATTER = new DateFormatter();

	/**
	 * Formatter for the file size.
	 */
	public static final SizeFormatter SIZE_FORMATTER = new SizeFormatter();

	/**
	 * Thinlet outlet for the status label.
	 */
	public Object statusText;

	public Object statusBar;

	public Object barText;
	
	/**
	 * Thinlet outlet for the search box.
	 */
	public Object fileSearchBox;

	/**
	 * Thinlet outlet for the file listing.
	 */
	public Object fileList;

	/**
	 * Thinlet outlet for the spare listing.
	 */
	public Object spareList;

	/**
	 * Thinlet outlet for the directory text field.
	 */
	public Object dirComponent;

	public Object fileField;
	
	/**
	 * Thinlet outlet for the "Ok" button.
	 */
	public Object okButton;

	/**
	 * Thinlet outlet for the activity light panel.
	 */
	public Object activity;
	
	public Object activityButton;

	/**
	 * Indicates whether or not the dialog is trying to load a new directory.
	 */
	public boolean loading;

	/**
	 * ArrayController for the spare listing.
	 */
	public final ArrayController sc;

	/**
	 * AutoFillController for the search box.
	 */
	public final AutoFillController afc;

	/**
	 * The text in the file text field.
	 */
	protected String filePath;

	/**
	 * Indicates whether or not the spare listing is being shown.
	 */
	protected boolean showSpare = true;

	/**
	 * String that describes an I/O error.
	 */
	public String ioError;

	/**
	 * Construct a FileDialogController with the given value.
	 * 
	 * @param netlab
	 *            The NetlabClient object this controller is working for.
	 */
	protected FileDialogController(NetlabClient netlab) {
		super(netlab);

		/* Setup the filter keys for this controller and */
		this.addFilterKeys(FILTER_KEYS);
		this.setFillKey("name");
		/*
		 * ... restrict the number of files to display. (XXX Otherwise the VM
		 * will crash!)
		 */
		this.setMaximumArraySize(512);

		/* Make the spare controller and */
		this.sc = new ArrayController(this.thinlet);
		this.sc.addFilterKey("");

		/* ... the search controller. */
		this.afc = new AutoFillController(this.thinlet);

		/*
		 * Add an observer for the file listing selection that updates the text
		 * field containing the selected file name.
		 */
		this.tkv.observeKeyValue(this, "selection",
				new ThinletKeyValueObserver() {
					public void observedKeyValue(Object obj, String keyPath,
							int flags, Object oldValue, Object newValue) {
						if (!showSpare) {
							Object fpValue = null;

							if (newValue != ThinletController.NO_SELECTION) {
								fpValue = tkv.getKeyValue(newValue, "name");
								tkv.setKeyValue(obj, "filePath", fpValue);
							}
						}
					}
				});

		/* Add key/value dependencies for some methods. */
		this.tkv.addKeyDependency(this, "canAccept", "filePath");

		this.tkv.addKeyDependency(this, "canSelectFiles", "showSpare");
		this.tkv.addKeyDependency(this, "canViewFiles", "showSpare");
		this.tkv.addKeyDependency(this, "canViewSpare", "showSpare");
		this.tkv.addKeyDependency(this, "isDirectoryValid", "showSpare");
		this.tkv.addKeyDependency(this, "isDirectoryInvalid", "showSpare");

		this.tkv.addKeyDependency(this, "canDisplayFiles", "isWithinMaximum");
		this.tkv.addKeyDependency(this, "fileListEmpty", "isWithinMaximum");

		this.tkv.addKeyDependency(this, "canDisplayFiles", "ioError");
		this.tkv.addKeyDependency(this, "isDirectoryValid", "ioError");
		this.tkv.addKeyDependency(this, "isDirectoryInvalid", "ioError");

		this.tkv.addKeyDependency(this, "canViewFiles", "loading");
		this.tkv.addKeyDependency(this, "canViewSpare", "loading");
		this.tkv.addKeyDependency(this, "isDirectoryValid", "loading");
		this.tkv.addKeyDependency(this, "isDirectoryInvalid", "loading");
		this.tkv.addKeyDependency(this, "canDisplayFiles", "loading");
		this.tkv.addKeyDependency(this, "canSelectFiles", "loading");
	}

	/**
	 * @param filePath
	 *            The text in the file text field.
	 */
	public void setFilePath(String filePath) {
		this.filePath = filePath;
	}

	/**
	 * @return The text in the file text field.
	 */
	public String getFilePath() {
		return this.filePath;
	}

	/**
	 * @param showSpare
	 *            Indicates whether or not the spare listing is being shown.
	 */
	public void setShowSpare(boolean showSpare) {
		this.showSpare = showSpare;
	}

	/**
	 * @return Indicates whether or not the spare listing is being shown.
	 */
	public boolean getShowSpare() {
		return this.showSpare;
	}
	
	public String getFileListEmpty() {
		String retval;
		
		if (this.isWithinMaximum)
			retval = "No files";
		else
			retval = "Too many files to display, use the search box to reduce the number.";
		
		return retval;
	}

	public Controller modelChanged(Object component) {
		return this;
	}

	/**
	 * Extended to sort the elements so that directories are listed above
	 * regular files.
	 * 
	 * @see thinlet.ArrayController#arrangeObjects(java.util.List)
	 */
	protected List arrangeObjects(List list) {
		List retval = super.arrangeObjects(list);

		Collections.sort(retval);

		return retval;
	}

	/**
	 * Method called just before the dialog box is shown on the screen. Does
	 * nothing at the moment.
	 */
	public void showDialog() {
	}

	/**
	 * Method called when the user clicks on the "Ok" button to test whether or
	 * not the action will cause an existing file to be overwritten.
	 * 
	 * @return false
	 */
	public boolean isOverwriting() {
		return false;
	}

	/**
	 * Method called when the components contained in the dialog will be
	 * focusable. Currently, it will set the focus to the file/spare listing.
	 */
	public void focusable() {
		if (!this.thinlet.requestFocus(this.fileList))
			this.thinlet.requestFocus(this.spareList);
	}

	/**
	 * Method called when the user hits return or tabs out of the directory text
	 * field. This implementation will clear the search box since its contents
	 * presumably applied to the previous directory and not the current.
	 * 
	 * @param text
	 *            The new text in the directory text field.
	 * @param reload
	 *            Non-zero if the directory is to be reloaded.
	 */
	public void commitDirectory(String text, int reload) {
		if (reload == 0) {
			this.thinlet.setString(this.fileSearchBox, "text", null);
		}
	}

	/**
	 * Method called when the spare button is clicked. Currently, it will just
	 * set showSpare to true.
	 */
	public void spare() {
		this.tkv.setKeyValue(this, "showSpare", true);
		this.thinlet.requestFocus(this.spareList);
	}

	/**
	 * Method called to cancel any in-progress directory loading operations.
	 * Currently, this method does nothing.
	 * 
	 * @param sender
	 *            The object that called this method, can be a Thinlet component
	 *            or this controller.
	 */
	public void cancelLoading(Object sender) {
	}

	/**
	 * Method called when the cancel button is clicked, removes the dialog box
	 * and cancels and in-progress directory loading operations.
	 */
	public void cancel() {
		if (this.loading) {
			this.cancelLoading(this);
		}
		this.thinlet.remove(this.view);
	}

	/**
	 * @return True if there is a file selected.
	 */
	public boolean canAccept() {
		return (this.filePath != null) && !"".equals(this.filePath);
	}

	/**
	 * @return True if the file listing is visible and the controller is not in
	 *         the process of loading the directory.
	 */
	public boolean canSelectFiles() {
		return !this.showSpare && !this.loading;
	}

	/**
	 * @return True if the file listing is visible, the controller is not in the
	 *         process of loading the directory, and there were no errors in
	 *         loading the directory.
	 */
	public boolean isDirectoryValid() {
		return !this.showSpare && !this.loading && (this.ioError == null);
	}

	/**
	 * @return True if the file listing is visible, the controller is not in the
	 *         process of loading the directory, and there was an error in
	 *         loading the directory.
	 */
	public boolean isDirectoryInvalid() {
		return !this.showSpare && !this.loading && (this.ioError != null);
	}

	/**
	 * @return True if the spare listing is being displayed.
	 */
	public boolean canViewSpare() {
		return this.showSpare && !this.loading;
	}
	
	public boolean canViewFiles() {
		return !this.showSpare && !this.loading;
	}

	/**
	 * @return True if the number of files is within the maximum displayable
	 *         allowed, the controller is not in the process of loading the
	 *         directory, and there were no errors while loading.
	 */
	public boolean canDisplayFiles() {
		return (super.isWithinMaximum() && !this.loading && (this.ioError == null));
	}

	/**
	 * @see java.lang.Object#toString()
	 */
	public String toString() {
		return "FileDialogController[" + super.toString() + "; statusText="
				+ this.statusText + "; fileSearchBox=" + this.fileSearchBox
				+ "; fileList=" + this.fileList + "; spareList="
				+ this.spareList + "; dirComponent=" + this.dirComponent
				+ "; okButton=" + this.okButton + "; activity=" + this.activity
				+ "; loading=" + this.loading + "; sc=" + this.sc + "; afc="
				+ this.afc + "; filePath=" + this.filePath + "; showSpare="
				+ this.showSpare + "; ioError=" + this.ioError + "]";
	}
}
