/*
 * LinkController.java
 *
 * Copyright (c) 2005 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */
package net.emulab.netlab.client;

import java.awt.Font;
import java.awt.Image;
import java.util.Iterator;
import java.util.List;

import net.emulab.ns.LinkTraffic;
import net.emulab.ns.NSAggregate;
import net.emulab.ns.NSInterface;
import net.emulab.ns.NSLink;
import net.emulab.ns.NSObject;
import net.emulab.ns.NSTopology;
import thinlet.ThinletKeyValueObserver;

/**
 * Controller for individual NSLink elements.
 */
public class LinkController extends LanLinkController {
	public static final String DEFAULT_LINK_NAME = "link";

	/**
	 * The NSLink this controller is managing.
	 */
	public NSLink nsl;

	/**
	 * Construct a LinkController with the given argument.
	 * 
	 * @param netlab
	 *            The NetlabClient instance this controller is attached to.
	 */
	public LinkController(NetlabClient netlab) {
		super(netlab);

		this.tkv.addKeyDependency(this, "nsl", "nsobject");
		this.tkv.addKeyDependency(this, "icon", "netlab");
		this.tkv.addKeyDependency(this, "tooltip", "nsl");
	}

	/**
	 * @return The appropriate link icon that corresponds to its current state.
	 */
	public Image getIcon() {
		Image retval;

		retval = this.netlab.getImageManager().getObjectImage("link-none");
		return retval;
	}

	public void delete(Object sender) {
		TopologyController tc = this.netlab.getTopologyController();
		NSTopology nst;
		NSLink nslink;
		Iterator iter;
		List list;
		
		nslink = (NSLink) this.netlab.getProperty(sender, "model");
		tc.getNamer().remove(nslink.getName());
		this.netlab.remove(sender);
		tc.removeNSObject(nslink);
		nst = tc.getNSTopology();
		list = nst.findAggregates(nslink, LinkTraffic.class);
		for (iter = list.iterator(); iter.hasNext(); ) {
			NSAggregate nsa = (NSAggregate)iter.next();
			
			nst.removeAggregate(nsa);
			if (nsa instanceof NSObject)
				nst.getNamer().remove(((NSObject)nsa).getName());
		}
	}

	public Object copyElement(NSObject component) {
		return null;
	}

	public boolean verifyElement(ProgressController pc, Object component) {
		boolean retval = true;

		return retval;
	}

	public void menuProperties(Object sender) throws Exception {
		this.showView(this.netlab.getParent(sender));
	}

	public void setElement(Object component) {
		if (this.nsl != null)
			throw new IllegalStateException();

		super.setElement(component);

		this.tkv.setKeyValue(this, "nsl", this.nsobject);
		this.tkv.observeKeyValue(this.nsl, "", new ThinletKeyValueObserver() {
			public void observedKeyValue(Object obj, String keyPath, int flags,
					Object oldValue, Object newValue) {
				LinkController.this.tkv.notifyKeyValue(LinkController.this,
						"nsl." + keyPath, flags, oldValue, newValue);
			}
		});
		this.tkv.observeKeyValue(this.nsl.getLeftMember(), "",
				new ThinletKeyValueObserver() {
					public void observedKeyValue(Object obj, String keyPath,
							int flags, Object oldValue, Object newValue) {
						LinkController.this.tkv.notifyKeyValue(
								LinkController.this.nsl, "leftMember."
										+ keyPath, flags, oldValue, newValue);
					}
				});
	}

	/**
	 * @return The tooltip string for this link. This value changes based on the
	 *         current settings of the link: If the link has the default
	 *         configuration, " (default config) " is displayed; Otherwise, any
	 *         deviations from the default are displayed.
	 */
	public String tooltip() {
		String retval = "";

		if (this.nsl != null) {
			String fullName, shortName, lIP, rIP;
			NSInterface left, right;
			Font newFont;

			fullName = this.nsl.getName().toString();
			shortName = this.nsl.getName().toShortString(
					this.nc.getShortNameLength());

			left = this.nsl.getLeftMember();
			right = this.nsl.getRightMember();
			lIP = (String) this.tkv.getKeyValue(this.nsl,
					"leftMember.IPv4Address");
			rIP = (String) this.tkv.getKeyValue(this.nsl,
					"rightMember.IPv4Address");

			if (fullName.length() > shortName.length())
				retval += "  " + fullName;
			if (!"".equals(lIP))
				retval += "  " + left.getNSObject().getName() + ":" + lIP;
			if (!"".equals(rIP))
				retval += "  " + right.getNSObject().getName() + ":" + rIP;
			if (left.getBandwidth() > 0) {
				retval += ("  BW: " + InterfacePropertiesController.BW_FORMATTER.
						stringForObjectValue(this.netlab, null, new Long(left.getBandwidth())));
			}
			if (left.getLatency() > 0.0) {
				retval += "  Lat: " + left.getLatency() + "ms";
			}
			if (left.getLossRate() > 0.0) {
				retval += "  PLR: " + (left.getLossRate() * 100.0) + "%";
			}

			if (retval.length() == 0) {
				retval = "  (default config)  ";
				newFont = this.nc.getTooltipFont();
			} else {
				retval += "  ";
				newFont = this.nc.getPlainFont();
			}
			this.tkv.setKeyValue(this, "tooltipFont", newFont);
		}
		return retval;
	}

	public String toString() {
		return "LinkController[" + super.toString() + "; nsl=" + this.nsl
				+ "; tooltipFont=" + this.tooltipFont + "]";
	}
}
