/*
 * LocalFileDialogController.java
 *
 * Copyright (c) 2005 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

package net.emulab.netlab.client;

import java.awt.Cursor;
import java.io.File;
import java.io.IOException;
import java.util.List;
import java.util.Vector;

import thinlet.ThinletKeyValueObserver;

/**
 * Controller for the local file system browser.
 */
public class LocalFileDialogController
    extends FileDialogController
{
    /**
     * The dialog box title.
     */
    public static final String DIALOG_TITLE = "Select a file to open";

    /**
     * The face text for the "spare" button, which corresponds to the list of
     * recently used directories.
     */
    public static final String SPARE_TEXT = "Recent";

    /**
     * The tooltip text for the "spare" button, which corresponds to the list
     * of recently used directories.
     */
    public static final String SPARE_TOOLTIP =
	"Show the list of recently used directories";

    /**
     * The column header for the "spare" listing.
     */
    public static final String SPARE_HEADER = "Path";

    /**
     * The number of "recent" directories to keep track of.
     */
    private static final int RECENT_DIR_COUNT = 10;

    /**
     * The default list of "recent" directories, basically, the user's home and
     * current directory.
     */
    private static final String DEFAULT_RECENT_DIRS[] = {
	System.getProperty("user.home"),
	System.getProperty("user.dir"),
    };

    /**
     * The flow object that triggered this controller.
     */
    private Object flow;

    /**
     * The state to put the flow object in when the "Ok" button is clicked.
     */
    private String okState;

    /**
     * The "direction" of the file operation: "save" = saving a file, "open" =
     * opening a file.
     */
    private String direction;

    /**
     * The File object representing the currently displayed directory listing
     * or null if the spare listing is being displayed.
     */
    private File dir;

    /**
     * Construct a LocalFileDialogController with the given value.
     *
     * @param netlab The NetlabClient this controller is working with.
     */
    public LocalFileDialogController(NetlabClient netlab)
    {
	super(netlab);

	NetlabPreferences np;
	Vector sparePaths;
	int lpc;
	
	np = this.netlab.getPreferences();
	sparePaths = np.getVector(NetlabPreferences.RECENT_DIR_PREFIX,
				  RECENT_DIR_COUNT);
	if( sparePaths.size() == 0 )
	{
	    for( lpc = 0; lpc < DEFAULT_RECENT_DIRS.length; lpc++ )
	    {
		sparePaths.add(DEFAULT_RECENT_DIRS[lpc]);
	    }
	}
	this.tkv.setKeyValue(this.sc, "content", sparePaths);

	this.tkv.observeKeyValue(
		this, "dir", new ThinletKeyValueObserver() {
			public void observedKeyValue(Object obj,
						     String keyPath,
						     int flags,
						     Object oldValue,
						     Object newValue)
			{
			    Vector newContent = new Vector();
			    String newDirPath = "";

			    tkv.setKeyValue(obj, "showSpare", Boolean.FALSE);
			    if( newValue != null )
			    {
				File dir, files[];
				int lpc;

				dir = (File)newValue;
				thinlet.setCursor(Cursor.getPredefinedCursor(
					Cursor.WAIT_CURSOR));
				try
				{
				    newDirPath = dir.getCanonicalPath();
				}
				catch(IOException e)
				{
				    // nothing
				}
				if( (files = dir.listFiles()) != null )
				{
				    for( lpc = 0; lpc < files.length; lpc++ )
				    {
					DirectoryEntry de;

					de = new DirectoryEntry(files[lpc]);
					newContent.add(de);
				    }
				}
				thinlet.setCursor(Cursor.getPredefinedCursor(
					Cursor.DEFAULT_CURSOR));
			    }
			    thinlet.setObject(dirComponent,
					      "text",
					      newDirPath);
			    tkv.setKeyValue(obj, "content", newContent);
			}
		    });
    }

    /**
     * @param dir The File object representing the currently displayed
     * directory listing or null if the spare listing is being displayed.
     */
    public void setDir(File dir)
    {
	this.dir = dir;
    }

    /**
     * @return The File object representing the currently displayed directory
     * listing or null if the spare listing is being displayed.
     */
    public File getDir()
    {
	return this.dir;
    }

    public Controller triggeredBy(Object component)
	throws Exception
    {
	this.flow = this.thinlet.findFlow(component);
	this.okState = (String)this.thinlet.getProperty(component, "state");
	this.direction = (String)
	    this.thinlet.getProperty(component, "direction");
	
	super.triggeredBy(component);
	
	return this;
    }

    /**
     * @return True if this is a save dialog and the selected file exists.
     */
    public boolean isOverwriting()
    {
	boolean retval = false;

	if( "save".equals(this.direction) )
	{
	    File newPath = new File(this.dir, this.filePath);

	    retval = newPath.exists();
	}
	return retval;
    }

    /**
     * Method called when the user hits return or tabs out of the directory
     * text field.  If the directory in the text field is different from the
     * displayed listing, it will be updated accordingly.
     *
     * @param dir The text in the directory field.
     */
    public void commitDirectory(String dirPath)
    {
	File newDir = new File(dirPath);

	if( !newDir.equals(this.dir) )
	{
	    this.tkv.setKeyValue(this, "dir", newDir);
	    super.commitDirectory(dirPath,0);
	}
    }

    /**
     * Method called when the user double clicks on the file listing.  If the
     * selected object is a directory, the browser will descend into that
     * directory.  Otherwise, it will behave as if the user clicked the "Ok"
     * button.
     */
    public void acceptFile()
    {
	DirectoryEntry de = (DirectoryEntry)this.selection;
	File newPath = new File(this.dir, de.name);

	switch( de.type )
	{
	case 'd':
	    this.tkv.setKeyValue(this, "filePath", "");
	    this.tkv.setKeyValue(this, "dir", newPath);
	    this.thinlet.setString(this.fileSearchBox, "text", null);
	    this.thinlet.requestFocus(this.fileList);
	    break;
	default:
	    this.thinlet.invoke(this.okButton, null, "action");
	    break;
	}
    }

    /**
     * Method called when the user double clicks on the spare listing.  The
     * only behavior is that the browser will switch to displaying the contents
     * of the selected directory.
     */
    public void acceptSpare()
    {
	this.tkv.setKeyValue(this,
			     "dir",
			     new File((String)this.sc.selection()));
	this.thinlet.requestFocus(this.fileList);
    }
    
    /**
     * Method called when the user clicks the "Ok" button.  Currently, this
     * will close the window, update the "file" property and "state" of the
     * flow that triggered this controller, and update the list of recently
     * used directories.
     */
    public void ok()
    {
	File newPath = new File(this.dir, this.filePath);
	
	this.thinlet.remove(this.view);

	try
	{
	    String recentDir;
	    List sparePaths;

	    /* Get the path as a string and */
	    recentDir = newPath.getParentFile().getCanonicalPath();
	    sparePaths = this.sc.contentArray();
	    /* ... check if its already been added. */
	    if( !sparePaths.contains(recentDir) )
	    {
		NetlabPreferences np = this.netlab.getPreferences();

		if( sparePaths.size() == RECENT_DIR_COUNT )
		{
		    /*
		     * Remove elements _after_ the default set, since those are
		     * always handy to have around.
		     */
		    this.sc.removeObject(sparePaths.
					 get(DEFAULT_RECENT_DIRS.length));
		}
		/* Add our new dir, */
		this.sc.addObject(recentDir);
		/* ... update the preferences, and */
		np.setVector(NetlabPreferences.RECENT_DIR_PREFIX, sparePaths);
		/* ... commit to disk. */
		np.store();
	    }
	}
	catch(IOException e)
	{
	    // ignore...
	    e.printStackTrace(System.err);
	}

	this.thinlet.putProperty(this.flow, "file", newPath);
	this.thinlet.setString(this.flow, "state", this.okState);
    }

    /**
     * Method called when the user clicks the "Recent" button, causing the
     * directory text field to be cleared and the listing to be switched from
     * displaying files to the listing of recent directories.
     */
    public void spare()
    {
	this.tkv.setKeyValue(this, "dir", null);
	super.spare();
    }

    /**
     * Method called when the user clicks the "Parent" button, causing the file
     * listing to display the parent directory.
     */
    public void parent()
    {
	this.tkv.setKeyValue(this, "dir", this.dir.getParentFile());
    }

    public String toString()
    {
	return "LocalFileDialogController["
	    + super.toString()
	    + "; flow="
	    + this.flow
	    + "; okState="
	    + this.okState
	    + "; direction="
	    + this.direction
	    + "; dir="
	    + this.dir
	    + "]";
    }
}
