/*
 * NetlabClient.java
 *
 * Copyright (c) 2005 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

package net.emulab.netlab.client;

import java.applet.Applet;
import java.awt.Color;
import java.awt.Cursor;
import java.awt.Dimension;
import java.awt.Rectangle;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.PrintStream;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.Arrays;
import java.util.List;
import java.util.Vector;

import net.emulab.ns.NSFile;
import net.emulab.ns.NSFile1;
import net.emulab.ns.NSInterface;
import net.emulab.ns.NSLan;
import net.emulab.ns.NSLanLink;
import net.emulab.ns.NSLink;
import net.emulab.ns.NSNode;
import net.emulab.ns.NSObject;
import net.emulab.util.EnvironmentVariables;
import net.emulab.util.ServerProperties;
import net.emulab.xmlrpc.AutoXmlRpcClient;
import thinlet.AutoFillController;
import thinlet.FrameLauncher;
import thinlet.ObjectFactory;
import thinlet.ObjectPool;
import thinlet.Thinlet;
import thinlet.ThinletDragInfo;
import thinlet.ThinletDragSource;
import thinlet.ThinletDragSourceBase;
import thinlet.ThinletDropDestination;
import thinlet.ThinletDropDestinationBase;
import thinlet.ThinletKeyValue;

public class NetlabClient extends Thinlet {

	public static final int PROXIMITY = 27;
	public static final Color LAN_COLOR = new Color(0x57, 0x8A, 0xBC);
	private final NetlabPreferences nbp;

	public static final ThinletDropDestination DROP_DST = new ThinletDropDestinationBase() {
		public boolean draggingEntered(Thinlet thinlet, ThinletDragInfo tdi) {
			// thinlet.setBoolean(tdi.getDropZone(), "border", true);
			return true;
		}

		public void draggingExited(Thinlet thinlet, ThinletDragInfo tdi) {
			// thinlet.setBoolean(tdi.getDropZone(), "border", false);
		}

		public boolean prepareForDrop(Thinlet thinlet, ThinletDragInfo tdi) {
			return true;
		}

		public void performDrop(Thinlet thinlet, ThinletDragInfo tdi) {
			NetlabClient nc = (NetlabClient)thinlet;

			// thinlet.setBoolean(tdi.getDropZone(), "border", false);
			thinlet.add(tdi.getDropZone(), tdi.getSource());
			try {
				nc.deselect(null);
				nc.showView(tdi.getSource());
			}
			catch (Exception e) {
				e.printStackTrace();
			}
		}

		public void finishDrop(Thinlet thinlet, ThinletDragInfo tdi) {
		}
	};
	


	public static final ThinletDragSource PALETTE_DRAG_SOURCE = new ThinletDragSourceBase() {
		public Object dragStart(Thinlet thinlet, Object sender) {
			Object retval;

			retval = thinlet.create("togglebutton");

			thinlet.setString(retval, "text", thinlet.getString(sender, "text"));
			thinlet.setColor(retval, "background", thinlet.getColor(sender, "background"));
			thinlet.setColor(retval, "foreground", thinlet.getColor(sender, "foreground"));
			thinlet.setChoice(retval, "style", thinlet.getChoice(sender, "style"));
			thinlet.setBoolean(retval, "usegradient", true);
			thinlet.setIcon(retval, "icon", thinlet.getIcon(sender, "icon"));
			thinlet.putProperty(retval, "controller",
					thinlet.getProperty(sender, "controller"));
			
			return retval;
		}
	};

	private Dimension minSize = new Dimension(300, 250);

	private final ObjectPool pool = new ObjectPool();
	private final TopologyController tc = new TopologyController(this);
	private final NetlabConfiguration nc = new NetlabConfiguration();
	private final ControllerManager cm = new ControllerManager(this);
	private final ImageManager im = new ImageManager(this);
	private final NoticeController noticeController;
	private final ProgressController progressController;
	
	public String version = "NetlabClient v0.2.0"; // XXX
	public Object deck;
	public Object field;
	public Object showFileDialog;
	public Object showFileText;
	public Object openAction;
	
	public String defaultPid;
	public String defaultEid;

	private final Object propertiesDialog;
	private Object currentPropertiesPanel;
	private int propertiesIndex = -1;
	private final Object propertyPanels[];
	private PropertiesController propertyControllers[];

	private ClipInfo ci[] = {};
	
	public final AutoFillController sc;
	
	public NetlabClient(Applet applet, String args[]) throws Exception {
		this.tkv.setKeyValue(this.pool, "applet", applet);
		this.tkv.setKeyValue(this.pool, "environ", new EnvironmentVariables());
		this.tkv.setKeyValue(this.pool, "server", new ServerProperties(this.tkv, this.pool));
		this.tkv.setKeyValue(this.pool, "xmlrpcClientFactory",
				new ObjectFactory(this.tkv, this.pool, AutoXmlRpcClient.class));

		this.sc = new AutoFillController(this);
		
		this.progressController = new ProgressController(this);
		this.setPoolValue("server.observer", this.progressController);
		this.nbp = new NetlabPreferences();
		
		this.add(this.parse("netlab.xml"));
		// this.add(this.deck, this.parse("palette.xml"), 0);
		
		this.showFileDialog = this.parse("showfile.xml");

		this.noticeController = new NoticeController(this);
		
		this.propertyControllers = new PropertiesController[] {
				(PropertiesController)this.cm.getController(ExpPropertiesController.class.getName()),
				(PropertiesController)this.cm.getController(NodePropertiesController.class.getName()),
				(PropertiesController)this.cm.getController(PLCPropertiesController.class.getName()),
				(PropertiesController)this.cm.getController(HMIPropertiesController.class.getName()),
				(PropertiesController)this.cm.getController(MTUPropertiesController.class.getName()),
				(PropertiesController)this.cm.getController(RTUPropertiesController.class.getName()),
				(PropertiesController)this.cm.getController(IEDPropertiesController.class.getName()),
				(PropertiesController)this.cm.getController(historianPropertiesController.class.getName()),
				(PropertiesController)this.cm.getController(IOSPropertiesController.class.getName()),
				(PropertiesController)this.cm.getController(firewallPropertiesController.class.getName()),
				(PropertiesController)this.cm.getController(RAPPropertiesController.class.getName()),
				(PropertiesController)this.cm.getController(IDSPropertiesController.class.getName()),
				(PropertiesController)this.cm.getController(LinkPropertiesController.class.getName()),
				(PropertiesController)this.cm.getController(LanPropertiesController.class.getName()),
				(PropertiesController)this.cm.getController(InterfacePropertiesController.class.getName()),
				(PropertiesController)this.cm.getController(CoSPropertiesController.class.getName())
		};
		
		this.propertyPanels = new Object[] {
				this.parse("expproperties.xml", this.propertyControllers[0]),
				this.parse("nodeproperties.xml", this.propertyControllers[1]),
				this.parse("PLCproperties.xml", this.propertyControllers[2]),
				this.parse("HMIproperties.xml", this.propertyControllers[3]),
				this.parse("MTUproperties.xml", this.propertyControllers[4]),
				this.parse("RTUproperties.xml", this.propertyControllers[5]),
				this.parse("IEDproperties.xml", this.propertyControllers[6]),
				this.parse("historianproperties.xml", this.propertyControllers[7]),
				this.parse("IOSproperties.xml", this.propertyControllers[8]),
				this.parse("firewallproperties.xml", this.propertyControllers[9]),
				this.parse("RAPproperties.xml", this.propertyControllers[10]),
				this.parse("IDSproperties.xml", this.propertyControllers[11]),
				this.parse("linkproperties.xml", this.propertyControllers[12]),
				this.parse("lanproperties.xml", this.propertyControllers[13]),
				this.parse("intproperties.xml", this.propertyControllers[14]),
				this.parse("CoSproperties.xml", this.propertyControllers[15]),
		};
		
		for (int lpc = 0; lpc < this.propertyPanels.length; lpc++) {
			if (this.propertyPanels[lpc] != null)
				this.propertyControllers[lpc].setView(this.propertyPanels[lpc]);
		}
		
		this.propertiesDialog = this.parse("properties.xml");
		
		this.add(this.deck, this.propertiesDialog, 0);
		
		Dimension s = getSize();
		setRectangle(content, "bounds", 0, 0, s.width, s.height);
		for (int lpc = this.propertyPanels.length - 1; lpc >= 0; lpc--) {
			this.tkv.setKeyValue(this, "propertiesIndex", lpc);
			doLayout(content);
		}
		
		if (args != null && args.length == 2) {
			this.defaultPid = args[0];
			this.defaultEid = args[1];
		}
		
		if (applet != null) {
			this.defaultPid = applet.getParameter("pid");
			this.defaultEid = applet.getParameter("eid");
		}
		
		new Thread() {
			public void run() {
				progressController.waitFor(false);
				if (defaultPid != null && defaultEid != null) {
					invoke(openAction, null, "method");
				}
			}
		}.start();
	}
	
	public NetlabPreferences getPreferences()
    {
		return this.nbp;
    }
	
	public NetlabClient(Applet applet) throws Exception {
		this(applet, null);
	}
	
	public ObjectPool getPool() {
		return this.pool;
	}
	
	public Object getPoolValue(String key) {
		return this.tkv.getKeyValue(this.pool, key);
	}
	
	public void setPoolValue(String key, Object value) {
		this.tkv.setKeyValue(this.pool, key, value);
	}
	
	public ImageManager getImageManager() {
		return this.im;
	}
	
	public ControllerManager getControllerManager() {
		return this.cm;
	}

	public TopologyController getTopologyController() {
		return this.tc;
	}
	
	public NetlabConfiguration getConfiguration() {
		return this.nc;
	}
	
	public NoticeController getNoticeController() {
		return this.noticeController;
	}
	
	public ProgressController getProgressController() {
		return this.progressController;
	}

	public ProgressController showProgressDialog(String title, boolean oneShot) {
		this.progressController.waitFor(true);
		this.progressController.setOneShot(oneShot);
		this.progressController.setTitle(title);
		this.add(this.progressController.getDialog());
		return this.progressController;
	}

	public ProgressController showProgressDialog(String title) {
		return this.showProgressDialog(title, true);
	}
	
	public void doSelectByName(Object box, String selectByName) throws Exception {
		if  (selectByName.length() > 0) {
			Object items[] = this.getItems(this.field);
			int lpc;
			
			for (lpc = 0; lpc < items.length; lpc++) {
				NSObject nso = (NSObject)this.getProperty(items[lpc], "model");
				
				if (nso instanceof NSInterface) {
				}
				else {
					String text = nso.getName().toString();
					boolean selected = false;
					
					if (text.indexOf(selectByName) != -1) {
						selected = true;
					}
					this.setBoolean(items[lpc], "selected", selected);
				}
			}
			
			this.sc.addUniqueString(selectByName);
			this.setString(box, "text", "");
			
			this.elementAction(null);
		}
	}
	
	public void newFile(Object sender) {
		this.tc.clear();
		this.removeAll(this.field);
	}
	
	public Dimension getMinimumSize() {
		return minSize;
	}
	
	public boolean isCopyable() {
		return this.getSelectedIndex(this.field) != -1;
	}
	
	public void doCutCopy(Object sender, int copy) {
		Object sel[];
		
		if ((sel = this.getSelectedItems(this.field)) != null) {
			boolean hit = false;
			int lpc;
			
			this.ci = new ClipInfo[sel.length];
			for (lpc = 0; lpc < sel.length; lpc++) {
				NSObject nso;
				
				nso = (NSObject) this.getProperty(sel[lpc], "model");
				if ((nso instanceof NSNode) || (nso instanceof NSLan)) {
					ClipInfo nci = new ClipInfo();
					
					nci.ec = (ElementController) this.getProperty(sel[lpc], "nsocontroller");
					if (nso instanceof NSLan)
						((NSLan) nso).getMembers().clear();
					nci.obj = (NSObject) nso.clone();
					nci.obj.setName(null);
					nci.baseName = nso.getName().getBase();
					nci.pasteBounds = (Rectangle) (this.getRectangle(sel[lpc], "bounds").clone());
					if (copy == 0) {
						nci.obj.setName(nso.getName());
						nci.ec.delete(sel[lpc]);
					}
					
					this.ci[lpc] = nci;
					hit = true;
				}
			}
			
			if (hit) {
				this.tkv.notifyKeyValue(this, "isPastable", ThinletKeyValue.CHANGE_SETTING, null, Boolean.TRUE);
				if (copy == 0)
					this.tkv.setKeyValue(this.tc, "selection", null);
			}
			else {
				this.ci = new ClipInfo[0];
				
				this.tkv.notifyKeyValue(this, "isPastable", ThinletKeyValue.CHANGE_SETTING, null, Boolean.FALSE);
			}
		}
	}

	private int nodeIndex() {
		boolean done = false;
		int lpc, retval = 0;
		Object items[];

		items = this.getItems(this.field);
		for (lpc = 0; (lpc < items.length) && !done; lpc++) {
			NSObject nso;

			nso = (NSObject) this.getProperty(items[lpc], "model");
			if (nso instanceof NSNode) {
				retval = lpc;
				done = true;
			}
		}
		
		return retval;
	}
	
	public boolean isPastable() {
		return (this.ci.length > 0);
	}

	public void doPaste(Object sender) throws Exception {
		int lpc;
		
		for (lpc = 0; lpc < this.ci.length; lpc++) {
			ClipInfo curr = this.ci[lpc];
			ElementController ec;
			NSObject nso;
			Object obj;
			
			if (curr != null) {
				nso = (NSObject) curr.obj.clone();

				if ((curr.obj.getName() != null) && !this.tc.getNamer().contains(curr.obj.getName())) {
					this.tc.getNamer().put(curr.obj.getName());
					nso.setName(curr.obj.getName());
				} else {
					nso.setName(this.tc.getNamer().suggest(curr.baseName));
				}
				ec = (ElementController)this.cm.getController(curr.ec.getClass().getName());
				obj = ec.copyElement(nso);
				curr.pasteBounds.x += 40;
				curr.pasteBounds.y += 40;
				this.setRectangle(obj, "bounds", curr.pasteBounds.x,
						curr.pasteBounds.y, curr.pasteBounds.width,
						curr.pasteBounds.height);
				this.add(this.field, obj, this.nodeIndex());
			}
		}
	}
	
	public void deselect(Object sender) {
		Object selected[];

		this.tkv.notifyKeyValue(this, "isCopyable", ThinletKeyValue.CHANGE_SETTING, null, Boolean.FALSE);
		
		if ((selected = this.getSelectedItems(this.field)) != null) {
			int lpc;

			for (lpc = 0; lpc < selected.length; lpc++) {
				this.setBoolean(selected[lpc], "selected", false);
			}
		}
		
		this.tkv.setKeyValue(this.tc, "selection", null);
	}
	
	public void elementAction(Object sender) throws Exception {
		Object selected[];

		selected = this.getSelectedItems(this.field);
		if (selected.length == 0) {
			this.deselect(sender);
		} else if (selected.length == 1) {
			this.tkv.setKeyValue(this.tc, "selection", selected[0]);
			this.showView(selected[0]);
		} else {
			List list = Arrays.asList(selected);

			this.tkv.setKeyValue(this.tc, "selection", list);
			this.showView(list);
		}
	}

	public Object gotConnect(Object left, Object right) throws Exception {
		String cname, lname, rname;
		NSObject lobj, robj;
		Object parent;

		parent = this.getParent(left);
		lname = this.getString(left, "name");
		rname = this.getString(right, "name");
		if (lname.compareTo(rname) < 0)
			cname = lname + "|" + rname;
		else
			cname = rname + "|" + lname;
		lobj = (NSObject) this.getProperty(left, "model");
		robj = (NSObject) this.getProperty(right, "model");
		if (this.find(parent, cname) != null) {
			this.noticeController.showDialog("Too many links", "Multiple links between nodes is not supported.");
		} else if ((lobj instanceof NSLan) && (robj instanceof NSLan)) {
			this.noticeController.showDialog("Cannot link LANs", "Use a router node to link LANs together.");
		} else {
			ElementController controller;
			Object connector;
			NSInterface nsi;
			NSLanLink nsll;

			this.deselect(null);

			if ((lobj instanceof NSNode) && (robj instanceof NSNode)) {
				Vector members;

				nsll = new NSLink();
				nsll.setName(this.tc.getNamer().suggest(LinkController.DEFAULT_LINK_NAME));
				members = nsll.getMembers();
				members.add(new NSInterface().setNSObject(lobj));
				members.add(new NSInterface().setNSObject(robj));
				controller = (ElementController) this.cm
						.getController("net.emulab.netlab.client.LinkController");
				connector = this.parse("nslink.xml", controller);
				this.setIcon(connector, "icon", this.im
						.getObjectImage("link-none"));
				this.setString(connector, "text", nsll.getName().toShortString(
						this.nc.getShortNameLength()));
				this.putProperty(connector, "model", nsll);
				this.tc.addNSObject(nsll);
				this.add(connector, this.parse("linkmenu.xml", controller));
			} else {
				NSNode nsn;

				if (lobj instanceof NSNode) {
					nsn = (NSNode) lobj;
					nsi = new NSInterface().setNSObject(lobj);
					nsll = (NSLan) robj;
				} else {
					nsn = (NSNode) robj;
					nsll = (NSLan) lobj;
					nsi = new NSInterface().setNSObject(robj);
				}
				nsi.setTemplate(((NSLan) nsll).getTemplateInterface());
				if (!nsll.isMember(nsn)) // This may be a bind
					nsll.addMember(nsi);
				controller = (ElementController) this.cm
						.getController("net.emulab.netlab.client.InterfaceController");
				connector = this.parse("nslanlink.xml", controller);
				this.putProperty(connector, "model", nsi);
				this.putProperty(connector, "lan", nsll);
				this.add(connector, this.parse("intmenu.xml", controller));
				this.putProperty(connector, "view", "intproperties.xml");
				this.putProperty(connector, "controller",
								"net.emulab.netlab.client.InterfacePropertiesController");
			}
			this.setString(connector, "left", lname);
			this.setString(connector, "right", rname);
			this.setString(connector, "name", cname);
			this.setMethod(connector, "action", "elementAction(this)", connector,
					this);
			this.setMethod(connector, "focusgained", "elementAction(this)", connector,
					this);
			this.setMethod(connector, "delete", "deleteNSObject(this)", connector,
					this);
			this.putProperty(connector, "nsocontroller", controller);

			this.add(parent, connector, 0);

			this.tkv.setKeyValue(controller, "element", connector);
			controller.modelChanged(connector);
			
			this.setBoolean(connector, "selected", true);
			this.elementAction(connector);

			return connector;
		}
		return null;
	}

	public void deleteNSObject(Object sender) {
		Object sel[];

		this.tkv.notifyKeyValue(this, "isCopyable", ThinletKeyValue.CHANGE_SETTING, null, Boolean.FALSE);
		
		if ((sel = this.getSelectedItems(field)) != null) {
			int lpc;
			
			for (lpc = 0; lpc < sel.length; lpc++) {
				ElementController ec = (ElementController)this.getProperty(sel[lpc], "nsocontroller");
				
				ec.delete(sel[lpc]);
			}
		}
		
		this.tkv.setKeyValue(this.tc, "selection", null);
	}

	public void showView(Object sender) throws Exception {
		this.tkv.notifyKeyValue(this, "isCopyable", ThinletKeyValue.CHANGE_SETTING, null, Boolean.TRUE);
		
		this.cm.showView(sender);
	}
	
	protected void handleException(Object component, Throwable th) {
		th.printStackTrace();
		this.noticeController.showDialog(th);
	}

	private synchronized void setPropertiesPanel(Object panel) {
		Object oldPanel;
		
		oldPanel = this.currentPropertiesPanel;
		this.remove(this.currentPropertiesPanel);
		/*
		 * XXX The remove can change focus and cause this method to be called
		 * again...
		 */
		if (oldPanel == this.currentPropertiesPanel) {
			this.currentPropertiesPanel = panel;
			this.add(this.propertiesDialog, this.currentPropertiesPanel);
		}
	}
	
	public int propertyPanelIndex(Object panel) {
		int retval;
		
		for (retval = 0; retval < this.propertyPanels.length; retval++) {
			if (this.propertyPanels[retval] == panel)
				return retval;
		}
		
		throw new IllegalArgumentException();
	}
	
	public void setPropertiesIndex(int index) {
		if (index >= this.propertyPanels.length)
			throw new IllegalArgumentException();

		if (this.propertiesIndex != index) {
			this.propertiesIndex = index;
			this.setPropertiesPanel(this.propertyPanels[index]);
		}
	}
	
	public int getPropertiesIndex() {
		return this.propertiesIndex;
	}

	/**
	 * Open a link in a browser window.
	 * 
	 * @param sender
	 *            The Thinlet component that triggered this call and,
	 *            presumably, has the URL in an "aref" property.
	 * @throws Exception
	 *             if there was a problem opening the URL.
	 */
	public void openLinkInObject(Object sender) throws Exception {
		this.openLink((String) this.getProperty(sender, "aref"));
	}

	public void openLinkInObjectAndWindow(Object sender, String windowID) throws Exception {
		this.openLink((String) this.getProperty(sender, "aref"), windowID);
	}

	public void openLink(String aref) throws Exception {
		this.openLink(aref, "netlab-doc");
	}
	
	public URL createURL(Applet applet, String aref) {
		URL retval;
		
		try {
			if (!aref.startsWith("http://") || !aref.startsWith("https://")) {
				retval = new URL(applet.getCodeBase(), aref);
			}
			else {
				retval = new URL(aref);
			}
		}
		catch (MalformedURLException e) {
			throw new InternalError();
		}
		
		return retval;
	}
	
	public void openLink(String aref, String windowID) throws Exception {
		Applet applet = (Applet)this.getPoolValue("applet");
		
		URL url = createURL(applet, aref);
		applet.getAppletContext().showDocument(url, windowID);
	}

	protected void updateLinkStatus(Object component, boolean entered) {
		String aref = (String)getProperty(component, "aref");
		Applet applet = (Applet)this.getPoolValue("applet");
		
		if (applet != null && aref != null) {
			String status = "";
			
			if (entered) {
				
				status = ("Open \""
						+ createURL(applet, aref).toExternalForm()
						+ "\" in a new window");
			}
			
			applet.getAppletContext().showStatus(status);
		}
	}

	public void showFile(Object sender) throws IOException{
		ByteArrayOutputStream baos = new ByteArrayOutputStream();
		NSFile nsf = new NSFile();
		byte file[];

		setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
		nsf.toStream(new PrintStream(baos),
				this.getTopologyController().getNSTopology());
		
		file = baos.toByteArray();
		
		this.setString(this.showFileText, "text", new String(file));
		if (this.getParent(this.showFileDialog) == null)
			this.add(this.showFileDialog);
		setCursor(Cursor.getPredefinedCursor(Cursor.DEFAULT_CURSOR));
	}
	
	public void showFile1(Object sender, int wToF, String loc) throws IOException{
		ByteArrayOutputStream baos = new ByteArrayOutputStream();
		NSFile1 nsf = new NSFile1();
		byte file[];

		setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
		nsf.toStream(new PrintStream(baos),
				this.getTopologyController().getNSTopology(),wToF, loc);
		
		file = baos.toByteArray();
		
		this.setString(this.showFileText, "text", new String(file));
		if (this.getParent(this.showFileDialog) == null)
			this.add(this.showFileDialog);
		setCursor(Cursor.getPredefinedCursor(Cursor.DEFAULT_CURSOR));
	}
	
	public void quit(Object sender) {
		System.exit(0);
	}

	public static void main(String args[]) throws Throwable {
		int width = 1024, height = 768;
		NetlabClient nb;

		System.setProperty("java.util.prefs.PreferencesFactory",
				"net.emulab.netlab.client.DisabledPreferencesFactory");

		nb = new NetlabClient(null, args);

		if (nb.getWidth() != 0)
			width = nb.getWidth();
		if (nb.getHeight() != 0)
			height = nb.getHeight();

		new FrameLauncher("NetlabClient", nb, width, height);
	}
}
