/*
 * NetlabPreferences.java
 *
 * Copyright (c) 2005 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

package net.emulab.netlab.client;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.List;
import java.util.Properties;
import java.util.Vector;

import net.emulab.netlab.client.emulab.SshConfiguration;

public class NetlabPreferences
    extends Properties
{
    protected static final SimpleDateFormat DATE_FORMAT =
	new SimpleDateFormat("yyyy-MM-dd HH:mm:ss");
    
    public static final String DEFAULT_SSH_CONFIG_TEMPLATE = "\n"
	+ "# {0} -> User name\n"
	+ "# {1} -> Path to known hosts file\n"
	+ "# {2} -> Forward X11 (yes|no)\n"
	+ "\n"
	+ "Protocol 1,2\n"
	+ "PasswordAuthentication no\n"
	+ "ChallengeResponseAuthentication yes\n"
	+ "Compression yes\n"
	+ "User {0}\n"
	+ "BatchMode yes\n"
	+ "UserKnownHostsFile {1}\n"
	+ "ForwardX11 {2}\n"
	+ "StrictHostKeyChecking yes\n";

    private static final String USER_HOME =
	System.getProperty("user.home");
    public static final File USER_SSH =
	new File(USER_HOME + File.separator + ".ssh");
    private static final File USER_SSH_CONFIG =
	new File(USER_SSH, "config");
    public static final String NETLAB_HOME =
	USER_HOME + File.separator + ".netlab-client";
    private static final String NETLAB_PROPERTIES =
	NETLAB_HOME + File.separator + "properties";
    public static final File NETLAB_SSH_CONFIG_TEMPLATE =
	new File(NETLAB_HOME + File.separator + "ssh-config.template");

    private static final String NETLAB_PROPERTIES_HEADER = ""
	+ "#\n"
	+ "# NetlabClient Properties\n"
	+ "#\n"
	+ "\n";
    
    public static final String EMULAB_USER =
	"net.emulab.netlab.client.emulab.user";
    public static final String EMULAB_HOST =
	"net.emulab.netlab.client.emulab.host";
    public static final String EMULAB_PATH =
	"net.emulab.netlab.client.emulab.path";
    
    public static final String EMULAB_WEB =
	"net.emulab.netlab.client.emulab.web";
    
    public static final String LAST_NEWS =
	"net.emulab.netlab.client.last_news";
    
    public static final String LAST_PROJECT =
	"net.emulab.netlab.client.last_project";
    public static final String LAST_GROUP =
	"net.emulab.netlab.client.last_group";
    
    public static final String RECENT_DIR_PREFIX =
	"net.emulab.netlab.client.recent.dir.";

    public static final String XTERM_COMMAND =
	"net.emulab.netlab.client.xterm";
    public static final String BROWSER_COMMAND =
	"com.Ostermiller.util.Browser.open";

    private SshConfiguration sc;
    
    public NetlabPreferences()
	throws IOException
    {
	File home = new File(NETLAB_HOME);
	
	if( !home.exists() && !home.mkdir() )
	{
	    throw new IOException("Unable to make NetlabClient home: "
				  + NETLAB_HOME);
	}
    }

    public void load()
	throws IOException
    {
	try
	{
	    this.load(new FileInputStream(NETLAB_PROPERTIES));
	}
	catch(FileNotFoundException e)
	{
	    // First time...
	}
    }

    public void store()
	throws IOException
    {
	this.store(new FileOutputStream(NETLAB_PROPERTIES),
		   NETLAB_PROPERTIES_HEADER);
    }

    public String getEmulabUserName()
    {
	return this.getProperty(NetlabPreferences.EMULAB_USER,
				System.getProperty("user.name"));
    }

    public String getEmulabHost()
    {
	return this.getProperty(NetlabPreferences.EMULAB_HOST,
				"boss.emulab.net");
    }
    
    public String getEmulabWeb()
    {
	return this.getProperty(NetlabPreferences.EMULAB_WEB,
				"http://www.emulab.net");
    }
    
    public String getEmulabDomain()
    {
	String tmp, retval;

	tmp = this.getEmulabHost();
	retval = tmp.substring(tmp.indexOf('.') + 1);
	return retval;
    }
    
    public String getEmulabPath()
    {
	return this.getProperty(NetlabPreferences.EMULAB_PATH, "xmlrpc");
    }

    public NetlabPreferences setDateProperty(String property, Date date)
    {
	super.setProperty(property, DATE_FORMAT.format(date));
	return this;
    }

    public Date getDateProperty(String property)
    {
	Date retval = null;

	try
	{
	    String value;
	    
	    if( (value = this.getProperty(property)) != null )
	    {
		retval = DATE_FORMAT.parse(value);
	    }
	}
	catch(ParseException e)
	{
	    this.setProperty(property, null);
	}
	return retval;
    }

    public String getLastProject()
    {
	return this.getProperty(NetlabPreferences.LAST_PROJECT, "");
    }

    public String getLastGroup()
    {
	return this.getProperty(NetlabPreferences.LAST_GROUP, "");
    }

    public String getXtermCommand()
    {
	return this.getProperty(NetlabPreferences.XTERM_COMMAND, "xterm");
    }

    public String getBrowserCommand()
    {
	return this.getProperty(NetlabPreferences.BROWSER_COMMAND, "");
    }

    public String getSshTemplate()
    {
	String retval = DEFAULT_SSH_CONFIG_TEMPLATE;

	if( NETLAB_SSH_CONFIG_TEMPLATE.exists() )
	{
	    FileInputStream fis = null;
	    
	    try
	    {
		StringBuffer sb = new StringBuffer();
		byte buffer[] = new byte[1024];
		int rc;
		
		fis = new FileInputStream(NETLAB_SSH_CONFIG_TEMPLATE);
		while( (rc = fis.read(buffer)) != -1 )
		{
		    sb.append(new String(buffer, 0, rc));
		}
		retval = sb.toString();
	    }
	    catch(IOException e)
	    {
		e.printStackTrace(System.err);
	    }
	    finally
	    {
		if( fis != null )
		{
		    try
		    {
			fis.close();
		    }
		    catch(IOException e)
		    {
		    }
		}
	    }
	}
	return retval;
    }

    public synchronized void clearSshConfiguration()
    {
	this.sc = null;
    }

    public synchronized String getSshConfiguration()
    {
	String retval;

	try
	{
	    if( this.sc == null )
	    {
		this.sc = new SshConfiguration(this);
		this.sc.setup();
	    }
	    retval = this.sc.getConfigFile().toString();
	}
	catch(IOException e)
	{
	    retval = null;
	}
	return retval;
    }

    public Vector getVector(String prefix, int max)
    {
	Vector retval = new Vector();
	int lpc;
	
	for( lpc = 0; lpc < max; lpc++ )
	{
	    String dir;

	    dir = this.getProperty(prefix + lpc, "");
	    if( !"".equals(dir) )
	    {
		retval.addElement(dir);
	    }
	}
	return retval;
    }

    public void setVector(String prefix, List list)
    {
	int lpc, size;

	size = list.size();
	for( lpc = 0; lpc < size; lpc++ )
	{
	    String value = (String)list.get(lpc);
	    
	    this.setProperty(prefix + lpc, value);
	}
    }
    
    public String toString()
    {
	return "NetlabPreferences["
	    + super.toString()
	    + "]";
    }
}
