/*
 * NewsController.java
 *
 * Copyright (c) 2005 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

package net.emulab.netlab.client;

import java.awt.Cursor;
import java.util.Date;
import java.util.Hashtable;
import java.util.Vector;

import net.emulab.netlab.client.emulab.EmulabEmulabProxy;
import thinlet.AutoFillController;
import thinlet.DateFormatter;

/**
 * Controller for the Emulab news listing.
 */
public class NewsController
    extends NetlabArrayController
{
    /**
     * The keys to scan when filtering out entries.
     */
    private static final String FILTER_KEYS[] = {
	"author",
	"subject"
    };
    
    /**
     * The last date the user read the news.
     */
    private static Date LAST_READ;

    /**
     * Formatter for the news item dates.
     */
    public static DateFormatter DATE_FORMATTER =
	new DateFormatter(DateFormatter.DEFAULT_DATE_FORMAT, null);

    /**
     * Update the time when the user last read news.
     *
     * @param np The preferences file to read date when the user last read
     * news.
     */
    static synchronized void updateLastRead(NetlabPreferences np)
    {
	LAST_READ = np.getDateProperty(NetlabPreferences.LAST_NEWS);
    }

    /**
     * Thinlet outlet for the news listing.
     */
    public Object newslist;

    /**
     * The search box controller.
     */
    public final AutoFillController sc;

    /**
     * Indicates whether or not the controller is synced with Emulab.
     */
    protected boolean synced = false;

    /**
     * Construct a NewsController with the given value.
     *
     * @param netlab The NetlabClient instance this controller is working with.
     */
    public NewsController(NetlabClient netlab)
    {
	super(netlab);
	
	if( LAST_READ == null )
	    updateLastRead(this.netlab.getPreferences());
	
	this.addFilterKeys(FILTER_KEYS);

	this.sc = new AutoFillController(netlab);
    }

    public Controller modelChanged(Object component)
    {
	return this;
    }

    /**
     * Munge the news listing retrieved from Emulab and set it as this
     * controller's content.
     *
     * @param news A vector of news items as returned by Emulab.
     */
    private void setNewsList(Vector news)
    {
	int lpc, size;

	size = news.size();
	for( lpc = 0; lpc < size; lpc++ )
	{
	    Hashtable ht = (Hashtable)news.get(lpc);
	    Date date;

	    date = (Date)ht.get("date");
	    if( (LAST_READ == null) || (date.compareTo(LAST_READ) > 0) )
	    {
		ht.put("read", this.netlab.getObjectImage("blue"));
	    }
	}
	this.tkv.setKeyValue(this, "content", news);
	this.synced = true;
    }

    /**
     * Method called just before the news window is shown on the screen.  If
     * the controller is not synced with Emulab, it will attempt to download
     * the news items.
     */
    public void showPanel()
    {
	if( !this.synced )
	{
	    final ProgressController pc;

	    pc = this.netlab.showProgressDialog("Downloading news items...");
	    new Thread() {
		public void run()
		{
		    try
		    {
			NetlabPreferences nbp;
			EmulabEmulabProxy eep;
			Date starting;

			starting = new Date(System.currentTimeMillis() -
					    (365L * 24L * 60L * 60L * 1000L));
			nbp = NewsController.this.netlab.getPreferences();
			eep = new EmulabEmulabProxy(nbp, pc);
			NewsController.this.
			    setNewsList(eep.news(starting, null));
			nbp.setDateProperty(NetlabPreferences.LAST_NEWS,
					    new Date());
			nbp.store();
		    }
		    catch(Throwable th)
		    {
			System.err.println("Downloader failed with:");
			th.printStackTrace(System.err);
			if( !pc.wasHalted() )
			{
			    pc.stop();
			    NewsController.this.netlab.
				showErrorDialog(th.getMessage(), null);
			}
			NewsController.this.netlab.
			    remove(NewsController.this.view);
		    }
		}
	    }.start();
	}
    }

    /**
     * Method called just after the window is shown and its contents are
     * focusable, in this case we set the focus to the news listing.
     */
    public void focusable()
    {
	this.thinlet.requestFocus(this.newslist);
    }

    /**
     * Method called when a user double clicks on a news item, causing a
     * browser window to open and display the news item.
     */
    public void openLink()
	throws Exception
    {
	String msgid;
	
	this.netlab.
	    setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
	this.tkv.setKeyValue(this, "selection.read", null);
	msgid = this.tkv.getKeyValue(this, "selection.msgid").toString();
	BrowserController.displayURL(this.netlab,
				     this.netlab.getPreferences().
				     getEmulabWeb()
				     + "/news.php3#"
				     + msgid);
	this.netlab.
	    setCursor(Cursor.getPredefinedCursor(Cursor.DEFAULT_CURSOR));
    }

    /**
     * Method called when the user clicks the "Ok" button, just closes the
     * window.
     */
    public void ok()
    {
	this.netlab.remove(this.view);
    }
    
    /**
     * @see java.lang.Object#toString()
     */
    public String toString()
    {
	return "NewsController["
	    + super.toString()
	    + "; newslist="
	    + this.newslist
	    + "; sc="
	    + this.sc
	    + "; synced="
	    + this.synced
	    + "]";
    }
}
