/*
 * ProgressController.java
 *
 * Copyright (c) 2005 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

package net.emulab.netlab.client;

import java.io.IOException;
import java.net.URL;

import net.emulab.util.ProgressHalter;
import net.emulab.util.ProgressObserver;
import thinlet.NumberFormatter;
import thinlet.SizeFormatter;
import thinlet.Thinlet;
import thinlet.ThinletKeyValue;

/**
 * Controller for the progress dialog.
 */
public class ProgressController implements ProgressObserver {
	public static final NumberFormatter SIZE_FORMATTER = new SizeFormatter("", "Bytes");

	/**
	 * The NetlabClient instance this controller is attached to.
	 */
	protected final Thinlet netbuild;
	
	protected final ThinletKeyValue tkv;
	
	protected boolean inProgress;

	/**
	 * The progress dialog window.
	 */
	protected Object dialog;

	/**
	 * The (optional) object that can halt progress of an operation.
	 */
	protected ProgressHalter ph;

	/**
	 * The maximum value for the progress bar.
	 */
	protected long maximum;

	/**
	 * Indicates whether or not the stop button was clicked.
	 */
	protected boolean halted;

	/**
	 * Indicates whether or not the window should be closed automatically after
	 * there is no content remaining.
	 */
	protected boolean oneShot;

	public int completed;
	
	/**
	 * Construct a ProgressController with the given value.
	 * 
	 * @param netbuild
	 *            The NetlabClient instance this controller is attached to.
	 */
	public ProgressController(Thinlet netbuild) throws IOException {
		this.netbuild = netbuild;
		this.tkv = this.netbuild.getNamespace();
		
		this.dialog = this.netbuild.parse("/net/emulab/netlab/client/progressdialog.xml", this);
	}

	/**
	 * @param oneShot
	 *            Indicates whether or not the window should be closed
	 *            automatically after there is no content remaining.
	 * @return this
	 */
	public ProgressController setOneShot(boolean oneShot) {
		this.oneShot = oneShot;
		return this;
	}

	/**
	 * @return Indicates whether or not the window should be closed
	 *         automatically after there is no content remaining.
	 */
	public boolean getOneShot() {
		return this.oneShot;
	}

	/**
	 * @return True if the user clicked the "stop" button, false otherwise.
	 */
	public boolean wasHalted() {
		return this.halted;
	}
	
	public synchronized void waitFor(boolean take) {
		while (this.inProgress) {
			try {
				this.wait();
			}
			catch (InterruptedException e) {
			}
		}
		this.inProgress = take;
	}

	/**
	 * Method called right before the dialog is shown on the screen, just resets
	 * all the views to an empty state.
	 * 
	 * @param dialog
	 *            The dialog window.
	 * @param status
	 *            The status message in the dialog.
	 * @param progressBar
	 *            The progress bar in the dialog.
	 * @param stop
	 *            The stop button, disabled when there is no halter set.
	 */
	public void showDialog(Object dialog, Object status, Object progressBar, Object stop) {
		this.ph = null;
		this.halted = false;
		this.netbuild.setString(status, "text", "");
		this.netbuild.setInteger(progressBar, "minimum", 0);
		this.netbuild.setInteger(progressBar, "maximum", 1);
		this.tkv.setKeyValue(this, "completed", -1);
		this.netbuild.setBoolean(stop, "enabled", false);
	}
	
	public Object getDialog() {
		return this.dialog;
	}

	/**
	 * Set the title of the progress dialog.
	 * 
	 * @param title
	 *            The title to set.
	 * @return this
	 */
	public ProgressController setTitle(String title) {
		this.netbuild.setString(this.dialog, "text", title);
		return this;
	}

	/**
	 * If the "ph" argument is non-null, this will enable the "stop" button.
	 * 
	 * @see net.emulab.netlab.client.ProgressObserver#setProgressHalter(java.net.URL,
	 *      net.emulab.netlab.client.ProgressHalter)
	 */
	public void setProgressHalter(URL url, ProgressHalter ph) {
		this.ph = ph;
		this.netbuild.setBoolean(this.netbuild.find(this.dialog, "stop"),
				"enabled", this.ph != null);
	}

	/**
	 * Just updates the "status" label in the window.
	 * 
	 * @see net.emulab.netlab.client.ProgressObserver#setStage(java.net.URL,
	 *      java.lang.String)
	 */
	public void setStage(URL url, String stageName) {
		Object status;

		status = this.netbuild.find(this.dialog, "status");
		this.netbuild.setString(status, "text", stageName);
	}

	/**
	 * Resets the progress bar values to handle the new content.
	 * 
	 * @see net.emulab.netlab.client.ProgressObserver#setContentLength(java.net.URL,
	 *      long)
	 */
	public void setContentLength(URL url, long contentLength) {
		Object progressBar;

		this.maximum = contentLength;
		progressBar = this.netbuild.find(this.dialog, "progressbar");
		this.netbuild.setInteger(progressBar, "minimum", 0);
		this.tkv.setKeyValue(this, "completed", 0);
		this.netbuild.setInteger(progressBar, "maximum", (int) contentLength);
	}

	/**
	 * Updates the progress bar, and optionally closes the window if there is no
	 * more content and oneShot is true.
	 * 
	 * @see net.emulab.netlab.client.ProgressObserver#updateContentRemaining(java.net.URL,
	 *      long)
	 */
	public void updateContentRemaining(URL url, long contentRemaining) {
		this.tkv.setKeyValue(this, "completed", (int) (this.maximum - contentRemaining));
		if ((contentRemaining == 0) && this.oneShot) {
			this.done();
		}
	}

	/**
	 * Try to close the dialog, if it hasn't already been.
	 */
	public synchronized void done() {
		if (!this.halted) {
			this.halted = true;
			this.netbuild.remove(this.dialog);
			this.inProgress = false;
			this.notify();
		}
	}

	/**
	 * Method called when the "Stop" button is clicked, just halts the progress
	 * and closes the window.
	 */
	public synchronized void stop() {
		if (!this.halted) {
			this.halted = true;
			if (this.ph != null)
				this.ph.haltProgress();
			this.netbuild.remove(this.dialog);
			this.inProgress = false;
			this.notify();
		}
	}

	/**
	 * @see java.lang.Object#toString()
	 */
	public String toString() {
		return "ProgressController[dialog=" + this.dialog + "; ph=" + this.ph
				+ "; maximum=" + this.maximum + "; halted=" + this.halted + "]";
	}
}
