/*
 * RPMFileListController.java
 *
 * Copyright (c) 2005 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

package net.emulab.netlab.client;

import java.util.Iterator;
import java.util.List;
import java.util.SortedSet;
import java.util.TreeSet;
import java.util.Vector;

import net.emulab.netlab.client.emulab.EmulabFSProxy;
import net.emulab.netlab.client.emulab.EmulabXmlRpcException;
import net.emulab.util.NullProgressObserver;

import org.apache.xmlrpc.XmlRpcClientException;

import thinlet.AutoFillController;
import thinlet.MutableString;
import thinlet.ThinletKeyValue;

/**
 * Controller for the RPM list dialog.
 * 
 * TODO: Handle ordering... What if one RPM depends on another...
 */
public class RPMFileListController extends NetlabArrayController {
	/**
	 * Sorted list RPM files (which are just strings).
	 */
	protected SortedSet ss;

	/**
	 * Auto-fill controller for the RPM file textfield.
	 */
	public final AutoFillController rc;

	public Object objController;

	public boolean syncing;
	private boolean synced;

	/**
	 * Construct an RPMFileListController with the given value.
	 * 
	 * @param nc
	 *            The NetlabClient instance this controller is attached to.
	 */
	public RPMFileListController(NetlabClient netlab) {
		super(netlab);

		this.rc = new FileAutoFillController(netlab);

		this.setObjectClass(MutableString.class);
	}

	/**
	 * Method called when this controller is activated by another component.
	 * Currently, it will get the list of RPM files to manipulate from the
	 * triggering component and make a copy to be this controller's content
	 * array.
	 * 
	 * @see net.emulab.netlab.client.Controller#triggeredBy(java.lang.Object)
	 */
	public Controller triggeredBy(Object component) throws Exception {
		Vector newContent;
		Iterator iterator;

		/* Get the list of RPMs, */
		this.objController = this.netlab
				.getProperty(component, "objcontroller");
		this.ss = (SortedSet)this.tkv.getKeyValueAsSingleton(this.objController,
				"selection.RPMFiles");

		/* ... make a new content array, and */
		newContent = new Vector(this.ss.size());
		/* ... make each string "mutable" before adding it to the content. */
		for (iterator = this.ss.iterator(); iterator.hasNext();) {
			newContent.add(new MutableString((String) iterator.next()));
		}
		this.tkv.setKeyValue(this, "content", newContent);

		super.triggeredBy(component);

		return this;
	}

	public void showDialog() {
		if (!this.synced) {
			this.syncRPMFiles();
		}
	}
	
	private synchronized void syncRPMFiles() {
		if (syncing)
			return;
		
		tkv.setKeyValue(RPMFileListController.this, "syncing", true);
		
		Thread th = new Thread() {
			public void run() {
				try {
					EmulabFSProxy fs;
					Iterator iter;
					List exports;
					Object oldpo;

					oldpo = netlab.getPoolValue("server.observer");
					netlab.setPoolValue("server.observer", NullProgressObserver.SINGLETON);
					fs = new EmulabFSProxy(tkv, netlab.getPool());
					netlab.setPoolValue("server.observer", oldpo);
					
					exports = fs.exports();
					for (iter = exports.iterator(); iter.hasNext(); ) {
						try {
							String path = (String)iter.next();
							
							path += "/rpms";
							if (fs.access(path, "execute")) {
								DirectoryEntry de[] = fs.listdir(path);
								int lpc;
								
								for (lpc = 0; lpc < de.length; lpc++) {
									String fullPath = path + "/" + de[lpc].name;
									
									if (de[lpc].type == 'f' && de[lpc].name.endsWith(".rpm")) {
										rc.addObject(fullPath);
									}
								}
							}
						}
						catch (EmulabXmlRpcException e) {
							e.printStackTrace();
						}
					}
					synced = true;
				}
				catch (Throwable th) {
					netlab.handleException(th);
				}
				finally {
					tkv.setKeyValue(RPMFileListController.this, "syncing", false);
				}
			}
		};
		
		th.setDaemon(true);
		th.start();
	}

	/**
	 * @see thinlet.ArrayController#newObject()
	 * 
	 * @return A new MutableString object with a non-empty initial value.
	 */
	public Object newObject() {
		return new MutableString("/path/to.rpm");
	}

	/**
	 * Method called when the user clicks the "Ok" button, just updates the
	 * model with all non-empty RPM file names and closes the window.
	 */
	public void ok() {
		TreeSet nts = new TreeSet();
		int lpc, size;
		List list;

		/* Out with the old and */
		this.ss.clear();
		list = this.contentArray();
		size = list.size();
		/* ... in with the (non-empty) new. */
		for (lpc = 0; lpc < size; lpc++) {
			String value = list.get(lpc).toString();

			if (!"".equals(value))
				nts.add(value);
		}

		this.tkv.setKeyValue(this.objController, "selection.RPMFiles", nts);

		/* Get rid of the dialog box. */
		this.netlab.remove(this.view);
		this.ss = null;
	}

	/**
	 * Method called when the "Cancel" button is clicked, just clears some state
	 * and closes the window.
	 * 
	 * @param panel
	 *            The dialog window.
	 */
	public void cancel() {
		this.netlab.remove(this.view);
		this.ss = null;
	}

	/**
	 * @see java.lang.Object#toString()
	 */
	public String toString() {
		return "RPMFileListController[" + super.toString() + "; ss=" + this.ss
				+ "]";
	}
}
