/*
 * TarFileListController.java
 *
 * Copyright (c) 2005 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

package net.emulab.netlab.client;

import java.util.Arrays;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;
import java.util.TreeSet;
import java.util.Vector;

import net.emulab.netlab.client.emulab.EmulabFSProxy;
import net.emulab.netlab.client.emulab.EmulabXmlRpcException;
import net.emulab.ns.NSTarFile;
import net.emulab.util.NullProgressObserver;

import org.apache.xmlrpc.XmlRpcClientException;

import thinlet.ArrayController;
import thinlet.AutoFillController;

/**
 * Controller for the Tar file list dialog.
 */
public class TarFileListController extends NetlabArrayController {
	
	private static final String SUFFIXES[] = {
		".tar",
		".tar.gz",
		".tar.Z",
		".tar.bz2"
	};
	
	private static final boolean isTarFile(String name) {
		int lpc;
		
		for (lpc = 0; lpc < SUFFIXES.length; lpc++) {
			if (name.endsWith(SUFFIXES[lpc]))
				return true;
		}
		
		return false;
	}
	
	/**
	 * Initial list of directories when doing auto fill.
	 */
	private static final List DST_AUTO_FILL = Arrays.asList(new String[] { "/",
			"/usr", "/usr/site", "/usr/local", });

	/**
	 * The TreeSet of Tar files being manipulated.
	 */
	protected TreeSet ts;
	
	public boolean syncing;
	protected boolean synced;

	/**
	 * Thinlet outlet for the directory text field.
	 */
	public Object dirComponent;

	/**
	 * Thinlet outlet for the tar file text field.
	 */
	public Object fileComponent;

	/**
	 * The controller of the object that owns the tar file list currently being
	 * manipulated. XXX We need this so we can notify the object that the tar
	 * file list has changed.
	 */
	public Object objController;

	/**
	 * Auto fill controller for the directory combo box.
	 */
	public final AutoFillController dc;

	/**
	 * Auto fill controller for the tar file combo box.
	 */
	public final AutoFillController tc;

	/**
	 * Construct a TarFileListController with the given value.
	 * 
	 * @param netlab
	 *            The NetlabClient instance this controller is attached to.
	 */
	public TarFileListController(NetlabClient netlab) {
		super(netlab);

		this.dc = new AutoFillController(netlab, new Vector(DST_AUTO_FILL));
		this.tc = new FileAutoFillController(netlab);

		this.setObjectClass(NSTarFile.class);
	}

	/**
	 * Method called when this controller is activated by another component.
	 * Currently, it will get the list of tar files to manipulate from the
	 * triggering component and make a copy to be this controller's content
	 * array.
	 * 
	 * @see net.emulab.netlab.client.Controller#triggeredBy(java.lang.Object)
	 */
	public Controller triggeredBy(Object component) throws Exception {
		/* Get the data set to manipulate and */
		this.objController = this.netlab
				.getProperty(component, "objcontroller");
		this.ts = (TreeSet) this.tkv.getKeyValueAsSingleton(this.objController,
				"selection.tarFiles");

		/* ... make a copy to be our contentArray. */
		try {
			List ca;

			ca = ArrayController.deepCopy(NSTarFile.class, this.ts);
			this.tkv.setKeyValue(this, "content", ca);
		} catch (CloneNotSupportedException e) {
			throw new InternalError();
		}

		super.triggeredBy(component);

		return this;
	}
	
	public void showDialog() {
		if (!this.synced) {
			this.syncTarFiles();
		}
	}
	
	private void syncTarFiles() {
		if (syncing)
			return;
		
		Thread th = new Thread() {
			public void run() {
				try {
					Object oldpo;
					
					EmulabFSProxy fs;
					List exports;
					Iterator iter;

					tkv.setKeyValue(TarFileListController.this, "syncing", true);
					
					oldpo = netlab.getPoolValue("server.observer");
					netlab.setPoolValue("server.observer", NullProgressObserver.SINGLETON);
					fs = new EmulabFSProxy(tkv, netlab.getPool());
					netlab.setPoolValue("server.observer", oldpo);
					
					exports = fs.exports();
					for (iter = exports.iterator(); iter.hasNext(); ) {
						try {
							String path = (String)iter.next();
							
							path += "/tarfiles";
							if (fs.access(path, "execute")) {
								DirectoryEntry de[] = fs.listdir(path);
								int lpc;
								
								for (lpc = 0; lpc < de.length; lpc++) {
									String fullPath = path + "/" + de[lpc].name;
									
									if (de[lpc].type == 'f' && isTarFile(de[lpc].name)) {
										tc.addObject(fullPath);
									}
								}
							}
						}
						catch (EmulabXmlRpcException e) {
							e.printStackTrace();
						}
					}
					synced = true;
				}
				catch (Throwable th) {
					netlab.handleException(th);
				}
				finally {
					tkv.setKeyValue(TarFileListController.this, "syncing", false);
				}
			}
		};
		
		th.setDaemon(true);
		th.start();
	}
	
	/**
	 * Method called when the "Ok" button is clicked, it will verify the data,
	 * update the model, and close the window if everything checks out.
	 * Otherwise, it will popup an error dialog and ask the user to fix any
	 * problems.
	 */
	public void ok() {
		List list = this.contentArray();
		boolean good2go = true;
		int lpc, size;

		/* Check for broken items in the list, */
		size = list.size();
		for (lpc = 0; (lpc < size) && good2go; lpc++) {
			NSTarFile nstf = (NSTarFile) list.get(lpc);
			boolean badDir, badFile;

			badDir = "".equals(nstf.getDestination());
			badFile = "".equals(nstf.getTarFile());
			if (badDir || badFile) {
				this.tkv.setKeyValue(this, "selection", nstf);
				this.netlab.getNoticeController().showDialog("Incomplete entry, missing "
						+ (badDir ? "destination." : "tar file name."));
				good2go = false;
			}
		}

		if (good2go) {
			TreeSet nts = new TreeSet();
			
			/* Out with the old and */
			this.ts.clear();
			/* ... in with the new. */
			nts.addAll((Collection) this.contentArray());

			this.tkv.setKeyValue(this.objController, "selection.tarFiles", nts);

			this.netlab.remove(this.view);
			this.ts = null;
		}
	}

	/**
	 * Method called when the "Cancel" button is clicked, just clears some state
	 * and closes the window.
	 */
	public void cancel() {
		this.netlab.remove(this.view);
		this.ts = null;
	}

	/**
	 * @see java.lang.Object#toString()
	 */
	public String toString() {
		return "TarFileListController[" + super.toString() + "; ts=" + this.ts
				+ "; dirComponent=" + this.dirComponent + "; fileComponent="
				+ this.fileComponent + "]";
	}
}
