/*
 * EmulabExperimentProxy.java
 *
 * Copyright (c) 2005 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

package net.emulab.netlab.client.emulab;

import java.io.IOException;
import java.net.MalformedURLException;
import java.util.Hashtable;
import java.util.List;
import java.util.Vector;

import net.emulab.netlab.client.ExperimentParameters;

import org.apache.xmlrpc.XmlRpcException;

import thinlet.ObjectPool;
import thinlet.ThinletKeyValue;

/**
 * Client-side XML-RPC proxy for accessing the Emulab "experiment" module.
 */
public class EmulabExperimentProxy extends EmulabProxy {
	/**
	 * The "startexp" method name.
	 */
	private static final String METHOD_STARTEXP = "experiment.startexp";

	/**
	 * The "endexp" method name.
	 */
	private static final String METHOD_ENDEXP = "experiment.endexp";

	/**
	 * The "swapexp" method name.
	 */
	private static final String METHOD_SWAPEXP = "experiment.swapexp";

	/**
	 * The "nscheck" method name.
	 */
	private static final String METHOD_NSCHECK = "experiment.nscheck";

	/**
	 * The "state" method name.
	 */
	private static final String METHOD_STATE = "experiment.state";

	/**
	 * The "info" method name.
	 */
	private static final String METHOD_INFO = "experiment.info";

	/**
	 * The "getlist" method name.
	 */
	private static final String METHOD_GETLIST = "experiment.getlist";

	/**
	 * The "virtual_topology" method name.
	 */
	private static final String METHOD_VIRTUAL_TOPOLOGY = "experiment.virtual_topology";

	/**
	 * The "delay_config" method name.
	 */
	private static final String METHOD_DELAY_CONFIG = "experiment.delay_config";

	/**
	 * The "getviz" method name.
	 */
	private static final String METHOD_GETVIZ = "experiment.getviz";

	/**
	 * The "thumbnail" method name.
	 */
	private static final String METHOD_THUMBNAIL = "experiment.thumbnail";

	/**
	 * The "statewait" method name.
	 */
	private static final String METHOD_STATEWAIT = "experiment.statewait";

	private static final String METHOD_METADATA = "experiment.metadata";
	
	private static final String METHOD_NSFILE = "experiment.nsfile";
	
	private static final String METHOD_MODIFY = "experiment.modify";

	private static final String METHOD_HISTORY = "experiment.history";

	/**
	 * Construct an EmulabUserProxy with the given values.
	 * 
	 * @param nbp
	 *            The NetlabPreferences to use when constructing the URL for the
	 *            RPC server.
	 * @param po
	 *            The ProgressObserver that is interested in this proxy.
	 * @throws MalformedURLException
	 *             if the preferences specified by the user form a bad URL.
	 */
	public EmulabExperimentProxy(ThinletKeyValue tkv, ObjectPool op)
			throws MalformedURLException {
		super(tkv, op);
	}

	/**
	 * Start a new experiment in Emulab.
	 * 
	 * @param ep
	 *            The experiment parameters (e.g. pid, gid, eid, etc...).
	 * @throws XmlRpcException
	 *             if there was an error while trying to execute the method.
	 * @throws IOException
	 *             if there was an error while communicating with the server.
	 */
	public synchronized void startexp(ExperimentParameters ep)
			throws XmlRpcException, IOException {
		this._invoke(METHOD_STARTEXP, ep);
	}

	/**
	 * Terminate an experiment in Emulab.
	 * 
	 * @param proj
	 *            The project where the experiment is located.
	 * @param exp
	 *            The name of the experiment.
	 * @param wait
	 *            Wait for the termination to complete before returning.
	 * @throws XmlRpcException
	 *             if there was an error while trying to execute the method.
	 * @throws IOException
	 *             if there was an error while communicating with the server.
	 */
	public synchronized void endexp(String proj, String exp, boolean wait)
			throws XmlRpcException, IOException {
		Hashtable args;

		args = new Hashtable();
		args.put("proj", proj);
		args.put("exp", exp);
		args.put("wait", wait ? "true" : "false");
		this._invoke(METHOD_ENDEXP, args);
	}

	/**
	 * Swap in/out an experiment.
	 * 
	 * @param proj
	 *            The project where the experiment is located.
	 * @param exp
	 *            The name of the experiment.
	 * @param direction
	 *            The swap direction, "in" or "out".
	 * @param wait
	 *            Wait for the swap to complete before returning.
	 * @throws XmlRpcException
	 *             if there was an error while trying to execute the method.
	 * @throws IOException
	 *             if there was an error while communicating with the server.
	 */
	public synchronized void swapexp(String proj, String exp, String direction,
			boolean wait) throws XmlRpcException, IOException {
		Hashtable args;

		args = new Hashtable();
		args.put("proj", proj);
		args.put("exp", exp);
		args.put("direction", direction);
		args.put("wait", wait ? "true" : "false");
		this._invoke(METHOD_SWAPEXP, args);
	}

	/**
	 * Get the state of the experiment.
	 * 
	 * @param proj
	 *            The project where the experiment is located.
	 * @param exp
	 *            The name of the experiment.
	 * @return The state of the experiment (e.g. "swapped", "activating",
	 *         "active").
	 * @throws XmlRpcException
	 *             if there was an error while trying to execute the method.
	 * @throws IOException
	 *             if there was an error while communicating with the server.
	 */
	public synchronized String state(String proj, String exp)
			throws XmlRpcException, IOException {
		Hashtable args;
		String retval;
		Object rc;

		if (proj == null)
			throw new IllegalArgumentException("proj is null");
		if (exp == null)
			throw new IllegalArgumentException("exp is null");

		args = new Hashtable();
		args.put("proj", proj);
		args.put("exp", exp);

		rc = this._invoke(METHOD_STATE, args);

		if (!(rc instanceof String)) {
			throw new XmlRpcException(0, "Bad response: " + rc);
		}

		retval = (String) rc;

		return retval;
	}

	/**
	 * Perform a syntax check on an NS file.
	 * 
	 * @param nsfilestr
	 *            The NS file to check.
	 * @return True if the check succeeded, false otherwise ...
	 * @throws XmlRpcException
	 *             if there was an error while trying to execute the method.
	 * @throws IOException
	 *             if there was an error while communicating with the server.
	 */
	public synchronized boolean nscheck(String nsfilestr)
			throws XmlRpcException, IOException {
		Hashtable args = new Hashtable();
		boolean retval = true;

		args.put("nsfilestr", nsfilestr);
		try {
			this._invoke(METHOD_NSCHECK, args);
		} catch (EmulabXmlRpcException e) {
			if (e.getCode() != EmulabResponse.ERROR) {
				throw e;
			} else {
				e.printStackTrace(System.err);
				retval = false;
			}
		}
		return retval;
	}

	/**
	 * Get information about an experiment.
	 * 
	 * @param proj
	 *            The project where the experiment is located.
	 * @param exp
	 *            The name of the experiment.
	 * @param aspect
	 *            The aspect of the experiment to get information on, either
	 *            "mapping" or "links".
	 * @return A table describing the describing.
	 * @throws XmlRpcException
	 *             if there was an error while trying to execute the method.
	 * @throws IOException
	 *             if there was an error while communicating with the server.
	 */
	public synchronized Hashtable info(String proj, String exp, String aspect)
			throws XmlRpcException, IOException {
		Hashtable args, retval;
		Object rc;

		if (proj == null)
			throw new IllegalArgumentException("proj is null");
		if (exp == null)
			throw new IllegalArgumentException("exp is null");

		args = new Hashtable();
		args.put("proj", proj);
		args.put("exp", exp);
		args.put("aspect", aspect);

		rc = this._invoke(METHOD_INFO, args);

		if (!(rc instanceof Hashtable)) {
			throw new XmlRpcException(0, "Bad response: " + rc);
		}

		retval = (Hashtable) rc;

		return retval;
	}

	/**
	 * Get the list of experiments where this user is the head.
	 * 
	 * @param format
	 *            Specifies the amount of format of the information to be
	 *            returned:
	 *            <ul>
	 *            <li>"brief" - Only retrieve the pid/gid/eid for each
	 *            experiment.
	 *            <li>"full" - Retrieve everything about the experiment.
	 *            </ul>
	 * @return A table indexed by project name that refers to a table indexed by
	 *         group name that contains a vector of experiment names.
	 * @throws XmlRpcException
	 *             if there was an error while trying to execute the method.
	 * @throws IOException
	 *             if there was an error while communicating with the server.
	 */
	public synchronized Hashtable getlist(String format)
			throws XmlRpcException, IOException {
		Hashtable args, retval;
		Object rc;

		args = new Hashtable();
		args.put("format", format);
		rc = this._invoke(METHOD_GETLIST, args);
		if (!(rc instanceof Hashtable)) {
			throw new XmlRpcException(0, "Bad response: " + rc);
		}

		retval = (Hashtable) rc;

		return retval;
	}

	/**
	 * Get the list of experiments where this user is the head and the result
	 * should be in the "brief" format.
	 * 
	 * @return A table indexed by project name that refers to a table indexed by
	 *         group name that contains a vector of experiment names.
	 * @throws XmlRpcException
	 *             if there was an error while trying to execute the method.
	 * @throws IOException
	 *             if there was an error while communicating with the server.
	 */
	public synchronized Hashtable getlist() throws XmlRpcException, IOException {
		return this.getlist("brief");
	}

	/**
	 * Get the "virtual" topology of an experiment.
	 * 
	 * @param proj
	 *            The project where the experiment is located.
	 * @param exp
	 *            The name of the experiment.
	 * @throws XmlRpcException
	 *             if there was an error while trying to execute the method.
	 * @throws IOException
	 *             if there was an error while communicating with the server.
	 */
	public synchronized Hashtable virtual_topology(String proj, String exp, List exclude)
			throws XmlRpcException, IOException {
		Hashtable args, retval;
		Vector vector;
		Object rc;

		args = new Hashtable();
		args.put("proj", proj);
		args.put("exp", exp);
		args.put("exclude", exclude);
		rc = this._invoke(METHOD_VIRTUAL_TOPOLOGY, args);
		if (!(rc instanceof Vector)) {
			throw new XmlRpcException(0, "Bad response: " + rc);
		}

		vector = (Vector) rc;

		retval = (Hashtable) vector.elementAt(0);
		retval = (Hashtable) retval.get("experiment");

		return retval;
	}

	/**
	 * Change the delay configuration for a link.
	 * 
	 * @param proj
	 *            The project where the experiment is located.
	 * @param exp
	 *            The name of the experiment.
	 * @param link
	 *            The name of the link to change.
	 * @param params
	 *            The link parameters to change and their new values.
	 * @param persist
	 *            True if the changes should persist across swapins or False if
	 *            the changes should only be for this instantiation.
	 */
	public synchronized void delay_config(String proj, String exp, String link,
			Hashtable params, boolean persist) throws XmlRpcException,
			IOException {
		Hashtable args;

		args = new Hashtable();
		args.put("proj", proj);
		args.put("exp", exp);
		args.put("link", link);
		args.put("params", params);
		args.put("persist", Boolean.valueOf(persist));
		this._invoke(METHOD_DELAY_CONFIG, args);
	}

	/**
	 * Get the visualization data for an experiment. This data is the two
	 * dimensional layout of the nodes and lans in an experiment.
	 * 
	 * @param proj
	 *            The project where the experiment is located.
	 * @param exp
	 *            The name of the experiment.
	 */
	public synchronized Hashtable getviz(String proj, String exp)
			throws XmlRpcException, IOException {
		Hashtable args, retval;
		Object rc;

		args = new Hashtable();
		args.put("proj", proj);
		args.put("exp", exp);
		rc = this._invoke(METHOD_GETVIZ, args);
		if (!(rc instanceof Hashtable)) {
			throw new XmlRpcException(0, "Bad response: " + rc);
		}

		retval = (Hashtable) rc;

		return retval;
	}

	/**
	 * Get the thumbnail image for an experiment.
	 * 
	 * @param proj
	 *            The project where the experiment is located.
	 * @param exp
	 *            The name of the experiment.
	 * @return A PNG image.
	 */
	public synchronized byte[] thumbnail(String proj, String exp)
			throws XmlRpcException, IOException {
		Hashtable args;
		byte retval[];
		Object rc;

		args = new Hashtable();
		args.put("proj", proj);
		args.put("exp", exp);
		rc = this._invoke(METHOD_THUMBNAIL, args);
		if (!(rc instanceof byte[])) {
			throw new XmlRpcException(0, "Bad response: " + rc);
		}

		retval = (byte[]) rc;

		return retval;
	}

	public synchronized String statewait(String proj, String exp,
			String states[], int timeout) throws XmlRpcException, IOException {
		Hashtable args;
		String retval;
		Object rc;

		args = new Hashtable();
		args.put("proj", proj);
		args.put("exp", exp);
		args.put("state", states);
		if (timeout != -1)
			args.put("timeout", new Integer(timeout));
		rc = this._invoke(METHOD_STATEWAIT, args);
		if (!(rc instanceof String)) {
			throw new XmlRpcException(0, "Bad response: " + rc);
		}

		retval = (String) rc;

		return retval;
	}

	public synchronized String statewait(String proj, String exp,
			String states[]) throws XmlRpcException, IOException {
		return this.statewait(proj, exp, states, -1);
	}

	public Hashtable metadata(String proj, String exp) throws XmlRpcException, IOException {
		Hashtable args, retval;
		Object rc;

		args = new Hashtable();
		args.put("proj", proj);
		args.put("exp", exp);
		rc = this._invoke(METHOD_METADATA, args);
		if (!(rc instanceof Hashtable)) {
			throw new XmlRpcException(0, "Bad response: " + rc);
		}

		retval = (Hashtable) rc;

		return retval;
	}

	public String nsfile(String proj, String exp) throws XmlRpcException, IOException {
		Hashtable args;
		String retval;
		Object rc;

		args = new Hashtable();
		args.put("proj", proj);
		args.put("exp", exp);
		rc = this._invoke(METHOD_NSFILE, args);
		if (!(rc instanceof String)) {
			throw new XmlRpcException(0, "Bad response: " + rc);
		}

		retval = (String) rc;

		return retval;
	}

	public void modify(String proj, String exp, String nsfile, boolean wait) throws XmlRpcException, IOException {
		Hashtable args;

		args = new Hashtable();
		args.put("proj", proj);
		args.put("exp", exp);
		args.put("nsfilestr", nsfile);
		args.put("wait", Boolean.valueOf(wait));
		this._invoke(METHOD_MODIFY, args);
	}

	public Hashtable history(String proj, String exp) throws XmlRpcException, IOException {
		Hashtable args, retval;
		Object rc;

		args = new Hashtable();
		args.put("proj", proj);
		args.put("exp", exp);
		rc = this._invoke(METHOD_HISTORY, args);
		if (!(rc instanceof Hashtable)) {
			throw new XmlRpcException(0, "Bad response: " + rc);
		}
		
		retval = (Hashtable) rc;
		
		return retval;
	}
	
	/**
	 * @see java.lang.Object#toString()
	 */
	public String toString() {
		return "EmulabExperimentProxy[" + super.toString() + "]";
	}
}
