/*
 * EmulabFSProxy.java
 *
 * Copyright (c) 2005 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */
package net.emulab.netlab.client.emulab;

import java.io.IOException;
import java.net.MalformedURLException;
import java.util.Hashtable;
import java.util.Vector;

import net.emulab.netlab.client.DirectoryEntry;

import org.apache.xmlrpc.XmlRpcException;

import thinlet.ObjectPool;
import thinlet.ThinletKeyValue;

/**
 * Client-side XML-RPC proxy for accessing the Emulab "user" module.
 */
public class EmulabFSProxy extends EmulabProxy {
	/**
	 * The "access" method name.
	 */
	private static final String METHOD_ACCESS = "fs.access";

	/**
	 * The "listdir" method name.
	 */
	private static final String METHOD_LISTDIR = "fs.listdir";

	/**
	 * THe "exports" method name.
	 */
	private static final String METHOD_EXPORTS = "fs.exports";

	/**
	 * Construct an EmulabFSProxy with the given values.
	 * 
	 * @param nbp
	 *            The NetlabPreferences to use when constructing the URL for the
	 *            RPC server.
	 * @param po
	 *            The ProgressObserver that is interested in this proxy.
	 * @throws MalformedURLException
	 *             if the preferences specified by the user form a bad URL.
	 */
	public EmulabFSProxy(ThinletKeyValue tkv, ObjectPool op)
			throws MalformedURLException {
		super(tkv, op);
	}

	/**
	 * Perform an access check on a path that should be exported to the
	 * experimental nodes via NFS.
	 * 
	 * @param path
	 *            The path to check.
	 * @param permission
	 *            The permission to check for, one of: read, write, execute, and
	 *            exists.
	 * @return True if the file is accessible, false otherwise.
	 * @throws XmlRpcException
	 *             if there was an error while trying to execute the method.
	 * @throws IOException
	 *             if there was an error while communicating with the server.
	 */
	public boolean access(String path, String permission)
			throws XmlRpcException, IOException {
		boolean retval;
		Hashtable args;
		Object rc;

		if (path == null)
			throw new IllegalArgumentException("path is null");
		if (permission == null)
			throw new IllegalArgumentException("permission is null");

		args = new Hashtable();
		args.put("path", path);
		args.put("permission", permission);
		rc = this._invoke(METHOD_ACCESS, args);
		if (!(rc instanceof Boolean)) {
			throw new XmlRpcException(0, "Bad response: " + rc);
		}

		retval = ((Boolean) rc).booleanValue();

		return retval;
	}

	/**
	 * Perform a read access check on a path that should be exported to the
	 * experimental nodes via NFS.
	 * 
	 * @param path
	 *            The path to check.
	 * @return True if the file is readable, false otherwise.
	 * @throws XmlRpcException
	 *             if there was an error while trying to execute the method.
	 * @throws IOException
	 *             if there was an error while communicating with the server.
	 */
	public boolean access(String path) throws XmlRpcException, IOException {
		return this.access(path, "r");
	}

	/**
	 * Get a directory listing for an NFS exported path on Emulab.
	 * 
	 * @param path
	 *            The directory path to query.
	 * @return An array of DirectoryEntry's that describe the contents of the
	 *         given path.
	 * @throws XmlRpcException
	 *             if there was an error while trying to execute the method.
	 * @throws IOException
	 *             if there was an error while communicating with the server.
	 */
	public synchronized DirectoryEntry[] listdir(String path)
			throws XmlRpcException, IOException {
		DirectoryEntry retval[];
		Hashtable args;
		Vector entries;
		Object rc;
		int lpc;

		if (path == null)
			throw new IllegalArgumentException("path is null");

		args = new Hashtable();
		args.put("path", path);
		rc = this._invoke(METHOD_LISTDIR, args);
		if (!(rc instanceof Vector)) {
			throw new XmlRpcException(0, "Bad response: " + rc);
		}

		entries = (Vector) rc;
		retval = new DirectoryEntry[entries.size()];
		for (lpc = 0; lpc < retval.length; lpc++) {
			Vector entry = (Vector) entries.elementAt(lpc);

			retval[lpc] = new DirectoryEntry(entry);
		}

		return retval;
	}

	/**
	 * Get a listing of all the directories that may be exported to the user's
	 * experimental nodes.
	 * 
	 * @return A vector of directory names.
	 * @throws XmlRpcException
	 *             if there was an error while trying to execute the method.
	 * @throws IOException
	 *             if there was an error while communicating with the server.
	 */
	public synchronized Vector exports() throws XmlRpcException, IOException {
		Vector retval;
		Object rc;

		rc = this._invoke(METHOD_EXPORTS);
		if (!(rc instanceof Vector)) {
			throw new XmlRpcException(0, "Bad response: " + rc);
		}

		retval = (Vector) rc;

		return retval;
	}

	/**
	 * @see java.lang.Object#toString()
	 */
	public String toString() {
		return "EmulabFSProxy[" + super.toString() + "]";
	}
}
