/*
 * EmulabNodeProxy.java
 *
 * Copyright (c) 2005 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */
package net.emulab.netlab.client.emulab;

import java.io.IOException;
import java.net.MalformedURLException;
import java.util.Hashtable;

import org.apache.xmlrpc.XmlRpcException;

import thinlet.ObjectPool;
import thinlet.ThinletKeyValue;

/**
 * Client-side XML-RPC proxy for accessing the Emulab "node" module.
 */
public class EmulabNodeProxy extends EmulabProxy {
	/**
	 * The "available" method name.
	 */
	private static final String METHOD_AVAILABLE = "node.available";

	/**
	 * The "reboot" method name.
	 */
	private static final String METHOD_REBOOT = "node.reboot";

	/**
	 * The "console" method name.
	 */
	private static final String METHOD_CONSOLE = "node.console";

	/**
	 * The "hostkeys" method name.
	 */
	private static final String METHOD_HOSTKEYS = "node.hostkeys";

	private static final String METHOD_SSHDESCRIPTION = "node.sshdescription";
	
	private static final String METHOD_TYPEINFO = "node.typeinfo";

	/**
	 * Construct an EmulabNodeProxy with the given values.
	 * 
	 * @param nbp
	 *            The NetlabPreferences to use when constructing the URL for the
	 *            RPC server.
	 * @param po
	 *            The ProgressObserver that is interested in this proxy.
	 * @throws MalformedURLException
	 *             if the preferences specified by the user form a bad URL.
	 */
	public EmulabNodeProxy(ThinletKeyValue tkv, ObjectPool op)
			throws MalformedURLException {
		super(tkv, op);
	}

	/**
	 * @return The number of free PCs.
	 * @throws XmlRpcException
	 *             if there was an error while trying to execute the method.
	 * @throws IOException
	 *             if there was an error while communicating with the server.
	 */
	public synchronized int available() throws XmlRpcException, IOException {
		int retval;
		Object rc;

		rc = this._invoke(METHOD_AVAILABLE);
		if (!(rc instanceof Integer)) {
			throw new XmlRpcException(0, "Bad response: " + rc);
		}

		retval = ((Integer) rc).intValue();

		return retval;
	}

	/**
	 * Reboot a node.
	 * 
	 * @param node
	 *            The node ID to reboot. This should be something like "pc123"
	 *            and not the name of the node in the experiment.
	 * @param wait
	 *            Wait for the reboot to complete.
	 * @throws XmlRpcException
	 *             if there was an error while trying to execute the method.
	 * @throws IOException
	 *             if there was an error while communicating with the server.
	 */
	public synchronized void reboot(String node, boolean wait)
			throws XmlRpcException, IOException {
		Hashtable args;

		if (node == null)
			throw new IllegalArgumentException("node is null");

		args = new Hashtable();
		args.put("nodes", node);
		args.put("wait", Boolean.valueOf(wait));

		this._invoke(METHOD_REBOOT, args);
	}

	/**
	 * Get the console ACL for a node.
	 * 
	 * @param node
	 *            The node ID to get the ACL for. This should be something like
	 *            "pc123" and not the name of the node in the experiment.
	 * @return A table containing the ACL values.
	 * @throws XmlRpcException
	 *             if there was an error while trying to execute the method.
	 * @throws IOException
	 *             if there was an error while communicating with the server.
	 */
	public synchronized Hashtable console(String node) throws XmlRpcException,
			IOException {
		Hashtable args, retval;
		Object rc;

		args = new Hashtable();
		args.put("node", node);
		rc = this._invoke(METHOD_CONSOLE, args);

		if (!(rc instanceof Hashtable)) {
			throw new XmlRpcException(0, "Bad response: " + rc);
		}

		retval = (Hashtable) rc;

		return retval;
	}

	public synchronized Hashtable hostkeys(String node) throws XmlRpcException,
			IOException {
		Hashtable args, retval;
		Object rc;

		args = new Hashtable();
		args.put("node", node);
		rc = this._invoke(METHOD_HOSTKEYS, args);

		if (!(rc instanceof Hashtable)) {
			throw new XmlRpcException(0, "Bad response: " + rc);
		}

		retval = (Hashtable) rc;

		return retval;
	}

	public synchronized Hashtable sshdescription(String node)
			throws XmlRpcException, IOException {
		Hashtable args, retval;
		Object rc;

		args = new Hashtable();
		args.put("node", node);
		rc = this._invoke(METHOD_SSHDESCRIPTION, args);

		if (!(rc instanceof Hashtable)) {
			throw new XmlRpcException(0, "Bad response: " + rc);
		}

		retval = (Hashtable) rc;

		return retval;
	}

	public synchronized Hashtable typeinfo()
			throws XmlRpcException, IOException {
		Hashtable retval;
		Object rc;

		rc = this._invoke(METHOD_TYPEINFO);

		if (!(rc instanceof Hashtable)) {
			throw new XmlRpcException(0, "Bad response: " + rc);
		}

		retval = (Hashtable) rc;

		return retval;
	}

	/**
	 * @see java.lang.Object#toString()
	 */
	public String toString() {
		return "EmulabNodeProxy[" + super.toString() + "]";
	}
}
