/*
 * EmulabProxy.java
 *
 * Copyright (c) 2005 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

package net.emulab.netlab.client.emulab;

import java.io.IOException;
import java.net.MalformedURLException;
import java.util.Hashtable;
import java.util.Vector;

import org.apache.xmlrpc.XmlRpcClient;
import org.apache.xmlrpc.XmlRpcException;

import thinlet.ObjectPool;
import thinlet.ThinletKeyValue;

/**
 * Base class for client side proxies of the Emulab RPC server.
 */
public class EmulabProxy {
	/**
	 * Emulab XML-RPC version number.
	 */
	private static final Double XMLRPC_VERSION = new Double(0.1);

	/**
	 * Hashtable used for methods that take no arguments.
	 */
	protected static final Hashtable EMPTY_ARGS = new Hashtable();

	/**
	 * Client connection to the server.
	 */
	protected final XmlRpcClient xrc;

	/**
	 * Construct an EmulabProxy from the given values.
	 */
	public EmulabProxy(ThinletKeyValue tkv, ObjectPool op) {
		this.xrc = (XmlRpcClient)tkv.getKeyValue(op, "xmlrpcClientFactory.whitehole");
	}

	/**
	 * Invoke a method on the server.
	 * 
	 * @param methodName
	 *            The name of the method.
	 * @param args
	 *            The arguments to pass to the method.
	 * @return The "value" part of the EmulabResponse object.
	 * @throws XmlRpcException
	 *             if there was an error while executing the method.
	 * @throws IOException
	 *             if there was an error while communicating with the server.
	 */
	protected Object _invoke(String methodName, Hashtable args)
			throws XmlRpcException, IOException {
		Vector argv = new Vector();
		Object rc, retval;

		if (methodName == null)
			throw new IllegalArgumentException("methodName is null");

		argv.setSize(2);
		argv.set(0, XMLRPC_VERSION);
		argv.set(1, args);

		rc = this.xrc.execute(methodName, argv);
		retval = EmulabResponse.parseResponse(rc);

		return retval;
	}

	/**
	 * Invoke a method on the server that takes no arguments.
	 * 
	 * @param methodName
	 *            The name of the method.
	 * @return The "value" part of the EmulabResponse object.
	 * @throws XmlRpcException
	 *             if there was an error while executing the method.
	 * @throws IOException
	 *             if there was an error while communicating with the server.
	 */
	protected Object _invoke(String methodName) throws XmlRpcException,
			IOException {
		return this._invoke(methodName, EMPTY_ARGS);
	}

	/**
	 * @see java.lang.Object#toString()
	 */
	public String toString() {
		return "EmulabProxy[xrc=" + this.xrc + "]";
	}
}
