/*
 * EmulabResponse.java
 *
 * Copyright (c) 2005 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */
package net.emulab.netlab.client.emulab;

import java.util.Hashtable;

import org.apache.xmlrpc.XmlRpcException;

/**
 * Static-only class used to parse the EmulabResponse objects returned by Emulab
 * XML-RPC methods.
 */
public class EmulabResponse {
	/**
	 * The method executed successfully.
	 */
	public static final int SUCCESS = 0;

	/**
	 * The arguments to the method were malformed.
	 */
	public static final int BADARGS = 1;

	/**
	 * There was an error while executing the method.
	 */
	public static final int ERROR = 2;

	/**
	 * The user does not have privileges to execute the method.
	 */
	public static final int FORBIDDEN = 3;

	/**
	 * The request has a bad version number.
	 */
	public static final int BADVERSION = 4;

	/**
	 * The server experienced an error while executing the method.
	 */
	public static final int SERVERERROR = 5;

	/**
	 * One or more arguments to the method is too large.
	 */
	public static final int TOOBIG = 6;

	/**
	 * There was a service outage.
	 */
	public static final int REFUSED = 7;

	/**
	 * The operation timed out.
	 */
	public static final int TIMEDOUT = 8;

	/**
	 * String mappings for the response integer values.
	 */
	protected static final String STRINGS[] = { "Success", "Bad Arguments",
			"Error", "Forbidden", "Bad Version", "Server Error",
			"Request Too Big", "Refused", "Operation Timed Out", };

	/**
	 * Convert a response code to a human readable string.
	 * 
	 * @param code
	 *            The response code to convert.
	 */
	public static String codeToString(int code) {
		String retval;

		if ((code < 0) || (code >= STRINGS.length)) {
			retval = "Unknown";
		} else {
			retval = STRINGS[code];
		}
		return retval;
	}

	/**
	 * Handle a response from an Emulab XML-RPC method.
	 * 
	 * @param response
	 *            The object returned by the method.
	 * @return The "value" portion of the response.
	 * @throws XmlRpcException
	 *             if the response was not well formed.
	 * @throws EmulabXmlRpcException
	 *             if the response was something other than #SUCCESS.
	 */
	public static Object parseResponse(Object response) throws XmlRpcException {
		Object code, retval;
		Hashtable ht;

		if (response == null) {
			throw new XmlRpcException(0, "No response");
		}

		if (!(response instanceof Hashtable)
				|| ((code = ((Hashtable) response).get("code")) == null)
				|| !(code instanceof Integer)) {
			throw new XmlRpcException(0, "Malformed Response: " + response);
		}

		ht = (Hashtable) response;

		if (((Integer) code).intValue() != EmulabResponse.SUCCESS) {
			throw new EmulabXmlRpcException(((Integer) code).intValue(), ht
					.get("output"));
		}

		retval = ht.get("value");
		return retval;
	}

	/**
	 * Private constructor, this is a static-only class at the moment.
	 */
	private EmulabResponse() {
	}
}
