/*
 * EmulabUserProxy.java
 *
 * Copyright (c) 2005 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */
package net.emulab.netlab.client.emulab;

import java.io.IOException;
import java.net.MalformedURLException;
import java.util.Hashtable;

import org.apache.xmlrpc.XmlRpcException;

import thinlet.ObjectPool;
import thinlet.ThinletKeyValue;

/**
 * Client-side XML-RPC proxy for accessing the Emulab "user" module.
 */
public class EmulabUserProxy extends EmulabProxy {
	/**
	 * String that denotes permission to read information about a project/group.
	 */
	public static final String PERMISSION_READ_INFO = "readinfo";

	/**
	 * String that denotes permission to make a new group.
	 */
	public static final String PERMISSION_MAKE_GROUP = "makegroup";

	/**
	 * String that denotes permission to create a new experiment.
	 */
	public static final String PERMISSION_CREATE_EXPERIMENT = "createexpt";

	/**
	 * String that denotes permission to make a new OS ID.
	 */
	public static final String PERMISSION_MAKE_OSID = "makeosid";

	/**
	 * String that denotes permission to make a new Image ID.
	 */
	public static final String PERMISSION_MAKE_IMAGEID = "makeimageid";

	/**
	 * The "membership" method name.
	 */
	private static final String METHOD_MEMBERSHIP = "user.membership";

	/**
	 * The "nodecount" method name.
	 */
	private static final String METHOD_NODECOUNT = "user.nodecount";

	/**
	 * Construct an EmulabUserProxy with the given values.
	 * 
	 * @param nbp
	 *            The NetlabPreferences to use when constructing the URL for the
	 *            RPC server.
	 * @param po
	 *            The ProgressObserver that is interested in this proxy.
	 * @throws MalformedURLException
	 *             if the preferences specified by the user form a bad URL.
	 */
	public EmulabUserProxy(ThinletKeyValue tkv, ObjectPool op)
			throws MalformedURLException {
		super(tkv, op);
	}

	/**
	 * Get the user's project/group membership where they are allowed to perform
	 * some operation.
	 * 
	 * @param permission
	 *            The operation the user wishes to perform.
	 * @return The membership table, indexed by project names that refer to
	 *         vectors of group names.
	 * @throws XmlRpcException
	 *             if there was an error while trying to execute the method.
	 * @throws IOException
	 *             if there was an error while communicating with the server.
	 */
	public synchronized Hashtable membership(String permission)
			throws XmlRpcException, IOException {
		Hashtable args = EMPTY_ARGS, retval;
		Object rc;

		if (permission != null) {
			args = new Hashtable();
			args.put("permission", permission);
		}
		rc = this._invoke(METHOD_MEMBERSHIP, args);
		if (!(rc instanceof Hashtable)) {
			throw new XmlRpcException(0, "Bad response: " + rc);
		}

		retval = (Hashtable) rc;

		return retval;
	}

	/**
	 * @return The number of nodes currently allocated by this user.
	 * @throws XmlRpcException
	 *             if there was an error while trying to execute the method.
	 * @throws IOException
	 *             if there was an error while communicating with the server.
	 */
	public synchronized int nodecount() throws XmlRpcException, IOException {
		int retval;
		Object rc;

		rc = this._invoke(METHOD_NODECOUNT);
		if (!(rc instanceof Integer)) {
			throw new XmlRpcException(0, "Bad response: " + rc);
		}

		retval = ((Integer) rc).intValue();

		return retval;
	}

	/**
	 * @see java.lang.Object#toString()
	 */
	public String toString() {
		return "EmulabUserProxy[" + super.toString() + "]";
	}
}
