/*
 * DBState.java
 *
 * Copyright (c) 2005 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

package net.emulab.ns;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.List;
import java.util.StringTokenizer;
import java.util.Vector;
import java.io.IOException;

import net.emulab.util.FileDiff;
import thinlet.ThinletController;
import thinlet.ThinletKeyValue;

/**
 * Class used to convert between an NSTopology and the Emulab DB version of a
 * virtual topology.
 */
public class DBState {
	private static final String OBJECT_TYPES[] = {
		"vtypes", "nodes", "lans", "programs", "trafgens", "event_groups",
		"lan_settings", "events",
	};
	
	private static final Hashtable ROUTING_PROTOCOL_MAP = new Hashtable();
	
	static {
		ROUTING_PROTOCOL_MAP.put("static", "Static");
		ROUTING_PROTOCOL_MAP.put("ospf", "Session");
		ROUTING_PROTOCOL_MAP.put("manual", "Manual");
		ROUTING_PROTOCOL_MAP.put("static-old", "Static-old");
		ROUTING_PROTOCOL_MAP.put("static-ddijk", "Static-ddijk");
	}
	
	public static Hashtable NODE_TYPE_INFO;

	protected final Hashtable memberToPair = new Hashtable();
	protected ThinletKeyValue tkv;
	protected NSTopology nst;
	protected String file;
	protected List warnings = new ArrayList();
	protected List diffs = new ArrayList();

	private static final class NodeInterfacePair {
		private final NSObject nso;

		private final String ipAddress;

		NodeInterfacePair(NSObject nso, String ipAddress) {
			this.nso = nso;
			this.ipAddress = ipAddress;
		}

		public NSObject getNSObject() {
			return this.nso;
		}

		public String getIPAddress() {
			return this.ipAddress;
		}

		public String toString() {
			return "NodeInterfacePair[nso=" + this.nso + "; ipAddress="
					+ this.ipAddress + "]";
		}
	}

	public DBState() {
	}

	public List getDiffs() {
		return diffs;
	}

	public List getWarnings() {
		return warnings;
	}

	public Hashtable invertVector(Vector vector) {
		Hashtable retval;
		int lpc, size;

		retval = new Hashtable();
		size = vector.size();
		for (lpc = 0; lpc < size; lpc++) {
			Hashtable obj = (Hashtable) vector.elementAt(lpc);
			Object key;

			if (((key = obj.get("vname")) != null)
					|| ((key = obj.get("name")) != null)) {
				Object value;

				if ((value = retval.get(key)) != null) {
					Vector collision;

					if (value instanceof Vector) {
						collision = (Vector) value;
					} else {
						collision = new Vector();
						collision.addElement(value);
						retval.put(key, collision);
					}
					collision.addElement(obj);
				} else {
					retval.put(key, obj);
				}
			} else {
				/* Object has no name. */
			}
		}
		return retval;
	}
	
	private UniqueName uniqueName(String name) {
		UniqueName retval;
		
		retval = new UniqueName(name);
		this.nst.getNamer().put(retval);
		
		return retval;
	}
	
	private UniqueName uniqueName(Hashtable obj, String key) {
		return this.uniqueName((String)obj.get(key));
	}

	private NSVType instantiateVType(Hashtable vtype) {
		String name = (String) vtype.get("name");
		NSSoftVType retval;
		
		if (name.equals("pc1gen") || name.equals("pc2gen"))
			return null;

		retval = new NSSoftVType();
		retval.setName(this.uniqueName(vtype, "name"));
		retval.setWeight(((Double) vtype.get("weight")).doubleValue());

		StringTokenizer st = new StringTokenizer((String) vtype.get("members"));

		while (st.hasMoreTokens()) {
			retval.addMember(st.nextToken());
		}

		return retval;
	}

	private NSNode instantiateNode(NSTopology nst, Hashtable node) {
		String str, vname = (String) node.get("vname");
		StringTokenizer st;
		NSNode retval;

		if (vname.equals(nst.getFirewallName()))
			return null;
		
		retval = new NSNode();
		retval.setName(this.uniqueName(node, "vname"));
		retval.setOS((String) node.get("osname"));
		retval.setFailureAction((String) node.get("failureaction"));
		// retval.setStartup((String) node.get("startupcmd")); // XXX valid?
		retval.setFixedNode((String) node.get("fixed"));
		if ("".equals(retval.getFixedNode())) {
			str = (String) node.get("type");
			if (!"pc".equals(str)) // XXX fucked up
				retval.setHardware(str);
		}

		st = new StringTokenizer((String) node.get("tarfiles"), ";");
		while (st.hasMoreTokens()) {
			String tar = st.nextToken();
			int index;
			
			if ((index = tar.indexOf(" ")) != -1) {
				NSTarFile tf = new NSTarFile();
				
				tf.setDestination(tar.substring(0, index));
				tf.setTarFile(tar.substring(index + 1));
				retval.getTarFiles().add(tf);
			}
		}

		st = new StringTokenizer((String) node.get("rpms"), ";");
		while (st.hasMoreTokens()) {
			String rpm = st.nextToken();
			
			retval.getRPMFiles().add(rpm);
		}
		
		st = new StringTokenizer((String) node.get("ips"), " :");
		while (st.hasMoreTokens()) {
			String iface = st.nextToken();
			String ip = st.nextToken();

			this.memberToPair.put(vname + ":" + iface, new NodeInterfacePair(
					retval, ip));
		}

		return retval;
	}

	private NSInterface instantiateInterface(Hashtable nic) {
		NodeInterfacePair nip;
		NSInterface retval;

		retval = new NSInterface();
		retval.setBandwidth((((Integer) nic.get("bandwidth")).intValue()) * 1000);
		retval.setLatency(((Double) nic.get("delay")).doubleValue());
		retval.setLossRate(((Double) nic.get("lossrate")).doubleValue());

		nip = (NodeInterfacePair) this.memberToPair.get(nic.get("member"));
		retval.setNSObject(nip.getNSObject());
		if (this.file.indexOf("tb-set-ip") != -1)
			retval.setIPv4Address(nip.getIPAddress());

		return retval;
	}

	private NSLink instantiateLink(Vector link) {
		Hashtable queue, left, right;
		NSInterface nsleft, nsright;
		NSLink retval;

		left = (Hashtable) link.elementAt(0);
		right = (Hashtable) link.elementAt(1);
		queue = left; // XXX

		nsleft = this.instantiateInterface(left);
		nsright = this.instantiateInterface(right);
		nsleft.setLatency(nsleft.getLatency() + nsright.getLatency());
		nsleft.setLossRate(NSInterface.lossOneToMany(nsleft.getLossRate()));
		
		retval = new NSLink();
		retval.setName(this.uniqueName(queue, "vname"));
		retval.addMember(nsleft);
		retval.addMember(nsright);
		if (((Integer) queue.get("q_droptail")).intValue() == 1)
			retval.setQueueing("DropTail");
		else if (((Integer) queue.get("q_gentle")).intValue() == 1)
			retval.setQueueing("GRED");
		else if (((Integer) queue.get("q_red")).intValue() == 1)
			retval.setQueueing("RED");
		else {
			retval.setQueueing("DropTail");
			
			this.warnings.add("Assuming 'DropTail' queueing for link '"
					+ retval.getName().toString()
					+ "'.");
		}

		if (tkv.getKeyValueAsBoolean(link, "traced")) {
			int sl;
			
			if ((sl = tkv.getKeyValueAsInteger(link, "trace_snaplen")) > 0)
				retval.setSnapLen(sl);
			retval.setTracingFilter(tkv.getKeyValueAsString(link, "trace_expr"));
			retval.setTracingType(tkv.getKeyValueAsString(link, "trace_type"));
		}
		
		return retval;
	}

	private NSLan instantiateLan(Vector lan) {
		Object sameBandwidth, sameLatency, sameLoss;
		int lpc, size;
		NSLan retval;

		retval = new NSLan();
		retval.setName(this.uniqueName((Hashtable) lan.elementAt(0), "vname"));

		sameBandwidth = tkv.getKeyValueAsSingleton(lan, "bandwidth");
		sameLatency = tkv.getKeyValueAsSingleton(lan, "delay");
		sameLoss = tkv.getKeyValueAsSingleton(lan, "lossrate");
		
		size = lan.size();
		for (lpc = 0; lpc < size; lpc++) {
			Hashtable member = (Hashtable) lan.elementAt(lpc);
			NSInterface nsi = this.instantiateInterface(member);

			if (sameBandwidth != ThinletController.MULTIPLE_SELECTION) {
				retval.getTemplateInterface().setBandwidth(
						((Number)sameBandwidth).longValue() * 1000);
				nsi.setBandwidth(-1);
			}
			if (sameLatency != ThinletController.MULTIPLE_SELECTION) {
				retval.getTemplateInterface().setLatency(
						((Number)sameLatency).doubleValue() * 2.0);
				nsi.setLatency(-1.0);
			}
			if (sameLoss != ThinletController.MULTIPLE_SELECTION) {
				double iloss = ((Number)sameLoss).doubleValue();
				
				retval.getTemplateInterface().setLossRate(
						NSInterface.lossOneToMany(iloss));
				nsi.setLossRate(-1.0);
			}
			if (tkv.getKeyValueAsBoolean(member, "is_accesspoint")) {
				retval.setAccessPoint(nsi.getNSObject().getName().toString());
			}
			retval.addMember(nsi);
		}

		if (tkv.getKeyValueAsBoolean(lan, "traced")) {
			int sl;
			
			if ((sl = tkv.getKeyValueAsInteger(lan, "trace_snaplen")) > 0)
				retval.setSnapLen(sl);
			retval.setTracingFilter(tkv.getKeyValueAsString(lan, "trace_expr"));
			retval.setTracingType(tkv.getKeyValueAsString(lan, "trace_type"));
			// XXX more still
		}
		
		retval.setProtocol(tkv.getKeyValueAsString(lan, "protocol"));
		
		return retval;
	}

	private NSLanLink instantiateLanLink(Vector lanlink) {
		NSLanLink retval = null;

		if (lanlink.size() == 2 && "ethernet".equals(tkv.getKeyValueAsString(lanlink, "protocol"))) {
			retval = this.instantiateLink(lanlink);
		} else {
			retval = this.instantiateLan(lanlink);
		}
		return retval;
	}
	
	private static final Double NEG_ONE = new Double(-1.0);
	private static final Integer ZERO = new Integer(0);

	private NSProgram instantiateProgram(Hashtable progAgent) {
		String name = (String)progAgent.get("vname");
		NSNode node = (NSNode)nst.findObject((String)progAgent.get("vnode"));
		String cmd = (String)progAgent.get("command");
		NSProgram retval = null;
		
		if (name.equals(node.getName().toString() + "_startcmd")) {
			node.setStartup(cmd.substring(1, cmd.lastIndexOf(';') - 1));
		}
		else {
			retval = new NSProgram();
			
			retval.setName(this.uniqueName(progAgent, "vname"));
			retval.setDirectory((String)progAgent.get("dir"));
			retval.setCommand(cmd);
			retval.setTimeout(tkv.getKeyValueAsNumber(progAgent, "timeout", NEG_ONE).intValue());
			retval.setExpectedExitCode(tkv.getKeyValueAsNumber(progAgent, "expected_exit_code", ZERO).intValue());
			retval.setNode(node);
		}
		
		return retval;
	}

	private NSAgent instantiateTrafgen(Hashtable trafAgent) {
		NSAgent retval = new NSAgent();
		String role, proto, vnode;
		NSNode node;
		
		retval.setName(this.uniqueName(trafAgent, "vname"));
		role = this.tkv.getKeyValueAsString(trafAgent, "role");
		proto = this.tkv.getKeyValueAsString(trafAgent, "proto");
		vnode = this.tkv.getKeyValueAsString(trafAgent, "vnode");
		node = (NSNode)this.nst.findObject(vnode);
		
		if ("sink".equals(role)) {
			
			if ("udp".equals(proto))
				retval.setType("Agent/Null");
			else if ("tcp".equals(proto))
				retval.setType("Agent/TCPSink");
			retval.setAttachPoint(node);
		}
		else if ("source".equals(role)) {
			NSAgent trans = new NSAgent();
			Hashtable ht;
			
			trans.setName(this.uniqueName(trafAgent.get("vname").toString() + "src"));
			if ("udp".equals(proto))
				trans.setType("Agent/UDP");
			else if ("tcp".equals(proto))
				trans.setType("Agent/TCP");
			trans.setAttachPoint(node);
			this.nst.add(trans);
			retval.setType("Application/Traffic/CBR");
			retval.setAttachPoint(trans);
			retval.getAttributes().putAll(NSAgent.DEFAULT_CBR_ATTRIBUTES);
		}
		
		return retval;
	}

	private NSEventGroup instantiateEventGroup(Hashtable trafAgent) {
		NSEventGroup retval = null;
		NSEventGroup nseg;
		String name = (String) trafAgent.get("group_name");
		NSObject member = this.nst.findObject((String) trafAgent.get("agent_name"));
		
		if ((nseg = (NSEventGroup)this.nst.findObject(name)) == null) {
			retval = nseg = new NSEventGroup();
			retval.setName(this.uniqueName(name));
		}
		
		nseg.addMember(member);
		
		return retval;
	}
	
	private void connectTrafgen(Hashtable trafAgent) {
		String role, target;
		
		role = this.tkv.getKeyValueAsString(trafAgent, "role");
		target = this.tkv.getKeyValueAsString(trafAgent, "target_vname");
		
		if ("source".equals(role)) {
			NSAgent src, sink;
			
			src = (NSAgent)this.nst.findObject(trafAgent.get("vname").toString());
			sink = (NSAgent)this.nst.findObject(target);
			src.setConnectPoint(sink);
		}
	}
	
	private void addLanSettings(Hashtable ls) {
		String vname = this.tkv.getKeyValueAsString(ls, "vname");
		String capkey = this.tkv.getKeyValueAsString(ls, "capkey");
		String capval = this.tkv.getKeyValueAsString(ls, "capval");
		NSLan nsl = (NSLan)this.nst.findObject(vname);
		
		nsl.getSettings().add(new Pair(capkey, capval));
	}
	
	private NSEvent instantiateEvent(Hashtable event) {
		String parentName;
		NSEvent retval;
		
		retval = new NSEvent();
		retval.setEventType(tkv.getKeyValueAsInteger(event, "eventtype"));
		retval.setObjectType(tkv.getKeyValueAsInteger(event, "objecttype"));
		retval.parseArguments((String)event.get("arguments"));
		retval.setNSObject(nst.findObject((String)event.get("vname")));
		parentName = (String)event.get("parent");
		if ("__ns_sequence".equals(parentName)) {
			if (retval.getNSObject() instanceof NSAgent) {
				NSAgent nsa = (NSAgent)retval.getNSObject();
				
				nsa.initFromEvent(this.warnings, retval);
				retval = null;
			}
		}
		else {
			retval.setParent(nst.findObject(parentName));
			if (retval.getNSObject() instanceof NSAgent) {
				NSAgent nsa = (NSAgent)retval.getNSObject();
				
				if (nsa.initFromEvent(this.warnings, retval)) {
					retval = null;
				}
			}
		}
		
		return retval;
	}
	
	private void instantiateNSObjects(NSTopology nst, String type,
			Vector objects) {
		Enumeration en;

		if (objects == null)
			return;
		
		if ("lans" == type) {
			en = this.invertVector(objects).elements();
		} else {
			en = objects.elements();
		}

		while (en.hasMoreElements()) {
			Object object = en.nextElement();

			if ("vtypes" == type)
				nst.add(this.instantiateVType((Hashtable) object));
			else if ("nodes" == type)
				nst.add(this.instantiateNode(nst, (Hashtable) object));
			else if ("lans" == type)
				nst.add(this.instantiateLanLink((Vector) object));
			else if ("programs" == type)
				nst.add(this.instantiateProgram((Hashtable) object));
			else if ("trafgens" == type)
				nst.add(this.instantiateTrafgen((Hashtable) object));
			else if ("event_groups" == type)
				nst.add(this.instantiateEventGroup((Hashtable) object));
			else if ("lan_settings" == type)
				this.addLanSettings((Hashtable) object);
			else if ("events" == type)
				nst.addEvent(this.instantiateEvent((Hashtable) object));
			else
				throw new InternalError();
		}
	}

	/**
	 * Convert the virtual topology received via XML-RPC to an NSTopology
	 * object.
	 * 
	 * @param nst
	 *            The destination object for the decoded topology.
	 * @param topology
	 *            The Emulab DB state of the virtual topology.
	 */
	public void fromVirtualTopology(final ThinletKeyValue tkv, NSTopology nst, Hashtable topology, String file)
	 throws IOException{
		int next_rule = 100, lpc, size;
		String str, fwname;
		List olist, list;

		this.tkv = tkv;
		this.nst = nst;
		this.file = file;
		
		str = tkv.getKeyValueAsString(topology, "nodes.routertype");
		if ((str = (String)ROUTING_PROTOCOL_MAP.get(str)) != null)
			nst.setRoutingProtocol(str);
		
		fwname = tkv.getKeyValueAsString(topology, "virt_firewalls.fwname", "");
		if (fwname != null && !"".equals(fwname)) {
			nst.setFirewallName(fwname);
			nst.setFirewallStyle(tkv.getKeyValueAsString(topology, "virt_firewalls.style"));
			nst.setFirewallType(tkv.getKeyValueAsString(topology, "virt_firewalls.type"));
		}
		
		olist = nst.getFirewallRules();
		olist.clear();
		
		list = tkv.getKeyValueAsList(topology, "firewall_rules", null);		
		if (list != null) {
			Collections.sort(list, new Comparator() {
				public int compare(Object arg0, Object arg1) {
					int i0, i1, retval;
					
					i0 = tkv.getKeyValueAsInteger(arg0, "ruleno");
					i1 = tkv.getKeyValueAsInteger(arg1, "ruleno");
					
					if (i0 < i1)
						retval = -1;
					else if (i0 == i1)
						retval = 0;
					else
						retval = 1;
					
					return retval;
				}
			});
			size = list.size();
			for (lpc = 0; lpc < size; lpc++) {
				FirewallRule fr = new FirewallRule();
				Object obj = list.get(lpc);
				int ruleno = tkv.getKeyValueAsInteger(obj, "ruleno");
				
				if (ruleno == next_rule)
					next_rule += 1;
				else
					fr.setNumber(ruleno);
				fr.setRule(tkv.getKeyValueAsString(obj, "rule"));
				olist.add(fr);
			}
		}
		
		olist = nst.getCapturedParameters();
		olist.clear();
		
		list = tkv.getKeyValueAsList(topology, "user_environment", null);
		if (list != null) {
			Collections.sort(list, new Comparator() {
				public int compare(Object arg0, Object arg1) {
					int i0, i1, retval;
				
					i0 = tkv.getKeyValueAsInteger(arg0, "idx");
					i1 = tkv.getKeyValueAsInteger(arg1, "idx");
					
					if (i0 < i1)
						retval = -1;
					else if (i0 == i1)
						retval = 0;
					else
						retval = 1;
					
					return retval;
				}
			});
			size = list.size();
			for (lpc = 0; lpc < size; lpc++) {
				Pair cp = new Pair();
				Object obj = list.get(lpc);
				
				cp.setKey(tkv.getKeyValueAsString(obj, "name"));
				cp.setValue(tkv.getKeyValueAsString(obj, "value"));
				olist.add(cp);
			}
		}
		
		nst.setForceEndNodeShaping(tkv.getKeyValueAsBoolean(topology, "settings.forcelinkdelays"));
		nst.setUseEndNodeShaping(tkv.getKeyValueAsBoolean(topology, "settings.uselinkdelays"));
		nst.setDelayOS(tkv.getKeyValueAsString(topology, "settings.delay_osname"));
		nst.setJailOS(tkv.getKeyValueAsString(topology, "settings.jail_osname"));
		if (file.indexOf("tb-set-sync-server") != -1)
			nst.setSyncServer(tkv.getKeyValueAsString(topology, "settings.sync_server"));
		for (lpc = 0; lpc < OBJECT_TYPES.length; lpc++) {
			String type = OBJECT_TYPES[lpc];

			this.instantiateNSObjects(nst, type, (Vector) topology.get(type));
		}
		
		list = (List)topology.get("trafgens");
		size = list.size();
		for (lpc = 0; lpc < size; lpc++) {
			this.connectTrafgen((Hashtable)list.get(lpc));
		}
		this.memberToPair.clear();
		
		this.nst.unflatten();
		
		NSFile nsf = new NSFile();
		String newFile = nsf.toNSString(nst);
		
		this.diffs = FileDiff.diff(file, newFile);
	}

	/**
	 * @see java.lang.Object#toString()
	 */
	public String toString() {
		return "DBState[memberToPair=" + this.memberToPair + "]";
	}
}
