/*
 * MutableResourceBundle.java
 *
 * Copyright (c) 2005 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

package net.emulab.ns; // TODO: Move to net.emulab.util

import java.util.Enumeration;
import java.util.Hashtable;
import java.util.ResourceBundle;

/**
 * A "run-time mutable" resource bundle implementation.
 */
public class MutableResourceBundle extends ResourceBundle {
	/**
	 * Container for keys that will override those in the parent resource
	 * bundle.
	 */
	protected final Hashtable keyToResource = new Hashtable();

	/**
	 * Construct a MutableResourceBundle that pulls default values from the
	 * given bundle.
	 * 
	 * @param parent
	 *            The parent for this bundle, all values will initially be
	 *            pulled from this object, until it is overridden using set().
	 */
	public MutableResourceBundle(ResourceBundle parent) {
		this.setParent(parent);
	}

	/**
	 * @see java.util.ResourceBundle#handleGetObject(java.lang.String)
	 */
	protected Object handleGetObject(String key) {
		return this.keyToResource.get(key);
	}

	/**
	 * @see java.util.ResourceBundle#getKeys()
	 */
	public Enumeration getKeys() {
		return this.parent.getKeys();
	}

	/**
	 * @param key
	 *            The key to override, must be a valid key in the parent bundle.
	 * @param obj
	 *            The value to pair with the key.
	 */
	public synchronized void set(String key, Object obj) {
		if (key == null)
			throw new IllegalArgumentException("key is null");
		if (obj == null)
			throw new IllegalArgumentException("obj is null");

		this.parent.getObject(key);

		this.keyToResource.put(key, obj);
	}

	/**
	 * Reset a key override so that the parent's value is returned from this
	 * point on.
	 * 
	 * @param key
	 *            The key name to reset, must be a valid key in the parent
	 *            bundle.
	 */
	public synchronized void reset(String key) {
		if (key == null)
			throw new IllegalArgumentException("key is null");

		this.parent.getObject(key);

		this.keyToResource.remove(key);
	}

	/**
	 * @see java.lang.Object#toString
	 */
	public String toString() {
		return "MutableResourceBundle[" + super.toString() + "; keyToResource="
				+ this.keyToResource + "]";
	}
}
