/*
 * NSBase.java
 *
 * Copyright (c) 2005 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

package net.emulab.ns;

import java.io.Serializable;

/**
 * Base class for most "normal" implementations of NSObject.
 */
public class NSBase implements NSObject, Serializable, Cloneable {
	/**
	 * Globally unique name for this object.
	 */
	protected UniqueName name;

	/**
	 * NS syntax to prepend to this definition.
	 */
	protected String nsPrepend = "";

	/**
	 * NS syntax to append to this definition.
	 */
	protected String nsAppend = "";

	/**
	 * Construct an NSBase with the default values.
	 */
	public NSBase() {
	}

	/**
	 * @see net.emulab.netlab.client.ns.NSObject#setName
	 */
	public NSObject setName(UniqueName name) {
		this.name = name;
		return this;
	}

	/**
	 * @see net.emulab.netlab.client.ns.NSObject#getName
	 */
	public UniqueName getName() {
		return this.name;
	}

	/**
	 * @see net.emulab.netlab.client.ns.NSObject#setNSPrepend
	 */
	public NSObject setNSPrepend(String prepend) {
		if (prepend == null)
			throw new IllegalArgumentException("prepend is null");

		this.nsPrepend = prepend;
		return this;
	}

	/**
	 * @see net.emulab.netlab.client.ns.NSObject#getNSPrepend
	 */
	public String getNSPrepend() {
		return this.nsPrepend;
	}

	/**
	 * @see net.emulab.netlab.client.ns.NSObject#setNSAppend
	 */
	public NSObject setNSAppend(String append) {
		if (append == null)
			throw new IllegalArgumentException("append is null");

		this.nsAppend = append;
		return this;
	}

	/**
	 * @see net.emulab.netlab.client.ns.NSObject#getNSAppend
	 */
	public String getNSAppend() {
		return this.nsAppend;
	}

	/**
	 * Compare against another NSObject. The comparison is done based on each
	 * object's unique name.
	 * 
	 * @see java.lang.Comparable#compareTo
	 */
	public int compareTo(Object obj) {
		NSObject nso;
		int retval;

		if (!(obj instanceof NSObject))
			throw new IllegalArgumentException("obj is not a NSObject");

		nso = (NSObject) obj;
		if (this.name != null)
			retval = this.name.compareTo(nso.getName());
		else if (nso.getName() == null)
			retval = 0;
		else
			retval = -1;
		return retval;
	}
	
	public boolean equals(Object obj) {
		boolean retval = false;
		
		if (obj instanceof NSBase) {
			NSBase nsb = (NSBase)obj;
			UniqueName un;
			
			un = nsb.getName();
			if (un != null)
				retval = un.equals(this.name);
		}
		
		return retval;
	}

	/**
	 * @see java.lang.Object#clone
	 */
	public Object clone() {
		try {
			Object retval = super.clone();
			NSBase nsb;

			nsb = (NSBase) retval;
			if (nsb.name != null)
				nsb.name = new UniqueName(nsb.name.getBase(), "");
			return retval;
		} catch (CloneNotSupportedException e) {
			throw new InternalError();
		}
	}

	/**
	 * @see java.lang.Object#toString
	 */
	public String toString() {
		return "NSBase[name=" + this.name + "; nsPrepend=" + this.nsPrepend
				+ "; nsAppend=" + this.nsAppend + "]";
	}
}
