/*
 * NSInterface.java
 *
 * Copyright (c) 2005 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

package net.emulab.ns;

import java.io.Serializable;
import java.util.StringTokenizer;

/**
 * Class that represents a network interface card and its standard NS properties
 * (bandwidth, latency) along with its Emulab-specific properties (IP address).
 */
public class NSInterface implements NSObject, Comparable, Serializable {
	/**
	 * Unique identifier for serialization.
	 */
	static final long serialVersionUID = 4302943738340209756L;

	/**
	 * 100 Mbps ethernet.
	 */
	public static final long ETH_100 = 100;

	public static final long KBPS = 1000;

	/**
	 * Megabits per second.
	 */
	public static final long MBPS = KBPS * KBPS;
	
	public static final long GBPS = MBPS * MBPS;

	/**
	 * Maximum bandwidth for a NIC.
	 * 
	 * TODO: Don't assume, get from Emulab.
	 */
	public static final long MAX_BANDWIDTH = 1 * GBPS;

	/**
	 * Number of bytes in an IPv4 address.
	 */
	public static final int IPV4_BYTES = 4;

	/**
	 * Minimum value for a byte in an IPv4 address.
	 */
	public static final int IPV4_BYTE_MIN = 0;

	/**
	 * Maximum value for a byte in an IPv4 address.
	 */
	public static final int IPV4_BYTE_MAX = 255;

	/**
	 * Array of acceptable IPv4 prefixes.
	 */
	private static final String ACCEPTABLE_PREFIXES[] = { "192.168.", "10.", };

	/**
	 * English description of the acceptable IPv4 prefixes.
	 */
	private static final String ACCEPTABLE_PREFIXES_STRING = "192.168.x.x or 10.x.x.x";

	/**
	 * Default template interface, any unspecified values are derived from this
	 * object.
	 */
	private static final NSInterface DEFAULT_TEMPLATE = new NSInterface();

	static {
		DEFAULT_TEMPLATE.setBandwidth(ETH_100 * MBPS);
		DEFAULT_TEMPLATE.setLatency(0.0);
		DEFAULT_TEMPLATE.setLossRate(0.0);
	}

	/**
	 * Check for the validity of an IPv4 adddress. If anything is found to be
	 * "unacceptable", an IllegalArgumentException will be thrown.
	 * 
	 * @param address
	 *            The address to check.
	 */
	public static void checkIPv4Address(String address) {
		String token = null;
		StringTokenizer st;
		int lpc;

		/* Check the string for four tokens */
		st = new StringTokenizer(address, ".");
		if (st.countTokens() != IPV4_BYTES) {
			throw new IllegalArgumentException("Invalid IPv4 address: "
					+ address);
		}

		/* ... that have sane values. */
		try {
			for (lpc = 0; lpc < IPV4_BYTES; lpc++) {
				int value;

				token = st.nextToken();
				value = Integer.parseInt(token);
				if ((value < IPV4_BYTE_MIN) || (value > IPV4_BYTE_MAX)) {
					throw new IllegalArgumentException(
							"Invalid IPv4 component: " + token);
				}
			}
		} catch (NumberFormatException e) {
			throw new IllegalArgumentException("Invalid IPv4 component: "
					+ token);
		}
	}
	
	public static double lossOneToMany(double loss) {
		double retval;
		
		retval = 1 - Math.pow(1 - loss, 2);
		retval = Math.floor(retval * 100000) / 100000;
		
		return retval;
	}

	/**
	 * The template interface, any unspecified values are derived from this
	 * object.
	 */
	protected NSInterface template = DEFAULT_TEMPLATE;

	/**
	 * The NSObject this interface is attached to.
	 */
	protected NSObject nso;

	/**
	 * The IPv4 address for this interface. An empty string means the value
	 * should be generated automatically.
	 */
	protected String address = "";

	/**
	 * Bandwidth, in Kbps, for this interface. A value of -1 means the value
	 * should be derived from the template interface.
	 */
	protected long bandwidth = -1;

	/**
	 * Latency, in milliseconds, for this interface. A value of -1 means the
	 * value should be derived from the template interface.
	 */
	protected double latency = -1.0;

	/**
	 * Packet loss rate, a probability between 0.0 and 1.0. A value of -1.0
	 * means the value should be derived from the template interface.
	 */
	protected double lossRate = -1.0;

	/** ADDED by Bela
	 * The unit cost of bandwidth, and of 3 security conduits.
	 */
	protected double bandwidthCost = -1;
	protected double conduit1Cost = -1;
	protected double conduit2Cost = -1;
	protected double conduit3Cost = -1;

	/**
	 * Construct an NSInterface object with the default values.
	 */
	public NSInterface() {
	}

	/**
	 * @param template
	 *            The template interface, any unspecified values are derived
	 *            from this object.
	 * @return this
	 */
	public NSInterface setTemplate(NSInterface template) {
		this.template = template;
		if (this.template != null) {
			if (this.bandwidth == this.template.bandwidth)
				this.bandwidth = -1;
			if (this.latency == this.template.latency)
				this.latency = -1.0;
			if (this.lossRate == this.template.lossRate)
				this.lossRate = -1.0;
		}
		return this;
	}

	/**
	 * @return The template interface, any unspecified values are derived from
	 *         this object.
	 */
	public NSInterface getTemplate() {
		return this.template;
	}

	/**
	 * @see net.emulab.netlab.client.ns.NSObject#setName
	 */
	public NSObject setName(UniqueName un) {
		new Throwable().printStackTrace();
		throw new InternalError();
	}

	/**
	 * @see net.emulab.netlab.client.ns.NSObject#getName
	 */
	public UniqueName getName() {
		new Throwable().printStackTrace();
		throw new InternalError();
	}

	/**
	 * @see net.emulab.netlab.client.ns.NSObject#setNSPrepend
	 */
	public NSObject setNSPrepend(String prepend) {
		throw new InternalError();
	}

	/**
	 * @see net.emulab.netlab.client.ns.NSObject#getNSPrepend
	 */
	public String getNSPrepend() {
		throw new InternalError();
	}

	/**
	 * @see net.emulab.netlab.client.ns.NSObject#setNSAppend
	 */
	public NSObject setNSAppend(String append) {
		throw new InternalError();
	}

	/**
	 * @see net.emulab.netlab.client.ns.NSObject#getNSAppend
	 */
	public String getNSAppend() {
		throw new InternalError();
	}

	/**
	 * @param nso
	 *            The NSObject this interface is attached to.
	 * @return this
	 */
	public NSInterface setNSObject(NSObject nso) {
		this.nso = nso;
		return this;
	}

	/**
	 * @return The NSObject this interface is attached to.
	 */
	public NSObject getNSObject() {
		return this.nso;
	}

	/**
	 * @param address
	 *            The IPv4 address for this interface. An empty string means the
	 *            value should be generated automatically.
	 * @return this
	 */
	public NSInterface setIPv4Address(String address) {
		if (!"".equals(address)) {
			boolean acceptablePrefix = false;
			int lpc;

			checkIPv4Address(address);

			/* ... form an IPv4 address with an acceptable prefix. */
			for (lpc = 0; (lpc < ACCEPTABLE_PREFIXES.length)
					&& !acceptablePrefix; lpc++) {
				acceptablePrefix = address.startsWith(ACCEPTABLE_PREFIXES[lpc]);
			}
			if (!acceptablePrefix) {
				throw new IllegalArgumentException(
						"Address must use an internal prefix (e.g. "
								+ ACCEPTABLE_PREFIXES_STRING + "): " + address);
			}
		}

		this.address = address;

		return this;
	}

	/**
	 * @return The IPv4 address for this interface. An empty string means the
	 *         value should be generated automatically.
	 */
	public String getIPv4Address() {
		return this.address;
	}

	/**
	 * @param bandwidth
	 *            Bandwidth, in Kbps, for this interface. A value of -1 means
	 *            the value should be derived from the template interface.
	 * @return this
	 */
	public NSInterface setBandwidth(long bandwidth) {
		if (bandwidth < -1) {
			throw new IllegalArgumentException("Negative bandwidth: "
					+ bandwidth);
		}
		if (bandwidth > MAX_BANDWIDTH) {
			throw new IllegalArgumentException("Bandwidth larger than "
					+ MAX_BANDWIDTH + " Kbps - " + bandwidth + " Kbps");
		}

		this.bandwidth = bandwidth;
		return this;
	}

	/**
	 * @return Bandwidth, in Kbps, for this interface. A value of -1 means the
	 *         value should be derived from the template interface.
	 */
	public long getBandwidth() {
		return this.bandwidth;
	}

	/**
	 * @param latency
	 *            Latency, in milliseconds, for this interface. A value of -1
	 *            means the value should be derived from the template interface.
	 * @return this
	 */
	public NSInterface setLatency(double latency) {
		if (latency < -1.0) {
			throw new IllegalArgumentException("Negative latency: " + latency);
		}

		this.latency = latency;
		return this;
	}

	/**
	 * @return Latency, in milliseconds, for this interface. A value of -1 means
	 *         the value should be derived from the template interface.
	 */
	public double getLatency() {
		return this.latency;
	}

	/**
	 * @param lossRate
	 *            Packet loss rate, a probability between 0.0 and 1.0. A value
	 *            of -1.0 means the value should be derived from the template
	 *            interface.
	 * @return this
	 */
	public NSInterface setLossRate(double lossRate) {
		if ((lossRate != -1.0) && ((lossRate < 0.0) || (lossRate > 1.0))) {
			throw new IllegalArgumentException(
					"Loss rate not between 0 and 1: " + lossRate);
		}

		this.lossRate = lossRate;
		return this;
	}

	/**
	 * @return Packet loss rate, a probability between 0.0 and 1.0. A value of
	 *         -1.0 means the value should be derived from the template
	 *         interface.
	 */
	public double getLossRate() {
		return this.lossRate;
	}

	/**
	 * @return Bandwidth, in Kbps, for this interface. If this interface's value
	 *         is -1, the value from the template is returned.
	 */
	public long getActualBandwidth() {
		long retval;

		if (this.bandwidth == -1)
			retval = this.template.getActualBandwidth();
		else
			retval = this.bandwidth;
		return retval;
	}

	/**
	 * @return Latency, in milliseconds, for this interface. If this interface's
	 *         value is -1, the value from the template is returned.
	 */
	public double getActualLatency() {
		double retval;

		if (this.latency == -1.0)
			retval = this.template.getActualLatency();
		else
			retval = this.latency;
		return retval;
	}
	
	public double getLanLatency() {
		return this.getActualLatency() / 2;
	}

	/**
	 * @return Packet loss rate, a probability between 0.0 and 1.0, for this
	 *         interface. If this interface's value is -1.0, the value from the
	 *         template is returned.
	 */
	public double getActualLossRate() {
		double retval;

		if (this.lossRate == -1.0)
			retval = this.template.getActualLossRate();
		else
			retval = this.lossRate;
		return retval;
	}
	
	public double getLanLossRate() {
		return (1 - Math.sqrt(1 - this.getActualLossRate()));
	}

	/**
	 * @return True if the template NIC is shaped or if this NIC has shaping
	 *         (e.g. non-default bandwidth/latency/loss-rate).
	 */
	public boolean isShaped() {
		boolean retval;

		retval = ((this.template != null) && this.template.isShaped())
				|| (this.bandwidth != -1) || (this.latency != -1.0)
				|| (this.lossRate != -1.0);

		return retval;
	}


	/**
	 * Added by BELA.
	 */
	public NSInterface setBandwidthCost(double bandwidthCost) {
                this.bandwidthCost = bandwidthCost;
                return this;
        }
        public double getBandwidthCost() {
                return this.bandwidthCost;
        }
	public NSInterface setConduit1Cost(double conduit1Cost) {
                this.conduit1Cost = conduit1Cost;
                return this;
        }
        public double getConduit1Cost() {
                return this.conduit1Cost;
        }
	public NSInterface setConduit2Cost(double conduit2Cost) {
                this.conduit2Cost = conduit2Cost;
                return this;
        }
        public double getConduit2Cost() {
                return this.conduit2Cost;
        }
	public NSInterface setConduit3Cost(double conduit3Cost) {
                this.conduit3Cost = conduit3Cost;
                return this;
        }
        public double getConduit3Cost() {
                return this.conduit3Cost;
        }





	/**
	 * @see java.lang.Comparable#compareTo
	 */
	public int compareTo(Object obj) {
		NSInterface nsi;
		int retval;

		if (!(obj instanceof NSInterface))
			throw new IllegalArgumentException("obj is not an NSInterface");

		nsi = (NSInterface) obj;
		retval = this.address.compareTo(nsi.address);
		if (retval == 0) {
			retval = this.nso.compareTo(nsi.nso);
		}
		return retval;
	}

	/**
	 * @see java.lang.Object#clone
	 */
	public Object clone() {
		try {
			Object retval = super.clone();

			return retval;
		} catch (CloneNotSupportedException e) {
			throw new InternalError();
		}
	}

	/**
	 * @see java.lang.Object#toString
	 */
	public String toString() {
		return "NSInterface[template=" + this.template + "; nso=" + this.nso
				+ "; address=" + this.address + "; bandwidth=" + this.bandwidth
				+ "; latency=" + this.latency + "; lossRate=" + this.lossRate
				+ "]";
	}
}
