/*
 * NSLanLink.java
 *
 * Copyright (c) 2005 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

package net.emulab.ns;

import java.util.Iterator;
import java.util.Vector;

/**
 * Combined base class for NSLans and NSLinks.
 */
public class NSLanLink extends NSBase {
	/**
	 * Unique identifier for serialization.
	 */
	static final long serialVersionUID = -239043042267925628L;

	/**
	 * References to the NSInterfaces of this LAN/link.
	 */
	protected Vector members = new Vector();

	/**
	 * The netmask for this LAN/link.
	 */
	protected String netmask = "";
	
	protected String tracingType = "";
	
	protected String tracingFilter = "";
	
	protected int snapLen = -1;

	/**
	 * Construct an NSLanLink with the default values.
	 */
	public NSLanLink() {
	}

	/**
	 * @param members
	 *            The NSInterface members of this LAN/link.
	 * @return this
	 */
	public NSLanLink setMembers(Vector members) {
		if (members == null)
			throw new IllegalArgumentException("members is null");

		this.members = members;
		return this;
	}

	/**
	 * @return The members of this LAN/link.
	 */
	public Vector getMembers() {
		return this.members;
	}

	/**
	 * Add a member to this LAN/link.
	 * 
	 * @param nsi
	 *            The member to add.
	 * @return this
	 */
	public NSLanLink addMember(NSInterface nsi) {
		if (nsi == null)
			throw new IllegalArgumentException("nsi is null");
		if (nsi.getNSObject() == null)
			throw new IllegalArgumentException("nsi has no NSObject");
		if (this.isMember(nsi.getNSObject())) {
			throw new IllegalArgumentException("nsi is already a member: "
					+ nsi);
		}

		this.members.addElement(nsi);
		return this;
	}

	/**
	 * Try to remove the NSInterface in this LAN/link that is part of the given
	 * NSObject.
	 * 
	 * @param nso
	 *            The NSObject whose NSInterface should be removed.
	 * @return True if the NSNode was found and removed, false otherwise.
	 */
	public boolean removeMember(NSObject nso) {
		boolean retval = false;
		Iterator iterator;

		iterator = this.members.iterator();
		while (iterator.hasNext()) {
			NSInterface nsi = (NSInterface) iterator.next();

			if (nsi.getNSObject().equals(nso)) {
				iterator.remove();
				retval = true;
			}
		}
		return retval;
	}

	/**
	 * @param nso
	 *            The NSObject to check for membership.
	 * @return True if the given NSObject is a member of this LAN/link.
	 */
	public boolean isMember(NSObject nso) {
		boolean retval = false;
		Iterator iterator;

		iterator = this.members.iterator();
		while (iterator.hasNext() && !retval) {
			NSInterface nsi = (NSInterface) iterator.next();

			if (nsi.getNSObject().equals(nso))
				retval = true;
		}
		return retval;
	}

	/**
	 * @param netmask
	 *            The netmask for this LAN/link.
	 * @return this
	 */
	public NSLanLink setNetMask(String netmask) {
		if (netmask == null)
			throw new IllegalArgumentException("netmask is null");
		if (!"".equals(netmask))
			NSInterface.checkIPv4Address(netmask);

		this.netmask = netmask;
		return this;
	}

	/**
	 * @return The netmask for this LAN/link.
	 */
	public String getNetMask() {
		return this.netmask;
	}

	public int getSnapLen() {
		return snapLen;
	}

	public void setSnapLen(int snapLen) {
		this.snapLen = snapLen;
	}

	public String getTracingFilter() {
		return tracingFilter;
	}

	public void setTracingFilter(String tracingFilter) {
		this.tracingFilter = tracingFilter;
	}

	public String getTracingType() {
		return tracingType;
	}

	public void setTracingType(String tracingType) {
		this.tracingType = tracingType;
	}

	/**
	 * @see java.lang.Object#clone
	 */
	public Object clone() {
		Object retval = super.clone();
		NSLanLink nsll;

		nsll = (NSLanLink) retval;
		nsll.members = new Vector();
		return retval;
	}

	/**
	 * @see java.lang.Object#toString
	 */
	public String toString() {
		return "NSLanLink[" + super.toString() + "; members=" + this.members
				+ "; netmask=" + this.netmask + "]";
	}
}
