/*
 * NSLink.java
 *
 * Copyright (c) 2005 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

package net.emulab.ns;

/**
 * Class that represents a "duplex-link" in NS, with Emulab extensions.
 */
public class NSLink extends NSLanLink {
	/**
	 * Unique identifier for serialization.
	 */
	static final long serialVersionUID = -5986816478293677854L;

	/**
	 * The packet queueing algorithm to run on the link.
	 */
	protected String queueing = "DropTail";

	/**
	 * Flag that indicates whether or not this link can be multiplexed with
	 * other virtual links over a single physical link.
	 */
	protected boolean multiplexed = false;

	/**
	 * Construct an NSLink with the default values.
	 */
	public NSLink() {
	}

	/**
	 * @param queueing
	 *            The packet queueing algorithm to run on the link.
	 * @return this
	 */
	public NSLink setQueueing(String queueing) {
		if (queueing == null)
			throw new IllegalArgumentException("queueing is null");

		this.queueing = queueing;
		return this;
	}

	/**
	 * @return The packet queueing algorithm to run on the link.
	 */
	public String getQueueing() {
		return this.queueing;
	}

	/**
	 * @param multiplexed
	 *            Flag that indicates whether or not this link can be
	 *            multiplexed with other virtual links over a single physical
	 *            link.
	 * @return this
	 */
	public NSLink setMultiplexed(boolean multiplexed) {
		this.multiplexed = multiplexed;
		return this;
	}

	/**
	 * @return Flag that indicates whether or not this link can be multiplexed
	 *         with other virtual links over a single physical link.
	 */
	public boolean getMultiplexed() {
		return this.multiplexed;
	}

	/**
	 * @param nsi
	 *            Add a NIC to this virtual link. Note that a link may contain
	 *            only two members, if any more than that are added an
	 *            IllegalStateException will be thrown.
	 * @return this
	 */
	public NSLanLink addMember(NSInterface nsi) {
		if (nsi == null)
			throw new IllegalArgumentException("nsi is null");

		if (super.members.size() >= 2) {
			throw new IllegalStateException("Links can only have two members");
		}

		return super.addMember(nsi);
	}

	/**
	 * @return The "left" side of this link.
	 */
	public NSInterface getLeftMember() {
		if (this.members.size() != 2)
			throw new IllegalStateException("Link members not initialized");

		return (NSInterface) this.members.firstElement();
	}

	/**
	 * @return The "right" side of this link.
	 */
	public NSInterface getRightMember() {
		if (this.members.size() != 2)
			throw new IllegalStateException("Link members not initialized");

		return (NSInterface) this.members.lastElement();
	}

	/**
	 * @return True if either the left or right members of this link are shaped.
	 */
	public boolean isShaped() {
		boolean retval;

		retval = this.getLeftMember().isShaped()
				|| this.getRightMember().isShaped();
		return retval;
	}

	/**
	 * @see java.lang.Object#toString
	 */
	public String toString() {
		return "NSLink[" + super.toString() + "; queueing=" + this.queueing
				+ "; multiplexed=" + this.multiplexed + "]";
	}
}
