/*
 * NSNode.java
 *
 * Copyright (c) 2005 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

package net.emulab.ns;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.SortedSet;
import java.util.TreeSet;

import thinlet.ArrayController;

/**
 * Class that represents a "node" in NS, with Emulab extensions.
 */
public class NSNode extends NSBase {
	/**
	 * Unique identifier for serialization.
	 */
	static final long serialVersionUID = 1967922132030022055L;

	/**
	 * The identifier for the operating system to install on this node.
	 */
	protected String os = "";

	/**
	 * The type of hardware to allocate for this node.
	 */
	protected String hardware = "";

	/**
	 * The script to run after the node starts up.
	 */
	protected String startup = "";

	/**
	 * Sorted list of TAR files to install on this node.
	 */
	protected TreeSet tarFiles = new TreeSet();

	/**
	 * Sorted list of RPM files to install on this node.
	 */
	protected TreeSet rpmFiles = new TreeSet();

	/**
	 * The name of the physical node to bind this virtual node to.
	 */
	protected String fixedNode = "";

	/**
	 * Action to take when a failure occurs while setting up this node.
	 * Currently, the supported values are "fatal" and "nonfatal".
	 */
	protected String failureAction = "";
	
	protected TreeSet programs = new TreeSet();
	
	/**
	 * Desires for this node.
	 */
	protected List desires = new ArrayList();



	/**
         * Added by Bela: node cost
         */
        protected double nodeCost = 0.0;
	protected double securityZone1Cost = -1;
	protected double securityZone2Cost = -1;
	protected double securityZone3Cost = -1;

	/**
	 * Construct an NSNode object with default values.
	 */
	public NSNode() {
	}

	/**
	 * @param os
	 *            The identifier for the operating system to install on this
	 *            node.
	 * @return this
	 */
	public NSNode setOS(String os) {
		this.os = os;
		return this;
	}

	/**
	 * @return The identifier for the operating system to install on this node.
	 */
	public String getOS() {
		return this.os;
	}

	/**
	 * @param hardware
	 *            The type of hardware to allocate for this node.
	 * @return this
	 */
	public NSNode setHardware(String hardware) {
		this.hardware = hardware;
		return this;
	}

	/**
	 * @return The type of hardware to allocate for this node.
	 */
	public String getHardware() {
		return this.hardware;
	}

	/**
	 * @param startup
	 *            The script to run after the node starts up.
	 * @return this
	 */
	public NSNode setStartup(String startup) {
		this.startup = startup;
		return this;
	}

	/**
	 * @return The script to run after the node starts up.
	 */
	public String getStartup() {
		return this.startup;
	}

	/**
	 * @return Sorted list of TAR files to install on this node.
	 */
	public SortedSet getTarFiles() {
		return this.tarFiles;
	}

	public void setTarFiles(TreeSet tarFiles) {
		this.tarFiles = tarFiles;
	}
	
	/**
	 * @return Sorted list of RPM files to install on this node.
	 */
	public SortedSet getRPMFiles() {
		return this.rpmFiles;
	}

	public void setRPMFiles(TreeSet rpmFiles) {
		this.rpmFiles = rpmFiles;
	}
	
	public SortedSet getPrograms() {
		return this.programs;
	}
	
	public void setPrograms(TreeSet programs) {
		this.programs = programs;
	}

	/**
	 * @param fixedNode
	 *            The name of the physical node to bind this virtual node to.
	 * @return this
	 */
	public NSNode setFixedNode(String fixedNode) {
		this.fixedNode = fixedNode;
		return this;
	}

	/**
	 * @return The name of the physical node to bind this virtual node to.
	 */
	public String getFixedNode() {
		return this.fixedNode;
	}

	/**
	 * @param failureAction
	 *            Action to take when a failure occurs while setting up this
	 *            node. Currently, the supported values are "fatal" and
	 *            "nonfatal".
	 * @return this
	 */
	public NSNode setFailureAction(String failureAction) {
		this.failureAction = failureAction;
		return this;
	}

	/**
	 * @return Action to take when a failure occurs while setting up this node.
	 *         Currently, the supported values are "fatal" and "nonfatal".
	 */
	public String getFailureAction() {
		return this.failureAction;
	}

	public List getDesires() {
		return this.desires;
	}
	
	public void setDesires(List list) {
		this.desires = list;
	}




 	/**
         * Added by Bela
         */
        public NSNode setNodeCost(double nodeCost) {
                if (nodeCost < 0) {
                        throw new IllegalArgumentException(
                                        "Node cost cannot be negative: " + nodeCost);
                }

                this.nodeCost = nodeCost;
                return this;
        }
        public double getNodeCost() {
                return this.nodeCost;
        }

	public NSNode setSecurityZone1Cost(double securityZone1Cost) {
                if (securityZone1Cost < -1) {
                        throw new IllegalArgumentException(
                                        "Security zone 1 cost cannot be lower than -1: " + securityZone1Cost);
                }

                this.securityZone1Cost = securityZone1Cost;
                return this;
        }
        public double getSecurityZone1Cost() {
                return this.securityZone1Cost;
        }

	public NSNode setSecurityZone2Cost(double securityZone2Cost) {
                if (securityZone2Cost < -1) {
                        throw new IllegalArgumentException(
                                        "Security zone 2 cost cannot be lower than -1: " + securityZone2Cost);
                }

                this.securityZone2Cost = securityZone2Cost;
                return this;
        }
        public double getSecurityZone2Cost() {
                return this.securityZone2Cost;
        }

	public NSNode setSecurityZone3Cost(double securityZone3Cost) {
                if (securityZone3Cost < -1) {
                        throw new IllegalArgumentException(
                                        "Security zone 3 cost cannot be lower than -1: " + securityZone3Cost);
                }

                this.securityZone3Cost = securityZone3Cost;
                return this;
        }
        public double getSecurityZone3Cost() {
                return this.securityZone3Cost;
        }







	/**
	 * @see java.lang.Object#clone
	 */
	public Object clone() {
		Object retval = super.clone();
		Iterator iter;
		NSNode nsnode;

		nsnode = (NSNode) retval;
		nsnode.tarFiles = (TreeSet) this.tarFiles.clone();
		nsnode.rpmFiles = (TreeSet) this.rpmFiles.clone();
		try {
			nsnode.desires = ArrayController.deepCopy(Pair.class, this.desires);
		}
		catch (CloneNotSupportedException e) {
			throw new InternalError();
		}
		for (iter = this.programs.iterator(); iter.hasNext(); ) {
			NSProgram nsp = (NSProgram)iter.next();
			
			nsp = (NSProgram)nsp.clone();
			nsp.setNode(nsnode);
			nsnode.programs.add(nsp);
		}
		
		return retval;
	}

	/**
	 * @see java.lang.Object#toString
	 */
	public String toString() {
		return "NSNode[" + super.toString() + "; os=" + this.os + "; hardware="
				+ this.hardware + "; startup=" + this.startup + "; tarFiles="
				+ this.tarFiles + "; rpmFiles=" + this.rpmFiles
				+ "; fixedNode=" + this.fixedNode + "; failureAction="
				+ this.failureAction + "]";
	}
}
