/*
 * NSTarFile.java
 *
 * Copyright (c) 2005 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */
package net.emulab.ns;

import java.io.Serializable;

/**
 * Class that represents a Tar file and the directory where it should be
 * installed.
 */
public class NSTarFile implements Serializable, Comparable, Cloneable {
	/**
	 * Unique identifier for serialization.
	 */
	static final long serialVersionUID = 6896951544977205180L;

	/**
	 * The path, on NFS, to the Tar file.
	 */
	protected String tarFile;

	/**
	 * The path, on the node, where the Tar should be installed.
	 */
	protected String destination;

	/**
	 * Construct an NSTarFile with the given values.
	 * 
	 * @param tarFile
	 *            The path, on NFS, to the Tar file.
	 * @param destination
	 *            The path, on the node, where the Tar should be installed.
	 */
	public NSTarFile(String tarFile, String destination) {
		this.setTarFile(tarFile);
		this.setDestination(destination);
	}

	/**
	 * Construct an NSTarFile with an empty tar file and destination directory.
	 */
	public NSTarFile() {
		this("/path/to/tar.tgz", "/");
	}

	/**
	 * @param tarFile
	 *            The path, on NFS, to the Tar file.
	 * @return this
	 */
	public NSTarFile setTarFile(String tarFile) {
		this.tarFile = tarFile;
		return this;
	}

	/**
	 * @return The path, on NFS, to the Tar file.
	 */
	public String getTarFile() {
		return this.tarFile;
	}

	/**
	 * @param destination
	 *            The path, on the node, where the Tar should be installed.
	 * @return this
	 */
	public NSTarFile setDestination(String destination) {
		this.destination = destination;
		return this;
	}

	/**
	 * @return The path, on the node, where the Tar should be installed.
	 */
	public String getDestination() {
		return this.destination;
	}

	/**
	 * @see java.lang.Comparable#compareTo
	 */
	public int compareTo(Object obj) {
		NSTarFile nstf;
		int retval;

		if (!(obj instanceof NSTarFile))
			throw new IllegalArgumentException("obj is not an NSTarFile");

		nstf = (NSTarFile) obj;
		retval = this.destination.compareTo(nstf.destination);
		if (retval == 0) {
			retval = this.tarFile.compareTo(nstf.tarFile);
		}
		return retval;
	}

	/**
	 * @see java.lang.Object#hashCode
	 */
	public int hashCode() {
		return this.tarFile.hashCode() + this.destination.hashCode();
	}

	/**
	 * @see java.lang.Object#equals
	 */
	public boolean equals(Object obj) {
		boolean retval = false;

		if (obj instanceof NSTarFile) {
			retval = (this.compareTo(obj) == 0);
		}
		return retval;
	}

	/**
	 * @see java.lang.Object#clone
	 */
	public Object clone() {
		try {
			return super.clone();
		} catch (CloneNotSupportedException e) {
			throw new InternalError();
		}
	}

	/**
	 * @see java.lang.Object#toString
	 */
	public String toString() {
		return "NSTarFile[tarFile=" + this.tarFile + "; destination="
				+ this.destination + "]";
	}
}
