/*
 * NSTopology.java
 *
 * Copyright (c) 2005 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

package net.emulab.ns;

import java.util.ArrayList;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;
import java.util.SortedSet;
import java.util.TreeSet;
import java.util.Vector;

/**
 * This class is the root for the other objects and provides a holding place for
 * attributes that apply to the whole network topology.
 */
public class NSTopology {
	/**
	 * The default routing protocol. Currently set to static routing.
	 */
	private static final String DEFAULT_ROUTING_PROTOCOL = "Static";
	
	protected final Namer namer = new Namer();

	/**
	 * Sorted list of nodes that make up the file.
	 */
	protected final List nodes = new ArrayList();
	
	protected final List programs = new ArrayList();

	/**
	 * Sorted list of links that make up the file.
	 */
	protected final List links = new ArrayList();

	/**
	 * Sorted list of LANs that make up the file.
	 */
	protected final List lans = new ArrayList();

	/**
	 * Sorted list of virtual node types.
	 */
	protected final List vtypes = new ArrayList();

	/**
	 * Sorted list of agents.
	 */
	protected final List agents = new ArrayList();

	protected final List eventGroups = new ArrayList();

	protected final List attachments = new ArrayList();

	protected final List connectors = new ArrayList();

	protected final List aggregates = new Vector();
	
	protected final Hashtable name2object = new Hashtable();

	/**
	 * The routing protocol to use for this topology.
	 */
	protected String routingProtocol;
	
	protected String firewallName = "ncfw";

	protected String firewallType = "";

	protected String firewallStyle = "basic";
	
	protected List firewallRules = new ArrayList();
	
	protected List capturedParameters = new ArrayList();
	
	protected List events = new ArrayList();
	
	/**
	 * Flag that indicates whether or not to use "linkdelays" for shaping
	 * traffic directly on the end nodes.
	 */
	protected boolean useEndNodeShaping;
	
	/**
	 * Force traffic to go through "linkdelays". XXX
	 */
	protected boolean forceEndNodeShaping;

	protected String syncServer = "";
	
	protected String delayOS = "";
	
	protected String jailOS = "";
	
	/**
	 * Construct an empty NSTopology.
	 */
	public NSTopology() {
		this.clear();
	}

	public Namer getNamer() {
		return this.namer;
	}

	/**
	 * @return The set of nodes contained in this topology.
	 */
	public List getNodes() {
		return this.nodes;
	}
	
	public List getPrograms() {
		return this.programs;
	}
	
	/**
	 * @return The set of links contained in this topology.
	 */
	public List getLinks() {
		return this.links;
	}

	/**
	 * @return The set of LANs contained in this topology.
	 */
	public List getLans() {
		return this.lans;
	}

	/**
	 * @return The set of virtual types made for this topology.
	 */
	public List getVTypes() {
		return this.vtypes;
	}

	public List getAgents() {
		return this.agents;
	}

	public List getEventGroups() {
		return this.eventGroups;
	}

	public List getAttachments() {
		return this.attachments;
	}

	public List getConnectors() {
		return this.connectors;
	}
	
	public NSObject findObject(String name) {
		return (NSObject) this.name2object.get(name);
	}
	
	public List findObjectsByName(String start) {
		List retval = new ArrayList();
		Enumeration en;
		
		en = this.name2object.keys();
		while (en.hasMoreElements()) {
			String key = (String)en.nextElement();
			
			if (key.startsWith(start)) {
				retval.add(this.name2object.get(key));
			}
		}
		
		return retval;
	}
	
	public NSLink findLink(NSNode left, NSNode right) {
		NSLink retval = null;
		Iterator iterator;

		iterator = this.links.iterator();
		while (iterator.hasNext() && (retval == null)) {
			NSLink nsl = (NSLink) iterator.next();

			if (nsl.isMember(left) && nsl.isMember(right))
				retval = nsl;
		}
		return retval;
	}

	public List findAggregates(NSObject key, Class type) {
		List retval = new Vector();
		Iterator iterator;

		iterator = this.aggregates.iterator();
		while (iterator.hasNext()) {
			NSAggregate nsa = (NSAggregate) iterator.next();

			if (nsa.contains(key) && (nsa.getClass() == type)) {
				retval.add(nsa);
			}
		}
		return retval;
	}

	public List getAggregates() {
		return this.aggregates;
	}

	public boolean isAttachedTo(NSObject a, NSObject b) {
		boolean retval = false;

		if (a instanceof NSAttachableObject) {
			NSAttachableObject nsao = (NSAttachableObject) a;
			NSObject point = nsao.getAttachPoint();

			if ((point != null)
					&& ((point == b) || this.isAttachedTo(point, b))) {
				retval = true;
			}
		}
		return retval;
	}

	public SortedSet findEventGroup(SortedSet ss, NSObject member) {
		SortedSet retval = ss;
		Iterator iterator;

		iterator = this.eventGroups.iterator();
		while (iterator.hasNext()) {
			NSEventGroup nseg = (NSEventGroup) iterator.next();

			if (nseg.getMembers().contains(member))
				retval.add(nseg);
		}
		return retval;
	}

	public SortedSet findEventGroup(NSObject member) {
		return this.findEventGroup(new TreeSet(), member);
	}

	public SortedSet findConnectors(SortedSet ss, NSObject nso) {
		SortedSet retval = ss;
		Iterator iterator;

		iterator = this.connectors.iterator();
		while (iterator.hasNext()) {
			NSConnectableObject nsco = (NSConnectableObject) iterator.next();

			if ((nsco.getConnectPoint() != null)
					&& this.isAttachedTo(nsco, nso)) {
				retval.add(nsco);
			}
		}
		return retval;
	}

	public SortedSet findConnectors(NSObject nso) {
		return this.findConnectors(new TreeSet(), nso);
	}

	/**
	 * @param routingProtocol
	 *            The routing protocol name to use for this topology.
	 * @return this
	 */
	public NSTopology setRoutingProtocol(String routingProtocol) {
		this.routingProtocol = routingProtocol;
		return this;
	}

	/**
	 * @return The routing protocol name to use for this topology.
	 */
	public String getRoutingProtocol() {
		return this.routingProtocol;
	}

	/**
	 * @return this
	 */
	public NSTopology setUseEndNodeShaping(boolean useEndNodeShaping) {
		this.useEndNodeShaping = useEndNodeShaping;
		return this;
	}

	public boolean getUseEndNodeShaping() {
		return this.useEndNodeShaping;
	}
	
	public NSTopology setSyncServer(String ss) {
		this.syncServer = ss;
		return this;
	}
	
	public String getSyncServer() {
		return this.syncServer;
	}

	public String getDelayOS() {
		return delayOS;
	}

	public void setDelayOS(String delayOS) {
		this.delayOS = delayOS;
	}

	public String getJailOS() {
		return jailOS;
	}

	public void setJailOS(String jailOS) {
		this.jailOS = jailOS;
	}

	public List getFirewallRules() {
		return firewallRules;
	}

	public void setFirewallRules(List firewallRules) {
		this.firewallRules = firewallRules;
	}

	public String getFirewallName() {
		return firewallName;
	}

	public void setFirewallName(String firewallName) {
		this.firewallName = firewallName;
	}

	public String getFirewallStyle() {
		return firewallStyle;
	}

	public void setFirewallStyle(String firewallStyle) {
		this.firewallStyle = firewallStyle;
	}

	public String getFirewallType() {
		return firewallType;
	}

	public void setFirewallType(String firewallType) {
		this.firewallType = firewallType;
	}

	/**
	 * @return this
	 */
	public NSTopology setForceEndNodeShaping(boolean forceEndNodeShaping) {
		this.forceEndNodeShaping = forceEndNodeShaping;
		return this;
	}

	public boolean getForceEndNodeShaping() {
		return this.forceEndNodeShaping;
	}

	public List getCapturedParameters() {
		return capturedParameters;
	}

	public void setCapturedParameters(List capturedParameters) {
		this.capturedParameters = capturedParameters;
	}

	public Hashtable getHardwareUsage() {
		Hashtable retval;
		Iterator iter;

		retval = new Hashtable();
		iter = this.nodes.iterator();
		while (iter.hasNext()) {
			NSNode nsn = (NSNode) iter.next();
			Vector collision;

			if ((collision = (Vector) retval.get(nsn.getHardware())) == null) {
				collision = new Vector();
				retval.put(nsn.getHardware(), collision);
			}
			collision.addElement(nsn);
		}
		return retval;
	}

	public Vector getShapedLinks() {
		Vector retval = new Vector();
		Iterator iter;

		iter = this.links.iterator();
		while (iter.hasNext()) {
			NSLink nsl = (NSLink) iter.next();

			if (nsl.isShaped())
				retval.addElement(nsl);
		}
		return retval;
	}

	/**
	 * Clear the internal state of this NSTopology.
	 */
	public void clear() {
		this.nodes.clear();
		this.links.clear();
		this.lans.clear();
		this.vtypes.clear();
		this.agents.clear();
		this.eventGroups.clear();
		this.attachments.clear();
		this.connectors.clear();
		this.routingProtocol = DEFAULT_ROUTING_PROTOCOL;
		this.useEndNodeShaping = false;
		this.forceEndNodeShaping = false;
	}

	/**
	 * Add an NSObject to the topology.
	 * 
	 * @param nso
	 *            The NSObject to add to the topology.
	 */
	public void add(NSObject nso) {
		if (nso == null)
			return; // Ignore
		else if (nso instanceof NSNode)
			this.nodes.add(nso);
		else if (nso instanceof NSProgram) {
			NSProgram nsp = (NSProgram)nso;

			nsp.getNode().getPrograms().add(nso);
			this.programs.add(nso);
		}
		else if (nso instanceof NSLink)
			this.links.add(nso);
		else if (nso instanceof NSLan)
			this.lans.add(nso);
		else if (nso instanceof NSInterface)
			; // Ignore
		else if (nso instanceof NSVType)
			this.vtypes.add(nso);
		else if (nso instanceof NSAgent)
			this.agents.add(nso);
		else if (nso instanceof NSEventGroup)
			this.eventGroups.add(nso);
		else
			throw new IllegalArgumentException("Unhandled: " + nso);

		if (nso instanceof NSAttachableObject)
			this.attachments.add(nso);
		if (nso instanceof NSConnectableObject)
			this.connectors.add(nso);
		
		this.name2object.put(nso.getName().toString(), nso);
	}

	/**
	 * Remove an NSObject from this topology.
	 * 
	 * @param nso
	 *            The NSObject to remove from the topology.
	 */
	public void remove(NSObject nso) {
		if (nso instanceof NSNode) {
			NSNode nsn = (NSNode)nso;
			Iterator iter;
			
			this.nodes.remove(nso);
			
			/* XXX */
			for (iter = nsn.getPrograms().iterator(); iter.hasNext(); ) {
				NSProgram nsp = (NSProgram)iter.next();
				
				this.programs.remove(nsp);
			}
		}
		else if (nso instanceof NSProgram) {
			NSProgram nsp = (NSProgram)nso;

			nsp.getNode().getPrograms().remove(nso);
			this.programs.remove(nso);
		}
		else if (nso instanceof NSLink)
			this.links.remove(nso);
		else if (nso instanceof NSLan)
			this.lans.remove(nso);
		else if (nso instanceof NSInterface)
			; // Ignore
		else if (nso instanceof NSVType)
			this.vtypes.remove(nso);
		else if (nso instanceof NSAgent)
			this.agents.remove(nso);
		else if (nso instanceof NSEventGroup)
			this.eventGroups.remove(nso);
		else
			throw new IllegalArgumentException("Unhandled: " + nso);

		if (nso instanceof NSAttachableObject)
			this.attachments.remove(nso);
		if (nso instanceof NSConnectableObject)
			this.connectors.remove(nso);
		
		this.name2object.remove(nso.getName().toString());
	}

	public void addAggregate(NSAggregate nsa) {
		this.aggregates.add(nsa);
	}

	public void removeAggregate(NSAggregate nsa) {
		this.aggregates.remove(nsa);
	}
	
	public List findEvents(NSObject parent) {
		List retval = new ArrayList();
		Iterator iter;
		
		iter = this.events.iterator();
		while (iter.hasNext()) {
			NSEvent nse = (NSEvent)iter.next();
			
			if (nse.getParent() == parent) {
				retval.add(nse);
			}
		}
		
		return retval;
	}
	
	public void addEvent(NSEvent nse) {
		if (nse != null)
			this.events.add(nse);
	}
	
	public void removeEvent(NSEvent nse) {
		if (nse != null)
			this.events.remove(nse);
	}

	public void flatten() {
		Iterator iter;
		
		for (iter = this.aggregates.iterator(); iter.hasNext(); ) {
			NSAggregate nsa = (NSAggregate) iter.next();

			nsa.ungroup(this);
		}
	}

	public void unflatten() {
		List list;
		
		list = LinkTraffic.group(this);
		list.removeAll(this.aggregates);
		this.aggregates.addAll(list);
		list = NodeTraffic.group(this);
		list.removeAll(this.aggregates);
		this.aggregates.addAll(list);
	}

	/**
	 * @see java.lang.Object#toString
	 */
	public String toString() {
		return "NSTopology[nodes=" + this.nodes + "; links=" + this.links + "; lans=" + this.lans + "; vtypes="
				+ this.vtypes + "; agents=" + this.agents + "; attachments="
				+ this.attachments + "; connectors=" + this.connectors
				+ "; aggregates=" + this.aggregates + "; routingProtocol="
				+ this.routingProtocol + "; useEndNodeShaping="
				+ this.useEndNodeShaping + "; forceEndNodeShaping="
				+ this.forceEndNodeShaping + "]";
	}
}
