/*
 * Namer.java
 *
 * Copyright (c) 2005 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

package net.emulab.ns;

import java.util.Hashtable;
import java.util.Iterator;
import java.util.TreeSet;

/**
 * Class used to manage unique names in an NS file.
 */
public class Namer {
	/**
	 * Mapping from base names to a set of #UniqueNames.
	 */
	private final Hashtable ht = new Hashtable();

	/**
	 * Construct an empty Namer object.
	 */
	public Namer() {
	}

	/**
	 * Clear all of the names from this object.
	 */
	public synchronized void clear() {
		this.ht.clear();
	}

	/**
	 * Put a new UniqueName into this object.
	 * 
	 * @param un
	 *            The UniqueName to add.
	 * @throws IllegalArgumentException
	 *             if un is null or the name already exists.
	 */
	public synchronized void put(UniqueName un) {
		TreeSet ts;

		if (un == null)
			throw new IllegalArgumentException("un is null");

		if ((ts = (TreeSet) this.ht.get(un.getBase())) == null) {
			/* First time we've seen this base name, add a new set. */
			ts = new TreeSet();
			ts.add(un);
			this.ht.put(un.getBase(), ts);
		} else if (ts.contains(un)) {
			System.out.println("ts " + ts);
			throw new IllegalArgumentException("Name is not unique: " + un);
		} else {
			ts.add(un);
		}
	}

	/**
	 * Check if this namer contains the given UniqueName.
	 * 
	 * @param un
	 *            The UniqueName to check for.
	 * @return True if the name is contained in this object, false otherwise.
	 */
	public synchronized boolean contains(UniqueName un) {
		boolean retval = false;
		TreeSet ts;

		if (un == null)
			throw new IllegalArgumentException("un is null");

		if (((ts = (TreeSet) this.ht.get(un.getBase())) != null)
				&& ts.contains(un)) {
			retval = true;
		}
		return retval;
	}

	/**
	 * Remove a UniqueName from this object.
	 * 
	 * @param un
	 *            The name to remove.
	 */
	public synchronized void remove(UniqueName un) {
		TreeSet ts;

		if (un == null)
			throw new IllegalArgumentException("un is null");

		if ((ts = (TreeSet) this.ht.get(un.getBase())) == null) {
			throw new IllegalArgumentException("un not in table: " + un);
		} else if (!ts.contains(un)) {
			throw new IllegalArgumentException("un not in set: " + un);
		} else {
			ts.remove(un);
			if (ts.size() == 0) {
				// Last one, dump the set.
				this.ht.remove(un.getBase());
			}
		}
	}

	/**
	 * Suggest a new UniqueName with a given base and a unique suffix.
	 * 
	 * @param base
	 *            The base of the name.
	 * @return A new UniqueName.
	 */
	public synchronized UniqueName suggest(String base) {
		UniqueName retval = null;
		TreeSet ts;

		if (base == null)
			throw new IllegalArgumentException("base is null");
		if ("".equals(base))
			throw new IllegalArgumentException("base is empty");

		if ((ts = (TreeSet) this.ht.get(base)) == null) {
			/* Its a new base. */
			retval = new UniqueName(base, "0");
		} else {
			UniqueName last = null;
			Iterator iter;

			if (ts.size() == 0)
				throw new InternalError();

			/* The base exists, search for a unique suffix. */
			iter = ts.iterator();
			while (iter.hasNext()) {
				UniqueName un = (UniqueName) iter.next();

				/* Look for a gap in the last name and the current one. */
				if ((last != null) && (un.compareTo(last) > 1)) {
				} else {
					last = un; // Record the last name encountered.
				}
			}

			/*
			 * At this point we have either a reference to a name that is at the
			 * start of a gap, or, a reference to the last element in the set.
			 */
			retval = last.increment();
		}

		/* Finally, add the new name to this object. */
		this.put(retval);

		return retval;
	}

	/**
	 * @see java.lang.Object#toString()
	 */
	public String toString() {
		return "Namer[ht=" + this.ht + "]";
	}
}
