/*
 * UniqueName.java
 *
 * Copyright (c) 2005 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

package net.emulab.ns;

import java.io.Serializable;

/**
 * Class that represents a unique variable name in the NS file.
 */
public class UniqueName implements Comparable, Serializable {
	/**
	 * Unique identifier for serialization.
	 */
	static final long serialVersionUID = -2479884330696614552L;

	// public static final UniqueName ANONYMOUS_NAME = new UniqueName("", "", "(anonymous)");

	public static final String INTERNAL_NAME = "ncinternal";
	
	/**
	 * Find the index of the "suffix", where suffix is defined as one or more
	 * digits at the end of a string. For example:
	 * <p>
	 * <code>
	 * indexOfSuffix("Foo0") == 3<br>
	 * indexOfSuffix("Foo") == -1<br>
	 * indexOfSuffix("Bar23") == 3<br>
	 * indexOfSuffix("Bar1a23") == 5<br>
	 * </code>
	 * 
	 * @param str
	 *            The string to analyze.
	 * @return The index of the first character in the suffix or -1 if no suffix
	 *         was found.
	 */
	public static int indexOfSuffix(String str) {
		int lpc, length, retval = -1;
		boolean done = false;

		if (str == null)
			throw new IllegalArgumentException("str is null");

		length = str.length();
		/* Look backwards for the integer suffix. */
		for (lpc = length - 1; (lpc >= 0) && !done; lpc--) {
			char ch;

			ch = str.charAt(lpc);
			if (Character.isDigit(ch)) {
				retval = lpc;
			} else {
				/*
				 * We have found the end of the base name, although that does
				 * not necessarily mean we have found a suffix.
				 */
				done = true;
			}
		}
		return retval;
	}

	/**
	 * Split the base of the name off from its suffix.
	 * 
	 * @param str
	 *            The string to split.
	 * @return The base part of the given string.
	 */
	public static String splitBase(String str) {
		String retval;
		int index;

		if ((index = indexOfSuffix(str)) == -1) {
			retval = str;
		} else {
			retval = str.substring(0, index);
		}
		return retval;
	}

	/**
	 * Split the suffix of the name off from its base.
	 * 
	 * @param str
	 *            The string to split.
	 * @return The suffix part of the given string or the empty string if there
	 *         is no suffix.
	 */
	public static String splitSuffix(String str) {
		String retval;
		int index;

		if ((index = indexOfSuffix(str)) == -1) {
			retval = "";
		} else {
			retval = str.substring(index);
		}
		return retval;
	}

	/**
	 * The base part of the unique name.
	 */
	private final String base;

	/**
	 * The suffix part of the unique name, needs to be a number.
	 */
	private final String suffix;

	/**
	 * The full name, as a single string.
	 */
	private final String fullName;

	/**
	 * Construct a UniqueName object with the given values.
	 * 
	 * @param base
	 *            The base part of the unique name.
	 * @param suffix
	 *            The suffix part of the unique name, needs to be a number.
	 * @param fullName
	 *            The full name (i.e. base + suffix).
	 * @throws IllegalArgumentException
	 *             if base starts with a digit or is empty.
	 */
	private UniqueName(String base, String suffix, String fullName) {
		if (!"(anonymous)".equals(fullName)) {
			if (base == null)
				throw new IllegalArgumentException("base is null");
			if ("".equals(base))
				throw new IllegalArgumentException("base is empty");
			if (!Character.isLetter(base.charAt(0))) {
				throw new IllegalArgumentException(
						"Name must start with a letter");
			}

			if (suffix == null)
				throw new IllegalArgumentException("suffix is null");
			if (!"".equals(suffix)) {
				try {
					Long.parseLong(suffix);
				} catch (NumberFormatException e) {
					throw new IllegalArgumentException(
							"suffix is not a number: " + suffix);
				}
			}
		}

		this.base = base;
		this.suffix = suffix;
		this.fullName = fullName;
	}

	/**
	 * Construct a UniqueName object with the given values.
	 * 
	 * @param base
	 *            The base part of the unique name.
	 * @param suffix
	 *            The suffix part of the unique name, needs to be a number.
	 * @throws IllegalArgumentException
	 *             if base starts with a digit or is empty.
	 */
	public UniqueName(String base, String suffix) {
		this(base, suffix, base + suffix);
	}

	/**
	 * Construct a UniqueName object from a well-formed string.
	 * 
	 * @param fullName
	 *            The name to split into a base and optional suffix.
	 */
	public UniqueName(String fullName) {
		this(splitBase(fullName), splitSuffix(fullName), fullName);
	}

	/**
	 * @return The base part of the unique name.
	 */
	public String getBase() {
		return this.base;
	}

	/**
	 * @return The suffix part of the unique name, should be a number or an
	 *         empty string if there is no suffix.
	 */
	public String getSuffix() {
		return this.suffix;
	}

	/**
	 * @return The suffix as a parsed long, or -1 if there is no suffix.
	 */
	public long getSuffixNumber() {
		long retval = -1L;

		if (!"".equals(this.suffix))
			retval = Long.parseLong(this.suffix);

		return retval;
	}

	/**
	 * The comparison is first done on the base part of the name, ignoring case,
	 * and then on the suffix interpreted as a number. In the second case the
	 * return value will be the difference of the suffixes. For example:
	 * <p>
	 * <code>
	 * un1 = new UniqueName("Foo1");<br>
	 * un2 = new UniqueName("Foo12");<br>
	 * <br>
	 * un1.compareTo(un2) == -11;<br>
	 * </code>
	 * 
	 * @see java.lang.Comparable#compareTo(java.lang.Object)
	 */
	public int compareTo(Object obj) {
		UniqueName un;
		int retval;

		if (!(obj instanceof UniqueName))
			throw new IllegalArgumentException("obj is not a UniqueName");

		un = (UniqueName) obj;
		retval = this.base.compareToIgnoreCase(un.base);
		if (retval == 0) {
			long thisNumber, objNumber;
			
			/* Bases are the same, compare the numbers. */
			thisNumber = this.getSuffixNumber();
			objNumber = un.getSuffixNumber();
			retval = (int) (thisNumber - objNumber);
			if (retval == 0) {
				retval = this.suffix.compareTo(un.suffix);
			}
		}

		return retval;
	}

	/**
	 * Produce a new UniqueName using this name as a base and incrementing the
	 * suffix by one. If this name's suffix is empty, the new name's suffix is
	 * "2".
	 * 
	 * @return The newly created unique name.
	 */
	public UniqueName increment() {
		UniqueName retval;

		if ("".equals(this.suffix)) {
			retval = new UniqueName(this.base, "2");
		} else {
			long number;

			number = this.getSuffixNumber();
			retval = new UniqueName(this.base, String.valueOf(number + 1));
		}
		return retval;
	}

	/**
	 * @see java.lang.Object#hashCode()
	 */
	public int hashCode() {
		return this.fullName.hashCode();
	}

	/**
	 * @see java.lang.Object#equals(java.lang.Object)
	 */
	public boolean equals(Object obj) {
		boolean retval = false;

		if (obj == this) {
			retval = true;
		} else if (obj instanceof UniqueName) {
			UniqueName un = (UniqueName) obj;

			if (this.fullName.equals(un.fullName)) {
				retval = true;
			}
		}
		return retval;
	}

	/**
	 * Produce a string that is suitable for a constrained display area. If the
	 * length of the name is smaller than the requested max, the name is
	 * returned as normal. Otherwise, the middle part of the name is replaced
	 * with two periods (".."). For example:
	 * <p>
	 * <code>
	 * new UniqueName("node0").toShortString(8) -> "node0"<br>
	 * <br>
	 * new UniqueName("reallylongname").toShortString(8) -> "rea..ame"<br>
	 * </code>
	 * 
	 * @param max
	 *            The maximum number of characters allowed in the returned
	 *            string.
	 * @return The short version of this unique name.
	 */
	public String toShortString(int max) {
		String retval;

		/* Try to use the regular name before */
		if (this.fullName.length() <= max) {
			retval = this.toString();
		}
		/* ... making a shortened version */
		else {
			int llen, lindex, rindex;
			String longString;

			longString = this.toString();
			llen = longString.length();
			lindex = (max / 2) - 1; // End of the left side
			rindex = llen - (max - lindex - 2); // Start of the right side
			retval = longString.substring(0, lindex) + ".."
					+ longString.substring(rindex);
		}
		return retval;
	}

	/**
	 * @return The unique name with no extra formatting.
	 * @see java.lang.Object#toString()
	 */
	public String toString() {
		return this.fullName;
	}
}
