/*
 * TipTrustManager.java
 *
 * Copyright (c) 2005 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */
package net.emulab.tip;

import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.security.cert.CertificateException;
import java.security.cert.X509Certificate;
import java.util.Arrays;

import javax.net.ssl.X509TrustManager;

/**
 * An X509TrustManager that checks a certification against the fingerprint
 * contained in the Emulab console ACL.
 */
public class TipTrustManager implements X509TrustManager {
	/**
	 * Radix for hexadecimal numbers.
	 */
	private static final int HEX_RADIX = 16;

	/**
	 * Convert a hexadecimal encoded string into a byte array.
	 * 
	 * TODO: Move elsewhere...
	 * 
	 * @param str
	 *            The hexadecimal encoded string to convert.
	 * @return The byte array that was encoded in the given string.
	 * @throws NumberFormatException
	 *             If there was a problem while decoding the string.
	 */
	private static byte[] hexToBytes(String str) throws NumberFormatException {
		byte retval[] = new byte[str.length() / 2];
		int lpc;

		for (lpc = 0; lpc < retval.length; lpc++) {
			int lo, hi;

			hi = Character.digit(str.charAt(lpc * 2), HEX_RADIX);
			lo = Character.digit(str.charAt(lpc * 2 + 1), HEX_RADIX);
			if (hi == -1 || lo == -1)
				throw new NumberFormatException(str);
			retval[lpc] = (byte) ((hi << 4) | lo);
		}
		return retval;
	}

	/**
	 * The SHA-1 fingerprint of the certificate, as obtained from the Emulab
	 * console ACL.
	 */
	private final byte certFingerPrint[];

	/**
	 * Construct a TipTrustManager with the given argument.
	 * 
	 * @param certFingerPrint
	 *            The SHA-1 fingerprint of the certificate, as obtained from the
	 *            Emulab console ACL.
	 */
	public TipTrustManager(byte certFingerPrint[]) {
		this.certFingerPrint = certFingerPrint;
	}

	/**
	 * 
	 * @param certFingerPrint
	 *            The SHA-1 fingerprint of the certificate, encoded as a
	 *            hexadecimal string.
	 */
	public TipTrustManager(String certFingerPrint) {
		this(hexToBytes(certFingerPrint));
	}

	/**
	 * @see javax.net.ssl.X509TrustManager#checkClientTrusted(java.security.cert.X509Certificate[],
	 *      java.lang.String)
	 */
	public void checkClientTrusted(X509Certificate chain[], String authType) {
	}

	/**
	 * @see javax.net.ssl.X509TrustManager#checkServerTrusted(java.security.cert.X509Certificate[],
	 *      java.lang.String)
	 */
	public void checkServerTrusted(X509Certificate chain[], String authType)
			throws CertificateException {
		try {
			MessageDigest md = MessageDigest.getInstance("SHA");
			byte certHash[];

			certHash = md.digest(chain[0].getEncoded());
			if (!Arrays.equals(certHash, this.certFingerPrint)) {
				System.out.println("no joy!");
				throw new CertificateException("Certificate for "
						+ chain[0].getSubjectDN()
						+ " does not match finger print");
			}
		} catch (NoSuchAlgorithmException e) {
			throw new CertificateException(
					"Unable to verify certificate because: " + e.getMessage());
		}
	}

	/**
	 * @see javax.net.ssl.X509TrustManager#getAcceptedIssuers()
	 */
	public X509Certificate[] getAcceptedIssuers() {
		return null;
	}

	/**
	 * @see java.lang.Object#toString()
	 */
	public String toString() {
		return "TipTrustManager[certFingerPrint=" + this.certFingerPrint + "]";
	}
}
