/*
 * TipTunnel.java
 *
 * Copyright (c) 2005 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

package net.emulab.tip;

import java.io.DataInputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.net.Socket;
import java.security.NoSuchAlgorithmException;
import java.security.SecureRandom;
import java.util.Hashtable;

import javax.net.ssl.SSLContext;
import javax.net.ssl.SSLSocket;
import javax.net.ssl.SSLSocketFactory;
import javax.net.ssl.TrustManager;

import net.emulab.util.ProgressHalter;
import net.emulab.util.ProgressObserver;
import net.emulab.util.UnixSecureRandom;

/**
 * Java version of the tiptunnel command.
 */
public class TipTunnel implements Runnable, ProgressHalter {
	/**
	 * Reply code from the server that indicates everything is OK to proceed.
	 */
	private static final byte CAPOK = 0;

	/**
	 * Reply code from the server that indicates that another connection is
	 * active.
	 */
	private static final byte CAPBUSY = 1;

	/**
	 * Reply code from the server that indicates that the given key was invalid.
	 */
	private static final byte CAPNOPERM = 2;

	/**
	 * The #ProgressObserver that is interested in this connection.
	 */
	private final ProgressObserver po;

	/**
	 * The domain name of the tip server.
	 */
	private final String server;

	/**
	 * The port number on the tip server that provides access for the node.
	 */
	private final int portNumber;

	/**
	 * The secret key that authorizes us to interact with the console.
	 */
	private final String keyData;

	/**
	 * The trust manager that verifies the tip server's SSL certificate.
	 */
	private final TipTrustManager ttm;

	/**
	 * The SSLSocket connected to capture on the tip server.
	 */
	private SSLSocket socket;

	/**
	 * Buffer used to send/receive data.
	 */
	private final byte buffer[] = new byte[260];

	/**
	 * The TunnelTelnetD instance that tunnels data between the user's telnet
	 * session and capture.
	 */
	private TunnelTelnetD ttd;

	/**
	 * Construct a TipTunnel object with the given values.
	 * 
	 * @param po
	 *            The #ProgressObserver that is interested in this connection.
	 * @param server
	 *            The domain name of the tip server.
	 * @param portNumber
	 *            The port number on the tip server that provides access for the
	 *            node.
	 * @param keyData
	 *            The secret key that authorizes us to interact with the
	 *            console.
	 * @param ttm
	 *            The trust manager that verifies the tip server's SSL
	 *            certificate.
	 */
	public TipTunnel(ProgressObserver po, String server, int portNumber,
			String keyData, TipTrustManager ttm) {
		this.po = po;
		this.server = server;
		this.portNumber = portNumber;
		this.keyData = keyData;
		this.ttm = ttm;

		this.po.setProgressHalter(null, this);
	}

	/**
	 * Construct a TipTunnel object with the given values.
	 * 
	 * @param po
	 *            The #ProgressObserver that is interested in this connection.
	 * @param acl
	 *            The console ACL, with entries for: "server", "portnum",
	 *            "keydata", and "certsha1".
	 */
	public TipTunnel(ProgressObserver po, Hashtable acl) {
		this(po, (String) acl.get("server"), ((Integer) acl.get("portnum"))
				.intValue(), (String) acl.get("keydata"), new TipTrustManager(
				(String) acl.get("certsha1")));
	}

	/**
	 * Marshal the secret key data to send to the server.
	 * 
	 * @param key
	 *            The key as a hexadecimal-encoded string.
	 * @return The length of the data to be sent to the server.
	 */
	public int marshalKey(String key) {
		int len = key.length();

		this.buffer[0] = (byte) (len);
		this.buffer[1] = (byte) (len >> 8);
		this.buffer[2] = (byte) (len >> 16);
		this.buffer[3] = (byte) (len >> 24);
		System.arraycopy(key.getBytes(), 0, this.buffer, 4, len);
		this.buffer[4 + len] = 0;

		return 4 + 256;
	}

	/**
	 * Connect and authenticate with the tip server.
	 * 
	 * TODO: Refactor and don't just throw Throwable.
	 * 
	 * @return The port number where the TunnelTelnetD is listening.
	 * @throws Throwable
	 *             if anything goes wrong during the connection.
	 */
	public int connect() throws Throwable {
		SecureRandom sr;

		this.po.setStage(null, "Initializing RNG...");
		try {
			sr = UnixSecureRandom.getInstance("SHA1PRNG");
		} catch (NoSuchAlgorithmException e) {
			sr = new SecureRandom();
		} catch (IOException e) {
			sr = SecureRandom.getInstance("SHA1PRNG");
		}
		sr.nextInt();
		this.po.setStage(null, "Connecting to tip server...");
		this.po.setContentLength(null, 1);

		Socket ps = new Socket(this.server, this.portNumber);
		this.po.setStage(null, "Getting SSLContext...");
		SSLContext sc = SSLContext.getInstance("SSL");
		OutputStream os = ps.getOutputStream();

		this.po.setStage(null, "Init SSLContext...");
		sc.init(null, new TrustManager[] { this.ttm }, sr);

		this.po.setStage(null, "Get factory...");
		SSLSocketFactory ssf = sc.getSocketFactory();

		os.write(this.buffer, 0, this.marshalKey("USESSL"));
		os.flush();

		this.po.setStage(null, "Switching to SSL...");

		this.socket = (SSLSocket) ssf.createSocket(ps, this.server,
				this.portNumber, true);

		this.po.setStage(null, "Sending key...");

		try {
			os = this.socket.getOutputStream();
			os.write(this.buffer, 0, this.marshalKey(this.keyData));
		} catch (IOException e) {
			// Ignore for now...
		}

		DataInputStream dis = new DataInputStream(this.socket.getInputStream());

		byte capret[] = new byte[4];

		this.po.setStage(null, "Waiting for response...");
		this.po.setContentLength(null, capret.length);

		dis.readFully(capret);

		this.po.updateContentRemaining(null, capret.length);

		switch (capret[0]) {
		case CAPOK:
			break;
		case CAPBUSY:
			throw new IOException("Another console is already running");
		case CAPNOPERM:
			throw new SecurityException(
					"You do not have permission to view this console");
		default:
			throw new IOException("Unknown response from the server: "
					+ capret[0]);
		}

		this.po.setStage(null, "Waiting for response...");

		this.ttd = new TunnelTelnetD(this.socket);

		return this.ttd.getServerSocket().getLocalPort();
	}

	/**
	 * Waits for the client to connect and then starts tunnelling.
	 * 
	 * @see java.lang.Runnable#run()
	 */
	public void run() {
		try {
			this.ttd.waitForClient();
			this.ttd.startTunnelling();
		} catch (IOException e) {
			e.printStackTrace(System.err);
		}
	}

	/**
	 * @see net.emulab.netlab.client.ProgressHalter#haltProgress()
	 */
	public void haltProgress() {
		if (this.socket != null) {
			try {
				this.socket.close();
			} catch (IOException e) {
				e.printStackTrace(System.err);
			}
		}
		if (this.ttd != null)
			this.ttd.stopTunnelling();
	}

	/**
	 * @see java.lang.Object#toString()
	 */
	public String toString() {
		return "TipTunnel[po=" + this.po + "; server=" + this.server
				+ "; portNumber=" + this.portNumber + "; keyData="
				+ this.keyData + "; ttm=" + this.ttm + "]";
	}
}
