/*
 * TunnelTelnetD.java
 *
 * Copyright (c) 2005 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

package net.emulab.tip;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.InetAddress;
import java.net.ServerSocket;
import java.net.Socket;
import java.net.UnknownHostException;

/**
 * Custom telnetd that acts as a tunnel between a telnet session on the user's
 * machine and the tip server.
 */
public class TunnelTelnetD {
	/**
	 * Initial set of telnet commands to send to the user's client.
	 */
	private static final byte INITIAL_COMMANDS[] = {
	// IAC WILL SUPPRESS GO AHEAD:
			(byte) 255, (byte) 251, 3,
			// IAC WILL ECHO:
			(byte) 255, (byte) 251, 1, };

	/**
	 * The ServerSocket to listen on for connections from the user.
	 */
	private ServerSocket ss;

	/**
	 * The socket connected to the tip server.
	 */
	private final Socket server;

	/**
	 * The socket connected to the user's telnet client.
	 */
	private Socket client;

	/**
	 * Construct a TunnelTelnetD object with the given value.
	 * 
	 * @param server
	 *            The server to tunnel data to/from.
	 * @throws IOException
	 *             if there is a problem creating the ServerSocket.
	 */
	public TunnelTelnetD(Socket server) throws IOException {
		try {
			this.server = server;
			this.ss = new ServerSocket(0, 1, InetAddress.getByName("127.0.0.1"));
		} catch (UnknownHostException e) {
			throw new InternalError();
		}
	}

	/**
	 * @return The ServerSocket to listen on for connections from the user.
	 */
	public ServerSocket getServerSocket() {
		return this.ss;
	}

	/**
	 * Wait for a telnet client to connect to our ServerSocket, then close the
	 * ServerSocket, and send the INITIAL_COMMANDS to the user's client.
	 * 
	 * @throws IOException
	 *             if there was a problem accepting the connection from the
	 *             client or sending the INITIAL_COMMANDS.
	 */
	public void waitForClient() throws IOException {
		this.client = this.ss.accept();
		this.ss.close();
		this.ss = null;
		this.client.getOutputStream().write(INITIAL_COMMANDS);
	}

	/**
	 * Starting tunnelling data between the user and the server. This method
	 * will return when either side of the tunnel disconnects.
	 */
	public void startTunnelling() {
		/* Create a thread to tunnel from the client to the server and */
		Thread clientToServer = new Thread() {
			public void run() {
				TunnelTelnetD ttd = TunnelTelnetD.this;

				try {
					ttd.tunnel(ttd.client.getInputStream(), ttd.server
							.getOutputStream());
				} catch (IOException e) {
					e.printStackTrace(System.err);
				} finally {
					ttd.stopTunnelling();
				}
			}
		};

		clientToServer.setDaemon(true);
		clientToServer.start();

		/* ... do the tunnel from the server to the client in this thread. */
		try {
			this.tunnel(this.server.getInputStream(), this.client
					.getOutputStream());
		} catch (IOException e) {
			e.printStackTrace(System.err);
		} finally {
			this.stopTunnelling();
		}

		/* Tunnel is shut down, wait for the other thread. */
		try {
			clientToServer.join();
		} catch (InterruptedException e) {
			e.printStackTrace(System.err);
		}
	}

	/**
	 * Tunnel data from one end of the connection to the other.
	 * 
	 * @param is
	 *            The stream to source end of the tunnel.
	 * @param os
	 *            The stream to sink end of the tunnel.
	 * @throws IOException
	 *             if there was a problem reading from the source or writing to
	 *             the sink.
	 */
	public void tunnel(InputStream is, OutputStream os) throws IOException {
		byte buffer[] = new byte[1024];
		boolean done = false;

		while (!done) {
			int rc;

			if ((rc = is.read(buffer)) > 0) {
				if ((buffer[0] & 0xff) != 255) {
					if (buffer[rc - 1] == 0)
						rc -= 1;
					os.write(buffer, 0, rc);
				}
			} else {
				done = true;
			}
		}
	}

	/**
	 * Stop tunnelling data.
	 */
	public void stopTunnelling() {
		try {
			this.server.close();
		} catch (IOException e) {
			e.printStackTrace(System.err);
		}
		/* Try closing the client. */
		if (this.client != null) {
			try {
				this.client.close();
			} catch (IOException e) {
				e.printStackTrace(System.err);
			}
		}
	}

	/**
	 * @see java.lang.Object#toString()
	 */
	public String toString() {
		return "TunnelTelnetD[" + "]";
	}
}
