package net.emulab.util;

import java.net.URL;

/**
 * Interface for classes that are interested in the progress of some operation.
 * The protocol for using this class is something like:
 * <p>
 * 
 * <pre>
 * po.setProgressHalter(url, ph); // Make sure the user can stop as early as
 * // possible.
 * po.setStage(url, &quot;Connecting to server...&quot;);
 * connect();
 * po.setStage(url, &quot;Receiving data...&quot;);
 * contentLength = readHeader();
 * po.setContentLength(url, contentLength);
 * while (moreData()) {
 * 	contentLength -= readData();
 * 	po.updateContentRemaining(url, contentLength);
 * }
 * </pre>
 */
public interface ProgressObserver {
	/**
	 * @param url
	 *            The URL of interest.
	 * @param ph
	 *            The ProgressHalter that will halt an operation in progress.
	 */
	void setProgressHalter(URL url, ProgressHalter ph);

	/**
	 * Tell the observer, using a "user comprehensible" string, that a new stage
	 * in the operation has been reached. For example, a downloader might use
	 * stage names like:
	 * 
	 * <ul>
	 * <li>Connecting to server...
	 * <li>Sending request...
	 * <li>Downloading content...
	 * <li>Disconnecting from server...
	 * </ul>
	 * 
	 * @param url
	 *            The URL of interest.
	 * @param stageName
	 *            The current stage of the operation in progress.
	 */
	void setStage(URL url, String stageName);

	/**
	 * @param url
	 *            The URL of interest.
	 * @param contentLength
	 *            The overall amount of data being transferred.
	 */
	void setContentLength(URL url, long contentLength);

	/**
	 * Update the observer on how much data is left to transfer.
	 * 
	 * @param url
	 *            The URL of interest.
	 * @param contentRemaining
	 *            The amount of data that has yet to be received.
	 */
	void updateContentRemaining(URL url, long contentRemaining);
}
