package net.emulab.util;

import java.io.FileInputStream;
import java.io.IOException;
import java.security.NoSuchAlgorithmException;
import java.security.SecureRandom;

/**
 * Helper class used to seed a SecureRandom from /dev/urandom rather than
 * waiting for the JDK's time consuming seed generator.
 */
public final class UnixSecureRandom {
	/**
	 * The random device name in the file system.
	 */
	private static final String DEV_URANDOM;

	static {
		DEV_URANDOM = System.getProperty("net.emulab.random", "/dev/urandom");
	}

	/**
	 * Get a SecureRandom instance that implements the given algorithm and is
	 * seeded from the random device.
	 * 
	 * @param algorithm
	 *            The SecureRandom algorithm.
	 * @return A seeded SecureRandom.
	 * @throws NoSuchAlgorithmException
	 *             if the requested algorithm does not exist.
	 * @throws IOException
	 *             if there was a problem
	 */
	public static SecureRandom getInstance(String algorithm)
			throws NoSuchAlgorithmException, IOException {
		byte seed[] = new byte[32];
		FileInputStream fis;
		SecureRandom retval;

		retval = SecureRandom.getInstance(algorithm);
		fis = new FileInputStream(DEV_URANDOM);
		try {
			if (fis.read(seed) == seed.length)
				retval.setSeed(seed);
			else
				throw new IOException("Short read from random device");
		} finally {
			fis.close();
		}
		return retval;
	}

	/**
	 * Private constructor for static-only class.
	 */
	private UnixSecureRandom() {
	}
}
