/*
 * SshConnection.java
 *
 * Copyright (c) 2005 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

package net.emulab.xmlrpc;

import java.io.FilePermission;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.ProtocolException;
import java.net.URL;
import java.net.URLConnection;
import java.security.Permission;

import net.emulab.util.ProgressHalter;
import net.emulab.util.ProgressObserver;

/**
 * A URLConnection class for ssh:// URLs.
 */
public class SshConnection extends URLConnection implements ProgressHalter {
	/**
	 * The permission required to open a connection to this URL.
	 */
	private final FilePermission permission;

	/**
	 * The SSH command string used to connect to the remote host.
	 */
	private final String command;

	/**
	 * The ProgressObserver that is interested in this connection.
	 */
	private final ProgressObserver po;

	/**
	 * The handle on the SSH process.
	 */
	private Process tunnel;

	/**
	 * Construct an SshConnection with the given values.
	 * 
	 * @param url
	 *            The URL to make a connection to.
	 * @param sshConfig
	 *            The SSH configuration to use for this connection.
	 * @param po
	 *            The ProgressObserver that is interested in this connection.
	 */
	protected SshConnection(URL url, String sshConfig, ProgressObserver po) {
		super(url);

		String user, flags = "-C ", path;

		if (this.url.getPort() != -1)
			flags += " -p " + this.url.getPort();
		if ((user = this.url.getUserInfo()) != null)
			flags += " -l " + user;
		if (sshConfig != null)
			flags += " -F " + sshConfig;
		if (this.url.getPath().length() == 1) {
			flags = " -t -t -X ";
			path = "worksheet.sh";
		} else {
			path = this.url.getPath().substring(1);
		}
		this.command = "ssh " + flags + " " + this.url.getHost() + " " + path;
		this.permission = new FilePermission(this.command, "execute");
		this.po = po;

		if (SshHandler.DEBUG) {
			System.err.println("new " + this.toString());
		}
	}

	public Process getTunnel() {
		return this.tunnel;
	}

	/**
	 * @see java.net.URLConnection#connect()
	 */
	public void connect() throws IOException {
		if (!this.connected) {
			Thread stderrThread;
			final Process proc;

			if (SshHandler.DEBUG) {
				System.err.println("connect " + this.toString());
			}

			this.tunnel = Runtime.getRuntime().exec(this.command);
			proc = this.tunnel;
			this.connected = true;

			stderrThread = new Thread() {
				public void run() {
					try {
						InputStream is = proc.getErrorStream();
						byte buffer[] = new byte[1024];
						int rc;

						while ((rc = is.read(buffer)) != -1) {
							System.err.write(buffer, 0, rc);
						}
					} catch (Throwable th) {
						th.printStackTrace();
					}
				}
			};
			stderrThread.setDaemon(true);
			stderrThread.start();

			if (SshHandler.DEBUG) {
				System.err.println("post connect " + this.toString());
			}
		}
	}

	/**
	 * @see java.net.URLConnection#getInputStream()
	 */
	public InputStream getInputStream() throws IOException {
		if (!doInput) {
			throw new ProtocolException(
					"Cannot open InputStream if doInput is false");
		}

		if (!connected)
			this.connect();

		return this.tunnel.getInputStream();
	}

	/**
	 * @see java.net.URLConnection#getOutputStream()
	 */
	public OutputStream getOutputStream() throws IOException {
		if (!doOutput) {
			throw new ProtocolException(
					"Cannot open OutputStream if doOutput is false");
		}

		if (!connected)
			this.connect();

		return this.tunnel.getOutputStream();
	}

	/**
	 * @see java.net.URLConnection#getPermission()
	 */
	public Permission getPermission() throws IOException {
		return this.permission;
	}

	/**
	 * @see net.emulab.netlab.client.ProgressHalter#haltProgress()
	 */
	public synchronized void haltProgress() {
		if (connected) {
			this.tunnel.destroy();
			try {
				this.tunnel.getInputStream().close();
			} catch (IOException e) {
				e.printStackTrace(System.err);
			}
			try {
				this.tunnel.getOutputStream().close();
			} catch (IOException e) {
				e.printStackTrace(System.err);
			}
			while (this.connected) {
				try {
					this.tunnel.waitFor();
					this.connected = false;
				} catch (InterruptedException e) {
				}
			}
			this.doInput = false;
			this.doOutput = false;
		}
	}

	/**
	 * @see java.lang.Object#toString()
	 */
	public String toString() {
		return "SshConnection[" + super.toString() + "; permission="
				+ this.permission + "; command=" + this.command
				// + "; po="
				// + this.po
				+ "; tunnel=" + this.tunnel + "]";
	}
}
