/*
 * SshHandler.java
 *
 * Copyright (c) 2005 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

package net.emulab.xmlrpc;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLConnection;
import java.net.URLStreamHandler;

import net.emulab.util.ProgressObserver;

/**
 * A URLStreamHandler for ssh:// URLs.
 */
public class SshHandler extends URLStreamHandler {
	/**
	 * Debugging flag for the net.emulab.xmlrpc package.
	 */
	static final boolean DEBUG = Boolean.getBoolean("net.emulab.xmlrpc.DEBUG");

	/**
	 * Check for legal characters in an ssh:// URL.
	 * 
	 * TODO: Needs more work.
	 * 
	 * @param str
	 *            The URL to check.
	 * @return True if the URL is legal, false otherwise.
	 */
	private static boolean legal(String str) {
		boolean retval = true;
		char array[];
		int lpc;

		array = str.toCharArray();
		for (lpc = 0; (lpc < array.length) && retval; lpc++) {
			if (!(Character.isLetterOrDigit(array[lpc]) || (array[lpc] == '/')
					|| (array[lpc] == '.') || (array[lpc] == '-') || (array[lpc] == '_'))) {
				retval = false;
			}
		}
		return retval;
	}

	/**
	 * The SSH configuration to use when making a connection.
	 */
	private final String sshConfig;

	/**
	 * The ProgressObserver to update when downloading data.
	 */
	private final ProgressObserver po;

	/**
	 * Construct an SshHandler with the given values.
	 * 
	 * @param sshConfig
	 *            The SSH configuration to use when making a connection.
	 * @param po
	 *            The ProgressObserver to update when downloading data.
	 * @throws FileNotFoundException
	 *             if the given sshConfig could not be found.
	 */
	public SshHandler(String sshConfig, ProgressObserver po)
			throws FileNotFoundException {
		if ((sshConfig != null) && !(new File(sshConfig).exists()))
			throw new FileNotFoundException(sshConfig);

		this.sshConfig = sshConfig;
		this.po = po;
	}

	/**
	 * Construct an SshHandler with the given values.
	 * 
	 * @param po
	 *            The ProgressObserver to update when downloading data.
	 */
	public SshHandler(ProgressObserver po) {
		/*
		 * Note: We do not use the two argument constructor since it throws an
		 * exception.
		 */
		this.sshConfig = null;
		this.po = po;
	}

	/**
	 * Construct an SshHandler with the default values.
	 */
	public SshHandler() {
		this(null);
	}

	/**
	 * @see java.net.URLStreamHandler#openConnection(java.net.URL)
	 */
	protected URLConnection openConnection(URL url) throws IOException {
		if (!"ssh".equals(url.getProtocol())) {
			throw new MalformedURLException("This handler only supports ssh: "
					+ url);
		}

		if (!legal(url.getPath())) {
			throw new MalformedURLException("Invalid path in URL: " + url);
		}

		return new SshConnection(url, this.sshConfig, this.po);
	}

	/**
	 * @see java.net.URLStreamHandler#getDefaultPort()
	 */
	protected int getDefaultPort() {
		return 22;
	}

	/**
	 * @return The name of the SshConfig file, or null if there is none.
	 */
	public String getSshConfig() {
		return this.sshConfig;
	}

	/**
	 * @see java.lang.Object#toString()
	 */
	public String toString() {
		return "SshHandler[sshConfig=" + this.sshConfig + "; po=" + this.po
				+ "]";
	}
}
