/*
 * SshXmlRpcTransport.java
 *
 * Copyright (c) 2005 The University of Utah and the Flux Group.
 * All rights reserved.
 *
 * This file is licensed under the terms of the GNU Public License.  
 * See the file "license.terms" for restrictions on redistribution 
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */

package net.emulab.xmlrpc;

import java.io.BufferedInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.URL;
import java.net.URLConnection;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.NoSuchElementException;
import java.util.StringTokenizer;

import net.emulab.util.ProgressHalter;
import net.emulab.util.ProgressObserver;

import org.apache.xmlrpc.XmlRpcClientException;
import org.apache.xmlrpc.XmlRpcTransport;

/**
 * An XmlRpcTransport that runs over SSH.
 */
public class SshXmlRpcTransport implements XmlRpcTransport {
	/**
	 * Cache of open URL connections.
	 */
	private static final Hashtable URL_TO_CONNECTIONS = new Hashtable();

	/**
	 * The number of minutes to wait between flushing open connections.
	 */
	private static final long FLUSH_TIME = 10; // minutes

	/**
	 * The content-length header key.
	 */
	private static final String CONTENT_LENGTH_HEADER = "content-length";

	/**
	 * The user-agent header key.
	 */
	private static final String USER_AGENT_HEADER = "user-agent";

	private static final String PROBE_HEADER = "probe";

	static {
		/*
		 * Start a "reaper" thread that cleans out the URL_TO_CONNECTIONS table
		 * every N minutes.
		 */
		Thread reaper = new Thread() {
			public void run() {
				while (true) {
					try {
						Thread.sleep(FLUSH_TIME * 60 * 1000);
						flushConnections();
					} catch (Throwable th) {
						th.printStackTrace(System.err);
					}
				}
			}
		};

		reaper.setDaemon(true);

		try {
			/* Add a shutdown hook to clean out any remaining connections. */
			Runtime.getRuntime().addShutdownHook(new Thread() {
				public void run() {
					flushConnections();
				}
			});
			reaper.start();
		}
		catch (SecurityException e) {
		}
	}

	private static Hashtable probe(URLConnection uc) throws IOException {
		Hashtable retval = new Hashtable();
		byte buffer[] = new byte[512];
		OutputStream os;
		InputStream is;
		int rc;

		os = uc.getOutputStream();
		is = uc.getInputStream();

		os.write((PROBE_HEADER + ": netbuild\r\n").getBytes());
		os.write("\r\n".getBytes());
		os.flush();

		while ((rc = is.read(buffer)) == buffer.length) {
		}

		try {
			String headers = new String(buffer, 0, rc);
			StringTokenizer st = new StringTokenizer(headers, "\n");

			while (st.hasMoreTokens()) {
				String line = st.nextToken();

				if (!"".equals(line)) {
					StringTokenizer line_st = new StringTokenizer(line, ":");

					System.out.println("line " + line);
					retval.put(line_st.nextToken().trim(), line_st.nextToken()
							.trim());
				}
			}
		} catch (NoSuchElementException e) {
		}

		return retval;
	}

	/**
	 * Get a connection to a given URL. If one is in the cache, it will be
	 * removed and returned instead of creating a new one.
	 * 
	 * @param url
	 *            The URL to get a connection for.
	 * @return An open URLConnection.
	 * @throws IOException
	 *             if there was an error while opening the connection.
	 */
	private static synchronized URLConnection getConnection(URL url)
			throws IOException {
		URLConnection retval;

		/* Try the cache before */
		if (SshHandler.DEBUG) {
			System.err.println("Has existing connection " + url);
		}
		if ((retval = (URLConnection) URL_TO_CONNECTIONS.remove(url)) != null) {
			if (retval.getDoOutput()) /* Make sure its still open. */
			{
				if (SshHandler.DEBUG) {
					System.err.println("Reusing " + retval);
				}
			} else {
				retval = null;
			}
		}

		/* ... making a new one. */
		if (retval == null) {
			retval = url.openConnection();
			retval.setDoOutput(true);
			retval.setDoInput(true);

			// Hashtable ht = probe(retval);

			// System.out.println("ht = " + ht);
		}

		return retval;
	}

	/**
	 * Flush all of the connections from the cache.
	 */
	private static synchronized void flushConnections() {
		Enumeration en;

		en = URL_TO_CONNECTIONS.keys();
		while (en.hasMoreElements()) {
			URL url = (URL) en.nextElement();
			URLConnection uc;

			uc = (URLConnection) URL_TO_CONNECTIONS.get(url);

			if (SshHandler.DEBUG) {
				System.err.println("Flushing " + uc);
			}

			try {
				uc.getInputStream().close();
			} catch (IOException e) {
				e.printStackTrace(System.err);
			}
			try {
				uc.getOutputStream().close();
			} catch (IOException e) {
				e.printStackTrace(System.err);
			}
		}

		URL_TO_CONNECTIONS.clear();
	}

	/**
	 * Put an open connection back in the cache.
	 * 
	 * @param uc
	 *            The connection to return to the cache.
	 */
	static synchronized void putConnection(URLConnection uc) {
		if (uc == null)
			throw new IllegalArgumentException("uc is null");

		URL_TO_CONNECTIONS.put(uc.getURL(), uc);
	}

	/**
	 * The URL of the RPC server.
	 */
	private final URL url;

	/**
	 * The ProgressObserver that is interested in this connection.
	 */
	private final ProgressObserver po;

	/**
	 * Buffer used when transferring data.
	 */
	private byte buffer[];

	/**
	 * Construct an SshXmlRpcTransport with the given values.
	 * 
	 * @param url
	 *            The URL of the RPC server.
	 * @param po
	 *            The ProgressObserver that is interested in this connection.
	 */
	SshXmlRpcTransport(URL url, ProgressObserver po) {
		this.url = url;
		this.po = po;
	}

	/**
	 * Gets a connection to the RPC server and sends the request, all the while
	 * updating the progress observer.
	 * 
	 * @see org.apache.xmlrpc.XmlRpcTransport#sendXmlRpc(byte[])
	 */
	public InputStream sendXmlRpc(byte request[]) throws IOException,
			XmlRpcClientException {
		String badLines = "", line;
		long contentLength = -1;
		BufferedInputStream bis;
		URLConnection uc;
		OutputStream os;

		if (this.po != null)
			this.po.setStage(this.url, "Opening connection");
		if ("ssh".equals(this.url.getProtocol())) {
			uc = getConnection(this.url);
			
			if (this.po != null)
				this.po.setProgressHalter(this.url, (ProgressHalter) uc);
			os = uc.getOutputStream();

			os.write((CONTENT_LENGTH_HEADER + ": " + request.length + "\r\n")
					.getBytes());
			os.write((USER_AGENT_HEADER + ": netlab-client v0.2.0\r\n").getBytes()); // XXX
			os.write("\r\n".getBytes());
		}
		else {
			uc = this.url.openConnection();
			uc.setDoInput(true);
			uc.setDoOutput(true);
			uc.setUseCaches(false);
			uc.setAllowUserInteraction(false);
			uc.setRequestProperty("content-length",
					String.valueOf(request.length));
			uc.setRequestProperty("content-type", "text/xml");
			uc.setRequestProperty(USER_AGENT_HEADER, "netlab-client v0.2.0");
			uc.connect();
			
			// XXX add a halter
			os = uc.getOutputStream();
		}

		if (this.po != null)
			this.po.setStage(this.url, "Sending request");
		os.write(request);
		os.flush();
		if (!"ssh".equals(this.url.getProtocol()))
			os.close();
		
		if (this.po != null)
			this.po.setStage(this.url, "Waiting for response");

		bis = new BufferedInputStream(uc.getInputStream());
		if ("ssh".equals(url.getProtocol())) {
			while (((line = this.readLine(bis)) != null) && !line.equals("")) {
				int index = line.indexOf(':');
				
				if (index == -1) {
					badLines += line;
				} else {
					String key, value;
					
					key = line.substring(0, index);
					value = line.substring(index + 1).trim();
					if (CONTENT_LENGTH_HEADER.equalsIgnoreCase(key))
						contentLength = Long.parseLong(value);
					else
						badLines += line;
				}
			}

			if (contentLength == -1) {
				os.close();
				bis.close();
				throw new IOException("Bad response from: "
						+ this.url.toExternalForm() + "\n" + badLines);
			}
		}
		else {
			contentLength = uc.getContentLength();
		}

		if (this.po != null) {
			this.po.setStage(this.url, "Receiving response");
			this.po.setContentLength(this.url, contentLength);
		}

		return new ServerInputStream(this.url, uc, bis, contentLength, this.po);
	}

	/**
	 * Read a line from the given InputStream.
	 * 
	 * @param is
	 *            The InputStream to read from.
	 * @return The next line in the InputStream.
	 * @throws IOException
	 *             if there was a problem reading from the connection.
	 */
	private String readLine(InputStream is) throws IOException {
		if (this.buffer == null) {
			this.buffer = new byte[2048];
		}
		int next;
		int count = 0;
		while (true) {
			next = is.read();
			if (next < 0 || next == '\n') {
				break;
			}
			if (next != '\r') {
				this.buffer[count++] = (byte) next;
			}
			if (count >= this.buffer.length) {
				throw new IOException("HTTP Header too long");
			}
		}
		return new String(this.buffer, 0, count);
	}

	public void endClientRequest() throws XmlRpcClientException {
	}
}
