package thinlet;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.Iterator;
import java.util.List;
import java.util.Vector;
import java.util.regex.Pattern;

/**
 * An ObjectController for mutable arrays.
 */
public class ArrayController extends ObjectController {
	/**
	 * Perform a one-level deep copy of a collection and return a List.
	 * 
	 * @param elementType
	 *            The type for the elements in the collection.
	 * @param collection
	 *            The collection to copy.
	 * @return A list containing a copy of the elements in the given collection.
	 * @throws CloneNotSupportedException
	 *             if the elementType class does not support the clone()
	 *             operation.
	 */
	public static List deepCopy(Class elementType, Collection collection)
			throws CloneNotSupportedException {
		Vector retval;

		try {
			Method cloneMethod;
			Iterator iterator;

			cloneMethod = elementType.getMethod("clone", null);
			retval = new Vector(collection.size());
			for (iterator = collection.iterator(); iterator.hasNext();) {
				Cloneable obj = (Cloneable) iterator.next();

				retval.add(cloneMethod.invoke(obj, null));
			}
		} catch (InvocationTargetException e) {
			throw new CloneNotSupportedException();
		} catch (NoSuchMethodException e) {
			throw new CloneNotSupportedException();
		} catch (IllegalAccessException e) {
			throw new CloneNotSupportedException();
		}

		return retval;
	}

	/**
	 * White space pattern used to split up the search string into individual
	 * items.
	 */
	private static final Pattern WHITE_SPACE_PATTERN = Pattern
			.compile("[ \r\n\t]");

	/**
	 * Typed version of the content object.
	 */
	protected List contentArray;

	/**
	 * The sorted and filtered version of the content array.
	 */
	protected List Objects;

	/**
	 * The key paths to scan when filtering out elements in the content array.
	 */
	protected final Vector filterKeys = new Vector();

	protected Object selection = ThinletController.NO_SELECTION;

	protected int selectedIndex = -1;

	/**
	 * The search string to use when filtering out elements in the content
	 * array.
	 */
	protected String searchKey;

	/**
	 * Tokenized and lowercased version of the searchKey.
	 */
	protected String searchTokens[];

	/**
	 * The active sort key.
	 */
	protected String activeKey;

	/**
	 * The direction to sort the elements, one of: ascent, descent, none.
	 */
	protected String sortDirection;

	/**
	 * The maximum number of elements allowed in the arrangedObjects array.
	 */
	protected int maximumArraySize = Integer.MAX_VALUE;

	protected boolean isWithinMaximum = true;

	/**
	 * Construct an ArrayController with the given values.
	 * 
	 * @param thinlet
	 *            The thinlet instance that this controller will be working
	 *            with.
	 * @param content
	 *            The initial content to manage.
	 */
	public ArrayController(Thinlet thinlet, List content) {
		super(thinlet, content);

		this.tkv.addKeyDependency(this, "isMultiple", "content");
		this.tkv.addKeyDependency(this, "canRemove", "selection");
		this.tkv.addKeyDependency(this, "canAccept", "selection");
		this.tkv.addKeyDependency(this, "canMoveDown", "selectedIndex");
		this.tkv.addKeyDependency(this, "canMoveUp", "selectedIndex");

		/* Relay changes to 'selection.<field>' to 'arrangedObjects.<field> */
		tkv.observeKeyValue(this, "selection", new ThinletKeyValueObserver() {
			public void observedKeyValue(Object obj, String keyPath, int flags,
					Object oldValue, Object newValue) {
				String akey = ArrayController.this.activeKey;
				
				if ("selection".equals(keyPath)) {
				} else if (akey != null && keyPath.regionMatches(10,
						akey, 0, akey.length())) {
					ArrayController.this.rearrangeObjects();
				} else if (selection != ThinletController.NO_SELECTION) {
					tkv.notifyKeyValue(obj, "arrangedObjects",
							ThinletKeyValue.CHANGE_REPLACE, null, new int[] {
									selectedIndex, selectedIndex + 1 });
				}
			}
		});

		tkv.observeKeyValue(this, "content", new ThinletKeyValueObserver() {
			public void observedKeyValue(Object obj, String keyPath, int flags,
					Object oldValue, Object newValue) {
				if (flags == ThinletKeyValue.CHANGE_SETTING) {
				} else {
					rearrangeObjects();
					if (keyPath.startsWith("content.")) {
						String sub = "selection." + keyPath.substring(8);
						
						tkv.notifyKeyValue(obj, sub, ThinletKeyValue.CHANGE_SETTING, null,
								tkv.getKeyValue(obj, sub));
					}
				}
			}
		});
		
		this.contentArray = (List) content;
		this.rearrangeObjects();
	}

	/**
	 * Construct an ArrayController with the given values and an empty content
	 * array.
	 * 
	 * @param thinlet
	 *            The thinlet instance that this controller will be working
	 *            with.
	 */
	public ArrayController(Thinlet thinlet) {
		this(thinlet, new Vector());
	}

	public void setSelection(Object selection) {
		this.selection = selection;
		if ((this.arrangedObjects != null)
			&& ((this.selectedIndex == -1) || (this.arrangedObjects
					.get(this.selectedIndex) != selection))) {
			if (selection == null) {
				new Throwable().printStackTrace();
			}
			this.tkv.setKeyValue(this, "selectedIndex", this.arrangedObjects
					.indexOf(selection));
		}
	}

	public Object selection() {
		return this.selection;
	}

	/**
	 * @param searchKey
	 *            The search string to use when filtering out elements in the
	 *            content array.
	 */
	public void setSearchKey(String searchKey) {
		this.searchKey = searchKey;
		if (searchKey != null) {
			int lpc;

			this.searchTokens = WHITE_SPACE_PATTERN.split(searchKey);
			for (lpc = 0; lpc < this.searchTokens.length; lpc++) {
				this.searchTokens[lpc] = this.searchTokens[lpc].toLowerCase();
			}
		} else {
			this.searchTokens = null;
		}
		this.rearrangeObjects();
	}

	/**
	 * @return The search string to use when filtering out elements in the
	 *         content array.
	 */
	public String searchKey() {
		return this.searchKey;
	}

	/**
	 * @return The vector of filter keys for this controller.
	 */
	public Vector getFilterKeys() {
		return this.filterKeys;
	}

	/**
	 * Add an element filter key to this controller and rearrange the objects to
	 * take into account the new path.
	 * 
	 * @param filterKey
	 *            The filter key path to add.
	 */
	public void addFilterKey(String filterKey) {
		this.filterKeys.add(filterKey);
		this.rearrangeObjects();
	}

	/**
	 * Add an array of element filter keys to this controller and rearrange the
	 * objects to take into account the new paths.
	 * 
	 * @param filterKeys
	 *            The array of filter keys to add.
	 */
	public void addFilterKeys(String filterKeys[]) {
		if (filterKeys != null) {
			int lpc;

			for (lpc = 0; lpc < filterKeys.length; lpc++) {
				this.filterKeys.add(filterKeys[lpc]);
			}
			this.rearrangeObjects();
		}
	}

	/**
	 * @param activeKey
	 *            The active sort key.
	 */
	public void setActiveKey(String activeKey) {
		this.activeKey = activeKey;
		this.rearrangeObjects();
	}

	/**
	 * @return The active sort key.
	 */
	public String activeKey() {
		return this.activeKey;
	}

	/**
	 * @param sortDirection
	 *            The direction to sort the elements, one of: ascent, descent,
	 *            none.
	 */
	public void setSortDirection(String sortDirection) {
		this.sortDirection = sortDirection;
		this.rearrangeObjects();
	}

	/**
	 * @return The direction to sort the elements, one of: ascent, descent,
	 *         none.
	 */
	public String sortDirection() {
		return this.sortDirection;
	}

	/**
	 * @param index
	 *            The index of the selected object in the arrangedObjects array
	 *            or -1 if there is no selection.
	 */
	public void setSelectedIndex(int index) {
		Object newSelection = ThinletController.NO_SELECTION;

		if ((index < -1) && (index >= this.arrangedObjects.size())) {
			throw new IllegalArgumentException("index '" + index
												+ "', out of range: " + -1
												+ " - "
												+ this.arrangedObjects.size());
		}

		if ((index != -1) && (this.arrangedObjects != null))
			newSelection = this.arrangedObjects.get(index);
		this.selectedIndex = index;
		this.tkv.setKeyValue(this, "selection", newSelection);
	}

	/**
	 * @return The index of the selected object in the arrangedObjects array or
	 *         -1 if there is no selection.
	 */
	public int selectedIndex() {
		return this.selectedIndex;
	}

	public void setContent(Object content) {
		if (content == ThinletController.NO_SELECTION)
			content = null;
		
		if ((content != null) && !(content instanceof List)) {
			throw new IllegalArgumentException("content is not a list: "
												+ content + " "
												+ content.getClass());
		}

		super.setContent(content);
		this.contentArray = (List) content;
		this.tkv.setKeyValue(this, "selection", ThinletController.NO_SELECTION);
		this.tkv.setKeyValue(this, "arrangedObjects", this
				.arrangeObjects(this.contentArray));
		if (content != null && this.arrangedObjects.size() > 0 && this.selectsNewContent()) {
			this.tkv.setKeyValue(this, "selection", this.arrangedObjects.get(0));
		}
	}

	public void bindContentTo(final Object obj, final String keyPath) {
		this.tkv.observeKeyValue(obj, keyPath, new ThinletKeyValueObserver() {
			public void observedKeyValue(Object obj, String changedKeyPath,
					int flags, Object oldValue, Object newValue) {
				Object newContent = null;

				try {
					newContent = tkv.getKeyValue(obj, keyPath);
					if (newContent == ThinletController.NO_SELECTION)
						newContent = null;
				} catch (IllegalArgumentException e) {
				}
				tkv.setKeyValue(ArrayController.this, "content", newContent);
			}
		});
	}

	/**
	 * @param maximumArraySize
	 *            The maximum number of elements allowed in the arrangedObjects
	 *            array.
	 */
	public void setMaximumArraySize(int maximumArraySize) {
		if (maximumArraySize < 0) {
			throw new IllegalArgumentException("negative maximumArraySize: "
												+ maximumArraySize);
		}

		this.maximumArraySize = maximumArraySize;
		this.rearrangeObjects();
	}

	/**
	 * @return The maximum number of elements allowed in the arrangedObjects
	 *         array.
	 */
	public int maximumArraySize() {
		return this.maximumArraySize;
	}

	/**
	 * @return The sorted/filtered version of the content array.
	 */
	public List arrangedObjects() {
		return this.arrangedObjects;
	}

	public void setArrangedObjects(List arrangedObjects) {
		this.arrangedObjects = arrangedObjects;
	}

	/**
	 * @return The contents of this controller returned as an array.
	 */
	public List contentArray() {
		return this.contentArray;
	}
	
	protected List filterObjects(List retval) {
		if ((this.searchKey != null) && !"".equals(this.searchKey)) {
			StringBuffer values = new StringBuffer();
			Vector filteredResult = new Vector();
			int listIndex, listSize;

			listSize = retval.size();
			for (listIndex = 0; listIndex < listSize; listIndex++) {
				int filterIndex, filterSize, searchIndex;
				Object value = retval.get(listIndex);
				boolean hit = false;

				/*
				 * For each element, we want to search through all of the
				 * filter key values at the same time. So, we get the value
				 * for each key and concatenate it onto a single string that
				 * we will search later.
				 */
				filterSize = this.filterKeys.size();
				for (filterIndex = 0; (filterIndex < filterSize) && !hit; filterIndex++) {
					Object keyValue = null;
					String filterKey;

					filterKey = (String) this.filterKeys.get(filterIndex);
					try {
						keyValue = this.tkv.getKeyValue(value, filterKey);
						if (keyValue != null) {
							values.append(' ');
							values.append(keyValue.toString().toLowerCase());
						}
					}
					catch (IllegalArgumentException e) {
					}
				}

				/*
				 * Now that we have a string with all of the values that
				 * matter, we check for each search token and add any
				 * results that completely match.
				 */
				hit = true;
				for (searchIndex = 0; (searchIndex < this.searchTokens.length)
										&& hit; searchIndex++) {
					if (values.indexOf(this.searchTokens[searchIndex]) == -1) {
						hit = false;
					}
				}
				if (hit)
					filteredResult.add(value);

				values.setLength(0);
			}
			retval = filteredResult;
		}
		
		return retval;
	}

	/**
	 * Method used to rearrange the elements of the content array. Subclasses
	 * should override this if they wish to do their own arranging of the
	 * elements or extend the existing behavior.
	 * 
	 * <p>
	 * Currently, this method will sort the elements based on the activeKey and
	 * the sort direction. Then, it will filter out any elements whose filter
	 * keys do not contain all of the search tokens. And finally, the resulting
	 * list will be checked against the maximum allowed arrangedObjects array
	 * size.
	 * 
	 * @param list
	 *            The content array.
	 * @return The sorted/filtered version of the content array or
	 *         ThinletController.TOO_MANY if the resulting array has more than
	 *         the maximum allowed number of elements.
	 */
	protected List arrangeObjects(List list) {
		boolean wasWithinMaximum = this.isWithinMaximum;
		List retval = null;

		if (list != null) {
			/*
			 * Start with a new vector, since we do not want to mess with the
			 * given one.
			 */
			retval = new Vector();
			/* Copy all of the elements in to the new vector, */
			retval.addAll(list);
			/* ... sort everything, */
			Collections.sort(retval, new Comparator() {
				public int compare(Object o1, Object o2) {
					Object so1, so2;
					int retval;

					if (activeKey != null) {
						/* Get the sub elements to compare. */
						so1 = tkv.getKeyValue(o1, activeKey);
						so2 = tkv.getKeyValue(o2, activeKey);
					} else {
						/* We're comparing the elements themselves. */
						so1 = o1;
						so2 = o2;
					}
					if (so1 instanceof String) {
						/*
						 * For strings, compare in a case-insensitive manner.
						 */
						if ("descent".equals(sortDirection)) {
							retval = ((String) so2)
									.compareToIgnoreCase((String) so1);
						} else {
							retval = ((String) so1)
									.compareToIgnoreCase((String) so2);
						}
					} else if (so1 instanceof Comparable) {
						if ("descent".equals(sortDirection)) {
							if (so2 == null)
								retval = -1;
							else
								retval = ((Comparable) so2).compareTo(so1);
						} else {
							if (so2 == null)
								retval = -1;
							else
								retval = ((Comparable) so1).compareTo(so2);
						}
					} else {
						/* No sorting to be done... */
						retval = 0;
					}
					return retval;
				}

				public boolean equals(Object obj) {
					return false;
				}
			});

			/* ... filter everything, and */
			retval = this.filterObjects(retval);

			/* ... make sure its within the maximum allowed size. */
			this.isWithinMaximum = (retval.size() < this.maximumArraySize);
		} else {
			this.isWithinMaximum = true;
		}
		if (wasWithinMaximum != this.isWithinMaximum) {
			tkv.notifyKeyValue(this, "isWithinMaximum",
					ThinletKeyValue.CHANGE_SETTING, Boolean
							.valueOf(wasWithinMaximum), Boolean
							.valueOf(this.isWithinMaximum));
		}
		if (!this.isWithinMaximum) {
			retval = ThinletController.TOO_MANY;
		}

		return retval;
	}

	/**
	 * Rearrange the objects in the content array and make the result the
	 * arrangedObjects array. The actual arranging of the objects is handled by
	 * the arrangeObjects method.
	 * 
	 * @see #arrangeObjects()
	 */
	public void rearrangeObjects() {
		if (this.contentArray != null) {
			int size, oldSize = 0, newSize, changeType;

			/* Remember the old size, so we can compare against the new. */
			if (this.arrangedObjects != null)
				oldSize = this.arrangedObjects.size();

			/* Recompute the array, */
			this.arrangedObjects = this.arrangeObjects(this.contentArray);

			newSize = this.arrangedObjects.size();
			size = Math.min(oldSize, newSize);

			/* ... notify everyone of potential replacements, */
			if (size > 0) {
				tkv.notifyKeyValue(this, "arrangedObjects",
						ThinletKeyValue.CHANGE_REPLACE, null, new int[] { 0,
								size });
			}
			/* ... changes in size, and */
			if (newSize == oldSize) {
			} else if (size == 0) {
				/*
				 * One of the sizes was zero, so its easier to just send a
				 * change in setting than insertion/removal.
				 */
				tkv.notifyKeyValue(this, "arrangedObjects",
						ThinletKeyValue.CHANGE_SETTING, null,
						this.arrangedObjects);
			} else {
				/*
				 * It is an incremental change in size, notify in terms of
				 * insertion/removal of a range of elements.
				 */
				if (oldSize > newSize)
					changeType = ThinletKeyValue.CHANGE_REMOVAL;
				else
					changeType = ThinletKeyValue.CHANGE_INSERTION;
				tkv.notifyKeyValue(this, "arrangedObjects", changeType, null,
						new int[] { size, Math.max(oldSize, newSize) });
			}
			/* ... selection index. */
			this.tkv.setKeyValue(this, "selectedIndex", this.arrangedObjects
					.indexOf(this.selection));
		}
	}

	private void notifyContentSize() {
		tkv.notifyKeyValue(this, "content@size",
				ThinletKeyValue.CHANGE_SETTING, null, new Integer(
						this.contentArray.size()));
	}

	public void addObject(Object value) {
		if (this.contentArray != null) {
			this.contentArray.add(value);
			this.notifyContentSize();
			this.rearrangeObjects();
			if (this.selectsInsertedObjects())
				this.tkv.setKeyValue(this, "selection", value);
		}
	}

	public void addObjects(Object valueArray[]) {
		int lpc;

		for (lpc = 0; lpc < valueArray.length; lpc++) {
			this.contentArray.add(valueArray[lpc]);
		}
		this.notifyContentSize();
		this.rearrangeObjects();
	}

	public void removeObject(Object value) {
		int oldIndex = -1;
		
		if (value == this.selection) {
			oldIndex = this.selectedIndex;
			this.tkv.setKeyValue(this, "selection",
					ThinletController.NO_SELECTION);
		}
		this.contentArray.remove(value);
		this.notifyContentSize();
		this.rearrangeObjects();
		if (oldIndex != -1 && (this.arrangedObjects.size() > 0)) {
			this.tkv.setKeyValue(this, "selectedIndex", Math.min(this.arrangedObjects.size() - 1, oldIndex));
		}
	}

	public void removeObjects(Object valueArray[]) {
		int lpc;

		for (lpc = 0; lpc < valueArray.length; lpc++) {
			this.contentArray.remove(valueArray[lpc]);
		}
		this.notifyContentSize();
		this.rearrangeObjects();
	}

	public void remove(Object sender) {
		this.removeObject(this.selection);
	}
	
	public void moveSelection(Object sender, int increment) {
		Collections.swap(this.contentArray, this.selectedIndex, this.selectedIndex + increment);
		this.rearrangeObjects();
	}
	
	public boolean selectsNewContent() {
		return false;
	}

	public boolean selectsInsertedObjects() {
		return true;
	}

	public boolean canInsert() {
		return true;
	}
	
	public boolean isMultiple() {
		boolean retval = false;
		
		if (this.contentArray != null && this.contentArray.size() > 1)
			retval = true;
		
		return retval;
	}

	public boolean canRemove() {
		return this.selection != ThinletController.NO_SELECTION;
	}

	public boolean canAccept() {
		return this.selection != ThinletController.NO_SELECTION;
	}

	public boolean canMoveUp() {
		return this.selectedIndex >= 1;
	}
	
	public boolean canMoveDown() {
		boolean retval = false;
		
		if (this.contentArray != null)
			retval = this.selectedIndex < (this.contentArray.size() - 1);
		
		return retval;
	}
	
	public boolean isWithinMaximum() {
		return this.isWithinMaximum;
	}

	public String toString() {
		return "ArrayController[" + super.toString() + "; arrangedObjects="
				+ this.arrangedObjects + "; filterKeys=" + this.filterKeys
				+ "; selection=" + this.selection + "; selectedIndex="
				+ this.selectedIndex + "; searchKey=" + this.searchKey
				+ "; searchTokens=" + this.searchTokens + "; activeKey="
				+ this.activeKey + "; sortDirection=" + this.sortDirection
				+ "]";
	}
}
