package thinlet;

import java.util.Vector;

/**
 * ArrayController that can be used for comboboxes that need to support
 * auto-fill behaviors. The contents of this controller is the "suggestion set"
 * that the user will be used to complete partial entries provided by the user.
 * For example, with a suggestion set of {"hello", "goodbye"}, when the user
 * types 'h' the text box will be completed to show "hello" and the combobox
 * will drop down to show the suggested choice.
 */
public class AutoFillController extends ArrayController {
	private String fillKey = "";

	/**
	 * Construct an AutoFillController with the given values.
	 * 
	 * @param thinlet
	 *            The Thinlet instance this controller will be working with.
	 * @param content
	 *            The initial suggestion set.
	 */
	public AutoFillController(Thinlet thinlet, Vector content) {
		super(thinlet, content);
	}

	/**
	 * Construct an AutoFillController with the given values and an empty
	 * suggestion set.
	 */
	public AutoFillController(Thinlet thinlet) {
		super(thinlet);
	}

	public String fillKey() {
		return this.fillKey;
	}

	public void setFillKey(String fillKey) {
		this.fillKey = fillKey;
	}
	
	public String prefixMatchesContent(String prefix, String content) {
		String retval = null;
		
		if (content.startsWith(prefix))
			retval = content;
		
		return retval;
	}

	/**
	 * Callback for the "insert" listener of the combobox, it will search for
	 * strings in the content array that start with the text entered into the
	 * combobox. If a match is found, it will update the textbox and selected
	 * choice accordingly. If a match is not found, the selection will be
	 * cleared.
	 * 
	 * @param cb
	 *            The combobox being controlled.
	 * @param text
	 *            The text in the combobox.
	 * @param caret
	 *            The position of the caret in the combobox.
	 */
	public void insertText(Object cb, String text, int caret) {
		if (text.length() == caret) {
			boolean found = false;
			int lpc, size;

			/* Search for the text in our content array. */
			size = this.arrangedObjects.size();
			for (lpc = 0; (lpc < size) && !found; lpc++) {
				Object value = this.arrangedObjects.get(lpc);
				String svalue, nvalue;

				svalue = this.tkv.getKeyValue(value, this.fillKey).toString();
				if ((nvalue = this.prefixMatchesContent(text, svalue)) != null) {
					if (nvalue != svalue)
						this.thinlet.setString(cb, "alt", svalue);
					else
						this.thinlet.setString(cb, "alt", null);
					/* Found a match, fill out the rest of the text box, */
					this.thinlet.setString(cb, "text", nvalue);
					/* ... select the part the user did not type, and */
					this.thinlet.setInteger(cb, "start", nvalue.length());
					/* ... update the selectedIndex if it changed. */
					if (this.selection != svalue) {
						this.tkv.setKeyValue(this, "selectedIndex", lpc);
					}
					found = true;
				}
			}
			if (!found) {
				/* No match was found, clear the selection. */
				this.tkv.setKeyValue(this, "selectedIndex", -1);
			}
		}
	}
	

	/**
	 * Callback for the "remove" listener of the combobox, just clears the
	 * selectedIndex of the controller.
	 * 
	 * @param cb
	 *            The combobox being controlled.
	 */
	public void removeText(Object cb) {
		this.tkv.setKeyValue(this, "selectedIndex", -1);
	}
	


	/**
	 * Add a string to the controller's content array if the string does not
	 * already exist.
	 * 
	 * @param value
	 *            The string to add to the content array.
	 */
	public void addUniqueString(String value) {
		if (!"".equals(value) && !this.contentArray.contains(value))
			this.addObject(value);
	}

	public String toString() {
		return "AutoFillController[" + super.toString() + "]";
	}
}
