package thinlet;

import java.util.Date;
import java.util.GregorianCalendar;

import java.text.DateFormat;
import java.text.FieldPosition;
import java.text.SimpleDateFormat;

/**
 * ThinletFormatter for java.util.Date objects.
 */
public class DateFormatter implements ThinletFormatter {
	/**
	 * The default date format, currently set to display the month as its
	 * unabbreviated name, the day, and year.
	 */
	public static final SimpleDateFormat DEFAULT_DATE_FORMAT = new SimpleDateFormat(
			" MMMMMMMM dd, yyyy");

	/**
	 * The default time format, currently set to display the hours and minutes
	 * separated by a colon and an "AM/PM" marker.
	 */
	public static final SimpleDateFormat DEFAULT_TIME_FORMAT = new SimpleDateFormat(
			"\t hh:mm a ");

	/**
	 * Constant used to indicate an unknown date.
	 */
	public static final Date UNKNOWN_DATE = new Date(0);

	/**
	 * The number of milliseconds in a day.
	 */
	private static final long MILLIS_PER_DAY = 24 * 60 * 60 * 1000;

	/**
	 * The DateFormat to use when formatting the month/day/year portion of
	 * dates.
	 */
	private final DateFormat df;

	/**
	 * The DateFormat to use when formatting the hour/minute portion of dates.
	 */
	private final DateFormat tf;

	/**
	 * Stores the current time so that we can compare it against the time to
	 * format and output "Future" instead of the month/day/year.
	 */
	private final Date current = new Date();

	/**
	 * Stores the starting time of "today" so that we can compare it against the
	 * time to format and output "Today" instead of the month/day/year.
	 */
	private final Date today = new Date();

	/**
	 * Stores the starting time of "yesterday" so that we can compare it against
	 * the time to format and output "Yesterday" instead of the month/day/year.
	 */
	private final Date yesterday = new Date();

	/**
	 * Construct a DateFormatter with the given value. <b>Note</b>: This
	 * formatter will substitute "Today" and "Yesterday" for the month/day/year
	 * portion of the date, where applicable.
	 * 
	 * @param df
	 *            The format to use when generating the month/day/year portion
	 *            of the date.
	 * @param tf
	 *            The format to use when generating the hour/minute portion of
	 *            the date.
	 */
	public DateFormatter(DateFormat df, DateFormat tf) {
		this.df = df;
		this.tf = tf;
	}

	/**
	 * Construct a DateFormatter with the default formats for month/day/year and
	 * hours/minutes. Examples of the format are:
	 * 
	 * <ul>
	 * <li>" May 25, 1976\t10:30 AM "
	 * <li>" March 24, 2004\t9:55 PM "
	 * </ul>
	 */
	public DateFormatter() {
		this(DEFAULT_DATE_FORMAT, DEFAULT_TIME_FORMAT);
	}

	public Object getObjectValue(String string) throws Exception {
		Object retval = null;

		if ((string != null) && !"".equals(string)) {
			Date date;

			/* TODO: Implement this for real. */
			date = this.df.parse(string);
			retval = date;
		}
		return retval;
	}

	public String stringForObjectValue(ThinletAPI thinlet, Object component,
			Object value) {
		String retval = null;

		if (value == UNKNOWN_DATE) {
			retval = " (unknown)";
		} else if (value != null) {
			long currentTime = System.currentTimeMillis();
			Date date = (Date) value;

			/* Update the current time, */
			this.current.setTime(currentTime);
			/* the start of today, and */
			currentTime -= (currentTime % MILLIS_PER_DAY);
			this.today.setTime(currentTime);
			/* ... yesterday. */
			this.yesterday.setTime(currentTime - MILLIS_PER_DAY);

			/*
			 * Test if we should display a relative time (e.g. "Today") before
			 * formatting it in an absolute form.
			 */
			if (date.after(this.current)) {
				retval = " Future";
			} else if (date.after(this.today)) {
				retval = " Today";
			} else if (date.after(this.yesterday)) {
				retval = " Yesterday";
			} else {
				retval = this.df.format(value);
			}
			if (this.tf != null)
				retval += this.tf.format(value);
		}
		return retval;
	}

	public boolean isPartialStringValid(ThinletAPI thinlet, Object component,
			StringBuffer partialString, FieldPosition fp) {
		return true;
	}

	public String toString() {
		return "DateFormatter[df=" + this.df + "; tf=" + this.tf + "; current="
				+ this.current + "; today=" + this.today + "; yesterday="
				+ this.yesterday + "]";
	}
}
