package thinlet;

import java.text.NumberFormat;
import java.text.FieldPosition;

/**
 * Thinlet formatter for numbers.
 */
public class NumberFormatter implements ThinletFormatter {
	public static final Number NEGATIVE_ONE = new Double(-1);
	
	private static NumberFormat getNumberFormatInstance() {
		NumberFormat retval;
		
		retval = NumberFormat.getNumberInstance();
		retval.setMaximumFractionDigits(8);
		
		return retval;
	}

	/**
	 * The localized formatter for the value.
	 */
	protected final NumberFormat nf;

	/**
	 * The minimum allowed value.
	 */
	protected double minimum = -Double.MAX_VALUE;

	/**
	 * The maximum allowed value.
	 */
	protected double maximum = Double.MAX_VALUE;

	/**
	 * The number to return when an empty string is encountered.
	 */
	protected Number empty = null;

	/**
	 * Construct a NumberFormatter with the given value.
	 * 
	 * @param nf
	 *            The NumberFormatter to use when formatting/parsing number
	 *            values.
	 */
	public NumberFormatter(NumberFormat nf) {
		this.nf = nf;
	}

	/**
	 * Construct a NumberFormatter with the default formatter as returned by the
	 * VM.
	 */
	public NumberFormatter() {
		this(getNumberFormatInstance());
	}

	public NumberFormatter setEmpty(Number number) {
		this.empty = number;
		return this;
	}

	public Number getEmpty() {
		return this.empty;
	}

	/**
	 * @param minimum
	 *            The minimum allowed value.
	 * @return this
	 */
	public NumberFormatter setMinimum(double minimum) {
		this.minimum = minimum;
		return this;
	}

	/**
	 * @return The minimum allowed value.
	 */
	public double getMinimum() {
		return this.minimum;
	}

	/**
	 * @param maximum
	 *            The maximum allowed value.
	 * @return this
	 */
	public NumberFormatter setMaximum(double maximum) {
		this.maximum = maximum;
		return this;
	}

	/**
	 * @return The maximum allowed value.
	 */
	public double getMaximum() {
		return this.maximum;
	}

	public Object getObjectValue(String string) throws Exception {
		Object retval;

		if ((string == null) || "".equals(string)) {
			if (this.empty == null) {
				throw new ThinletEntryException("Empty values are not allowed", "Enter a number.");
			}
			retval = this.empty;
		} else {
			Number number;
			double value;

			number = this.nf.parse(string);
			value = number.doubleValue();
			if (value < this.minimum) {
				throw new ThinletEntryException(
						"Value is too small",
						"Value must be greater than: " + this.minimum);
			} else if (value > this.maximum) {
				throw new ThinletEntryException(
						"Value is too large",
						"Value must be less than: " + this.maximum);
			}
			retval = number;
		}
		return retval;
	}

	public String stringForObjectValue(ThinletAPI thinlet, Object component,
			Object value) {
		String retval = null;

		if (value != null) {
			if (value.equals(this.empty)) {
				retval = "";
			}
			else {
				try {
					retval = this.nf.format(value);
				}
				catch (IllegalArgumentException e) {
					retval = "";
				}
			}
		}
		return retval;
	}

	public boolean isPartialStringValid(ThinletAPI thinlet, Object component,
			StringBuffer partialString, FieldPosition fp) {
		return true;
	}

	public String toString() {
		return "NumberFormatter[nf=" + this.nf + "; minimum=" + this.minimum
				+ "; maximum=" + this.maximum + "; empty=" + this.empty + "]";
	}
}
