package thinlet;

import java.util.Hashtable;

/**
 * Controller for plain objects.
 */
public class ObjectController extends ThinletControllerBase {
	/**
	 * The Class that represents the content type.
	 */
	protected Class objectClass = Hashtable.class;

	/**
	 * The content being managed by this controller.
	 */
	protected Object content;

	/**
	 * Indicates whether or not the content is editable.
	 */
	protected boolean editable = true;

	/**
	 * Construct an ObjectController with the given values.
	 * 
	 * @param thinlet
	 *            The Thinlet instance this controller is working with.
	 * @param content
	 *            The content being managed by this controller.
	 */
	public ObjectController(Thinlet thinlet, Object content) {
		super(thinlet);

		this.content = content;

		this.tkv.addKeyDependency(this, "hasContent", "content");
		this.tkv.addKeyDependency(this, "hasSelection", "selection");
	}

	/**
	 * Construct an ObjectController with the given value and an empty Hashtable
	 * as the content.
	 * 
	 * @param thinlet
	 *            The Thinlet instance this controller is working with.
	 */
	public ObjectController(Thinlet thinlet) {
		this(thinlet, new Hashtable());
	}

	/**
	 * @param objectClass
	 *            The Class that represents the content type.
	 */
	public void setObjectClass(Class objectClass) {
		this.objectClass = objectClass;
	}

	/**
	 * @return The Class that represents the content type.
	 */
	public Class objectClass() {
		return this.objectClass;
	}

	/**
	 * @param content
	 *            The content being managed by this controller.
	 */
	public void setContent(Object content) {
		this.content = content;
	}

	/**
	 * @return The content being managed by this controller.
	 */
	public Object content() {
		return this.content;
	}

	/**
	 * @return A new object of the objectClass type.
	 */
	public Object newObject() throws InstantiationException,
			IllegalAccessException {
		return this.objectClass.newInstance();
	}

	/**
	 * Add a new object to this controller using addObject().
	 * 
	 * @param sender
	 *            The object that called this method.
	 */
	public void add(Object sender) throws InstantiationException,
			IllegalAccessException {
		this.addObject(this.newObject());
	}

	/**
	 * Remove the content object.
	 * 
	 * @param sender
	 *            The object that called this method.
	 */
	public void remove(Object sender) {
		this.removeObject(this.content);
	}

	/**
	 * Add an object to this controller. A concrete instance of ObjectController
	 * will set the content to the given object.
	 * 
	 * @param obj
	 *            The object to add.
	 */
	public void addObject(Object obj) {
		this.tkv.setKeyValue(this, "content", obj);
	}

	/**
	 * Remove an object from this controller. A concrete instance of
	 * ObjectController will clear the value of content.
	 * 
	 * @param obj
	 *            The object to add.
	 */
	public void removeObject(Object obj) {
		this.tkv.setKeyValue(this, "content", null);
	}

	/**
	 * @return True if objects can be added to this controller, false otherwise.
	 */
	public boolean canAdd() {
		return true;
	}

	/**
	 * @return True if objects can be removed from this controller, false
	 *         otherwise.
	 */
	public boolean canRemove() {
		return (this.content != null);
	}

	public boolean hasContent() {
		return this.content != null;
	}

	public boolean isEditable() {
		return this.editable;
	}

	public void setEditable(boolean editable) {
		this.editable = editable;
	}

	public Object[] selectedObjects() {
		return new Object[] { this.content };
	}

	public Object selection() {
		Object retval;

		if (!this.isEditable()) {
			retval = ThinletController.NOT_APPLICABLE;
		} else {
			Object so[] = this.selectedObjects();

			switch (so.length) {
			case 0:
				retval = ThinletController.NO_SELECTION;
				break;
			case 1:
				retval = so[0];
				break;
			default:
				retval = ThinletController.MULTIPLE_SELECTION;
				break;
			}
		}
		return retval;
	}

	public boolean hasSelection() {
		return this.selection() != ThinletController.NO_SELECTION;
	}
	
	public String toString() {
		return "ObjectController[" + super.toString() + "; objectClass="
				+ this.objectClass + "; content=" + this.content + "]";
	}
}
