package thinlet;

import java.util.Map;
import java.util.Iterator;
import java.util.Properties;

/**
 * Controller for a Properties object that is used to store user preferences.
 */
public class PropertiesController extends ThinletControllerBase {
	/**
	 * The default settings for the properties.
	 */
	private final Properties userDefaults;

	/**
	 * The current values for the properties.
	 */
	private Properties values;

	/**
	 * Construct a PropertiesController with the given values.
	 * 
	 * @param thinlet
	 *            The Thinlet instance this controller will be working with.
	 * @param userDefaults
	 *            The defaults settings for the properties. These values will be
	 *            used when a "revert" is performed.
	 */
	public PropertiesController(Thinlet thinlet, Properties userDefaults) {
		super(thinlet);

		this.userDefaults = userDefaults;
		this.values = userDefaults;
	}

	/**
	 * @return True if any changes are immediately applied to the userDefaults
	 *         properties object.
	 */
	public boolean appliesImmediately() {
		return this.userDefaults == this.values;
	}

	/**
	 * @param ai
	 *            True if any changes should be immediately applied to the
	 *            userDefaults object. Otherwise, the values are kept in a
	 *            separate object so they can be restored using "revert" and
	 *            commited with a "save" operation.
	 */
	public void setAppliesImmediately(boolean ai) {
		Object newValues;

		if (ai)
			newValues = this.userDefaults;
		else
			newValues = this.userDefaults.clone();
		this.tkv.setKeyValue(this, "values", newValues);
	}

	/**
	 * @return The object containing the current values of the properties.
	 */
	public Properties values() {
		return this.values;
	}

	/**
	 * @param values
	 *            The object containing the current values of the properties.
	 */
	public void setValues(Properties values) {
		this.values = values;
	}

	/**
	 * Restore the "values" object to a copy of the userDefaults object.
	 */
	public void revert() {
		if (!this.appliesImmediately())
			this.tkv.setKeyValue(this, "values", this.userDefaults.clone());
	}

	/**
	 * Commit the changes in the "values" object to the userDefaults object.
	 */
	public void save() {
		if (!this.appliesImmediately())
			this.userDefaults.putAll(this.values);
	}

	public String toString() {
		return "PropertiesController[userDefaults=" + this.userDefaults
				+ "; values=" + this.values + "]";
	}
}
