package thinlet;

import java.text.FieldPosition;

/**
 * ThinletFormatter for numbers that should be considered byte counts. So, the
 * output will display 3000 bytes as 3K and so on. As a special case, if the
 * size is -1 the entry is considered to be a directory and the whole will have
 * a different foreground color.
 */
public class SizeFormatter extends NumberFormatter {
	/**
	 * The number of bytes in a kilo.
	 */
	long KILO = 1000;

	/**
	 * The number of bytes in a mega.
	 */
	long MEGA = KILO * KILO;

	/**
	 * The number of bytes in a giga.
	 */
	long GIGA = KILO * MEGA;

	private final String negativeOne;

	private String units;

	public SizeFormatter(String negativeOne, String units) {
		this.negativeOne = negativeOne;
		this.units = units;
	}

	/**
	 * Construct a SizeFormatter with the default values: blue for the directory
	 * color and black for plain entries.
	 */
	public SizeFormatter() {
		this("(dir)", "B");
	}

	public Object getObjectValue(String string) throws Exception {
		Object retval = null;

		if ((string == null) || "".equals(string)) {
			retval = super.getObjectValue(string);
		} else {
			int scaleIndex, length;
			Number number;
			String digits;

			length = string.length();
			for (scaleIndex = 0; (scaleIndex < length)
									&& (Character.isDigit(string
											.charAt(scaleIndex)) || (string
											.charAt(scaleIndex) == '.')); scaleIndex++)
				;
			if (scaleIndex == 0) {
				throw new IllegalArgumentException();
			}
			digits = string.substring(0, scaleIndex);
			retval = number = (Number) super.getObjectValue(digits);
			for (; (scaleIndex < length)
					&& Character.isWhitespace(string.charAt(scaleIndex)); scaleIndex++)
				;
			if ((scaleIndex < length)
				&& (string.charAt(scaleIndex) != this.units.charAt(0))) {
				double d = number.doubleValue();

				switch (string.charAt(scaleIndex)) {
				case 'g':
				case 'G':
					d *= GIGA;
					break;
				case 'm':
				case 'M':
					d *= MEGA;
					break;
				case 'k':
				case 'K':
					d *= KILO;
					break;
				default:
					throw new IllegalArgumentException(
							"Illegal unit: " + string.charAt(scaleIndex));
				}

				if (d < super.minimum) {
					throw new NumberFormatException(
							"Value is too small, it must be greater than:\n"
									+ super.minimum);
				} else if (d > super.maximum) {
					throw new NumberFormatException(
							"Value is too large, it must be less than:\n"
									+ super.maximum);
				}
				retval = new Long((long) d);
			}
		}
		return retval;
	}

	public String stringForObjectValue(ThinletAPI thinlet, Object component,
			Object value) {
		String retval = null;

		if (value != null) {
			Number number = (Number) value;
			long size;

			size = number.longValue();
			/* Check for a directory, */
			if (size == -1) {
				retval = this.negativeOne;
			}
			/* ... sub-kilo size, */
			else if (size < KILO) {
				retval = number.toString() + " " + this.units;
			}
			/* ... sub-mega size, */
			else if (size < MEGA) {
				retval = ("" + (size / KILO) + "." + ((size % KILO) / 100)
							+ " K" + this.units);
			}
			/* ... sub-giga size, and */
			else if (size < GIGA) {
				retval = ("" + (size / MEGA) + "."
							+ ((size % MEGA) / (100 * 1000)) + " M" + this.units);
			}
			/* ... giga size. */
			else {
				retval = ("" + (size / GIGA) + "."
							+ ((size % GIGA) / (100 * 1000 * 1000)) + " G" + this.units);
			}
		}
		return retval;
	}

	public boolean isPartialStringValid(ThinletAPI thinlet, Object component,
			StringBuffer partialString, FieldPosition fp) {
		return true;
	}

	public String toString() {
		return "SizeFormatter[negativeOne=" + this.negativeOne + "; units="
				+ this.units + "]";
	}
}
