package thinlet;

import java.awt.AlphaComposite;
import java.awt.Color;
import java.awt.Component;
import java.awt.Composite;
import java.awt.Font;
import java.awt.Image;
import java.awt.MediaTracker;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.Toolkit;
import java.awt.event.InputEvent;
import java.awt.event.KeyEvent;
import java.io.InputStream;
import java.net.URL;
import java.util.Hashtable;
import java.util.MissingResourceException;
import java.util.ResourceBundle;
import java.util.StringTokenizer;

public abstract class ThinletAPI extends ThinletBase {
	transient Font font;
	{
		setFont(new Font("SansSerif", Font.PLAIN, 12));
	}

	transient int block;

	transient Image hgradient, vgradient;

	transient Hashtable hgradients = new Hashtable();
	transient Hashtable vgradients = new Hashtable();
	
	transient ResourceBundle resourcebundle; // for internationalization

	static ResourceBundle langResource = null; // for I18N

	static ResourceBundle langResourceDefault = null; // for I18N

	transient boolean allI18n = false; // for I18N

	Color c_bg;

	Color c_text;

	Color c_textbg;

	Color c_border;

	Color c_resizeborder;

	Color c_disable;

	Color c_hover;

	Color c_press;

	Color c_focus;

	Color c_caret;

	Color c_empty;

	Color c_select;

	Color c_ctrl;

	Color c_tooltip;
	
	public Color c_default_button;

	Composite dragComposite;

	Composite modalComposite;
	{
		setColors(0xe6e6e6, 0x000000, 0xffffff, 0x909090, 0xb0b0b0, 0xededed,
				0xb9b9b9, 0x89899a, 0x99ff99, 0x4040ff);
	}

	/**
	 * Sets the 9 colors used for components, and repaints the whole UI
	 * 
	 * @param background
	 *            the backround of panels (dialogs, desktops), and disabled
	 *            controls, not editable texts, lines between list items (the
	 *            default value if <i>#e6e6e6</i>)
	 * @param text
	 *            for text, arrow foreground (<i>black</i> by default)
	 * @param textbackground
	 *            the background of text components, and lists (<i>white</i>
	 *            by default)
	 * @param border
	 *            for outer in inner borders of enabled components (<i>#909090</i>
	 *            by default)
	 * @param disable
	 *            for text, border, arrow color in disabled components (<i>#b0b0b0</i>
	 *            by default)
	 * @param hover
	 *            indicates that the mouse is inside a button area (<i>#ededed</i>
	 *            by default)
	 * @param press
	 *            for pressed buttons, gradient image is calculated using the
	 *            background and this press color (<i>#b9b9b9</i> by default)
	 * @param focus
	 *            for text caret and rectagle color marking the focus owner (<i>#89899a</i>
	 *            by default)
	 * @param select
	 *            used as the background of selected text, and list items, and
	 *            in slider (<i>#c5c5dd</i> by default)
	 */
	public void setColors(int background, int text, int textbackground,
			int border, int disable, int hover, int press, int focus,
			int select, int resizeBorder) {
		c_bg = new Color(background);
		c_text = new Color(text);
		c_textbg = new Color(textbackground);
		c_border = new Color(border);
		c_disable = new Color(disable);
		c_hover = new Color(hover);
		c_press = new Color(press);
		c_focus = new Color(focus);
		c_select = new Color(select);
		c_resizeborder = new Color(resizeBorder);
		c_tooltip = new Color(0xffff88);
		c_caret = Color.black;
		c_empty = Color.gray;
		c_default_button = new Color(135, 200, 255);
		hgradient = vgradient = null;
		dragComposite = AlphaComposite.getInstance(AlphaComposite.SRC_OVER,
				0.75f);
		modalComposite = AlphaComposite.getInstance(AlphaComposite.SRC_OVER,
				0.3f);
		repaint();
	}

	protected abstract void addGradient(Color c1, Color c2);
	
	/**
	 * Sets the only one font used everywhere, and revalidates the whole UI.
	 * Scrollbar width/height, spinbox, and combobox button width, and slider
	 * size is the same as the font height
	 * 
	 * @param font
	 *            the default font is <i>SansSerif</i>, <i>plain</i>, and
	 *            <i>12pt</i>
	 */
	public void setFont(Font font) {
		if (font == null)
			throw new IllegalArgumentException("font is null");

		block = getFontMetrics(font).getHeight();
		super.setFont(font);
		this.font = font;
		hgradient = vgradient = null;
		if (content != null)
			validate(content);
	}

	/**
	 * Set a bundle used in parse method, it replaces the parameter values
	 * starting with the 'i18n.' string with a value found in the given bundle
	 * 
	 * @param resourcebundle
	 *            a bundle for the next parsing or null to remove the current
	 *            one.
	 * @throws MissingResourceException
	 *             if no object for the given key can be found.
	 */
	public void setResourceBundle(ResourceBundle resourcebundle) {
		this.resourcebundle = resourcebundle;
	}

	/**
	 * 
	 */
	protected boolean setString(Object component, String key, String value,
			String defaultvalue) {
		if (allI18n && (langResource != null)
			&& ((key == "text") || (key == "tooltip"))) {
			putProperty(component, "i18n." + key, null); // for I18N
		}
		if (value != null && value.equals(defaultvalue))
			value = null;

		return set(component, key, value); // use defaultvalue
	}

	/**
	 * Sets the given property pair (key and value) for the component
	 */
	public void setString(Object component, String key, String value) {
		Object[] definition = getDefinition(getClass(component), key, null);

		if (("string" != definition[0]) && ("object" != definition[0])) {
			throw new IllegalArgumentException("unknown " + key + " for "
												+ getClass(component) + " " + definition[0]);
		}

		if (setString(component, (String) definition[1], value,
				(String) definition[3])) {
			update(component, definition[2]);
		}
	}

	/**
	 * 
	 */
	public String getString(Object component, String key, String defaultvalue) {
		Object rc = get(component, key);
		ThinletFormatter tf;
		String retval;

		if (rc == null) {
			retval = null;
		} else if (rc instanceof String) {
			retval = (String) rc;
		} else if ((tf = (ThinletFormatter) get(component, "formatter")) != null) {
			String fcache, cacheKey = keyForCache(key);

			if ((fcache = (String) get(component, cacheKey)) != null) {
				retval = fcache;
			} else {
				retval = tf.stringForObjectValue(this, component, rc);
				setReal(component, cacheKey, retval);
			}
		} else {
			retval = rc.toString();
		}

		return (rc == null) ? defaultvalue : getI18NString(component, key,
				retval);
	}

	/**
	 * Gets the property value of the given component by the property key
	 */
	public String getString(Object component, String key) {
		return getString(component, key, null);
	}

	public void setDelegate(Object component, String key, ThinletDelegate value) {
		Object[] definition = getDefinition(getClass(component), key,
				"delegate");

		if (set(component, definition[1], value)) {
			if (definition.length == 4) {
				// XXX clear caches here?
				update(component, definition[2]);
			}
		}
	}

	public ThinletDelegate getDelegate(Object component, String key) {
		Object[] definition = getDefinition(getClass(component), key,
				"delegate");

		return (ThinletDelegate) get(component, definition[1]);
	}

	/**
	 * 
	 * @throws java.lang.IllegalArgumentException
	 */
	protected boolean setChoice(Object component, String key, String value,
			String[] values, String defaultvalue) {
		if (value == null) {
			return set(component, key, defaultvalue);
		}

		for (int i = 0; i < values.length; i++) {
			if (value.equals(values[i])) {
				return set(component, key, values[i]);
			}
		}
		throw new IllegalArgumentException("unknown " + value + " for " + key);
	}

	/**
	 * Sets the given property pair (key and value) for the component
	 */
	public void setChoice(Object component, String key, String value) {
		Object[] definition = getDefinition(getClass(component), key, "choice");
		String[] values = (String[]) definition[3];

		if (setChoice(component, (String) definition[1], value, values,
				values[0])) {
			update(component, definition[2]);
		}
	}

	/**
	 * Gets the property value of the given component by the property key
	 */
	public String getChoice(Object component, String key) {
		Object[] definition = getDefinition(getClass(component), key, "choice");

		return getString(component, (String) definition[1],
				((String[]) definition[3])[0]);
	}

	public void setObject(Object component, String key, Object value) {
		Object[] definition = getDefinition(getClass(component), key, "object");

		if (set(component, key, value)) {
			setReal(component, keyForCache(key), null);
			update(component, definition[2]);
		}
	}

	public Object getObject(Object component, String key) {
		if (key.charAt(0) == ':') {
			throw new IllegalArgumentException("attempt to get internal key "
												+ key);
		}

		return get(component, key, "object");
	}

	/**
	 * Sets the given property pair (key and value) for the component
	 */
	public void setBoolean(Object component, String key, boolean value) {
		Object[] definition = getDefinition(getClass(component), key, "boolean");

		if (setBoolean(component, (String) definition[1], value, (definition[3]
				.equals(Boolean.TRUE)))) {
			update(component, definition[2]);
		}
	}

	/**
	 * Gets the property value of the given component by the property key
	 */
	public boolean getBoolean(Object component, String key) {
		return get(component, key, "boolean").equals(Boolean.TRUE);
	}

	/**
	 * 
	 */
	protected boolean setBoolean(Object component, String key, boolean value,
			boolean defaultvalue) {
		return set(component, key, (value == defaultvalue) ? null
				: (value ? Boolean.TRUE : Boolean.FALSE));
	}

	/**
	 * 
	 */
	protected boolean getBoolean(Object component, String key,
			boolean defaultvalue) {
		Object value = get(component, key);

		return (value == null) ? defaultvalue : ((Boolean) value)
				.booleanValue();
	}

	/**
	 * Sets the given property pair (key and value) for the component
	 */
	public void setInteger(Object component, String key, int value) {
		Object[] definition = getDefinition(getClass(component), key, "integer");

		if (setInteger(component, (String) definition[1], value,
				((Integer) definition[3]).intValue())) {
			update(component, definition[2]);
		}
	}

	/**
	 * Gets the property value of the given component by the property key
	 */
	public int getInteger(Object component, String key) {
		return ((Number) get(component, key, "integer")).intValue();
	}

	/**
	 * 
	 */
	protected boolean setInteger(Object component, String key, int value,
			int defaultvalue) {
		return set(component, key, (value == defaultvalue) ? null
				: new Integer(value));
	}

	/**
	 * 
	 */
	protected int getInteger(Object component, String key, int defaultvalue) {
		Object value = get(component, key);
		return (value == null || !(value instanceof Number)) ? defaultvalue : ((Number) value).intValue();
	}

	/**
	 * Sets the given property pair (key and value) for the component
	 */
	public void setIcon(Object component, String key, Image icon) {
		Object[] definition = getDefinition(getClass(component), key, "icon");
		if (set(component, definition[1], icon)) {
			update(component, definition[2]);
		}
	}

	/**
	 * Gets the property value of the given component by the property key
	 */
	public Image getIcon(Object component, String key) {
		return (Image) get(component, key, "icon");
	}

	/**
	 * 
	 */
	protected Image getIcon(Object component, String key, Image defaultvalue) {
		Object value = get(component, key);
		return (value == null) ? defaultvalue : (Image) value;
	}

	/**
	 * Creates an image, and loads it immediately by default
	 * 
	 * @param path
	 *            is relative to your thinlet instance or the classpath (if the
	 *            path starts with <i>'/'</i> character), or a full URL
	 * @return the loaded image or null
	 */
	public Image getIcon(String path) {
		return getIcon(path, true);
	}

	/**
	 * Creates an image from the specified resource. To speed up loading the
	 * same images use a cache (a simple hashtable). And flush the resources
	 * being used by an image when you won't use it henceforward
	 * 
	 * @param path
	 *            is relative to your thinlet instance or the classpath, or an
	 *            URL
	 * @param preload
	 *            waits for the whole image if true, starts loading (and
	 *            repaints, and updates the layout) only when required (painted,
	 *            or size requested) if false
	 * @return the loaded image or null
	 */
	public Image getIcon(String path, boolean preload) {
		if ((path == null) || (path.length() == 0)) {
			return null;
		}
		Image image = null; // (Image) imagepool.get(path);
		try {
			URL url = getClass().getResource(path); // ClassLoader.getSystemResource(path)
			if (url != null) { // contributed by Stefan Matthias Aust
				image = Toolkit.getDefaultToolkit().getImage(url);
			}
		} catch (Throwable e) {
		}
		if (image == null) {
			try {
				InputStream is = getClass().getResourceAsStream(path);
				// InputStream is = ClassLoader.getSystemResourceAsStream(path);
				if (is != null) {
					byte[] data = new byte[is.available()];
					is.read(data, 0, data.length);
					image = getToolkit().createImage(data);
					is.close();
				} else { // contributed by Wolf Paulus
					image = Toolkit.getDefaultToolkit().getImage(new URL(path));
				}
			} catch (Throwable e) {
			}
		}
		if (preload && (image != null)) {
			MediaTracker mediatracker = new MediaTracker(this);
			mediatracker.addImage(image, 1);
			try {
				mediatracker.waitForID(1, 5000);
			} catch (InterruptedException ie) {
			}
			// imagepool.put(path, image);
		}
		return image;
	}

	/**
	 * 
	 */
	protected void setKeystrokeImpl(Object component, String key, String value) {
		Long keystroke = null;

		if (value != null) {
			String token = value;
			try {
				int keycode = 0, modifiers = 0;
				StringTokenizer st = new StringTokenizer(value, " \r\n\t+");
				while (st.hasMoreTokens()) {
					token = st.nextToken().toUpperCase();
					try {
						modifiers = modifiers
									| InputEvent.class
											.getField(token + "_MASK").getInt(
													null);
					} catch (Exception exc) { // not mask value
						keycode = KeyEvent.class.getField("VK_" + token)
								.getInt(null);
					}
				}
				keystroke = new Long(((long) modifiers) << 32 | keycode);
			} catch (Exception exc) {
				throw new IllegalArgumentException(token);
			}
		}
		set(component, key, keystroke);
	}

	/**
	 * 
	 */
	public void setKeystroke(Object component, String key, String value) {
		Object[] definition = getDefinition(getClass(component), key,
				"keystroke");
		// TODO check if changed
		setKeystrokeImpl(component, (String) definition[1], value);
		update(component, definition[2]);
	}

	/**
	 * Get the AWT component of the given (currently <i>bean</i>) widget
	 * 
	 * @param component
	 *            a <i>bean</i> widget
	 * @param key
	 *            the identifier of the parameter
	 * @return an AWT component, or null
	 */
	public Component getComponent(Object component, String key) {
		return (Component) get(component, key, "bean");
	}

	/**
	 * Set custom font on a component
	 * 
	 * @param component
	 *            component to use the custom font
	 * @param font
	 *            custom font to use, or null to reset component to use default
	 *            font
	 */
	public void setFont(Object component, String key, Font font) {
		Object[] definition = getDefinition(getClass(component), key, "font");
		if (set(component, definition[1], font)) {
			update(component, definition[2]);
		}
	}

	public void setFont(Object component, String key, String value) {
		StringTokenizer st = new StringTokenizer(value);
		int style = 0, size = 0;
		String name = null;

		while (st.hasMoreTokens()) {
			String token = st.nextToken();
			if ("bold".equalsIgnoreCase(token)) {
				style |= Font.BOLD;
			} else if ("italic".equalsIgnoreCase(token)) {
				style |= Font.ITALIC;
			} else {
				try {
					size = Integer.parseInt(token);
				} catch (NumberFormatException nfe) {
					name = (name == null) ? token : (name + " " + token);
				}
			}
		}
		if (name == null) {
			name = font.getName();
		}
		if (size == 0) {
			size = font.getSize();
		}
		setFont(component, key, new Font(name, style, size));
	}

	public Font getFont(Object component, String key) {
		return (Font) get(component, key, "font");
	}

	/**
	 * Set custom color on a component. Notes: For "foreground" key, this sets
	 * the text color. For "background" key, on gradient-filled components (such
	 * as tabs, buttons etc) this will result in a component filled with solid
	 * background color, and not a new gradient. Also, Color.brighter() will be
	 * used for highlight, and Color.darker() will be used for pressed or not
	 * selected.
	 * 
	 * @param component
	 *            component to use for custom color
	 * @param key
	 *            currently "background" and "foreground" are supported
	 * @param color
	 *            custom color to use, or null to reset component to use default
	 *            color
	 */
	public void setColor(Object component, String key, Color color) {
		Object[] definition = getDefinition(getClass(component), key, "color");

		if (set(component, definition[1], color)) {
			update(component, definition[2]);
		}
	}

	public void setColor(Object component, String key, String value) {
		try {
			Color color = null;
			int colorInt = 0;

			if (value.startsWith("#")) {
				if (value.length() == (1 + 6))
					colorInt = Integer.parseInt(value.substring(1), 16);
				else
					throw new NumberFormatException();
			} else if (value.startsWith("0x")) {
				if (value.length() == (2 + 6))
					colorInt = Integer.parseInt(value.substring(2), 16);
				else
					throw new NumberFormatException();
			} else {
				try {
					color = (Color) Color.class.getField(value).get(null);
				} catch (Exception e) {
					// three separated integer including red, green, and blue
					StringTokenizer st = new StringTokenizer(value, " \r\n\t,");

					if (st.countTokens() == 3) {
						colorInt = 0xff000000
									| ((Integer.parseInt(st.nextToken()) & 0xff) << 16)
									| ((Integer.parseInt(st.nextToken()) & 0xff) << 8)
									| (Integer.parseInt(st.nextToken()) & 0xff);
					} else {
						throw new NumberFormatException();
					}
				}
			}
			if (color == null) {
				color = new Color(colorInt);
			}
			setColor(component, key, color);
		} catch (NumberFormatException e) {
			throw new IllegalArgumentException(
					"Color value for '"
							+ key
							+ "' needs to be in hex format (e.g. #efef66, 0xefef66), "
							+ "a color name (e.g. red), or a triple of ints representing "
							+ "the rgb value (e.g. '256,128,256'): " + value);
		}
	}

	public Color getColor(Object component, String key) {
		return (Color) get(component, key, "color");
	}

	public Color getColor(Object component, String key, Color defaultValue) {
		Color retval = (Color) get(component, key, "color");
		return (retval == null) ? defaultValue : retval;
	}

	/**
	 * 
	 */
	public void setRectangle(Object component, String key, int x, int y,
			int width, int height) {
		Rectangle rectangle = getRectangle(component, key);

		if (rectangle != null) {
			rectangle.x = x;
			rectangle.y = y;
			rectangle.width = width;
			rectangle.height = height;
		} else {
			set(component, key, new Rectangle(x, y, width, height));
		}
	}

	/**
	 * 
	 */
	public Rectangle getRectangle(Object component, String key) {
		return (Rectangle) get(component, key);
	}

	public void setPoint(Object component, String key, int x, int y) {
		Point point = getPoint(component, key);

		if (point != null) {
			point.x = x;
			point.y = y;
		} else {
			set(component, key, new Point(x, y));
		}
	}

	public Point getPoint(Object component, String key) {
		return (Point) get(component, key);
	}

	/**
	 * Set the AWT component for the given (currently <i>bean</i>) widget
	 * 
	 * @param component
	 *            a <i>bean</i> widget
	 * @param key
	 *            the identifier of the parameter
	 * @param bean
	 *            an AWT component, or null
	 */
	public void setComponent(Object component, String key, Component bean) {
		Object[] definition = getDefinition(getClass(component), key, "bean");

		if (set(component, definition[1], bean)) {
			update(component, definition[2]);
		}
	}

	// TODO add set/getComponent for popupmenu and header

	/**
	 * 
	 */
	public Object getWidget(Object component, String key) {
		if ("popupmenu".equals(key)) {
			return get(component, "popupmenu");
		} else if ("header".equals(key)) {
			return get(component, "header");
		} else {
			throw new IllegalArgumentException(key);
		}
	}

	/**
	 * Binds the specified key to the specified value, and stores in this
	 * component. <i>Null</i> value removes the property. Use the parameter tag
	 * in the xml resource to bind a string value, the format is:
	 * <i>parameter='key=value'</i>
	 * 
	 * @param component
	 *            the hashtable is binded to this component
	 * @param key
	 *            the client property key
	 * @param value
	 *            the new client property value
	 */
	public void putProperty(Object component, Object key, Object value) {
		Object table = get(component, ":bind");
		if (value != null) {
			if (table == null) {
				set(component, ":bind", table = new Hashtable());
			}
			((Hashtable) table).put(key, value);
		} else if (table != null) {
			((Hashtable) table).remove(key);
		}
	}

	/**
	 * Returns the value of the property with the specified key.
	 * 
	 * @param component
	 *            searches the hashtable of this component
	 * @param key
	 *            the client property key
	 * @return the value to which the key is mapped or null if the key is not
	 *         mapped to any value
	 */
	public Object getProperty(Object component, Object key) {
		Object table = get(component, ":bind");
		return (table != null) ? ((Hashtable) table).get(key) : null;
	}

	protected boolean supports(Object component, String key, String type) {
		try {
			Object definition[] = getDefinition(getClass(component), key, null);

			return definition[0] == type;
		} catch (IllegalArgumentException e) {
		}
		return false;
	}

	public boolean supportsMethod(Object component, String key) {
		return supports(component, key, "method");
	}

	public boolean supportsObject(Object component, String key) {
		return supports(component, key, "object");
	}

	public boolean supportsString(Object component, String key) {
		return supports(component, key, "string");
	}

	public boolean supportsBoolean(Object component, String key) {
		return supports(component, key, "boolean");
	}

	class ContentObserver implements ThinletKeyValueObserver {
		private final Object component;

		private final Object headerComponent;

		private final Object tableComponent;

		private final ThinletDelegate cellFormatter;

		private final String cellAlignment;

		private final ThinletContent tc;

		private final String classname;

		private final Object controller;

		private final String keyPath;
		
		public ContentObserver(Object component, ThinletContent tc,
				String classname, Object controller, String keyPath) {
			this.component = component;
			if ("table" == classname) {
				this.headerComponent = get(component, "header");
				this.tableComponent = null;
				this.cellFormatter = null;
				this.cellAlignment = null;
			} else if ("column" == classname) {
				this.headerComponent = getParent(component);
				this.tableComponent = getParent(headerComponent);
				this.cellFormatter = getDelegate(component, "cellFormatter");
				this.cellAlignment = getString(component, "cellAlignment");
			} else {
				this.headerComponent = null;
				this.tableComponent = null;
				this.cellFormatter = null;
				this.cellAlignment = null;
			}
			this.tc = tc;
			this.classname = classname;
			this.controller = controller;
			this.keyPath = keyPath;
		}

		private Object findRow(int index) {
			Object comp = (this.tableComponent != null) ? this.tableComponent
					: component;

			return getItem(comp, index);
		}

		private void fillRow(Object row, Object content) {
			int lpc, count = getCount(this.headerComponent);

			for (lpc = 0; lpc < count; lpc++) {
				Object cell = create("cell");

				set(cell, ":parent", row);
				set(cell, ":next", get(row, ":comp"));
				set(row, ":comp", cell);
			}
			if (tc != null)
				tc.initItem(ThinletAPI.this, row, content);
		}

		private void fillRow(Object row, int index) {
			this.fillRow(row, tkv.getKeyValue(controller, keyPath, index));
		}

		private boolean updateItem(boolean init, Object item, Object value) {
			boolean retval = false;
			Object ivalue, tvalue;

			if (tc != null) {
				if (init)
					tc.initItem(ThinletAPI.this, item, value);
				else
					tc.replaceItem(ThinletAPI.this, item, value);
			} else {
				value = tkv.reduceValueList(value);
				if (value instanceof Image) {
					ivalue = value;
					tvalue = null;
				} else {
					ivalue = null;
					tvalue = value;
				}
				retval = setReal(item, "icon", ivalue) || retval;
				retval = setReal(item, "text", tvalue) || retval;
			}
			setReal(item, keyForCache("text"), null);

			return retval;
		}

		private boolean updateItem(boolean init, Object item, int index) {
			return this.updateItem(init, item, tkv.getKeyValue(controller,
					keyPath, index));
		}

		public void observedKeyValue(Object obj, String changedKeyPath,
				int flags, Object oldValue, Object newValue) {
			Object changedObject = component;
			boolean changed = false;
			int columnIndex = -1;

			if (false) {
				System.out.println("content change in " + classname + " '"
									+ changedKeyPath + "' "
									+ Integer.toHexString(flags) + " = "
				/* + newValue */);
			}

			if (this.tableComponent != null) {
				columnIndex = getIndex(this.headerComponent, component);

				if ((flags & ThinletKeyValue.CHANGE_SETTING) != 0) {
					java.util.List list;

					list = tkv.getKeyValueAsList(obj, this.keyPath);
					if ((list != null) && (list.size() > 0)) {
						Object row, parent;
						int lpc, size;

						parent = getParent(component);
						set(component, ":parent", null);
						size = list.size();
						row = get(tableComponent, ":comp");
						for (lpc = 0; lpc < size; lpc++) {
							Object cell;

							cell = getItem(row, columnIndex);
							if (cellFormatter != null)
								set(cell, "formatter", cellFormatter);
							if (cellAlignment != null)
								set(cell, "alignment", cellAlignment);
							this.updateItem(false, cell, list.get(lpc));
							row = get(row, ":next");
						}
						set(component, ":parent", parent);
					}

					if (getBoolean(this.tableComponent, "visible")) {
						update(this.tableComponent, "validate");
					}
					return;
				}

				if ((flags & ThinletKeyValue.CHANGE_INSERTION) != 0) {
					int range[] = (int[]) newValue;
					Object row, parent;
					int lpc;

					parent = getParent(component);
					set(component, ":parent", null);
					row = findRow(range[0]);
					for (lpc = range[0]; lpc < range[1]; lpc++) {
						Object cell = getItem(row, columnIndex);

						if (cellFormatter != null)
							setDelegate(cell, "formatter", cellFormatter);
						if (cellAlignment != null)
							setChoice(cell, "alignment", cellAlignment);
						this.updateItem(false, cell, lpc);
						row = get(row, ":next");
					}
					set(component, ":parent", parent);

					if (getBoolean(this.tableComponent, "visible")) {
						update(this.tableComponent, "validate");
					}
					return;
				}

				if ((flags & ThinletKeyValue.CHANGE_REMOVAL) != 0) {
					return;
				}
			}

			if ((flags & ThinletKeyValue.CHANGE_SETTING) != 0
				&& (("table" != classname) || keyPath.startsWith(changedKeyPath))) {
				java.util.List list;

				list = tkv.getKeyValueAsList(obj, this.keyPath);
				set(component, ":comp", null);
				set(component, ":lead", null);
				if ((list != null) && (list.size() > 0)) {
					Object firstitem = null, lastitem = null;
					int lpc, size;

					size = list.size();
					for (lpc = 0; lpc < size; lpc++) {
						Object item, value;

						if ("list" == classname || "combobox" == classname) {
							item = create("item");
							this.updateItem(true, item, list.get(lpc));
						} else if ("table" == classname) {
							item = create("row");
							fillRow(item, list.get(lpc));
						} else {
							throw new InternalError();
						}
						if (firstitem == null)
							firstitem = changedObject = item;
						else
							set(lastitem, ":next", item);
						set(item, ":parent", component);
						lastitem = item;
					}
					if (firstitem != null) {
						set(component, ":comp", firstitem);
					}
				}
				changed = true;
			}

			if ((flags & ThinletKeyValue.CHANGE_REPLACE) != 0
				&& (("table" != classname) || keyPath.equals(changedKeyPath))) {
				int lpc, range[] = (int[]) newValue;
				Object row, parent;

				parent = getParent(component);
				set(component, ":parent", null);
				row = changedObject = findRow(range[0]);
				for (lpc = range[0]; lpc < range[1]; lpc++) {
					Object item;

					if ("list" == classname || "combobox" == classname)
						item = row;
					else if ("column" == classname)
						item = getItem(row, columnIndex);
					else if ("table" == classname)
						item = row;
					else
						throw new InternalError();
					changed = this.updateItem(false, item, lpc) || changed;
					row = get(row, ":next");
				}
				set(component, ":parent", parent);
			}

			if ((flags & ThinletKeyValue.CHANGE_INSERTION) != 0) {
				int range[] = (int[]) newValue;
				Object row;
				int lpc;

				row = changedObject = findRow(range[0] - 1);
				for (lpc = range[0]; lpc < range[1]; lpc++) {
					Object item;

					if ("list" == classname || "combobox" == classname) {
						item = create("item");
						this.updateItem(true, item, lpc);
					} else if ("table" == classname) {
						item = create("row");
						this.fillRow(item, lpc);
					} else {
						throw new InternalError();
					}
					set(item, ":parent", component);
					set(item, ":next", get(row, ":next"));
					if (row == null)
						set(component, ":comp", item);
					else
						set(row, ":next", item);
					row = item;
				}
				changed = true;
			}

			if ((flags & ThinletKeyValue.CHANGE_REMOVAL) != 0) {
				int lpc, range[] = (int[]) newValue;
				Object row;

				setLead(component, get(component, ":lead"), null);
				if (range[0] == 0) {
					row = findRow(range[1]);
					changedObject = get(component, ":comp");
					set(component, ":comp", row);
				} else {
					row = findRow(range[0] - 1);
					changedObject = get(row, ":next");
					set(row, ":next", findRow(range[1]));
				}
				changed = true;
			}

			if (changed && getBoolean(component, "visible")) {
				update(changedObject, "validate");
			}
		}
	}

	transient Object focusowner;

	protected abstract boolean setNextFocusable(Object comp, boolean b);

	public void bindAttribute(final Object component, String key, String value,
			Object handler) {
		final Object[] definition = getDefinition(getClass(component), key,
				null);
		final String classname = getClass(component);
		final Object controller;
		final String keyPath;

		if (value.charAt(0) != '@') {
			throw new IllegalArgumentException(
					"Binding for '" + key + "' does not start with an '@': "
							+ value);
		}

		if (value.charAt(1) == '(') {
			int cEnd = value.indexOf(')');

			if (cEnd == -1) {
				throw new IllegalArgumentException(
						"Syntax error in attribute binding for " + key
								+ ", closing paren is missing: " + value);
			}
			controller = tkv.getKeyValue(handler, value.substring(2, cEnd));
			keyPath = value.substring(cEnd + 1);
		} else {
			controller = handler;
			keyPath = value.substring(1);
		}

		if ("content" == definition[1]) {
			final ThinletContent tc = (ThinletContent) getDelegate(component,
					"contentmapper");

			tkv.observeKeyValue(controller, keyPath, new ContentObserver(
					component, tc, classname, controller, keyPath));
			tkv.notifyKeyValue(controller, keyPath,
					ThinletKeyValue.CHANGE_SETTING, null, tkv.getKeyValue(
							controller, keyPath));
			return;
		} else if ("selectedIndex" == definition[1]) {
			tkv.observeKeyValue(component, (String) definition[1],
					new ThinletKeyValueObserver() {
						public void observedKeyValue(Object obj,
								String changedKeyPath, int flags,
								Object oldValue, Object newValue) {
							if ((flags & ThinletKeyValue.CHANGE_SETTING) != 0) {
								tkv.setKeyValue(controller, keyPath, newValue);
							}
						}
					});
			tkv.observeKeyValue(controller, keyPath,
					new ThinletKeyValueObserver() {
						public void observedKeyValue(Object obj,
								String changedKeyPath, int flags,
								Object oldValue, Object newValue) {
							if ((flags & ThinletKeyValue.CHANGE_SETTING) != 0) {
								int index = tkv.getKeyValueAsInteger(
										controller, keyPath);
								Object oldLead = get(component, ":lead");
								Object item;
								int lpc;

								for (lpc = 0, item = get(component, ":comp"); item != null; lpc++, item = get(
										item, ":next")) {
									if (lpc == index) {
										if (setBoolean(item, "selected", true,
												false)
											&& isVisible(component)) {
											update(item, "paint");
											setLead(component, oldLead, item);
										}
									} else if (setBoolean(item, "selected",
											false, false)
												&& isVisible(component)) {
										repaint(component, null, item);
									}
								}
							}
						}
					});
		} else {
			tkv.observeKeyValue(controller, keyPath,
					new ThinletKeyValueObserver() {
						public void observedKeyValue(Object obj,
								String changedKeyPath, int flags,
								Object oldValue, Object newValue) {
							if (!changedKeyPath.equals(keyPath)) {
								newValue = tkv.getKeyValueAsSingleton(obj,
										keyPath);
							}
							if ("icon" == definition[0]) {
								if (!(newValue instanceof Image))
									newValue = null;
							} else if ("boolean" == definition[0]) {
								if (!(newValue instanceof Boolean))
									newValue = Boolean.FALSE;
							}
							// System.out.println("change " + definition[1] + "  " + keyPath + " " + newValue);
							if (setReal(component, definition[1], newValue)) {
								String cacheKey = keyForCache(definition[1]);
								if (cacheKey != null)
									setReal(component, cacheKey, null);
								if (false) {
									System.out.println(" update cl:"
														+ classname + " key:"
														+ definition[1]
														+ " rep:"
														// + definition[2]
														+ " nv:" + newValue);
								}
								if ("combobox" == classname && "text" == definition[1] && newValue != null) {
									int lpc = 0;
									
									for (Object choice = get(component, ":comp"); choice != null; choice = get(
											choice, ":next"), lpc++) {
										if (newValue.equals(getString(choice, "text", ""))) {
											// setReal(component, "selected", new Integer(lpc));
											setInteger(component, "selected", lpc);
											break;
										}
									}
								}
								update(component, definition[2]);
							}
							if ("enabled" == definition[1] && (focusowner == component) &&
									!((Boolean)newValue).booleanValue()) {
								setNextFocusable(component, false);
							}
						}
					});
			tkv.observeKeyValue(component, key, new ThinletKeyValueObserver() {
				public void observedKeyValue(Object obj, String changedKeyPath,
						int flags, Object oldValue, Object newValue) {
					tkv.setKeyValue(controller, keyPath, newValue);
				}
			});
		}

		Object ovalue = tkv.getKeyValue(controller, keyPath);
		if ("icon" == definition[0]) {
			if (!(ovalue instanceof Image))
				ovalue = null;
		} else if ("boolean" == definition[0]) {
			if (!(ovalue instanceof Boolean))
				ovalue = Boolean.FALSE;
		} else if ("value" == definition[1]) {
			if (false) {
				System.out.println("ov " + ovalue + "; " + keyPath + " "
									+ controller);
			}
		}
		set(component, key, ovalue);
	}

	/**
	 * Set current language resource bundle. This flushes all cached translated
	 * values, performs lazy loading of new values, and repaints the desktop.
	 * This implementation allows applications to switch language resources on
	 * the fly, without rebuilding/reloading components that use them. <br />
	 * The pseudo-code is as follows:
	 * 
	 * <ul>
	 * <li>if langResource and langResourceDefault are null, don't translate
	 * anything, no matter what other settings are. This behaviour provides
	 * compatibility with previous versions.</li>
	 * <li>if only langResourceDefault is set, use this when translation is
	 * required</li>
	 * <li>if allI18n is set to true:
	 * <ul>
	 * <li>if property "i18n" on a component is missing, or set to "true",
	 * translate</li>
	 * <li>if property "i18n" is present, and set to "false", do not translate</li>
	 * </ul>
	 * </li>
	 * <li>if allI18n is set to false:
	 * <ul>
	 * <li>if property "i18n" on a component is missing, or set to "false", do
	 * not translate</li>
	 * <li>if property "i18n" is present, and set to "true", translate</li>
	 * </ul>
	 * </li>
	 * </ul>
	 * 
	 * The "translate" step is applied only to values from "text" and "tooltip"
	 * properties (for now), and is applied as follows:
	 * 
	 * <ul>
	 * <li>use the value of "text" or "tooltip" as a lookup key</li>
	 * <li>use langResource to lookup the result value
	 * <ul>
	 * <li>if no value is found, use langResourceDefault for lookup
	 * <ul>
	 * <li>if no value is found, just return the original value of the
	 * property. Set a flag on component that prevents lookups in the future.
	 * This flag is cleared when langResource is changed.</li>
	 * </ul>
	 * </li>
	 * </ul>
	 * </li>
	 * <li>cache the result value, if any</li>
	 * </ul>
	 * 
	 * If translated value is found successfully, it is cached in the component.
	 * This cache is gradually flushed when setLangResource is called. Cached
	 * value is also flushed when setString() is called on a component.
	 * 
	 * @param res
	 *            resource bundle containing localized texts for "text" and
	 *            "tooltip"
	 */
	public void setLangResource(ResourceBundle res) { // for I18N
		langResource = res;
		doLayout(content);
		repaint(content);
	}

	/**
	 * Returns language resource bundle currently in use, or default bundle, or
	 * null.
	 */
	public static ResourceBundle getLangResource() { // for I18N
		return langResource;
	}

	/**
	 * Set default language resource bundle. Resources from this bundle will be
	 * used if they are missing in the current bundle.
	 * 
	 * @param res
	 *            resource bundle containing default localized texts for "text"
	 *            and "tooltip".
	 */
	public void setLangResourceDefault(ResourceBundle res) { // for I18N
		langResourceDefault = res;
		if (langResource == null)
			setLangResource(res);
	}

	/**
	 * Returns default language resource bundle, or null.
	 */
	public static ResourceBundle getLangResourceDefault() { // for I18N
		return langResourceDefault;
	}

	/**
	 * Sets the default behaviour of internationalization code. If set to
	 * "true", try to translate all components' "text" and "tooltip" values,
	 * unless explicitly prohibited by setting <code>i18n="false"</code> on a
	 * specific component. If set to "false", do not translate unless explicitly
	 * requested by setting <code>i18n="true"</code> on a specific component.
	 * <br />
	 * Default value is "false", to provide backwards compatibility.
	 * 
	 * @param val
	 *            if "true", translate by default; if "false", do not translate
	 *            by default.
	 */
	public void setAllI18n(boolean val) { // for I18N
		allI18n = val;
	}

	/**
	 * 
	 */
	String getI18NString(Object component, String key, String text) { // for
																		// I18N
		if (allI18n && (langResource != null)
			&& ((key == "text") || (key == "tooltip"))
			&& getBoolean(component, "i18n", true)) {
			String ikey = (String) getProperty(component, "i18n." + key);
			if (!"__NONE__".equals(ikey)) {
				if (ikey == null) { // initialize
					putProperty(component, "i18n." + key, ikey = text);
				}
				try {
					return langResource.getString(ikey);
				} catch (Exception exc) { // not found. Try default
					if (langResourceDefault != null) {
						try {
							return langResourceDefault.getString(ikey);
						} catch (Exception dexc) {
							putProperty(component, "i18n." + key, "__NONE__");
						}
					}
				}
			}
		}
		return text;
	}

	/**
	 * 
	 */
	protected Object get(Object component, String key, String type) {
		Object[] definition = getDefinition(getClass(component), key, type);
		Object value = get(component, definition[1]);

		return (value != null) ? value : definition[3];
	}

	protected abstract Object create(String classname);

	public abstract int getSelectedIndex(Object component);

	public abstract Object getSelectedItem(Object component);

	public abstract Object[] getSelectedItems(Object component);

	protected abstract void setLead(Object component, Object oldlead,
			Object lead);

	public abstract boolean isVisible(Object component);

	protected abstract String keyForCache(Object key);

	/**
	 * 
	 */
	protected abstract void update(Object component, Object mode);

	/**
	 * 
	 */
	protected abstract void doLayout(Object component);

	/**
	 * 
	 */
	protected abstract void repaint(Object component);

	protected abstract void repaint(Object component, Object classname,
			Object part);

	/**
	 * 
	 */
	protected abstract void validate(Object component);
}
