package thinlet;

import java.awt.Container;
import java.io.Serializable;

/**
 * Split Thinlet file containing basic model manipulation.
 * 
 * @version $Id: ThinletBase.java,v 1.7 2005/12/15 01:57:25 stack Exp $
 */
public abstract class ThinletBase extends Container implements Runnable,
		Serializable {
	static final boolean SHOW_REPAINT;

	static {
		boolean sr;

		try {
			sr = Boolean.getBoolean("thinlet.showRepaint");
		} catch (Throwable th) {
			sr = false;
		}
		SHOW_REPAINT = sr;
	}

	/**
	 * Object used to support key-value-coding.
	 */
	protected transient final ThinletKeyValue tkv = new ThinletKeyValue();

	/**
	 * The root thinlet object, all other objects will be added to under here.
	 */
	protected Object content = createImpl("desktop");

	public ThinletKeyValue getNamespace() {
		return this.tkv;
	}
	
	public void printCaller(StackTraceElement ste[]) {
		int lpc;
		
		for (lpc = 0; lpc < 6 && lpc < ste.length; lpc++) {
			System.out.println("    " + ste[lpc].toString());
		}
	}

	/**
	 * Get the topmost component
	 * 
	 * @return the root object (it is a <i>desktop </i>), never <i>null </i>
	 */
	public Object getDesktop() {// #
		return content;
	}

	// -----------------------------------------------------------------------
	/**
	 * Override this method to handle exceptions thrown by the invoked custom
	 * methods
	 * 
	 * @param throwable
	 *            the thrown exception by the business logic
	 */
	public void handleException(Throwable throwable) {
		throwable.printStackTrace();
	}

	protected void handleException(Object component, Throwable throwable) {
		handleException(throwable);
	}

	// -----------------------------------------------------------------------
	/**
	 * This method is called by the FrameLauncher if the window was closing, or
	 * AppletLauncher's destroy method. Overwrite it to e.g. save the
	 * application changes.
	 * 
	 * @return true to exit, and false to keep the frame and continue the
	 *         application
	 */
	public boolean destroy() {
		return true;
	}

	// -----------------------------------------------------------------------
	/**
	 * Gets the type of the given component
	 * 
	 * @param component
	 *            a widget
	 * @return the class name of the component (e.g. <i>button </i>)
	 */
	public static String getClass(Object component) { // #
		return (String) get(component, ":class");
	}

	public static void printcomp(Object component) {
		System.out.println("comp = " + component);
		for (Object[] entry = (Object[]) component; entry != null; entry = (Object[]) entry[2]) {
			System.out.println("tag = " + entry[0] + " = " + entry[1]);
			if (entry[0] == ":comp" || entry[0] == ":next") {
				System.out.println(">>>>");
				printcomp(entry[1]);
				System.out.println("<<<<");
			}
		}
	}

	protected abstract Object[] getDefinition(Object classname, String key,
			String type);

	/**
	 * 
	 */
	protected static Object createImpl(String classname) {
		return new Object[] { ":class", classname, null };
	}

	/**
	 * 
	 */
	protected boolean setReal(Object component, Object key, Object value) {
		Object[] previous = (Object[]) component;

		if (key == null)
			throw new IllegalArgumentException("key is null");

		for (Object[] entry = previous; entry != null; entry = (Object[]) entry[2]) {
			if (entry[0] == key) {
				if (value != null) { // set the row's value
					Object oldvalue = entry[1];
					entry[1] = value;
					return !value.equals(oldvalue);
				} else { // remove the row
					previous[2] = entry[2];
					entry[2] = null;
					return true;
				}
			}
			previous = entry;
		}
		if (value != null) { // append a new row
			if (previous != null)
				previous[2] = new Object[] { key, value, null };
			return true;
		}
		return false;
	}

	protected boolean set(Object component, Object key, Object value) {
		String skey = (String) key;
		boolean retval;

		if (key == null)
			throw new IllegalArgumentException("key is null");

		retval = this.setReal(component, key, value);
		if (retval && (component != null) && (skey.charAt(0) != ':')
			&& (tkv.findObservers(component) != null)) {
			if (value == null) {
				Object[] definition = getDefinition(getClass(component), skey,
						null);

				if (definition.length > 2)
					value = definition[3]; // reset to the default value
			}
			tkv.notifyKeyValue(component, skey, ThinletKeyValue.CHANGE_SETTING,
					null, value);
		}
		return retval;
	}

	/**
	 * Get a specified sub-component or component part. Key is e.g. "header"
	 * (table header), ":parent" (parent component), ":comp" (head of a
	 * component list), ":next" (next component in a list), "popupmenu" is a
	 * popupmenu, etc ...
	 */
	protected static Object get(Object component, Object key) {
		for (Object[] entry = (Object[]) component; entry != null; entry = (Object[]) entry[2]) {
			if (entry[0] == key) {
				return entry[1];
			}
		}
		return null;
	}

	/**
	 * 
	 */
	protected static int getItemCountImpl(Object component, String key) {
		int i = 0;
		for (Object comp = get(component, key); comp != null; comp = get(comp,
				":next")) {
			i++;
		}
		return i;
	}

	/**
	 * Referenced by DOM, replace by getItem for others
	 */
	protected static Object getItemImpl(Object component, Object key, int index) {
		int i = 0;

		if (index < -1)
			throw new IllegalArgumentException("Illegal index: " + index);

		for (Object item = get(component, key); item != null; item = get(item,
				":next")) {
			if (i == index) {
				return item;
			}
			i++;
		}
		return null;
	}

	/**
	 * Referenced by DOM
	 */
	protected void insertItemImpl(Object parent, Object key, Object component,
			int index) {
		Object item = parent, next = get(parent, key);

		if (index < -1)
			throw new IllegalArgumentException("Illegal index: " + index);

		if (get(component, ":parent") != null) {
			throw new IllegalArgumentException("component, "
												+ getClass(component)
												+ ", is already attached to "
												+ getClass(get(component,
														":parent")));
		}

		for (int i = 0;; i++) {
			if ((i == index) || (next == null)) {
				set(item, key, component);
				set(component, ":parent", parent);
				set(component, ":next", next);
				break;
			}
			next = get(item = next, key = ":next");
		}
	}

	/**
	 * Delete the give component from its parent list
	 * 
	 * @param parent
	 * @param component
	 */
	protected void removeItemImpl(Object parent, Object component) {
		Object previous = null; // the widget before the given component
		for (Object comp = get(parent, ":comp"); comp != null;) {
			Object next = get(comp, ":next");
			if (next == component) {
				previous = comp;
				break;
			}
			comp = next;
		}
		set((previous != null) ? previous : parent,
				(previous != null) ? ":next" : ":comp", get(component, ":next"));
		set(component, ":next", null);
		set(component, ":parent", null); // not required
	}

	/**
	 * Gets the count of sub-components in the list of the given component
	 * 
	 * @param component
	 *            a widget
	 * @return the number of components in this component
	 */
	public int getCount(Object component) {
		return getItemCountImpl(component, ":comp");
	}

	/**
	 * Gets the parent of this component
	 * 
	 * @param component
	 *            a widget
	 * @return the parent container of this component or item
	 */
	public Object getParent(Object component) {
		return get(component, ":parent");
	}

	/**
	 * Returns the subcomponent of the given component's specified list at the
	 * given index
	 * 
	 * @param component
	 *            a specified container
	 * @param index
	 *            the index of the component to get
	 * @return the index <sup>th </sup> component in this container
	 */
	public Object getItem(Object component, int index) {
		return getItemImpl(component, ":comp", index);
	}

	/**
	 * Gets all the components in this container
	 * 
	 * @param component
	 *            a specified container
	 * @return an array of all the components in this container
	 */
	public Object[] getItems(Object component) {
		Object[] items = new Object[getItemCountImpl(component, ":comp")];
		Object comp = get(component, ":comp");
		for (int i = 0; i < items.length; i++) {
			items[i] = comp;
			comp = get(comp, ":next");
		}
		return items;
	}

	/**
	 * 
	 */
	protected int getIndex(Object component, Object value) {
		int index = 0;
		for (Object item = get(component, ":comp"); item != null; item = get(
				item, ":next")) {
			if (value == item) {
				return index;
			}
			index++;
		}
		return -1;
	}
}
