package thinlet;

public abstract class ThinletFocus extends ThinletInvoke {
	transient boolean focusinside;

	/**
	 * This component can be traversed using Tab or Shift-Tab keyboard focus
	 * traversal, although 1.4 replaced this method by <i>isFocusable</i>, so
	 * 1.4 compilers write deprecation warning
	 * 
	 * @return true as focus-transverable component, overwrites the default
	 *         false value
	 */
	public boolean isFocusTraversable() {
		return true;
	}

	/**
	 * Requests that both the <i>Thinlet</i> component, and the given widget
	 * get the input focus
	 * 
	 * @param component
	 *            a focusable widget inside visible and enabled parents, and
	 *            tabbedpane's selected tab
	 * @return true, if the given component was focusable
	 */
	public boolean requestFocus(Object component) { // #
		if (isFocusable(component, true)) {
			setFocus(component);
			repaint(component);
			return true;
		}
		return false;
	}

	/**
	 * Request focus for the given component
	 * 
	 * @param component
	 *            a focusable component
	 * @return true if the focusowner was changed, otherwise false
	 */
	boolean setFocus(Object component) {
		if (!focusinside) { // request focus for the thinlet component
			requestFocus();
		}
		if (focusowner != component) {
			Object focused = focusowner;
			if (focusowner != null) {
				if (instance(getClass(focused), "textfield")) {
					try {
						if (focusformatter != null) {
							String text = getString(focused, "text");
							Object value;

							value = focusformatter.getObjectValue(text);
							setReal(focused, "text", value);
							tkv.notifyKeyValue(focused, "text",
									ThinletKeyValue.CHANGE_SETTING,
									null, value);
						}
					} catch (Exception e) {
						handleException(focused, e);
						return false;
					}
				}
				focusowner = null; // clear focusowner
				focusformatter = null;
				repaint(focused);
				// invoke the focus listener of the previously focused component
				invoke(focused, null, "focuslost");
			}
			if (focusowner == null) { // it won't be null, if refocused
				focusowner = component;
				// invoke the focus listener of the new focused component
				if (instance(getClass(component), "textfield")) {
					focusformatter = (ThinletFormatter) getDelegate(component,
							"formatter");
				}
				invoke(component, null, "focusgained");
			}
			return true;
		}
		return false;
	}

	/**
	 * @return next focusable component is found (not the first of the
	 *         desktop/dialog)
	 */
	protected boolean setNextFocusable(Object current, boolean outgo) {
		boolean didModal = false, consumed = true;
		for (Object next = null, component = current; true; component = next) {
			next = get(component, ":comp"); // check first subcomponent
			if (next == null) {
				next = get(component, ":next");
			} // check next component
			while (next == null) { // find the next of the parents, or the
									// topmost
				component = getParent(component); // current is not on the
													// desktop
				if (component == null) {
					return false;
				}
				String cclass = getClass(component);
				if ((component == content)
					|| ((cclass == "dialog") && (!outgo || getBoolean(
							component, "modal", false)))) {
					if (getBoolean(component, "modal", false)) {
						if (!didModal) {
							didModal = true;
						} else {
							/* Modal doesn't seem to have any objects. */
							focusinside = true;
							focusowner = component;
							if (instance(getClass(component), "textfield")) {
								focusformatter = (ThinletFormatter) getDelegate(
										component, "formatter");
							}
							return true;
						}
					}
					consumed = false; // find next focusable but does not
										// consume event
					next = component; // the topmost (desktop or modal dialog)
				} else {
					next = get(component, ":next");
				}
			}
			if (next == current) {
				return false;
			} // one fucusable, no loop
			if (isFocusable(next, false)) {
				setFocus(next);
				return consumed;
			}
		}
	}

	/**
	 * @return previous focusable component is found (not the last of the
	 *         desktop/dialog)
	 */
	boolean setPreviousFocusable(Object component, boolean outgo) {
		for (int i = 0; i < 2; i++) { // 0 is backward direction
			Object previous = getPreviousFocusable(component, null, true,
					false, (i == 0), outgo);
			if (previous != null) {
				setFocus(previous);
				return (i == 0);
			}
		}
		return false;
	}

	/**
	 * For the starting component search its parent direction for a focusable
	 * component, and then its next component (if not search backward from the
	 * component).<br />
	 * For its parent components check its first component, the current one, and
	 * its parent direction (backward search), or its parent, then next
	 * component (forward direction).<br />
	 * For the rest components check the next, then the first subcomponent
	 * direction, and finally check whether the component is focusable.
	 */
	Object getPreviousFocusable(Object component, Object block, boolean start,
			boolean upward, boolean backward, boolean outgo) {
		Object previous = null;
		if ((component != null) && (component != block)) {
			boolean go = ((getClass(component) != "dialog") || (outgo && !getBoolean(
					component, "modal", false)));
			if (!start && !upward && go) {
				previous = getPreviousFocusable(get(component, ":next"), block,
						false, false, backward, outgo);
			}
			if ((previous == null)
				&& ((upward && backward) || (!start && !upward))) {
				previous = getPreviousFocusable(get(component, ":comp"), block,
						false, false, backward, outgo);
				if ((previous == null) && isFocusable(component, false)) {
					previous = component;
				}
			}
			if ((previous == null) && (start || upward) && go) {
				previous = getPreviousFocusable(getParent(component),
						component, false, true, backward, outgo);
			}
			if ((previous == null) && (start || upward) && !backward && go) {
				previous = getPreviousFocusable(get(component, ":next"), block,
						false, false, backward, outgo);
			}
		}
		return previous;
	}

	/**
	 * Check whether the given widget can become focusowner
	 * 
	 * @param component
	 *            check this widget
	 * @param forced
	 *            splitpane is also checked (e.g. false for tab navigating, and
	 *            true for mouse selection or application request)
	 * @return true if focusable, otherwise false
	 */
	boolean isFocusable(Object component, boolean forced) {
		String classname = getClass(component);
		if ((classname == "button") || (classname == "checkbox")
			|| ("togglebutton" == classname) || (classname == "combobox")
			|| (classname == "textfield") || (classname == "passwordfield")
			|| (classname == "textarea") || (classname == "spinbox")
			|| (classname == "slider") || (classname == "list")
			|| (classname == "table") || (classname == "tree")
			|| (classname == "tabbedpane")
			|| (forced && (classname == "splitpane"))
			|| (classname == "dropzone")) {
			for (Object comp = component; comp != null;) {
				// component and parents are enabled and visible
				if (!getBoolean(comp, "enabled", true)
					|| !getBoolean(comp, "visible", true)
					|| !getBoolean(comp, "isfocusable", true)) {
					return false;
				}
				Object parent = getParent(comp);
				// inside the selected tabbedpane tab
				if ((getClass(comp) == "tab")
					&& (getItem(parent, getInteger(parent, "selected", 0)) != comp)) {
					return false;
				}
				comp = parent;
			}
			return true;
		}
		return false;
	}

}
