package thinlet;

import java.lang.reflect.Field;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.text.NumberFormat;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Stack;
import java.util.StringTokenizer;
import java.util.Vector;

public class ThinletKeyValue {
	public static boolean VERBOSE = false;

	private static final Object INVALIDATED_KEY_PATH = new Object();

	public static final int B_CHANGE_SETTING = 0;

	public static final int B_CHANGE_INSERTION = 1;

	public static final int B_CHANGE_REMOVAL = 2;

	public static final int B_CHANGE_REPLACE = 3;

	public static final int CHANGE_SETTING = (1 << B_CHANGE_SETTING);

	public static final int CHANGE_INSERTION = (1 << B_CHANGE_INSERTION);

	public static final int CHANGE_REMOVAL = (1 << B_CHANGE_REMOVAL);

	public static final int CHANGE_REPLACE = (1 << B_CHANGE_REPLACE);

	public static final int CHANGE_ANY = (CHANGE_SETTING | CHANGE_INSERTION
											| CHANGE_REMOVAL | CHANGE_REPLACE);

	private static final Hashtable CL_TO_METHODS = new Hashtable();

	private synchronized static Method[] getClassMethods(Class cl) {
		Method retval[];

		if ((retval = (Method[]) CL_TO_METHODS.get(cl)) == null) {
			retval = cl.getMethods();
			CL_TO_METHODS.put(cl, retval);
		}
		return retval;
	}

	private static final Hashtable CL_TO_FIELDS = new Hashtable();

	private synchronized static Field[] getClassFields(Class cl) {
		Field retval[];

		if ((retval = (Field[]) CL_TO_FIELDS.get(cl)) == null) {
			retval = cl.getFields();
			CL_TO_FIELDS.put(cl, retval);

			if (false) {
				int lpc;

				System.out.println("cl: " + cl);
				for (lpc = 0; lpc < retval.length; lpc++) {
					System.out.println("  " + retval[lpc]);
				}
			}
		}
		return retval;
	}

	protected static final int TYPE_NONE = 0;

	protected static final int TYPE_BOOLEAN = 1;

	protected static final int TYPE_BYTE = 2;

	protected static final int TYPE_SHORT = 3;

	protected static final int TYPE_INTEGER = 4;

	protected static final int TYPE_LONG = 5;

	protected static final int TYPE_FLOAT = 6;

	protected static final int TYPE_DOUBLE = 7;

	int mapPrimitiveType(Class cl) {
		int retval;

		if (!cl.isPrimitive())
			retval = TYPE_NONE;
		else if (cl == Boolean.TYPE)
			retval = TYPE_BOOLEAN;
		else if (cl == Byte.TYPE)
			retval = TYPE_BYTE;
		else if (cl == Short.TYPE)
			retval = TYPE_SHORT;
		else if (cl == Integer.TYPE)
			retval = TYPE_INTEGER;
		else if (cl == Long.TYPE)
			retval = TYPE_LONG;
		else if (cl == Float.TYPE)
			retval = TYPE_FLOAT;
		else if (cl == Double.TYPE)
			retval = TYPE_DOUBLE;
		else
			throw new InternalError("cl = " + cl);

		return retval;
	}

	interface Setter {
		void set(Object object, String key, Object value);
	}

	interface Getter {
		Object get(Object object, String key);
	}

	interface Accessor extends Setter, Getter {
		boolean hasSupportFor(Object object, String key);
	}

	public static final Object KV_NULL = new Object();

	private static final Accessor MAP_ACCESSOR = new Accessor() {
		public void set(Object object, String key, Object value) {
			Map map = (Map) object;

			if (value == null)
				value = KV_NULL;
			map.put(key, value);
		}

		public Object get(Object object, String key) {
			return ((Map) object).get(key);
		}

		public boolean hasSupportFor(Object object, String key) {
			return true;
		}
	};

	private final Accessor LIST_ACCESSOR = new Accessor() {
		public void set(Object object, String key, Object value) {
			List list = (List) object;

			if (list.size() > 0) {
				Accessor accessor;
				int lpc, size;

				accessor = resolveKey(list.get(0), key);
				size = list.size();
				for (lpc = 0; lpc < size; lpc++) {
					Object nvalue = value;
					
					if (value instanceof String) {
						nvalue = ((String)value).intern();
					}
					else if (value instanceof Cloneable) {
						Method meth;
						try {
							meth = value.getClass().getMethod("clone", null);
							nvalue = meth.invoke(value, null);
						} catch (SecurityException e) {
						} catch (NoSuchMethodException e) {
						} catch (IllegalArgumentException e) {
						} catch (IllegalAccessException e) {
						} catch (InvocationTargetException e) {
						}
					}
					accessor.set(list.get(lpc), key, nvalue);
					// System.out.println("lset " + list.get(lpc) + " -> " + nvalue);
				}
			}
		}

		public Object get(Object object, String key) {
			Vector retval = new Vector();
			List list = (List) object;

			if (list.size() > 0) {
				Accessor accessor;
				int lpc, size;

				accessor = resolveKey(list.get(0), key);
				size = list.size();
				for (lpc = 0; lpc < size; lpc++) {
					retval.add(accessor.get(list.get(lpc), key));
				}
			}
			return retval;
		}

		public boolean hasSupportFor(Object object, String key) {
			return true;
		}
	};

	private static final Accessor NULL_ACCESSOR = new Accessor() {
		public void set(Object object, String key, Object value) {
		}

		public Object get(Object object, String key) {
			return null;
		}

		public boolean hasSupportFor(Object object, String key) {
			return true;
		}
	};

	private static final Accessor NO_SELECTION_ACCESSOR = new Accessor() {
		public void set(Object object, String key, Object value) {
		}

		public Object get(Object object, String key) {
			return ThinletController.NO_SELECTION;
		}

		public boolean hasSupportFor(Object object, String key) {
			return true;
		}
	};

	private final Hashtable class2accessors = new Hashtable();

	private final Hashtable observers = new Hashtable();

	class PieceWiseAccessor implements Accessor {
		private final Hashtable setters = new Hashtable();

		private final Hashtable getters = new Hashtable();

		public PieceWiseAccessor() {
		}

		public void addSetterFor(Setter setter, String key) {
			if (setter != null)
				this.setters.put(key, setter);
		}

		public void addGetterFor(Getter getter, String key) {
			if (getter != null)
				this.getters.put(key, getter);
		}

		public void set(Object object, String key, Object value) {
			Setter setter;

			if ((setter = (Setter) this.setters.get(key)) != null)
				setter.set(object, key, value);
		}

		public Object get(Object object, String key) {
			Object retval = null;
			Getter getter;

			if ((getter = (Getter) this.getters.get(key)) != null)
				retval = getter.get(object, key);
			return retval;
		}

		public boolean hasSupportFor(Object object, String key) {
			return this.getters.containsKey(key);
		}

		public String toString() {
			return "PieceWiseAccessor[setters=" + this.setters + "; getters="
					+ this.getters + "]";
		}
	}

	public class MethodGetter implements Getter {
		private final Method method;

		MethodGetter(Method method) {
			this.method = method;
		}

		public Object get(Object obj, String key) {
			try {
				return this.method.invoke(obj, null);
			} catch (Throwable th) {
				th.printStackTrace();
				throw new InternalError();
			}
		}

		public String toString() {
			return "MethodGetter[method=" + this.method + "]";
		}
	}

	class MethodSetter implements Setter {
		private final Method validatorMethod;

		private final Method setterMethod;

		private final int primitiveType;
		
		private boolean isController;

		MethodSetter(Method validatorMethod, Method setterMethod) {
			this.validatorMethod = validatorMethod;
			this.setterMethod = setterMethod;
			this.primitiveType = mapPrimitiveType(setterMethod
					.getParameterTypes()[0]);
			this.isController = ThinletController.class.isAssignableFrom(
					this.setterMethod.getDeclaringClass());
		}

		public void set(Object obj, String key, Object value) {
			try {
				Object args[] = new Object[1];
				
				switch (this.primitiveType) {
				case TYPE_BOOLEAN:
					if (value instanceof Number) {
						value = Boolean
								.valueOf(((Number) value).byteValue() != 0);
					}
					break;
				case TYPE_BYTE:
					value = new Byte(((Number) value).byteValue());
					break;
				case TYPE_SHORT:
					value = new Short(((Number) value).shortValue());
					break;
				case TYPE_INTEGER:
					value = new Integer(((Number) value).intValue());
					break;
				case TYPE_LONG:
					value = new Long(((Number) value).longValue());
					break;
				case TYPE_FLOAT:
					value = new Float(((Number) value).floatValue());
					break;
				case TYPE_DOUBLE:
					value = new Double(((Number) value).doubleValue());
					break;
				}
				args[0] = value;
				if (this.validatorMethod != null)
					this.validatorMethod.invoke(obj, args);
				this.setterMethod.invoke(obj, args);
			} catch (InvocationTargetException ite) {
				ite.getTargetException().printStackTrace();
				throw new IllegalArgumentException(ite.getTargetException()
						.toString()
													+ ": "
													+ key
													+ " = "
													+ value);
			} catch (Throwable th) {
				System.out.println("failure to set " + key + " on "
									+ obj.getClass() + " to " + value + " ("
									+ value.getClass() + ")");
				th.printStackTrace();
				throw new InternalError();
			}
		}

		public String toString() {
			return "MethodSetter[validatorMethod=" + this.validatorMethod
					+ "; setterMethod=" + this.setterMethod + "]";
		}
	}

	class FieldGetter implements Getter {
		private final Field field;

		FieldGetter(Field field) {
			this.field = field;
		}

		public Object get(Object obj, String key) {
			try {
				return this.field.get(obj);
			} catch (Throwable th) {
				th.printStackTrace();
				throw new InternalError();
			}
		}

		public String toString() {
			return "FieldGetter[field=" + this.field + "]";
		}
	}

	class FieldSetter implements Setter {
		private final Field field;

		private final int primitiveType;

		private final Method validatorMethod;

		private final Object args[] = new Object[1];

		FieldSetter(Method validatorMethod, Field field) {
			this.validatorMethod = validatorMethod;
			this.field = field;
			this.primitiveType = mapPrimitiveType(field.getType());
		}

		public synchronized void set(Object obj, String key, Object value) {
			try {
				if (this.validatorMethod != null) {
					this.args[0] = value;
					this.validatorMethod.invoke(obj, this.args);
					this.args[0] = null;
				}
				if ((this.primitiveType != TYPE_NONE)
					&& (value instanceof Number)) {
					Number number = (Number) value;

					switch (this.primitiveType) {
					case TYPE_BOOLEAN:
						this.field.setBoolean(obj, number.byteValue() != 0);
						break;
					case TYPE_BYTE:
						this.field.setByte(obj, number.byteValue());
						break;
					case TYPE_SHORT:
						this.field.setShort(obj, number.shortValue());
						break;
					case TYPE_INTEGER:
						this.field.setInt(obj, number.intValue());
						break;
					case TYPE_LONG:
						this.field.setLong(obj, number.longValue());
						break;
					case TYPE_FLOAT:
						this.field.setFloat(obj, number.floatValue());
						break;
					case TYPE_DOUBLE:
						this.field.setDouble(obj, number.doubleValue());
						break;
					}
				} else {
					this.field.set(obj, value);
				}
			} catch (Throwable th) {
				th.printStackTrace();
				throw new InternalError();
			}
		}

		public String toString() {
			return "FieldSetter[validatorMethod=" + this.validatorMethod
					+ "; field=" + this.field + "]";
		}
	}

	private final Stack notifyStack = new Stack();
	private final Stack notifyObjStack = new Stack();

	private final Hashtable recentPaths = new Hashtable();
	
	public synchronized String[] tokenizePath(String keyPath) {
		String retval[];

		if (keyPath == null) {
			retval = new String[0];
		} else if ((retval = (String[]) this.recentPaths.get(keyPath)) == null) {
			Vector vector = new Vector();
			int sIndex = 0, eIndex;

			while (sIndex < keyPath.length()) {
				if (keyPath.charAt(sIndex) == '\'') {
					if ((eIndex = keyPath.indexOf('\'', sIndex + 1)) == -1) {
						throw new IllegalArgumentException(
								"KeyPath component not terminated: "
										+ keyPath.substring(sIndex));
					}
					vector.add(keyPath.substring(sIndex + 1, eIndex).intern());
					if ((eIndex + 1) < keyPath.length()) {
						if (keyPath.charAt(eIndex + 1) != '.') {
							throw new IllegalArgumentException(
									"KeyPath has garbage instead of '.' after "
											+ "quote: "
											+ keyPath.substring(eIndex + 1));
						}
						eIndex += 1;
					}
				} else {
					int atIndex, dotIndex;

					if (sIndex > 0 && (keyPath.charAt(sIndex - 1) == '@'))
						sIndex -= 1;
					atIndex = keyPath.indexOf('@', sIndex + 1);
					dotIndex = keyPath.indexOf('.', sIndex + 1);
					if ((atIndex == -1) && (dotIndex == -1))
						eIndex = keyPath.length();
					else if (atIndex == -1)
						eIndex = dotIndex;
					else if (dotIndex == -1)
						eIndex = atIndex;
					else
						eIndex = Math.min(atIndex, dotIndex);
					vector.add(keyPath.substring(sIndex, eIndex).intern());
				}
				sIndex = eIndex + 1;
			}

			retval = (String[]) vector.toArray(new String[vector.size()]);

			if (this.recentPaths.size() > 512) {
				this.recentPaths.clear(); // heavy handed
			}
			this.recentPaths.put(keyPath, retval);
		}

		return retval;
	}

	private Accessor resolveKeyInObject(Object object, String key) {
		Method validator = null;
		Getter getter = null;
		Setter setter = null;
		Method methods[];
		Accessor retval;
		Field fields[];
		char key0;
		Class cl;
		int lpc;

		cl = object.getClass();

		key0 = Character.toUpperCase(key.charAt(0));

		methods = getClassMethods(cl);
		for (lpc = 0; (lpc < methods.length) && (validator == null); lpc++) {
			int modifiers = methods[lpc].getModifiers();
			String name = methods[lpc].getName();
			final String validate = "validate";

			if (!Modifier.isPublic(modifiers)) {
				// Not visible.
			} else if (name.startsWith(validate)
						&& (name.length() > (validate.length() + 1))
						&& (name.charAt(validate.length()) == key0)
						&& name.substring(validate.length() + 1).equals(
								key.substring(1))) {
				validator = methods[lpc];
			}
		}

		for (lpc = 0; (lpc < methods.length)
						&& ((setter == null) || (getter == null)); lpc++) {
			int modifiers = methods[lpc].getModifiers();
			String name = methods[lpc].getName();
			final int ACC_LEN = 3;

			if (!Modifier.isPublic(modifiers)) {
				// Not visible.
			} else if (name.equals(key)) {
				switch (methods[lpc].getParameterTypes().length) {
				case 0:
					getter = new MethodGetter(methods[lpc]);
					break;
				case 1:
					setter = new MethodSetter(validator, methods[lpc]);
					break;
				default:
					break;
				}
			} else if ((name.length() > (ACC_LEN + 1))
						&& (name.charAt(ACC_LEN) == key0)
						&& name.substring(ACC_LEN + 1).equals(key.substring(1))) {
				if (name.startsWith("set") && methods[lpc].getParameterTypes().length == 1)
					setter = new MethodSetter(validator, methods[lpc]);
				else if (name.startsWith("get") && methods[lpc].getParameterTypes().length == 0)
					getter = new MethodGetter(methods[lpc]);
			}
		}

		fields = getClassFields(cl);
		for (lpc = 0; (lpc < fields.length)
						&& ((setter == null) || (getter == null)); lpc++) {
			int modifiers = fields[lpc].getModifiers();
			String name = fields[lpc].getName();

			if (!Modifier.isPublic(modifiers)) {
				// Not visible.
			} else if (name.equals(key)) {
				if (getter == null)
					getter = new FieldGetter(fields[lpc]);
				if (setter == null && !Modifier.isFinal(modifiers))
					setter = new FieldSetter(validator, fields[lpc]);
			}
		}

		if ((setter == null) && (getter == null)) {
			try {
				/*
				 * XXX Ugh, for some reason the JVM might not be returning all
				 * of the fields in getFields(), so we have to check manually
				 * with getField().
				 */

				Field field;

				field = cl.getField(key);

				// System.err.println("*** Hit JVM bug for - " + key);

				if (getter == null)
					getter = new FieldGetter(field);
				if (setter == null && !Modifier.isFinal(field.getModifiers()))
					setter = new FieldSetter(validator, field);
			} catch (NoSuchFieldException e) {
				if (VERBOSE) {
					System.out.println("no key " + key + " " + cl + " " + object);
					// new Throwable().printStackTrace();
				}
				throw new IllegalArgumentException("Cannot find key '" + key
													+ "', in " + cl);
			}
		}

		PieceWiseAccessor pwa;

		if ((pwa = (PieceWiseAccessor) this.class2accessors.get(cl)) == null) {
			pwa = new PieceWiseAccessor();
			this.class2accessors.put(cl, pwa);
		}
		pwa.addSetterFor(setter, key);
		pwa.addGetterFor(getter, key);

		retval = pwa;

		return retval;
	}

	private Accessor resolveKeyInMap(Object object, String key) {
		return MAP_ACCESSOR;
	}

	private Accessor resolveKey(Object object, String key) {
		Accessor retval;

		if (object == null) {
			retval = NULL_ACCESSOR;
		} else if (object == ThinletController.NO_SELECTION) {
			retval = NO_SELECTION_ACCESSOR;
		} else if (object instanceof List) {
			retval = LIST_ACCESSOR;
		} else if ((object instanceof Map) && ((Map) object).containsKey(key)) {
			retval = this.resolveKeyInMap(object, key);
		} else if (((retval = (Accessor) this.class2accessors.get(object
				.getClass())) == null)
					|| !retval.hasSupportFor(object, key)) {
			retval = this.resolveKeyInObject(object, key);
		}
		return retval;
	}

	public static void main(String args[]) {
		final Vector rvic = new Vector();

		rvic.add(new Integer(1));
		rvic.add(new Integer(2));

		final Hashtable ht = new Hashtable();

		ht.put("net.emulab.foo", "hello");
		ht.put("net.emulab.bar", "world");

		final Hashtable ht2 = new Hashtable();

		ht.put("net.emulab.foo", "world");
		ht.put("net.emulab.bar", "hello");

		ThinletKeyValue tkv = new ThinletKeyValue();
		final Object obj = new Object() {
			public int foo() {
				return 1;
			}

			public void foo(int value) {
				System.out.println("set to " + value);
			}

			public String toString() {
				return "obj";
			}
		};
		final Object obj2 = new Object() {
			public int getFoo() {
				return 2;
			}

			public void setFoo(int value) {
				System.out.println("set to " + value);
			}
		};
		final Object obj3 = new Object() {

			public int foo = 5;

			public Object bar = obj;

			public List list = rvic;

			public Hashtable map = ht;

			public String toString() {
				return "obj3";
			}
		};

		long start, end;

		System.out.println("obj.foo() = " + tkv.getKeyValue(obj, "foo"));
		tkv.setKeyValue(obj, "foo", new Integer(3));
		System.out.println("obj.foo() = " + tkv.getKeyValue(obj2, "foo"));
		tkv.setKeyValue(obj2, "foo", new Integer(4));
		System.out.println("obj.foo() = " + tkv.getKeyValue(obj3, "foo"));
		tkv.setKeyValue(obj3, "foo", new Integer(6));
		System.out.println("obj.foo() = " + tkv.getKeyValue(obj3, "foo"));
		System.out.println("obj3.list() = "
							+ tkv.getKeyValue(obj3, "list.doubleValue"));
		System.out.println("obj3.list().size() = "
							+ tkv.getKeyValue(obj3, "list@size"));
		System.out.println("obj3.map().foo() = "
							+ tkv.getKeyValue(obj3, "map.'net.emulab.foo'"));
		System.out.println("obj3.map().bar() = "
							+ tkv.getKeyValue(obj3, "map.'net.emulab.bar'"));

		System.out.println("obj.foo() = " + tkv.getKeyValue(obj3, "bar.foo"));
		tkv.setKeyValue(obj3, "bar.foo", new Integer(10));

		ThinletKeyValueObserver tkvo = new ThinletKeyValueObserver() {
			public void observedKeyValue(Object obj, String keyPath, int flags,
					Object oldValue, Object newValue) {
				System.out.println("changed " + keyPath);
			}
		};
		tkv.observeKeyValue(obj3, "map.'net.emulab.foo'", tkvo);

		tkv.setKeyValue(obj3, "map", ht2);
		tkv.setKeyValue(obj3, "map.'net.emulab.bar'", "busy");
		tkv.setKeyValue(obj3, "foo", new Integer(20));

		tkv.removeKeyValueObserver(obj3, "map.'net.emulab.foo'", tkvo);

		tkv.setKeyValue(obj3, "map.'net.emulab.foo'", "done");

		printStrings(tkv.tokenizePath("arrangedObjects@count"));

		start = System.currentTimeMillis();
		for (int lpc = 0; lpc < 1000000; lpc++) {
			tkv.getKeyValue(obj3, "map.'net.emulab.foo'");
		}
		end = System.currentTimeMillis();

		System.out.println("time = " + (end - start));
	}

	public static void printStrings(String args[]) {
		int lpc;

		System.out.println("size = " + args.length);
		for (lpc = 0; lpc < args.length; lpc++) {
			if (lpc > 0)
				System.out.print(" . ");
			System.out.print(args[lpc]);
		}
		System.out.println();
	}

	public Object objectFromPath(Object obj, String path[], int startingPoint,
			int stoppingPoint, int aggregateIndex) {
		Object retval = obj;
		int lpc;

		for (lpc = startingPoint; lpc < stoppingPoint; lpc++) {
			if (retval == ThinletController.NO_SELECTION) {
			} else if (path[lpc].charAt(0) == '@') {
				// Operation on an aggregate...
				if ("@sum".equals(path[lpc])
						|| "@max".equals(path[lpc])
						|| "@min".equals(path[lpc])) {
					if (retval instanceof Number) {
					}
					else if (stoppingPoint == path.length) {
						List list = (List) retval;
						double sum = 0.0, max = 0.0, min = Double.MAX_VALUE;
						int lpc2, size;
						double value;

						size = list.size();
						for (lpc2 = 0; lpc2 < size; lpc2++) {
							Object sub = this.objectFromPath(
									retval, path, lpc + 1, stoppingPoint, lpc2);
							
							if (sub instanceof Boolean) {
								Boolean bool = (Boolean) sub;
								
								if (bool.booleanValue())
									sum += 1;
							}
							else {
								Number number = (Number) sub;

								sum += number.doubleValue();
								min = Math.min(min, number.doubleValue());
								max = Math.max(max, number.doubleValue());
							}
						}
						if ("@sum".equals(path[lpc]))
							value = sum;
						else if ("@min".equals(path[lpc]))
							value = min;
						else
							value = max;
						if (value == Math.rint(value))
							retval = new Long((long)value);
						else
							retval = new Double(value);
					} else {
						retval = null;
					}
					lpc = stoppingPoint;
				} else if (path[lpc].startsWith("@equals|")) {
					if (retval == null) {
						retval = Boolean.FALSE;
					}
					else {
						String str = retval.toString();
						
						if (str.length() == path[lpc].length() - 8) {
							retval = Boolean.valueOf(str.regionMatches(0, path[lpc], 8,
									path[lpc].length() - 8));
						}
						else {
							retval = Boolean.FALSE;
						}
					}
				} else if ("@not".equals(path[lpc])) {
					if (retval != null && retval != KV_NULL) {
						if (retval instanceof Boolean) {
							retval = ((Boolean) retval).booleanValue() ? Boolean.FALSE
									: Boolean.TRUE;
						}
						else if (retval instanceof Number) {
							retval = Boolean.valueOf(((Number) retval).intValue() != 0);
						}
						else {
							retval = Boolean.FALSE;
						}
					} else {
						retval = Boolean.TRUE;
					}
				} else if ("@notnull".equals(path[lpc])) {
					if (retval != null && retval != KV_NULL) {
						retval = Boolean.TRUE;
					} else {
						retval = Boolean.FALSE;
					}
				} else if ("@size".equals(path[lpc])) {
					if (retval != null)
						retval = new Integer(((Collection) retval).size());
					else
						retval = new Integer(0);
				} else {
					String methodName;

					methodName = path[lpc].substring(1);
					retval = this.resolveKey(retval, methodName).get(retval,
							methodName);
				}
			} else if (retval instanceof Collection) {
				Collection coll = (Collection) retval;

				if (aggregateIndex != -1) {
					// Select instance of an aggregate and its path
					if (aggregateIndex < coll.size()) {
						retval = coll.toArray()[aggregateIndex]; // XXX
						retval = this.resolveKey(retval, path[lpc]).get(retval,
								path[lpc]);
					}
					else {
						retval = null;
					}

					aggregateIndex = -1;
				} else {
					// Collect paths into an aggregate
					Iterator iterator = coll.iterator();
					Vector vector = new Vector(coll.size());
					int collend = stoppingPoint;

					for (int lpc2 = lpc; lpc2 < stoppingPoint; lpc2++) {
						if ("@;".equals(path[lpc2])) {
							collend = lpc2;
							break;
						}
					}
					while (iterator.hasNext()) {
						Object base;

						base = iterator.next();
						try {
							vector.add(this.objectFromPath(base, path, lpc,
									collend, -1));
						}
						catch (IllegalArgumentException e) {
						}
					}
					retval = vector;
					lpc = collend;
					// break; // XXX
				}
			} else {
				// Get singleton path
				retval = this.resolveKey(retval, path[lpc]).get(retval,
						path[lpc]);
			}
		}
		if ((aggregateIndex != -1) && (retval instanceof List)) {
			List list = (List) retval;

			if (aggregateIndex < list.size())
				retval = list.get(aggregateIndex);
			else
				retval = null;
		}
		return retval;
	}

	public Vector findObservers(Object obj) {
		Vector retval = null;

		retval = (Vector) this.observers.get(obj);
		return retval;
	}

	private boolean matchPath(String keyPath1[], String keyPath2[]) {
		boolean retval = true;
		int lpc;

		for (lpc = 0; (lpc < keyPath1.length) && (lpc < keyPath2.length)
						&& retval; lpc++) {
			retval = keyPath1[lpc].equals(keyPath2[lpc]);
		}
		return retval;
	}

	public synchronized void notifyKeyValue(Object obj, String keyPath,
			int flags, Object oldValue, Object newValue) {
		Vector objObservers;
		int slot;

		if ((slot = this.notifyStack.indexOf(keyPath)) != -1) {
			if (this.notifyObjStack.get(slot) == obj) {
				System.out.println("invalidate " + keyPath);
				// new Throwable().printStackTrace(System.out);
				this.notifyStack.set(slot, INVALIDATED_KEY_PATH);
			}
		}
		this.notifyObjStack.push(obj);
		this.notifyStack.push(keyPath);
		if ((objObservers = this.findObservers(obj)) != null) {
			String kt[] = this.tokenizePath(keyPath);
			int lpc, size;

			if (VERBOSE) {
				System.out.println("not " + keyPath + " " + oldValue + " -> "
									+ newValue);
			}
			size = objObservers.size();
			for (lpc = 0; lpc < size; lpc++) {
				Object pair[] = (Object[]) objObservers.elementAt(lpc);

				if (VERBOSE)
					System.out.println("obs " + keyPath + "; " + pair[0]);
				if (matchPath(kt, (String[])pair[1])) {
					ThinletKeyValueObserver tkvo;

					if (VERBOSE)
						System.out.println("obs2 " + newValue);
					tkvo = (ThinletKeyValueObserver) pair[2];
					tkvo.observedKeyValue(obj, keyPath, flags, oldValue,
							newValue);
					if (this.notifyStack.peek() != keyPath) {
						if (!(obj instanceof Object[])) {
							try {
								newValue = this.getKeyValue(obj, keyPath);
							} catch (IllegalArgumentException e) {
								// oh well...
							}
						}
						this.notifyStack.pop();
						this.notifyStack.push(keyPath);
					}
				}
			}
		} else if (VERBOSE) {
			System.out.println("no observers - " + keyPath);
			System.out.println("observers - " + this.observers);
		}
		this.notifyStack.pop();
		this.notifyObjStack.pop();
	}

	public void setKeyValue(Object obj, String keyPath, Object value, int aggregateIndex) {
		String path[] = tokenizePath(keyPath);
		Object model, oldValue;

		model = this.objectFromPath(obj, path, 0, path.length - 1,
				aggregateIndex);
		if (VERBOSE)
			System.out.println("set " + keyPath + "; " + model + "; " + value);
		if (model != null) {
			if (model instanceof Map) {
				Map map = (Map) model;

				if (!map.containsKey(path[path.length - 1])) {
					if (VERBOSE) {
						System.out.println("1st time " + path[path.length - 1]);
					}
					map.put(path[path.length - 1], KV_NULL);
					if (VERBOSE) {
						System.out.println("1st time " + map);
					}
				}
			}
			oldValue = this.objectFromPath(model, path, path.length - 1,
					path.length, aggregateIndex);
			if (oldValue != value) {
				String key = path[path.length - 1];

				if (key.charAt(0) != '@') {
					String partialPath = "";
					Accessor accessor;
					int lpc;

					accessor = this.resolveKey(model, key);
					accessor.set(model, key, value);

					if (VERBOSE) {
						System.out.println("set " + keyPath + " " + oldValue
											+ " -> " + value + " "
											+ obj.hashCode());
					}
					for (lpc = path.length - 1; lpc >= 0; lpc--) {
						if (lpc < path.length - 1)
							partialPath = "." + partialPath;
						if (path[lpc].indexOf('.') == -1)
							partialPath = path[lpc] + partialPath;
						else
							partialPath = "'" + path[lpc] + "'" + partialPath;
						model = this.objectFromPath(obj, path, 0, lpc,
								aggregateIndex);
						this.notifyKeyValue(model, partialPath,
								ThinletKeyValue.CHANGE_SETTING,
								oldValue, value);
					}
				}
			}
		}
	}

	public void setKeyValue(Object obj, String keyPath, Object value) {
		this.setKeyValue(obj, keyPath, value, -1);
	}

	public void setKeyValue(Object obj, String keyPath, boolean value) {
		this.setKeyValue(obj, keyPath, Boolean.valueOf(value), -1);
	}

	public void setKeyValue(Object obj, String keyPath, byte value) {
		this.setKeyValue(obj, keyPath, new Byte(value), -1);
	}

	public void setKeyValue(Object obj, String keyPath, char value) {
		this.setKeyValue(obj, keyPath, new Character(value), -1);
	}

	public void setKeyValue(Object obj, String keyPath, short value) {
		this.setKeyValue(obj, keyPath, new Short(value), -1);
	}

	public void setKeyValue(Object obj, String keyPath, int value) {
		this.setKeyValue(obj, keyPath, new Integer(value), -1);
	}

	public void setKeyValue(Object obj, String keyPath, long value) {
		this.setKeyValue(obj, keyPath, new Long(value), -1);
	}

	public void setKeyValue(Object obj, String keyPath, float value) {
		this.setKeyValue(obj, keyPath, new Float(value), -1);
	}

	public void setKeyValue(Object obj, String keyPath, double value) {
		this.setKeyValue(obj, keyPath, new Double(value), -1);
	}

	public Object getKeyValue(Object obj, String keyPath, int aggregateIndex) {
		String path[] = tokenizePath(keyPath);
		Object retval;

		retval = this.objectFromPath(obj, path, 0, path.length, aggregateIndex);

		if (retval == KV_NULL)
			retval = null;

		return retval;
	}

	public Object getKeyValue(Object obj, String keyPath) {
		return this.getKeyValue(obj, keyPath, -1);
	}
	
	public Object reduceValueList(Object retval) {
		if (retval instanceof List) {
			List list = (List) retval;
			boolean same = true;
			int lpc, size;

			size = list.size();
			if (size == 0) {
				retval = ThinletController.NO_SELECTION;
			} else {
				for (lpc = 1; (lpc < size) && same; lpc++) {
					Object element = list.get(lpc);

					same = element.equals(list.get(lpc - 1));
				}
				if (same)
					retval = list.get(0);
				else
					retval = ThinletController.MULTIPLE_SELECTION;
			}
		}
		
		return retval;
	}

	public Object getKeyValueAsSingleton(Object obj, String keyPath, int index) {
		Object retval;

		retval = this.getKeyValue(obj, keyPath, index);
		if (VERBOSE) {
			System.out.println("single = " + keyPath + " -> " + retval);
		}
		retval = this.reduceValueList(retval);
		
		return retval;
	}

	public Object getKeyValueAsSingleton(Object obj, String keyPath) {
		return this.getKeyValueAsSingleton(obj, keyPath, -1);
	}
	
	public List getKeyValueAsList(Object obj, String keyPath) {
		Object retval = this.getKeyValue(obj, keyPath, -1);
		try {
			return (List) retval;
		}
		catch (ClassCastException e) {
			if (retval instanceof Collection)
				return new ArrayList((Collection)retval);
			else {
				System.out.println(" not a list " + retval);
				return new ArrayList(); // XXX
			}
		}
	}

	public String getKeyValueAsString(Object obj, String keyPath) {
		String retval = null;
		Object rc;
		
		if ((rc = this.getKeyValueAsSingleton(obj, keyPath)) != null)
			retval = rc.toString();
		
		return retval;
	}

	public boolean getKeyValueAsBoolean(Object obj, String keyPath) {
		boolean retval;
		Object rc;

		rc = this.getKeyValueAsSingleton(obj, keyPath);
		if (rc instanceof Boolean)
			retval = ((Boolean) rc).booleanValue();
		else if (rc instanceof Number)
			retval = ((Number) rc).intValue() != 0;
		else
			retval = (rc != null);
		return retval;
	}

	public byte getKeyValueAsByte(Object obj, String keyPath) {
		return ((Number) this.getKeyValueAsSingleton(obj, keyPath)).byteValue();
	}

	public char getKeyValueAsChar(Object obj, String keyPath) {
		return ((Character) this.getKeyValueAsSingleton(obj, keyPath)).charValue();
	}
	
	private static final Double ZERO = new Double(0.0);
	private static final NumberFormat DEFAULT_NUMBER_FORMAT = NumberFormat.getInstance();
	
	public Number getKeyValueAsNumber(Object obj, String keyPath, Number defaultValue) {
		Number retval = defaultValue;
		Object rc;
		
		if ((rc = this.getKeyValueAsSingleton(obj, keyPath)) != null) {
			if (rc instanceof Number) {
				retval = (Number)rc;
			}
			else if (rc instanceof String) {
				try {
					retval = DEFAULT_NUMBER_FORMAT.parse((String)rc);
				}
				catch (Throwable th) {
				}
			}
		}
		
		return retval;
	}

	public short getKeyValueAsShort(Object obj, String keyPath) {
		return this.getKeyValueAsNumber(obj, keyPath, ZERO).shortValue();
	}

	public int getKeyValueAsInteger(Object obj, String keyPath) {
		return this.getKeyValueAsNumber(obj, keyPath, ZERO).intValue();
	}

	public long getKeyValueAsLong(Object obj, String keyPath) {
		return this.getKeyValueAsNumber(obj, keyPath, ZERO).longValue();
	}

	public float getKeyValueAsFloat(Object obj, String keyPath) {
		return this.getKeyValueAsNumber(obj, keyPath, ZERO).floatValue();
	}

	public double getKeyValueAsDouble(Object obj, String keyPath) {
		return this.getKeyValueAsNumber(obj, keyPath, ZERO).doubleValue();
	}

	public List getKeyValueAsList(Object obj, String keyPath, List defaultValue) {
		try {
			return this.getKeyValueAsList(obj, keyPath);
		}
		catch (IllegalArgumentException e) {
			return defaultValue;
		}
	}

	public String getKeyValueAsString(Object obj, String keyPath, String defaultValue) {
		try {
			return this.getKeyValueAsString(obj, keyPath);
		}
		catch (IllegalArgumentException e) {
			return defaultValue;
		}
	}

	public int getKeyValueAsInteger(Object obj, String keyPath, int defaultValue) {
		try {
			return this.getKeyValueAsInteger(obj, keyPath);
		}
		catch (IllegalArgumentException e) {
			return defaultValue;
		}
	}

	public double getKeyValueAsDouble(Object obj, String keyPath, double defaultValue) {
		try {
			return this.getKeyValueAsDouble(obj, keyPath);
		}
		catch (IllegalArgumentException e) {
			return defaultValue;
		}
	}

	public synchronized void observeKeyValue(Object obj, String keyPath,
			ThinletKeyValueObserver tkvo) {
		String path[] = tokenizePath(keyPath);
		Object parent, value;
		Vector vector;

		if (obj == null)
			obj = KV_NULL;

		parent = this.objectFromPath(obj, path, 0, path.length - 1, -1);
		if (parent instanceof Map) {
			Map map = (Map) parent;

			if (!map.containsKey(path[path.length - 1]))
				map.put(path[path.length - 1], KV_NULL);
		}

		if (VERBOSE) {
			System.out.println("observe - " + obj + " " + keyPath + " "
								+ obj.hashCode());
		}

		if ((vector = (Vector) this.observers.get(obj)) == null) {
			vector = new Vector();
			this.observers.put(obj, vector);
		}
		vector.add(new Object[] { keyPath, tokenizePath(keyPath), tkvo });

		try {
			if ((value = this.getKeyValue(obj, keyPath)) != KV_NULL) {
				tkvo.observedKeyValue(obj, keyPath, CHANGE_SETTING, null,
						value);
			}
		} catch (IllegalArgumentException e) {
		}
	}

	public synchronized void observeKeyValue(Object obj, final String keyPath,
			final int expectedFlags, final Object handler, final String value) {
		try {
			StringTokenizer st = new StringTokenizer(value, "(), ");
			final String pnames[], methodName;
			final Method method;
			int pcount, lpc;
			Class ptypes[];

			methodName = st.nextToken();
			pcount = st.countTokens();
			if (pcount > 0) {
				ptypes = new Class[pcount];
				pnames = new String[pcount];

				for (lpc = 0; lpc < ptypes.length; lpc++) {
					String arg = st.nextToken();

					pnames[lpc] = arg;
					if ("object".equals(arg))
						ptypes[lpc] = Object.class;
					else if ("keyPath".equals(arg))
						ptypes[lpc] = String.class;
					else if ("old".equals(arg))
						ptypes[lpc] = Object.class;
					else if ("new".equals(arg))
						ptypes[lpc] = Object.class;
					else
						throw new InternalError();
				}

				method = handler.getClass().getMethod(methodName, ptypes);
			} else {
				pnames = null;
				method = null;
			}
			this.observeKeyValue(obj, keyPath, new ThinletKeyValueObserver() {
				public void observedKeyValue(Object obj, String nkeyPath,
						int flags, Object oldValue, Object newValue) {
					if ((flags & expectedFlags) == 0)
						return;
					if (newValue == ThinletController.NO_SELECTION) {
						ThinletKeyValue.this.setKeyValue(handler, value, null);
						return;
					}
					
					if ((flags & CHANGE_REPLACE) != 0) {
						int range[] = (int[])newValue;
						List list;
						int lpc;
						
						list = (List)getKeyValue(handler, value);
						for (lpc = range[0]; lpc < range[1]; lpc++) {
							list.set(lpc, getKeyValue(obj, keyPath, lpc));
						}

						notifyKeyValue(handler, value, flags, null, newValue);
						// System.out.println(" repl " + list);
						return;
					}
					else if ((flags & CHANGE_INSERTION) != 0) {
						int range[] = (int[])newValue;
						List list;
						int lpc;
						
						list = (List)getKeyValue(handler, value);
						for (lpc = range[0]; lpc < range[1]; lpc++) {
							list.add(getKeyValue(obj, keyPath, lpc));
						}
						
						notifyKeyValue(handler, value, flags, null, newValue);
						// System.out.println(" ins " + list);
						return;
					}
					else if ((flags & CHANGE_REMOVAL) != 0) {
						int range[] = (int[])newValue;
						List list;
						int lpc;
						
						list = (List)getKeyValue(handler, value);
						for (lpc = range[0]; lpc < range[1] && lpc < list.size(); lpc++) {
							list.remove(lpc);
						}
						
						notifyKeyValue(handler, value, flags, null, newValue);
						// System.out.println(" rem " + list);
						return;
					}
					
					newValue = getKeyValue(obj, keyPath);
					
					if (method == null) {
						ThinletKeyValue.this.setKeyValue(handler, value,
								newValue);
						return;
					}

					try {
						Object args[] = new Object[pnames.length];
						int lpc;

						for (lpc = 0; lpc < args.length; lpc++) {
							if ("object".equals(pnames[lpc]))
								args[lpc] = obj;
							else if ("keyPath".equals(pnames[lpc]))
								args[lpc] = keyPath;
							else if ("old".equals(pnames[lpc]))
								args[lpc] = oldValue;
							else if ("new".equals(pnames[lpc]))
								args[lpc] = newValue;
							else
								throw new InternalError();
						}
						method.invoke(handler, args);
					} catch (IllegalAccessException e) {
						e.printStackTrace();
					} catch (InvocationTargetException e) {
						e.printStackTrace();
					}
				}
			});
		} catch (NoSuchMethodException e) {
			e.printStackTrace();
			throw new InternalError();
		}
	}

	public synchronized void removeKeyValueObserver(Object obj, String keyPath,
			ThinletKeyValueObserver tkvo) {
		Vector vector;

		if (obj == null)
			obj = KV_NULL;
		if ((vector = (Vector) this.observers.get(obj)) != null) {
			boolean found = false;
			int lpc, size;

			size = vector.size();
			for (lpc = 0; (lpc < size) && !found; lpc++) {
				Object elem[] = (Object[]) vector.elementAt(lpc);

				if (elem[0].equals(keyPath) && (elem[1] == tkvo)) {
					vector.remove(lpc);
					found = true;
				}
			}
			if (vector.size() == 0) {
				this.observers.remove(obj);
			}
		}
	}

	public synchronized void addKeyDependency(final Object obj,
			final String targetPath, final Object obj2, final String dependencyPath) {
		this.observeKeyValue(obj2, dependencyPath,
				new ThinletKeyValueObserver() {
					public void observedKeyValue(Object obj, String keyPath,
							int flags, Object oldValue, Object newValue) {
						ThinletKeyValue tkv;

						tkv = ThinletKeyValue.this;
						tkv.notifyKeyValue(obj, targetPath,
								ThinletKeyValue.CHANGE_SETTING, null, tkv
										.getKeyValue(obj, targetPath));
					}
				});
	}

	public synchronized void addKeyDependency(final Object obj,
			final String targetPath, final String dependencyPath) {
		this.addKeyDependency(obj, targetPath, obj, dependencyPath);
	}
	
	public String toString() {
		return "ThinletKeyValue[observers=" + this.observers + "]";
	}
}
