package thinlet;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.Rectangle;
import java.awt.event.InputEvent;
import java.awt.event.KeyEvent;
import java.io.UnsupportedEncodingException;
import java.lang.reflect.Field;
import java.text.FieldPosition;
import java.util.StringTokenizer;
import java.util.Vector;

public abstract class ThinletModel extends ThinletAPI {
	transient ThinletFormatter focusformatter;

	transient final FieldPosition fp = new FieldPosition(0);

	/**
	 * Creates a new component
	 * 
	 * @param classname
	 *            the widget type (e.g. <i>button</i>)
	 * @return a new component, every component is simply an <i>Object</i>
	 * @throws java.lang.IllegalArgumentException
	 *             for unknown widget type
	 */
	public Object create(String classname) { // #
		if (classname == null)
			throw new IllegalArgumentException("Null classname");
		if ("".equals(classname))
			throw new IllegalArgumentException("Empty classname");

		for (int i = 0; i < dtd.length; i += 3) {
			if (dtd[i].equals(classname)) {
				return createImpl((String) dtd[i]);
			}
		}
		throw new IllegalArgumentException("Unknown widget type '" + classname
											+ "'");
	}

	/**
	 * Gets the index of the first selected item in the given component
	 * 
	 * @param component
	 *            a widget (combobox, tabbedpane, list, table, or tree)
	 * @return the first selected index or -1
	 */
	public int getSelectedIndex(Object component) {
		String classname = getClass(component);

		if ((classname == "combobox") || (classname == "tabbedpane")) {
			return getInteger(component, "selected",
					(classname == "combobox") ? -1 : 0);
		}

		if ((classname == "list") || (classname == "table")
			|| (classname == "header") || (classname == "tree")
			|| (classname == "dropzone")) {
			Object item = get(component, ":comp");
			for (int i = 0; item != null; i++) {
				if (getBoolean(item, "selected", false)) {
					return i;
				}
				item = get(item, ":next");
			}
			return -1;
		}

		throw new IllegalArgumentException(classname);
	}

	/**
	 * Gets the first selected item of the given component
	 * 
	 * @param component
	 *            a widget (combobox, tabbedpane, list, table, or tree)
	 * @return the first selected item or null
	 */
	public Object getSelectedItem(Object component) {
		String classname = getClass(component);

		if ((classname == "combobox") || (classname == "tabbedpane")) {
			int index = getInteger(component, "selected",
					(classname == "combobox") ? -1 : 0);

			return (index != -1) ? getItemImpl(component, ":comp", index)
					: null;
		}

		if ((classname == "list") || (classname == "table")
			|| (classname == "header") || (classname == "tree")
			|| (classname == "dropzone")) {
			for (Object item = findNextItem(component, classname, null); item != null; item = findNextItem(
					component, classname, item)) {
				if (getBoolean(item, "selected", false)) {
					return item;
				}
			}
			return null;
		}

		throw new IllegalArgumentException(classname);
	}

	/**
	 * Gets the selected item of the given component (list, table, or tree) when
	 * multiple selection is allowed
	 * 
	 * @param component
	 *            a widget
	 * @return the array of selected items, or a 0 length array
	 */
	public Object[] getSelectedItems(Object component) {
		String classname = getClass(component);
		Object[] selecteds = new Object[0];

		for (Object item = findNextItem(component, classname, null); item != null; item = findNextItem(
				component, classname, item)) {
			if (getBoolean(item, "selected", false)) {
				Object[] temp = new Object[selecteds.length + 1];
				System.arraycopy(selecteds, 0, temp, 0, selecteds.length);
				temp[selecteds.length] = item;
				selecteds = temp;
			}
		}

		return selecteds;
	}

	/**
	 * @return the first or the next item of the (list, table, or tree)
	 *         component
	 */
	Object findNextItem(Object component, String classname, Object item) {
		if (item == null) { // first item
			return get(component, ":comp");
		} else if ("tree" == classname) { // next tree node
			Object next = get(item, ":comp");
			if ((next == null) || !getBoolean(item, "expanded", true)) { // no
																			// subnode
																			// or
																			// collapsed
				while ((item != component)
						&& ((next = get(item, ":next")) == null)) {
					item = getParent(item); // next node of in backward path
				}
			}
			return next;
		} else { // next list or tree item
			return get(item, ":next");
		}
	}

	/**
	 * Removes all the components from this container's specified list
	 * 
	 * @param component
	 *            the specified container
	 */
	public void removeAll(Object component) {
		if (get(component, ":comp") != null) {
			set(component, ":comp", null);
			set(component, ":lead", null);
			update(component, "validate");
			// printcomp(component);
		}
	}

	/**
	 * Adds the specified component to the root desktop
	 * 
	 * @param component
	 *            a widget to be added
	 */
	public void add(Object component) {
		add(content, component, 0);
	}

	/**
	 * Adds the specified component to the end of the specified container
	 * 
	 * @param parent
	 *            a container
	 * @param component
	 *            a component to be added
	 */
	public void add(Object parent, Object component) {
		add(parent, component, -1);
	}

	/**
	 * Adds the specified component to the container at the given position
	 * 
	 * @param parent
	 *            a container
	 * @param component
	 *            a component to be inserted
	 * @param index
	 *            the position at which to insert the component, or -1 to insert
	 *            the component at the end
	 */
	public void add(Object parent, Object component, int index) {
		addImpl(parent, component, index);
		update(component, "validate");
		if (parent == content) {
			if (instance(getClass(component), "panel")
				&& getBoolean(component, "modal", false)) {
				set(component, ":postmodalfocus", focusowner);
				focusowner = null;
			}
			if (instance(getClass(component), "component")) {
				if (isVisible(component)) {
					if (!invoke(component, null, "focusable")) {
						setNextFocusable(component, false);
					}
				} else {
					setNextFocusable(component, false);
				}
			}
		}
	}

	/**
	 * Remove the specified component from its parent list, or delete
	 * component's popupmenu or table's header
	 * 
	 * @param component
	 *            the component to be removed
	 */
	public void remove(Object component) {
		update(component, "gparent");
		Object parent = getParent(component);
		Object classname = getClass(component);
		if ("dialog" == classname && getBoolean(component, "modal")) {
			repaint(content);
		}
		if (("popupmenu" == classname) || ("header" == classname)) {
			set(parent, classname, null);
		} else {
			removeItemImpl(parent, component);
			// request focus for its parent if the component (or subcomponent)
			// is currently focused
			// System.out.println("focus " + focusowner + " " + component);
			for (Object comp = focusowner; comp != null; comp = getParent(comp)) {
				if (comp == component) {
					Object postmodalfocus;

					postmodalfocus = get(component, ":postmodalfocus");
					if (postmodalfocus != null) {
						focusowner = postmodalfocus;
						if (instance(getClass(focusowner), "textfield")) {
							focusformatter = (ThinletFormatter) getDelegate(
									focusowner, "formatter");
						}
						set(component, ":postmodalfocus", null);
					} else {
						setNextFocusable(parent, true);
					}
					break;
				}
				// System.out.println("again");
			}
		}
	}

	/**
	 * Finds the first component from the root desktop by a specified name value
	 * 
	 * @param name
	 *            parameter value identifies the widget
	 * @return the first suitable component, or null
	 */
	public Object find(String name) {
		return find(content, name);
	}

	/**
	 * Finds the first component from the specified component by a name
	 * 
	 * @param component
	 *            the widget is searched inside this component
	 * @param name
	 *            parameter value identifies the widget
	 * @return the first suitable component, or null
	 */
	public Object find(Object component, String name) {
		if (name.equals(get(component, "name"))) {
			return component;
		}
		// otherwise search in its subcomponents
		Object found = null;
		for (Object comp = get(component, ":comp"); comp != null; comp = get(
				comp, ":next")) {
			if ((found = find(comp, name)) != null) {
				return found;
			}
		}
		// search in table header
		Object header = get(component, "header"); // if ("table" == classname)
		if ((header != null) && ((found = find(header, name)) != null)) {
			return found;
		}
		// search in component's popupmenu
		Object popupmenu = get(component, "popupmenu"); // if
														// instance(classname,
														// "component")
		if ((popupmenu != null) && ((found = find(popupmenu, name)) != null)) {
			return found;
		}
		return null;
	}

	/**
	 * @param component
	 * @param keycode
	 * @param modifiers
	 * @return true if the component has the given mnemonic
	 */
	boolean hasMnemonic(Object component, int keycode, int modifiers) {
		if (modifiers == InputEvent.ALT_MASK) {
			int index = getInteger(component, "mnemonic", -1);
			if (index != -1) {
				String text = getString(component, "text", null);
				return (text != null)
						&& (text.length() > index)
						&& (Character.toUpperCase(text.charAt(index)) == keycode);
			}
		}
		return false;
	}

	/**
	 * @param component
	 * @param keycode
	 * @param modifiers
	 * @return true if the component has the given accelerator
	 */
	boolean hasAccelerator(Object component, int keycode, int modifiers) {
		Object accelerator = get(component, "accelerator");
		if (accelerator != null) {
			long keystroke = ((Long) accelerator).longValue();
			return ((keystroke >> 32) == modifiers)
					&& ((keystroke & 0xffff) == keycode);
		}
		return false;
	}

	/**
	 * @param component
	 *            a menuitem
	 * @return key modifier strings and key text
	 */
	String getAccelerator(Object component) {
		Object accelerator = get(component, "accelerator");
		if (accelerator != null) {
			long keystroke = ((Long) accelerator).longValue();
			int keycode = (int) (keystroke >> 32);
			int modifiers = (int) (keystroke & 0xffff);
			return KeyEvent.getKeyModifiersText(keycode) + " "
					+ KeyEvent.getKeyText(modifiers);
		}
		return null;
	}

	public boolean isVisible(Object component) {
		boolean retval = false;

		while ((component != null) && !retval) {
			Boolean vis;

			if (((vis = (Boolean) get(component, "visible")) != null)
				&& !vis.booleanValue()) {
				return false;
			} else if (component == content) {
				retval = true;
			} else if ("tab" == getClass(component)
						&& (component != getSelectedItem(getParent(component)))) {
				return false;
			}
			component = getParent(component);
		}
		return retval;
	}

	void invokeShow(Object component) {
		String classname = getClass(component);

		if (instance(classname, "component") || ("column" == classname)
			|| ("header" == classname)) {
			invoke(component, null, "show");
		}
	}

	/**
	 * Add the component to the parent's ':comp' list, and set its ':parent' or
	 * set single components
	 * 
	 * @param index
	 *            add at the specified index
	 * @throws java.lang.IllegalArgumentException
	 */
	void addImpl(Object parent, Object component, int index) {
		String parentclass = getClass(parent);
		String classname = getClass(component);
		if ("flow" == classname) {
			/* Set parent now and */
			set(component, ":parent", parent);
			/* ... connect it later in the parser. */
		} else if ((("flow" == parentclass) || ("cond" == parentclass)
					|| ("option" == parentclass) || ("confirm" == parentclass))
					&& (("cond" == classname) || ("confirm" == classname)
						|| ("option" == classname) || ("action" == classname))) {
			/* XXX better checking here */
			insertItemImpl(parent, ":comp", component, index);
		} else if ((("combobox" == parentclass) && ("choice" == classname))
					|| (("tabbedpane" == parentclass) && ("tab" == classname))
					|| (("list" == parentclass) && ("item" == classname))
					|| (("table" == parentclass) && ("row" == classname))
					|| (("header" == parentclass) && ("column" == classname))
					|| (("row" == parentclass) && ("cell" == classname))
					|| ((("tree" == parentclass) || ("node" == parentclass)) && ("node" == classname))
					|| (("menubar" == parentclass) && ("menu" == classname))
					|| ((("menu" == parentclass) || ("popupmenu" == parentclass)) && (("menu" == classname)
																						|| ("menuitem" == classname)
																						|| ("checkboxmenuitem" == classname) || ("separator" == classname)))
					|| ((("panel" == parentclass) || ("dropzone" == parentclass) || ("desktop" == parentclass)
							|| ("splitpane" == parentclass)
							|| ("dialog" == parentclass) || ("tab" == parentclass))
						&& instance(classname, "component") && (classname != "popupmenu"))) {
			Rectangle bounds;

			insertItemImpl(parent, ":comp", component, index);
			if (instance(classname, "component") && isVisible(component)) {
				invokeShow(component);
			}
			if (("dialog" == parentclass)
				&& (bounds = getRectangle(parent, "bounds")) != null) {
				Dimension d = getPreferredSize(parent);

				// System.out.println("bounds = " + bounds);
				// System.out.println("gpbounds = " + gpbounds);
				// System.out.println("d = " + d);
				if (bounds.x != 0) {
					bounds.x += Math.abs(bounds.width) - d.width;
					if (bounds.x < 0)
						bounds.x = 0;
				}
				if (bounds.y != 0) {
					bounds.y += Math.abs(bounds.height) - d.height;
					if (bounds.y < 0)
						bounds.y = 0;
				}
				// System.out.println("diff = " + diff);
				// System.out.println("new x = " + bounds.x);
				if (getBoolean(parent, "resizable", false)) {
					if (d.width < bounds.width)
						d.width = bounds.width;
					if (d.height < bounds.height)
						d.height = bounds.height;
				}
				bounds.width = d.width;
				bounds.height = d.height;
			} else if (getRectangle(parent, "bounds") != null) {
				// System.out.println("+++ doing layout ***");
				// printcomp(parent);
				// doLayout(parent);
				// System.out.println("+++ after layout ***");
				// printcomp(parent);
			}
		} else if ((("table" == parentclass) && ("header" == classname))
					|| (("popupmenu" == classname) && instance(parentclass,
							"component"))) {
			set(parent, classname, component);
			set(component, ":parent", parent);
		} else if (("column" == parentclass)
					&& (get(parent, ":editor") == null)) {
			set(parent, ":editor", component);
			set(component, ":parent", parent);
		} else
			throw new IllegalArgumentException(classname + " add "
												+ parentclass);
	}

	/**
	 * 
	 */
	boolean instance(Object classname, Object extendclass) {
		if (classname == extendclass) {
			return true;
		}
		for (int i = 0; i < dtd.length; i += 3) {
			if (classname == dtd[i]) {
				return instance(dtd[i + 1], extendclass);
			}
		}
		return false;
	}

	/**
	 * 
	 */
	Object addElement(Object parent, String name) {
		Object component = create(name);
		addImpl(parent, component, -1);
		return component;
	}

	/**
	 * 
	 * @throws java.lang.IllegalArgumentException
	 */
	Vector addAttribute(final Object component, String key, String value,
			String encoding, Vector lasts, Object handler)
			throws UnsupportedEncodingException {
		// replace value found in the bundle
		if ((resourcebundle != null) && value.startsWith("i18n.")) {
			value = resourcebundle.getString(value.substring(5));
		}

		final Object[] definition = getDefinition(getClass(component), key,
				null);
		key = (String) definition[1];

		if ((value.length() > 0) && (value.charAt(0) == '@')) {
			if (lasts == null) {
				lasts = new Vector();
			}
			lasts.addElement(component);
			lasts.addElement(definition);
			lasts.addElement(value);
			// bindAttribute(component, key, value, handler);
			return lasts;
		}

		if ("string" == definition[0]) {
			value = (encoding == null) ? new String(value) : new String(value
					.getBytes(), 0, value.length(), encoding);
			setString(component, key, value, (String) definition[3]);
		} else if ("choice" == definition[0]) {
			String[] values = (String[]) definition[3];
			setChoice(component, key, value, values, values[0]);
		} else if ("boolean" == definition[0]) {
			if ("true".equals(value)) {
				if (definition[3].equals(Boolean.FALSE)) {
					set(component, key, Boolean.TRUE);
				}
			} else if ("false".equals(value)) {
				if (definition[3].equals(Boolean.TRUE)) {
					set(component, key, Boolean.FALSE);
				}
			} else
				throw new IllegalArgumentException(value);
		} else if ("integer" == definition[0]) {
			set(component, key, Integer.valueOf(value));
		} else if ("icon" == definition[0]) {
			set(component, key, getIcon(value));
		} else if (("method" == definition[0])
					|| ("component" == definition[0])) {
			if (lasts == null) {
				lasts = new Vector();
			}
			lasts.addElement(component);
			lasts.addElement(definition);
			lasts.addElement(value);
		} else if ("outlet" == definition[0]) {
			if (lasts == null) {
				lasts = new Vector();
			}
			lasts.add(0, component);
			lasts.add(1, definition);
			lasts.add(2, value);
		} else if ("content" == definition[0]) {
		} else if ("delegate" == definition[0]) {
			try {
				Field field = handler.getClass().getField(value);

				if (ThinletDelegate.class.isAssignableFrom(field.getType())) {
					set(component, definition[1], field.get(handler));
				} else {
					throw new IllegalArgumentException(
							"Delegate " + handler.getClass() + "." + value
									+ " is not a ThinletDelegate");
				}
			} catch (NoSuchFieldException e) {
				throw new IllegalArgumentException("Delegate "
													+ handler.getClass() + "."
													+ value
													+ " does not exist.");
			} catch (IllegalAccessException e) {
				throw new IllegalArgumentException("Delegate "
													+ handler.getClass() + "."
													+ value
													+ " is not accessible.");
			}
		} else if ("property" == definition[0]) {
			StringTokenizer st = new StringTokenizer(value, ";");
			while (st.hasMoreTokens()) {
				String token = st.nextToken();
				int equals = token.indexOf('=');
				if (equals == -1) {
					throw new IllegalArgumentException(token);
				}
				putProperty(component, new String(token.substring(0, equals)),
						new String(token.substring(equals + 1)));
			}
		} else if ("font" == definition[0]) {
			String name = null;
			boolean bold = false;
			boolean italic = false;
			int size = 0;
			StringTokenizer st = new StringTokenizer(value);
			while (st.hasMoreTokens()) {
				String token = st.nextToken();
				if ("bold".equalsIgnoreCase(token)) {
					bold = true;
				} else if ("italic".equalsIgnoreCase(token)) {
					italic = true;
				} else {
					try {
						size = Integer.parseInt(token);
					} catch (NumberFormatException nfe) {
						name = (name == null) ? new String(token)
								: (name + " " + token);
					}
				}
			}
			if (name == null) {
				name = font.getName();
			}
			if (size == 0) {
				size = font.getSize();
			}
			set(component, key, new Font(name, (bold ? Font.BOLD : 0)
												| (italic ? Font.ITALIC : 0),
					size));
		} else if ("color" == definition[0]) {
			int color = 0;
			if (value.startsWith("#")) {
				color = Integer.parseInt(value.substring(1), 16);
			} else if (value.startsWith("0x")) {
				color = Integer.parseInt(value.substring(2), 16);
			} else { // three separated integer including red, green, and
						// blue
				StringTokenizer st = new StringTokenizer(value, " \r\n\t,");
				color = 0xff000000
						| ((Integer.parseInt(st.nextToken()) & 0xff) << 16)
						| ((Integer.parseInt(st.nextToken()) & 0xff) << 8)
						| (Integer.parseInt(st.nextToken()) & 0xff);
			}
			set(component, key, new Color(color));
		} else if ("keystroke" == definition[0]) {
			setKeystrokeImpl(component, key, value);
		} else if ("bean" == definition[0]) {
			try {
				Object beanObject = Class.forName(value).newInstance();
				if (beanObject instanceof CustomComponent) {
					CustomComponent bean = (CustomComponent) beanObject;
					bean.setComponent(component);
				}
				set(component, key, beanObject);
			} catch (Exception exc) {
				throw new IllegalArgumentException(value);
			}
		} else if ("object" == definition[0]) {
			setString(component, key, value, (String) definition[3]);
		} else
			throw new IllegalArgumentException((String) definition[0]);
		return lasts;
	}

	/**
	 * 
	 * @throws java.lang.IllegalArgumentException
	 */
	protected Object[] getDefinition(Object classname, String key, String type) {
		Object currentname = classname;
		while (classname != null) {
			boolean detected = false;
			for (int i = 0; i < dtd.length; i += 3) {
				if (dtd[i] == classname) {
					detected = true;
					Object[][] attributes = (Object[][]) dtd[i + 2];
					if (attributes != null) {
						for (int j = 0; j < attributes.length; j++) {
							if (attributes[j][1].equals(key)) {
								if ((type != null)
									&& (type != attributes[j][0])) {
									throw new IllegalArgumentException(
											attributes[j][0].toString());
								}
								return attributes[j];
							}
						}
					}
					classname = dtd[i + 1];
					break;
				}
			}
			if (!detected)
				break;
		}
		throw new IllegalArgumentException("unknown " + key + " " + type
											+ " for " + currentname);
	}

	/**
	 * Gets the child class that populates this container style component. For
	 * example, passing in a tree will return an node.
	 * 
	 * @param parentclass
	 *            the parent clas name
	 * @return the child class name
	 * @throws IllegalArgumentException
	 *             if the parent is not a container
	 */
	String getChildClass(String parentclass) {
		if ("list" == parentclass) {
			return "item";
		} else if ("tree" == parentclass) {
			return "node";
		} else if ("table" == parentclass) {
			return "row";
		} else if ("combobox" == parentclass) {
			return "choice";
		} else if ("tabbedpane" == parentclass) {
			return "tab";
		} else {
			throw new IllegalArgumentException(parentclass + " has no item");
		}
	}

	public abstract boolean invoke(Object component, Object part, String event);

	/**
	 * 
	 */
	protected abstract boolean setNextFocusable(Object current, boolean outgo); // TODO:
																		// SJC:
																		// remove??

	abstract Dimension getPreferredSize(Object component);

	protected String keyForCache(Object key) {
		for (int lpc = 0; lpc < cacheKeys.length; lpc++) {
			if (cacheKeys[lpc][0].equals(key))
				return cacheKeys[lpc][1];
		}
		return null;
	}

	protected static String[][] cacheKeys;
	static {
		cacheKeys = new String[][] { { "text", ":textcache" },
				{ "empty", ":emptycache" }, };
	}

	protected static Object[] dtd;
	static {
		Integer integer_1 = new Integer(-1);
		Integer integer0 = new Integer(0);
		Integer integer1 = new Integer(1);
		String[] orientation = { "horizontal", "vertical" };
		String[] leftcenterright = { "left", "center", "right" };
		String[] fillleftcenterright = { "fill", "left", "center", "right" };
		String[] placement = { "q", "w", "e", "a", "s", "d", "z", "x", "c" };
		String[] selections = { "single", "interval", "multiple" }; // +none
		dtd = new Object[] {
				"component",
				null,
				new Object[][] {
						{ "string", "name", null, null },
						{ "boolean", "enabled", "paint", Boolean.TRUE },
						{ "boolean", "isfocusable", "paint", Boolean.TRUE },
						{ "boolean", "overlay", "validate", Boolean.FALSE },
						{ "choice", "placement", "validate", placement },
						{ "boolean", "visible", "gparent", Boolean.TRUE },
						{ "boolean", "i18n", "validate", Boolean.FALSE }, // for
																			// I18N
						{ "string", "tooltip", null, null },
						{ "font", "font", "validate", null },
						{ "font", "tooltipfont", "validate", null },
						{ "color", "foreground", "paint", null },
						{ "color", "background", "paint", null },
						{ "boolean", "usegradient", "paint", Boolean.FALSE },
						{ "integer", "width", "validate", integer0 },
						{ "integer", "height", "validate", integer0 },
						{ "integer", "colspan", "validate", integer1 },
						{ "integer", "rowspan", "validate", integer1 },
						{ "integer", "weightx", "validate", integer0 },
						{ "integer", "weighty", "validate", integer0 },
						{
								"choice",
								"halign",
								"validate",
								new String[] { "fill", "center", "left",
										"right" } },
						{
								"choice",
								"valign",
								"validate",
								new String[] { "fill", "center", "top",
										"bottom" } },
						// component class String null*
						// parent Object null
						// (bounds) Rectangle 0 0 0 0
						{ "property", "property", null, null },
						{ "choice", "selections", "paint", selections },
						{ "method", "init" }, { "method", "focuslost" },
						{ "method", "focusgained" }, { "method", "drop" },
						{ "method", "show" }, { "method", "focusable" },
						{ "outlet", "outlet" },
						{ "delegate", "dragsrc" }, { "delegate", "dropdst" },
						{ "delegate", "contentmapper" }, },
				"label",
				"component",
				new Object[][] {
						{ "string", "text", "validate", null },
						{ "icon", "icon", "validate", null },
						{ "choice", "alignment", "validate",
								fillleftcenterright },
						{ "integer", "mnemonic", "paint", integer_1 },
						{ "component", "for", null, null },
						{ "delegate", "formatter", "layout", null }, },
				"button",
				"label",
				new Object[][] {
						{ "choice", "alignment", "validate",
								fillleftcenterright },
						{ "method", "action" },
						{ "boolean", "connectable", "validate", Boolean.FALSE },
						{ "integer", "proximity", "validate", integer_1 },
						{ "string", "proximitytooltip", null, null },
						{ "method", "connect" },
						{ "method", "delete" },
						{ "choice", "style", "validate",
							new String[] { "normal", "round" } },
						{
								"choice",
								"type",
								"paint",
								new String[] { "normal", "default", "cancel",
										"link" } } },
				"checkbox",
				"label",
				new Object[][] {
						{ "boolean", "selected", "paint", Boolean.FALSE }, // ...group
						{ "string", "group", "paint", null }, // ...group
						{ "choice", "style", "paint",
								new String[] { "normal", "arrow", "blank" } },
						{ "method", "action" } },
				"togglebutton",
				"checkbox",
				new Object[][] {
						{ "choice", "style", "validate",
							new String[] { "normal", "round" } },
						{ "boolean", "connectable", "validate", Boolean.FALSE },
						{ "integer", "proximity", "validate", integer_1 },
						{ "string", "proximitytooltip", null, null },
						{ "method", "connect" }, { "method", "delete" }, },
				"connector",
				"togglebutton",
				new Object[][] { { "string", "left", "validate", null },
						{ "string", "right", "validate", null }, },
				"combobox",
				"textfield",
				new Object[][] {
						{ "icon", "icon", "validate", null },
						{ "boolean", "pickable", "paint", Boolean.TRUE },
						{ "choice", "style", "validate",
								new String[] { "normal", "round" } },
						{ "integer", "selected", "layout", integer_1 },
						{ "content", "content", }, },
				"choice",
				null,
				new Object[][] {
						{ "string", "name", null, null },
						{ "boolean", "enabled", "paint", Boolean.TRUE },
						{ "boolean", "i18n", "validate", Boolean.FALSE }, // for
																			// I18N
						{ "object", "text", "parent", null },
						{ "icon", "icon", "parent", null },
						{ "choice", "alignment", "parent", fillleftcenterright },
						{ "string", "tooltip", null, null },
						{ "font", "font", "validate", null },
						{ "color", "foreground", "paint", null },
						{ "color", "background", "paint", null },
						{ "property", "property", null, null },
						{ "outlet", "outlet" },
						{ "delegate", "formatter", "layout", null }, },
				"textfield",
				"component",
				new Object[][] { { "object", "text", "layout", "" },
						{ "integer", "columns", "validate", integer0 },
						{ "boolean", "editable", "paint", Boolean.TRUE },
						{ "choice", "alignment", "validate", leftcenterright },
						{ "boolean", "selectable", "paint", Boolean.TRUE },
						{ "integer", "start", "layout", integer0 },
						{ "integer", "end", "layout", integer0 },
						{ "string", "allowed", "layout", null },
						{ "string", "notallowed", "layout", null },
						{ "string", "empty", "layout", null },
						{ "object", "underline", "paint", null },
						{ "method", "action" }, { "method", "insert" },
						{ "method", "remove" }, { "method", "caret" },
						{ "method", "perform" },
						{ "boolean", "border", "validate", Boolean.TRUE },
						{ "delegate", "formatter", "layout", null },
						{ "object", "alt", null, null },
						{ "content", "content" }, },
				"passwordfield",
				"textfield",
				null,
				"textarea",
				"textfield",
				new Object[][] { { "integer", "rows", "validate", integer0 },
						{ "boolean", "wrap", "layout", Boolean.FALSE },
						{ "method", "sigint" }, },
				"tabbedpane",
				"component",
				new Object[][] {
						{
								"choice",
								"placement",
								"validate",
								new String[] { "top", "left", "bottom",
										"right", "stacked" } },
						{ "integer", "selected", "paint", integer0 },
						{ "method", "action" } }, // ...focus
				"tab",
				"choice",
				new Object[][] { { "integer", "mnemonic", "paint", integer_1 } },
				"panel",
				"component",
				new Object[][] {
						{ "integer", "columns", "validate", integer0 },
						{ "integer", "top", "validate", integer0 },
						{ "integer", "left", "validate", integer0 },
						{ "integer", "bottom", "validate", integer0 },
						{ "integer", "right", "validate", integer0 },
						{ "integer", "gap", "validate", integer0 },
						{ "integer", "radius", "paint", integer_1 },
						{ "string", "text", "validate", null },
						{ "icon", "icon", "validate", null },
						{ "boolean", "border", "validate", Boolean.FALSE },
						{ "boolean", "scrollable", "validate", Boolean.FALSE }, },
				"desktop",
				"component",
				null,
				"dialog",
				"panel",
				new Object[][] { { "integer", "x", "paint", null },
						{ "integer", "y", "paint", null },
						{ "boolean", "modal", null, Boolean.FALSE },
						{ "boolean", "resizable", "validate", Boolean.FALSE },
						{ "method", "close", },
						{ "boolean", "maximizable", "paint", Boolean.FALSE },
						{ "boolean", "iconifiable", "paint", Boolean.FALSE },
						{ "boolean", "xlock", "validate", Boolean.TRUE },
						{ "boolean", "ylock", "validate", Boolean.TRUE } },
				"dropzone",
				"panel",
				new Object[][] { { "content", "selection" }, { "method", "deselect" }, 
						{ "method", "selectchanged" },
						{ "method", "delete" }, },
				"spinbox",
				"textfield",
				new Object[][] {
						{ "integer", "minimum", null,
								new Integer(Integer.MIN_VALUE) },
						{ "integer", "maximum", null,
								new Integer(Integer.MAX_VALUE) },
						{ "integer", "step", null, integer1 },
						{ "integer", "value", null, integer0 } }, // == text?
																	// deprecated
				"progressbar",
				"component",
				new Object[][] {
						{ "choice", "orientation", "validate", orientation },
						{ "integer", "minimum", "paint", integer0 }, // ...checkvalue
						{ "integer", "maximum", "paint", new Integer(100) },
						{ "integer", "value", "paint", integer0 } },
				// change stringpainted
				"slider",
				"progressbar",
				new Object[][] { { "integer", "unit", null, new Integer(5) },
						{ "integer", "block", null, new Integer(25) },
						{ "method", "action" } },
				// minor/majortickspacing
				// inverted
				// labelincrement labelstart
				"splitpane",
				"component",
				new Object[][] {
						{ "choice", "orientation", "validate", orientation },
						{ "integer", "divider", "layout", integer_1 } },
				"list",
				"component",
				new Object[][] {
						{ "choice", "selection", "paint", selections },
						{ "method", "action" }, { "method", "perform" },
						{ "content", "content" },
						{ "integer", "selectedIndex" },
						{ "boolean", "line", "validate", Boolean.TRUE },
						{ "string", "empty", "layout", null }, },
				"item",
				"choice",
				new Object[][] { { "boolean", "selected", "paint",
						Boolean.FALSE } },
				"table",
				"list",
				new Object[][] {
						{ "choice", "sort", null,
								new String[] { "none", "ascent", "descent" } },
						{ "string", "key", null, null },
				/*
				 * { "choice", "selection", new String[] { "singlerow",
				 * "rowinterval", "multiplerow", "cell", "cellinterval",
				 * "singlecolumn", "columninterval", "multiplecolumn" } }
				 */},
				"header",
				null,
				new Object[][] { { "method", "action" },
						{ "boolean", "resizable", null, Boolean.FALSE },
						{ "boolean", "sortable", null, Boolean.FALSE },
						{ "outlet", "outlet" } },
				// reordering allowed
				// autoresize mode: off next (column boundries) subsequents last
				// all columns
				// column row selection
				// selection row column cell
				// editing row/column
				"column",
				"choice",
				new Object[][] {
						{ "integer", "width", "parent", new Integer(80) },
						{ "choice", "sort", null,
								new String[] { "none", "ascent", "descent" } },
						{ "boolean", "selected", null, Boolean.FALSE },
						{ "string", "key", null, null },
						{ "content", "content" },
						{ "method", "action" },
						{ "method", "show" },
						{ "delegate", "contentmapper" },
						{ "delegate", "cellFormatter" },
						{ "boolean", "editable", "parent", Boolean.FALSE },
						{ "boolean", "visible", "parent", Boolean.TRUE },
						{ "choice", "cellAlignment", "parent",
								fillleftcenterright }, },
				"row",
				null,
				new Object[][] {
						{ "boolean", "selected", "paint", Boolean.FALSE },
						{ "outlet", "outlet" } },
				"cell",
				"choice",
				null,
				"tree",
				"list",
				new Object[][] { { "boolean", "angle", null, Boolean.FALSE },
						{ "method", "expand" }, { "method", "collapse" } },
				"node",
				"choice",
				new Object[][] {
						{ "boolean", "selected", null, Boolean.FALSE },
						{ "boolean", "expanded", null, Boolean.TRUE } },
				"separator",
				"component",
				null,
				"menubar",
				"component",
				new Object[][] { { "choice", "placement", "validate",
						new String[] { "top", "bottom" } } },
				"menu",
				"choice",
				new Object[][] { { "integer", "mnemonic", "paint", integer_1 } },
				"menuitem",
				"choice",
				new Object[][] { { "keystroke", "accelerator", null, null },
						{ "method", "action" },
						{ "boolean", "visible", "parent", Boolean.TRUE },
						{ "integer", "mnemonic", "paint", integer_1 } },
				"checkboxmenuitem",
				"menuitem",
				new Object[][] {
						{ "boolean", "selected", "paint", Boolean.FALSE }, // ...group
						{ "string", "group", "paint", null } }, // ...group
				"popupmenu",
				"component",
				new Object[][] { { "method", "menushown" } }, // Post menu:
																// Shift+F10
				"bean",
				"component",
				new Object[][] { { "bean", "bean", null, null }, },
				"flow",
				null,
				new Object[][] { {"outlet", "outlet" }, { "string", "name", null, null },
						{ "string", "state", "reeval", "" },
						{ "property", "property", null, null }, },
				"cond",
				null,
				new Object[][] { { "method", "test" },
						{ "string", "state", null, "" },
						{ "property", "property", null, null }, },
				"confirm",
				null,
				new Object[][] { { "string", "title", null, " " },
						{ "string", "text", null, "" },
						{ "string", "explanation", null, "" },
						{ "property", "property", null, null }, },
				"option",
				null,
				new Object[][] { { "string", "text", null, "" },
						{ "property", "property", null, null }, },
				"action",
				null,
				new Object[][] { {"outlet", "outlet" }, { "method", "method" },
						{ "string", "state", null, null },
						{ "property", "property", null, null }, },

				"prim",
				"component",
				new Object[][] { { "string", "for", "validate", null },
						{ "boolean", "fill", "paint", Boolean.FALSE },
						{ "boolean", "border", "paint", Boolean.FALSE },
						{ "integer", "xoffset", "validate", integer0 },
						{ "integer", "yoffset", "validate", integer0 }, },

				"oval",
				"prim",
				null,
				"rect",
				"prim",
				null,
				"line",
				"prim",
				new Object[][] { { "string", "lfor", "validate", null },
						{ "string", "rfor", "validate", null }, },
				"note",
				"prim",
				new Object[][] { { "string", "text", "validate", null },
						{ "icon", "icon", "validate", null }, }, };
	}
}
