package thinlet;

import java.util.Vector;
import java.util.Hashtable;

import java.io.IOException;
import java.io.InputStream;
import java.io.Reader;
import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.io.UnsupportedEncodingException;

import java.net.URL;
import java.net.MalformedURLException;

public class ThinletXML extends ThinletPaint {
	/**
	 * Creates a component (and its subcomponents, and properties) from the
	 * given xml resource
	 * 
	 * @param path
	 *            is relative to your thinlet instance or the classpath (if the
	 *            path starts with an <i>/</i> character), or a full URL
	 * @return the root component of the parsed resource
	 * @throws java.io.IOException
	 */
	public Object parse(String path) throws IOException {
		return parse(path, this);
	}

	/**
	 * Creates a component from the given xml resource using the specified event
	 * handler
	 * 
	 * @param path
	 *            is relative to your application package or the classpath, or
	 *            an URL
	 * @param handler
	 *            bussiness methods are implemented in this object
	 * @return the parsed components' root
	 * @throws java.io.IOException
	 */
	public Object parse(String path, Object handler) throws IOException {
		InputStream inputstream = null;
		try {
			inputstream = getClass().getResourceAsStream(path);
			if (inputstream == null) {
				try {
					inputstream = new URL(path).openStream();
				} catch (MalformedURLException mfe) { /*
														 * thows
														 * nullpointerexception
														 */
				}
			}
		} catch (Throwable e) {
		}
		return parse(inputstream, handler);
	}

	/**
	 * Creates a component from the given stream
	 * 
	 * @param inputstream
	 *            e.g. <i>new URL("http://myserver/myservlet").openStream()</i>
	 * @return the root component of the parsed stream
	 * @throws java.io.IOException
	 */
	public Object parse(InputStream inputstream) throws IOException {
		return parse(inputstream, this);
	}

	/**
	 * Creates a component from the given stream and event handler
	 * 
	 * @param inputstream
	 *            read xml from this stream
	 * @param handler
	 *            event handlers are implemented in this object
	 * @return the parsed components' root
	 * @throws java.io.IOException
	 */
	public Object parse(InputStream inputstream, Object handler)
			throws IOException {
		return parse(inputstream, 'T', handler);
	}

	/**
	 * You can use the internal xml parser as a simple SAX-like parser, during
	 * the process it calls the <i>startElement </i>, <i>characters </i>, and
	 * <i>endElement </i> methods
	 * 
	 * @param inputstream
	 *            e.g. <i>new URL("http://myserver/myservlet").openStream() </i>
	 * @throws java.io.IOException
	 */
	protected void parseXML(InputStream inputstream) throws IOException {
		parse(inputstream, 'S', null);
	}

	/**
	 * The SAX-like parser calls this method, you have to overwrite it
	 * 
	 * @param name
	 *            of the tag
	 * @param attributelist
	 *            a list of attributes including keys and value pairs
	 */
	protected void startElement(String name, Hashtable attributelist) {
	}

	/**
	 * The SAX-like parser calls this method, you have to overwrite it
	 * 
	 * @param text
	 *            the content of a tag
	 */
	protected void characters(String text) {
	}

	/**
	 * The SAX-like parser calls this method, you have to overwrite it
	 */
	protected void endElement() {
	}

	/**
	 * You can use the internal xml parser as a simple DOM-like parser, use the
	 * <i>getDOMAttribute </i>, <i>getDOMText </i>, <i>getDOMCount </i>,
	 * <i>getDOMNode </i>, <i>getClass </i>, and <i>getParent </i> methods to
	 * analise the document
	 * 
	 * @param inputstream
	 *            e.g. <i>new URL("http://myserver/myservlet").openStream() </i>
	 * @return the root tag
	 * @throws java.io.IOException
	 */
	protected Object parseDOM(InputStream inputstream) throws IOException {
		return parse(inputstream, 'D', null);
	}

	/**
	 * Gets the attribute value by the specified key for a DOM tag
	 * 
	 * @param node
	 *            a specified tag
	 * @param key
	 *            a string to identify the value pair
	 * @return the value, or null
	 */
	protected static String getDOMAttribute(Object node, String key) {
		return (String) get(node, key.intern());
	}

	/**
	 * Gets the content string of a tag
	 * 
	 * @param node
	 *            a specified tag
	 * @return the value, or null
	 */
	protected static String getDOMText(Object node) {
		return (String) get(node, ":text");
	}

	/**
	 * Gets the number of tags in a tag by a specified tagname
	 * 
	 * @param node
	 *            a specified tag
	 * @param key
	 *            the searched tagname
	 * @return the number of tags
	 */
	protected static int getDOMCount(Object node, String key) {
		return getItemCountImpl(node, key.intern());
	}

	/**
	 * Gets the subtag of the specified tag by tagname and index
	 * 
	 * @param node
	 *            a specified tag
	 * @param key
	 *            the searched tagname
	 * @param index
	 *            the index of the requested subtag
	 * @return the found tag, or null
	 */
	protected static Object getDOMNode(Object node, String key, int index) {
		return getItemImpl(node, key.intern(), index);
	}

	StringBuffer clearStringBuffer(StringBuffer sb) {
		StringBuffer retval;

		if (sb.capacity() > 32) {
			retval = new StringBuffer();
		} else {
			sb.setLength(0);
			retval = sb;
		}
		return retval;
	}

	/**
	 * 
	 * @param inputstream
	 * @param validate
	 *            parse GUI from xml if true
	 * @param dom
	 *            parse an xml resoource
	 * @param handler
	 * @return
	 * @throws java.io.IOException
	 * @throws java.lang.IllegalArgumentException
	 */
	private Object parse(InputStream inputstream, char mode, Object handler)
			throws IOException {
		Reader reader = new BufferedReader(new InputStreamReader(inputstream));
		try {
			Object[] parentlist = null;
			Object current = null;
			Hashtable attributelist = null;
			Vector methods = null;
			StringBuffer text = new StringBuffer();
			String encoding = null; // encoding value of xml declaration
			for (int c = reader.read(); c != -1;) {
				if (c == '<') {
					if (text.length() > 0) {
						if (text.charAt(text.length() - 1) == ' ') {
							text.setLength(text.length() - 1);
						}
						if (mode == 'D') { // DOM parser
							set(current, ":text", text.toString());
						} else if (mode == 'S') { // SAX parser
							characters(text.toString());
						}
						text = clearStringBuffer(text);
					}
					if ((c = reader.read()) == '/') { // endtag
						String tagname = (String) parentlist[2];
						for (int i = 0; i < tagname.length(); i++) { // check
																		// current
																		// tag's
																		// name
							if ((c = reader.read()) != tagname.charAt(i)) {
								throw new IllegalArgumentException(tagname);
							}
						}
						while (" \t\n\r".indexOf(c = reader.read()) != -1)
							; // read whitespace
						if (c != '>')
							throw new IllegalArgumentException(); // read '>'
						c = reader.read();
						if (mode == 'S') {
							endElement();
						}
						if (parentlist[0] == null) {
							reader.close();
							finishParse(methods, current, handler);
							return current;
						}
						c = reader.read();
						current = parentlist[0];
						parentlist = (Object[]) parentlist[1];
					} else if (c == '!') { // doctype
						while ((c = reader.read()) != '>')
							; // read to '>'
						c = reader.read();
					} else { // start or standalone tag
						text = clearStringBuffer(text);
						boolean iscomment = false;
						while (">/ \t\n\r".indexOf(c) == -1) { // to next
																// whitespace or
																// '/'
							text.append((char) c);
							if ((text.length() == 3) && (text.charAt(0) == '!')
								&& (text.charAt(1) == '-')
								&& (text.charAt(2) == '-')) { // comment
								int m = 0;
								while (true) { // read to '-->'
									c = reader.read();
									if (c == '-') {
										m++;
									} else if ((c == '>') && (m >= 2)) {
										break;
									} else {
										m = 0;
									}
								}
								iscomment = true;
							}
							c = reader.read();
						}
						if (iscomment) {
							continue;
						}
						if (text.length() == 0)
							throw new IllegalArgumentException();
						boolean pi = (text.charAt(0) == '?'); // processing
																// instruction
						String tagname = text.toString();
						if (!pi) { // tagname is available
							parentlist = new Object[] { current, parentlist,
									tagname };
							if (mode == 'T') { // GUI parser
								current = (current != null) ? addElement(
										current, tagname) : create(tagname);
							} else if (mode == 'D') { // DOM parser
								Object parent = current;
								current = createImpl(tagname = tagname.intern());
								if (parent != null) {
									insertItemImpl(parent, tagname, current, -1);
									// XXX set(current, ":parent", null);
								}
							} else { // SAX parser
								current = new String(tagname);
							}
						}
						text = clearStringBuffer(text);
						while (true) { // read attributes
							boolean whitespace = false;
							while (" \t\n\r".indexOf(c) != -1) { // read
																	// whitespace
								c = reader.read();
								whitespace = true;
							}
							if (pi && (c == '?')) { // end of processing
													// instruction
								if ((c = reader.read()) != '>') {
									throw new IllegalArgumentException(); // read
																			// '>'
								}
							} else if (c == '>') { // end of tag start
								if (mode == 'S') { // SAX parser
									startElement((String) current,
											attributelist);
									attributelist = null;
								} else if ("flow" == getClass(current)) {
									String name = getString(current, "name");
									Object parent, def[];

									if (name == null) {
										throw new IllegalArgumentException(
												"flow must have a name");
									}
									parent = getParent(current);
									def = getDefinition(getClass(parent), name,
											null);
									set(parent, def[1], current);
								}
							} else if (c == '/') { // standalone tag
								if ((c = reader.read()) != '>') {
									throw new IllegalArgumentException(); // read
																			// '>'
								}
								if (mode == 'S') { // SAX parser
									startElement((String) current,
											attributelist);
									attributelist = null;
									endElement();
								}
								if (parentlist[0] == null) {
									reader.close();
									finishParse(methods, current, handler);
									return current;
								}
								current = parentlist[0];
								parentlist = (Object[]) parentlist[1];
							} else if (whitespace) {
								while ("= \t\n\r".indexOf(c) == -1) { // read
																		// to
																		// key's
																		// end
									text.append((char) c);
									c = reader.read();
								}
								String key = text.toString();
								text = clearStringBuffer(text);
								while (" \t\n\r".indexOf(c) != -1)
									c = reader.read();
								if (c != '=')
									throw new IllegalArgumentException();
								while (" \t\n\r".indexOf(c = reader.read()) != -1)
									;
								char quote = (char) c;
								if ((c != '\"') && (c != '\''))
									throw new IllegalArgumentException();
								while (quote != (c = reader.read())) {
									if (c == '&') {
										StringBuffer eb = new StringBuffer();
										while (';' != (c = reader.read()) && c != -1) {
											eb.append((char) c);
										}
										String entity = eb.toString();
										if ("lt".equals(entity)) {
											text.append('<');
										} else if ("gt".equals(entity)) {
											text.append('>');
										} else if ("amp".equals(entity)) {
											text.append('&');
										} else if ("quot".equals(entity)) {
											text.append('"');
										} else if ("apos".equals(entity)) {
											text.append('\'');
										} else if ("bull".equals(entity)) {
											text.append('\u2022');
										} else if ("nbsp".equals(entity)) {
											text.append(' ');
										} else if (entity.startsWith("#")) {
											boolean hexa = (entity.charAt(1) == 'x');
											text.append((char) Integer
													.parseInt(entity
															.substring(hexa ? 2
																	: 1),
															hexa ? 16 : 10));
										} else
											throw new IllegalArgumentException(
													"unknown " + "entity "
															+ entity);
									} else
										text.append((char) c);
								}
								if (pi) {
									if ("?xml".equals(tagname)
										&& "encoding".equals(key)) {
										try {
											String enc = text.toString();
											new String(new byte[0], 0, 0, enc);
											encoding = new String(enc);
										} catch (UnsupportedEncodingException uee) {
											System.err
													.println(uee.getMessage());
										}
									}
								} else if (mode == 'T') { // GUI parser
									methods = addAttribute(current, key, text
											.toString(), encoding, methods,
											handler);
								} else if (mode == 'D') { // DOM parser
									set(current, key.intern(), new String(text
											.toString()));
								} else { // SAX parser
									if (attributelist == null) {
										attributelist = new Hashtable();
									}
									attributelist.put(new String(key),
											new String(text.toString()));
								}
								// '<![CDATA[' ']]>'
								text = clearStringBuffer(text);
								c = reader.read();
								continue;
							} else
								throw new IllegalArgumentException();
							c = reader.read();
							break;
						}
					}
				} else {
					if (" \t\n\r".indexOf(c) != -1) {
						if ((text.length() > 0)
							&& (text.charAt(text.length() - 1) != ' ')) {
							text.append(' ');
						}
					} else {
						text.append((char) c);
					}
					c = reader.read();
				}
			}
			throw new IllegalArgumentException();
		} finally {
			if (reader != null) {
				reader.close();
			}
		}
	}

	/**
	 * 
	 */
	private void finishParse(Vector methods, Object root, Object handler) {
		if (methods != null)
			for (int i = 0; i < methods.size(); i += 3) {
				Object component = methods.elementAt(i);
				Object[] definition = (Object[]) methods.elementAt(i + 1);
				String value = (String) methods.elementAt(i + 2);

				if ((value.length() > 0) && (value.charAt(0) == '@')) {
					bindAttribute(component, (String) definition[1], value,
							handler);
				} else if ("method" == definition[0]) {
					ThinletMethod method = getMethod(component, value, root,
							handler);
					if ("init" == definition[1]) {
						invokeImpl(method, null);
					} else {
						set(component, definition[1], method);
					}
				} else if ("outlet" == definition[0]) {
					try {
						tkv.setKeyValue(handler, value, component);
					} catch (Throwable e) {
						throw new IllegalArgumentException(
								"Exception while setting outlet '" + value
										+ "': " + e.toString());
					}
				} else { // ("component" == definition[0])
					Object reference = find(root, value); // +start find from
															// the component
					if (reference == null)
						throw new IllegalArgumentException(value + " not found");
					set(component, definition[1], reference);
				}
			}
	}

}
